/*
 * CRC functions
 * Based on RFC 1952
 *
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <types.h>

#include <liberror.h>

/* Table of CRCs of all 8-bit messages.
 */
uint32_t libfmapi_crc_table[ 256 ];

/* Value to indicate the crc table been computed
 */
int libfmapi_crc_table_computed = 0;

/* Initializes the internal crc table
 * The table speeds up the crc calculation
 */
void libfmapi_initialize_crc_table(
      void )
{
	uint32_t crc                = 0;
	uint16_t crc_table_iterator = 0;
	uint8_t bit_iterator        = 0;

	for( crc_table_iterator = 0; crc_table_iterator < 256; crc_table_iterator++ )
	{
		crc = (uint32_t) crc_table_iterator;

		for( bit_iterator = 0; bit_iterator < 8; bit_iterator++ )
		{
			if( crc & 1 )
			{
				crc = (uint32_t) 0xedb88320L ^ ( crc >> 1 );
			}
			else
			{
				crc = crc >> 1;
			}
		}
		libfmapi_crc_table[ crc_table_iterator ] = crc;
	}
	libfmapi_crc_table_computed = 1;
}

/* Calculates the crc of a buffer
 * Use a previous key of 0 to calculate a new crc
 */
int libfmapi_crc_calculate(
     uint32_t previous_key,
     uint8_t *buffer,
     size_t size,
     uint32_t *crc,
     liberror_error_t **error )
{
	static char *function = "libfmapi_crc_calculate";
	size_t buffer_iterator = 0;

	if( buffer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid buffer.",
		 function );

		return( -1 );
	}
	if( size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( crc == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid crc.",
		 function );

		return( -1 );
	}
	*crc = previous_key ^ (uint32_t) 0xffffffffL;

        if( libfmapi_crc_table_computed == 0 )
	{
		libfmapi_initialize_crc_table();
	}
        for( buffer_iterator = 0; buffer_iterator < size; buffer_iterator++ )
	{
		*crc = libfmapi_crc_table[ ( *crc ^ buffer[ buffer_iterator ] ) & 0x000000ffL ] ^ ( *crc >> 8 );
        }
        *crc ^= 0xffffffffL;

	return( 1 );
}

/* Calculates the weak crc of a buffer
 * Returns 1 if successful or -1 on error
 */
int libfmapi_crc_weak_calculate(
     uint32_t previous_key,
     uint8_t *buffer,
     size_t size,
     uint32_t *crc,
     liberror_error_t **error )
{
	static char *function = "libfmapi_crc_weak_calculate";
	size_t buffer_iterator = 0;

	if( buffer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid buffer.",
		 function );

		return( -1 );
	}
	if( size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( crc == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid crc.",
		 function );

		return( -1 );
	}
	*crc = previous_key;

        if( libfmapi_crc_table_computed == 0 )
	{
		libfmapi_initialize_crc_table();
	}
        for( buffer_iterator = 0; buffer_iterator < size; buffer_iterator++ )
	{
		*crc = libfmapi_crc_table[ ( *crc ^ buffer[ buffer_iterator ] ) & 0x000000ffL ] ^ ( *crc >> 8 );
        }
        return( 1 );
}

