/*
 * Security identifier (SID) functions
 *
 * Copyright (c) 2009-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <types.h>

#include <liberror.h>

#include "libfmapi_security_identifier.h"

/* Converts a class identifier stored in a byte stream into a runtime version
 * Returns 1 if successful or -1 on error
 */
int libfmapi_security_identifier_from_byte_stream(
     libfmapi_security_identifier_t *security_identifier,
     uint8_t *byte_stream,
     size_t byte_stream_size,
     liberror_error_t **error )
{
	static char *function          = "libfmapi_security_identifier_from_byte_stream";
	uint8_t sub_authority_iterator = 0;

	if( security_identifier == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid class identifier.",
		 function );

		return( -1 );
	}
	if( byte_stream == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid byte stream.",
		 function );

		return( -1 );
	}
	if( byte_stream_size < 8 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( byte_stream_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: byte stream size exceeds maximum.",
		 function );

		return( -1 );
	}
	security_identifier->revision_number           = byte_stream[ 0 ];
	security_identifier->number_of_sub_authorities = byte_stream[ 1 ];

	if( byte_stream_size < (size_t) ( 8 + ( security_identifier->number_of_sub_authorities * 4 ) ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( security_identifier->number_of_sub_authorities > 10 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported security identifier contains more than 10 sub authoritites.",
		 function );

		return( -1 );
	}
	byte_stream_copy_to_uint64_big_endian(
	 byte_stream,
	 security_identifier->authority );

	/* The authority is stored as a 48-bit value
	 */
	security_identifier->authority &= 0x0000ffffffffffffLL;

	fprintf( stderr, "S-%" PRIu8 "-%" PRIu64 "",
	 security_identifier->revision_number,
	 security_identifier->authority );

	byte_stream += 8;

	for( sub_authority_iterator = 0;
	     sub_authority_iterator < security_identifier->number_of_sub_authorities;
	     sub_authority_iterator++ )
	{
		byte_stream_copy_to_uint32_little_endian(
		 byte_stream,
		 security_identifier->sub_authority[ sub_authority_iterator ] );

		fprintf( stderr, "-%" PRIu32 "",
		 security_identifier->sub_authority[ sub_authority_iterator ] );

		byte_stream += 4;
	}
	fprintf( stderr, "\n" );

	return( 1 );
}

