/*
 * Security identifier (SID) functions
 *
 * Copyright (c) 2009-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#if !defined( _LIBFMAPI_SECURITY_IDENTIFIER_H )
#define _LIBFMAPI_SECURITY_IDENTIFIER_H

#include <common.h>
#include <types.h>

#include <liberror.h>

#if defined( __cplusplus )
extern "C" {
#endif

/* SID has format as follows: S-1-5-21-7623811015-3361044348-030300820-1013
 *
 * S - The string is a SID.
 * 1 - The revision number (the version of the SID specification).
 * 5 - The identifier authority value.
 * 21-7623811015-3361044348-030300820 - domain or local computer identifier
 * 1013 – a Relative ID (RID). Any group or user that is not created by default will have a Relative ID of 1000 or greater. 
 *
 * The SID is stored as
 * 01       - 05      - 00 00 00 00 00 05 - aa aa aa aa - bb bb bb bb - cc cc cc cc - dd dd dd dd - ee ee ee ee 
 * uint8_t  - uint8_t - uint48_t          - uint32_t    - uin32_t     - uin32_t     - uin32_t     - uin32_t
 *                    - (big endian)      - (little endian)
 * revision - number  - authority         - sub authorities
 *
 * The second value contains the number of sub authorities in the SID
 */

enum LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITIES
{
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_NULL			= 0x00,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_WORLD			= 0x01,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_LOCAL			= 0x02,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_CREATOR			= 0x03,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_NON_UNIQUE		= 0x04,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_NT_AUTHORITY		= 0x05,
	LIBFMAPI_SECURITY_IDENTIFIER_AUTHORITY_EXCHANGE_2007_AUTHORITY	= 0x09
};

/* A runtime version of the NT security identifier (SID)
 */
typedef struct libfmapi_security_identifier libfmapi_security_identifier_t;

struct libfmapi_security_identifier
{
	/* The revision number
	 */
	uint8_t revision_number;

	/* The authority value
	 */
	uint64_t authority;

	/* The number of sub authoritties
	 */
	uint8_t number_of_sub_authorities;

	/* The sub autoritites
	 */
	uint32_t sub_authority[ 10 ];
};

int libfmapi_security_identifier_from_byte_stream(
     libfmapi_security_identifier_t *security_identifier,
     uint8_t *byte_stream,
     size_t byte_stream_size,
     liberror_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif

