/*
 * Attached file functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2006-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>

#include "libpff_attachment.h"
#include "libpff_attached_file.h"
#include "libpff_definitions.h"
#include "libpff_libbfio.h"
#include "libpff_types.h"

/* Initializes the attached file IO handle
 * Returns 1 if successful or -1 on error
 */
int libpff_attached_file_io_handle_initialize(
     libpff_attached_file_io_handle_t **attached_file_io_handle,
     liberror_error_t **error )
{
	static char *function = "libpff_attached_file_io_handle_initialize";

	if( attached_file_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid attached file IO handle.",
		 function );

		return( -1 );
	}
	if( *attached_file_io_handle == NULL )
	{
		*attached_file_io_handle = (libpff_attached_file_io_handle_t *) memory_allocate(
		                                                                 sizeof( libpff_attached_file_io_handle_t ) );

		if( *attached_file_io_handle == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create attached file IO handle.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *attached_file_io_handle,
		     0,
		     sizeof( libpff_attached_file_io_handle_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear attached file IO handle.",
			 function );

			memory_free(
			 *attached_file_io_handle );

			*attached_file_io_handle = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees the attached file IO handle and its attributes
 * Returns 1 if succesful or -1 on error
 */
int libpff_attached_file_io_handle_free(
     intptr_t *io_handle,
     liberror_error_t **error )
{
	static char *function = "libpff_attached_file_io_handle_free";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	memory_free(
	 io_handle );

	return( 1 );
}

/* Clones (duplicates) the attached file IO handle and its attributes
 * Returns 1 if succesful or -1 on error
 */
int libpff_attached_file_io_handle_clone(
     intptr_t **destination_io_handle,
     intptr_t *source_io_handle,
     liberror_error_t **error )
{
	static char *function = "libpff_attached_file_io_handle_clone";

	if( destination_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination IO handle.",
		 function );

		return( -1 );
	}
	if( *destination_io_handle != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: destination IO handle already set.",
		 function );

		return( -1 );
	}
	if( source_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid source IO handle.",
		 function );

		return( -1 );
	}
	if( libpff_attached_file_io_handle_initialize(
	     (libpff_attached_file_io_handle_t **) destination_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create attached file IO handle.",
		 function );

		return( -1 );
	}
	( (libpff_attached_file_io_handle_t *) *destination_io_handle )->attachment   = ( (libpff_attached_file_io_handle_t *) source_io_handle )->attachment;
	( (libpff_attached_file_io_handle_t *) *destination_io_handle )->access_flags = ( (libpff_attached_file_io_handle_t *) source_io_handle )->access_flags;

	return( 1 );
}

/* Opens the attached file IO handle
 * Returns 1 if successful or -1 on error
 */
int libpff_attached_file_open(
     intptr_t *io_handle,
     int flags,
     liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_open";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: IO handle already open.",
		 function );

		return( -1 );
	}
	if( ( flags & LIBBFIO_FLAG_READ ) == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported flags.",
		 function );

		return( -1 );
	}
	/* Currently only support for reading data
	 */
	if( ( flags & ~( LIBBFIO_FLAG_READ ) ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported flags.",
		 function );

		return( -1 );
	}
	attached_file_io_handle->access_flags = flags;
	attached_file_io_handle->is_open      = 1;

	return( 1 );
}

/* Closes the attached file IO handle
 * Returns 0 if successful or -1 on error
 */
int libpff_attached_file_close(
     intptr_t *io_handle,
     liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_close";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - not open.",
		 function );

		return( -1 );
	}
	attached_file_io_handle->is_open = 0;

	return( 0 );
}

/* Reads a buffer from the attached file IO handle
 * Returns the number of bytes read if successful, or -1 on errror
 */
ssize_t libpff_attached_file_read(
         intptr_t *io_handle,
         uint8_t *buffer,
         size_t size,
         liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_read";
	ssize_t read_count                                        = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - not open.",
		 function );

		return( -1 );
	}
	if( ( attached_file_io_handle->access_flags & LIBBFIO_FLAG_READ ) == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - no read access.",
		 function );

		return( -1 );
	}
	read_count = libpff_attachment_data_read_buffer(
	              attached_file_io_handle->attachment,
	              buffer,
	              size,
	              error );

	if( read_count <= -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read attachment data.",
		 function );

		return( -1 );
	}
	return( read_count );
}

/* Writes a buffer to the attached file IO handle
 * Returns the number of bytes written if successful, or -1 on errror
 */
ssize_t libpff_attached_file_write(
         intptr_t *io_handle,
         uint8_t *buffer,
         size_t size,
         liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_write";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - not open.",
		 function );

		return( -1 );
	}
	if( ( attached_file_io_handle->access_flags & LIBBFIO_FLAG_WRITE ) == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - no write access.",
		 function );

		return( -1 );
	}
	if( buffer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid buffer.",
		 function );

		return( -1 );
	}
	if( size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid size value exceeds maximum.",
		 function );

		return( -1 );
	}
	return( 0 );
}

/* Seeks a certain offset within the attached file IO handle
 * Returns the offset if the seek is successful or -1 on error
 */
off64_t libpff_attached_file_seek_offset(
         intptr_t *io_handle,
         off64_t offset,
         int whence,
         liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_seek_offset";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - not open.",
		 function );

		return( -1 );
	}
	offset = libpff_attachment_data_seek_offset(
	          attached_file_io_handle->attachment,
	          offset,
	          whence,
	          error );

	if( offset == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek offset in attachment data.",
		 function );

		return( -1 );
	}
	return( offset );
}

/* Function to determine if the attached file exists
 * Return 1 if the attached file exists, 0 if not or -1 on error
 */
int libpff_attached_file_exists(
     intptr_t *io_handle,
     liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_exists";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		return( 0 );
	}
	return( 1 );
}

/* Check if the attached file is open
 * Returns 1 if open, 0 if not or -1 on error
 */
int libpff_attached_file_is_open(
     intptr_t *io_handle,
     liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_is_open";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( attached_file_io_handle->attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid IO handle - missing attachment item.",
		 function );

		return( -1 );
	}
	if( attached_file_io_handle->is_open == 0 )
	{
		return( 0 );
	}
	return( 1 );
}

/* Retrieves the attached file size
 * Returns 1 if successful or -1 on error
 */
int libpff_attached_file_get_size(
     intptr_t *io_handle,
     size64_t *size,
     liberror_error_t **error )
{
	libpff_attached_file_io_handle_t *attached_file_io_handle = NULL;
	static char *function                                     = "libpff_attached_file_get_size";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid IO handle.",
		 function );

		return( -1 );
	}
	attached_file_io_handle = (libpff_attached_file_io_handle_t *) io_handle;

	if( libpff_attachment_get_data_size(
	     attached_file_io_handle->attachment,
	     size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve attachment data size.",
		 function );

		return( -1 );
	}
	return( 1 );
}

