/*
 * Data array functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_data_array.h"
#include "libpff_definitions.h"
#include "libpff_encryption.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"

/* Initializes the data array and its values
 * The data is not created if data size is 0
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_initialize(
     libpff_data_array_t **data_array,
     uint32_t descriptor_identifier,
     libpff_io_handle_t *io_handle,
     size64_t data_size,
     uint32_t number_of_entries,
     liberror_error_t **error )
{
	static char *function      = "libpff_data_array_initialize";
	uint8_t cached_entry_index = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( number_of_entries > (uint32_t) INT32_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid number of entries value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( *data_array == NULL )
	{
		*data_array = (libpff_data_array_t *) memory_allocate(
		                                       sizeof( libpff_data_array_t ) );

		if( *data_array == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create data array.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *data_array,
		     0,
		     sizeof( libpff_data_array_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear data array.",
			 function );

			memory_free(
			 *data_array );

			*data_array = NULL;

			return( -1 );
		}
		( *data_array )->descriptor_identifier = descriptor_identifier;
		( *data_array )->io_handle             = io_handle;
		( *data_array )->data_size             = data_size;

		if( number_of_entries > 0 )
		{
			( *data_array )->identifier = (uint64_t *) memory_allocate(
								    sizeof( uint64_t ) * number_of_entries );
	   
			if( ( *data_array )->identifier == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create data array identifiers.",
				 function );

				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			if( memory_set(
			     ( *data_array )->identifier,
			     0,
			     sizeof( uint64_t ) * number_of_entries ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear data array identifiers.",
				 function );

				memory_free(
				 ( *data_array )->identifier );
				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			( *data_array )->offset = (off64_t *) memory_allocate(
							       sizeof( off64_t ) * number_of_entries );
	   
			if( ( *data_array )->offset == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create data array offsets.",
				 function );

				memory_free(
				 ( *data_array )->identifier );
				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			if( memory_set(
			     ( *data_array )->offset,
			     0,
			     sizeof( off64_t ) * number_of_entries ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear data array offsets.",
				 function );

				memory_free(
				 ( *data_array )->offset );
				memory_free(
				 ( *data_array )->identifier );
				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			( *data_array )->size = (size_t *) memory_allocate(
							    sizeof( size_t ) * number_of_entries );

			if( ( *data_array )->size == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create data array sizes.",
				 function );

				memory_free(
				 ( *data_array )->offset );
				memory_free(
				 ( *data_array )->identifier );
				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			if( memory_set(
			     ( *data_array )->size,
			     0,
			     sizeof( size_t ) * number_of_entries ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear data array sizes.",
				 function );

				memory_free(
				 ( *data_array )->size );
				memory_free(
				 ( *data_array )->offset );
				memory_free(
				 ( *data_array )->identifier );
				memory_free(
				 *data_array );

				*data_array = NULL;

				return( -1 );
			}
			( *data_array )->number_of_entries = number_of_entries;
		}
		/* The size of a single entry is 8180 or 8176 so 8 KiB (8192)
		 * should be sufficient as an initial value
		 */
		( *data_array )->data_cache_size = 8192;

		for( cached_entry_index = 0;
		     cached_entry_index < LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES;
		     cached_entry_index++ )
		{
			( ( *data_array )->cached_entry )[ cached_entry_index ] = LIBPFF_DATA_ARRAY_CACHED_ENTRY_FREE;
		}
	}
	return( 1 );
}

/* Frees the array and its values
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_free(
     libpff_data_array_t **data_array,
     liberror_error_t **error )
{
	static char *function      = "libpff_data_array_free";
	uint8_t cached_entry_index = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( *data_array != NULL )
	{
		if( ( *data_array )->identifier != NULL )
		{
			memory_free(
			 ( *data_array )->identifier );
		}
		if( ( *data_array )->offset != NULL )
		{
			memory_free(
			 ( *data_array )->offset );
		}
		if( ( *data_array )->size != NULL )
		{
			memory_free(
			 ( *data_array )->size );
		}
		for( cached_entry_index = 0;
		     cached_entry_index < LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES;
		     cached_entry_index++ )
		{
			if( ( *data_array )->data_cache[ cached_entry_index ] != NULL )
			{
				memory_free(
				 ( *data_array )->data_cache[ cached_entry_index ] );
			}
		}
		memory_free(
		 *data_array );

		*data_array = NULL;
	}
	return( 1 );
}

/* Resizes the data array
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_resize(
     libpff_data_array_t *data_array,
     uint32_t number_of_entries,
     liberror_error_t **error )
{
	void *reallocation    = NULL;
	static char *function = "libpff_data_array_initialize";

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( number_of_entries > (uint32_t) INT32_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid number of entries value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( number_of_entries > data_array->number_of_entries )
	{
		reallocation = memory_reallocate(
		                data_array->identifier,
				sizeof( uint64_t ) * number_of_entries );

		if( reallocation == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to resize data array identifiers.",
			 function );

			return( -1 );
		}
		data_array->identifier = (uint64_t *) reallocation;

		if( memory_set(
		     &( data_array->identifier[ data_array->number_of_entries ] ),
		     0,
		     sizeof( uint64_t ) * ( number_of_entries - data_array->number_of_entries ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear data array identifiers.",
			 function );

			return( -1 );
		}
		reallocation = memory_reallocate(
		                data_array->offset,
		                sizeof( off64_t ) * number_of_entries );

		if( reallocation == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to resize data array offsets.",
			 function );

			return( -1 );
		}
		data_array->offset = (off64_t *) reallocation;

		if( memory_set(
		     &( data_array->offset[ data_array->number_of_entries ] ),
		     0,
		     sizeof( off64_t ) * ( number_of_entries - data_array->number_of_entries ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear data array offsets.",
			 function );

			return( -1 );
		}
		reallocation = memory_reallocate(
		                data_array->size,
		                sizeof( size_t ) * number_of_entries );

		if( reallocation == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to resize data array sizes.",
			 function );

			return( -1 );
		}
		data_array->size = (size_t *) reallocation;

		if( memory_set(
		     &( data_array->size[ data_array->number_of_entries ] ),
		     0,
		     sizeof( size_t ) * ( number_of_entries - data_array->number_of_entries ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear data array sizes.",
			 function );

			return( -1 );
		}
		data_array->number_of_entries = number_of_entries;

	}
	return( 1 );
}

/* Resizes the data cache
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_resize_data_cache(
     libpff_data_array_t *data_array,
     size_t data_cache_size,
     uint8_t required_cached_entries,
     liberror_error_t **error )
{
	static char *function      = "libpff_data_array_set_entry";
	void *reallocation         = NULL;
	uint8_t cached_entry_index = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_cache_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid data cache size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( required_cached_entries >= LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: required number of cached entries: %" PRIu8 " exceeds available number of cache entries: %d.",
		 function,
		 required_cached_entries,
		 LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES );

		return( -1 );
	}
	for( cached_entry_index = 0;
	     cached_entry_index < LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES;
	     cached_entry_index++ )
	{
		if( cached_entry_index < required_cached_entries )
		{
			if( ( data_array->data_cache[ cached_entry_index ] == NULL )
			 || ( data_cache_size > data_array->data_cache_size ) )
			{
				reallocation = memory_reallocate(
						data_array->data_cache[ cached_entry_index ],
						sizeof( uint8_t ) * data_cache_size );

				if( reallocation == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_MEMORY,
					 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to resize data cache.",
					 function );

					return( -1 );
				}
				data_array->data_cache[ cached_entry_index ] = (uint8_t *) reallocation;
			}
		}
		else if( data_array->data_cache[ cached_entry_index ] != NULL )
		{
			memory_free(
			 data_array->data_cache[ cached_entry_index ] );

			data_array->data_cache[ cached_entry_index ]   = NULL;
			data_array->cached_entry[ cached_entry_index ] = LIBPFF_DATA_ARRAY_CACHED_ENTRY_FREE;
		}
	}
	data_array->data_cache_size         = data_cache_size;
	data_array->available_cache_entries = required_cached_entries;

	if( data_array->next_cache_entry >= required_cached_entries )
	{
		data_array->next_cache_entry = 0;
	}
	return( 1 );
}

/* Set a data array entry
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_set_entry(
     libpff_data_array_t *data_array,
     uint32_t data_array_entry,
     uint64_t entry_identifier,
     off64_t entry_data_offset,
     size_t entry_data_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_array_set_entry";

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->identifier == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing identifier.",
		 function );

		return( -1 );
	}
	if( data_array->offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing offsets.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( data_array_entry >= data_array->number_of_entries )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: data array: %" PRIu32 " entry: %" PRIu32 " exceeds number of entries: %" PRIu32 ".",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry,
		 data_array->number_of_entries );

		return( -1 );
	}
	if( entry_data_offset < 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid entry data offset value zero or less.",
		 function );

		return( -1 );
	}
	if( entry_data_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid entry data size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( entry_data_size > data_array->data_cache_size )
	{
		if( libpff_data_array_resize_data_cache(
		     data_array,
		     entry_data_size,
		     data_array->available_cache_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize data cache.",
			 function );

			return( -1 );
		}
	}
	data_array->identifier[ data_array_entry ] = entry_identifier;
	data_array->offset[ data_array_entry ]     = entry_data_offset;
	data_array->size[ data_array_entry ]       = entry_data_size;

	return( 1 );
}

/* Reads a data array entry
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_read_entry(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint32_t data_array_entry,
     uint8_t *buffer,
     size_t buffer_size,
     uint8_t flags,
     liberror_error_t **error )
{
	static char *function        = "libpff_data_array_read_entry";
	ssize_t read_count           = 0;
	int encryption_type          = 0;
	uint16_t table_index_offset  = 0;
	uint8_t decrypt_data         = 0;
	uint8_t node_identifier_type = 0;
	uint8_t node_contains_table  = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing io handle.",
		 function );

		return( -1 );
	}
	if( data_array->identifier == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing identifier.",
		 function );

		return( -1 );
	}
	if( data_array->offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing offsets.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( data_array_entry >= data_array->number_of_entries )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: data array: %" PRIu32 " entry: %" PRIu32 " exceeds number of entries: %" PRIu32 ".",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry,
		 data_array->number_of_entries );

		return( -1 );
	}
	if( buffer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid buffer.",
		 function );

		return( -1 );
	}
	if( buffer_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid buffer size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( buffer_size < data_array->size[ data_array_entry ] )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: invalid buffer size value too small.",
		 function );

		return( -1 );
	}
	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     data_array->offset[ data_array_entry ],
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek array entry offset: %" PRIu64 ".",
		 function,
		 data_array->offset[ data_array_entry ] );

		return( -1 );
	}
	read_count = libbfio_handle_read(
		      file_io_handle,
		      buffer,
		      data_array->size[ data_array_entry ],
		      error );

	if( read_count != (ssize_t) data_array->size[ data_array_entry ] )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read data.",
		 function );

		return( -1 );
	}
	encryption_type = (int) data_array->io_handle->encryption_type;

	/* Check if the internal (unencrypted) flag in (data) offset index identifier is not set
	 */
	if( ( data_array->identifier[ data_array_entry ] & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) == 0 )
	{
		decrypt_data = 1;
	}
	/* Check if data is encrypted
	 * some 'invalid' PFF's have an encryption type of none
	 * but do actually contain encrypted data
	 * although they are considered invalid by Outlook
	 * it is still useful to be able to read them
	 */
	if( ( data_array->io_handle->encryption_type == LIBPFF_ENCRYPTION_TYPE_NONE )
	 && ( ( flags & LIBPFF_GET_ENTRY_FLAG_IGNORE_FORCE_DECRYPTION ) == 0 )
	 && ( data_array->size[ data_array_entry ] > 4 ) )
	{
		node_identifier_type = (uint8_t) ( data_array->descriptor_identifier & 0x0000001fUL );

		if( ( ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_INTERNAL )
		  && ( ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_MESSAGE_STORE )
		   || ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_NAME_TO_ID_MAP )
		   || ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2049 )
		   || ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2081 )
		   || ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2113 )
		   || ( data_array->descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_3073 ) ) )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_FOLDER )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SEARCH_FOLDER )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_MESSAGE )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_ASSOCIATED_CONTENT )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_FOLDERS )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_MESSAGES )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_ASSOCIATED_CONTENTS )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SEARCH_CONTENTS_TABLE )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_ATTACHMENTS )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_RECIPIENTS )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1718 )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1751 )
		 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1784 ) )
		{
			node_contains_table = 1;
		}
		/* Only check the first entry
		 * some table array contain the table type in every array entry but not all
		 */
		if( ( data_array_entry == 0 )
		 && ( node_contains_table != 0 ) )
		{
			/* Test if the data contains an unencrypted table
			 * a table consists of 0xec in the third byte
			 * and 0x6c, 0x7c, 0x8c, 0x9c, 0xa5, 0xac, 0xbc, 0xcc in the fourth
			 */
			if( ( buffer[ 2 ] != 0xec )
			 || ( ( buffer[ 3 ] != 0x6c )
			  && ( buffer[ 3 ] != 0x7c )
			  && ( buffer[ 3 ] != 0x8c )
			  && ( buffer[ 3 ] != 0x9c )
			  && ( buffer[ 3 ] != 0xa5 )
			  && ( buffer[ 3 ] != 0xac )
			  && ( buffer[ 3 ] != 0xbc )
			  && ( buffer[ 3 ] != 0xcc ) ) )
			{
#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: detected compressible encrypted data while encryption type is none - decryption forced.\n",
					 function );
				}
#endif
				encryption_type                         = LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE;
				decrypt_data                            = 1;
				data_array->io_handle->force_decryption = 1;
				data_array->decryption_forced           = 1;
			}
		}
		else if( data_array->io_handle->force_decryption == 1 )
		{
			/* Some of the last table array entries do not seem to be encrypted
			 */
			if( ( node_contains_table != 0 )
			 && ( data_array_entry == ( data_array->number_of_entries - 1 ) ) )
			{
				byte_stream_copy_to_uint16_little_endian(
				 buffer,
				 table_index_offset );

				if( (size_t) table_index_offset > data_array->size[ data_array_entry ] )
				{
#if defined( HAVE_VERBOSE_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: detected encrypted last table array entry - decryption forced.\n",
						 function );
					}
#endif

					encryption_type               = LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE;
					decrypt_data                  = 1;
					data_array->decryption_forced = 1;
				}
			}
			else
			{
#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: decryption forced.\n",
					 function );
				}
#endif

				encryption_type               = LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE;
				decrypt_data                  = 1;
				data_array->decryption_forced = 1;
			}
		}
	}
	/* Check if unencrypted flag in offset index identifier is not set
	 */
	if( decrypt_data != 0 )
	{
		if( libpff_encryption_decrypt(
		     encryption_type,
		     (uint32_t) data_array->identifier[ data_array_entry ],
		     buffer,
		     data_array->size[ data_array_entry ],
		     error ) != (ssize_t) data_array->size[ data_array_entry ] )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ENCRYPTION,
			 LIBERROR_ENCRYPTION_ERROR_DECRYPT_FAILED,
			 "%s: unable to decrypt entry value data.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves a data array entry, the data is read into the cache if necessary
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_get_entry_data(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint32_t data_array_entry,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error )
{
	static char *function      = "libpff_data_array_get_entry_data";
	size_t entry_data_offset   = 0;
	uint32_t cache_entry_index = 0;
	uint8_t entry_is_cached    = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( data_array_entry >= data_array->number_of_entries )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: data array: %" PRIu32 " entry: %" PRIu32 " exceeds number of entries: %" PRIu32 ".",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry,
		 data_array->number_of_entries );

		return( -1 );
	}
	if( entry_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry data.",
		 function );

		return( -1 );
	}
	if( entry_data_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry data size.",
		 function );

		return( -1 );
	}
	*entry_data_size = data_array->size[ data_array_entry ];

	/* Check if all entries are cached
	 */
	if( data_array->cached_entry[ 0 ] == LIBPFF_DATA_ARRAY_CACHED_ENTRY_ALL )
	{
		for( cache_entry_index = 0;
		     cache_entry_index < data_array_entry;
		     cache_entry_index++ )
		{
			entry_data_offset += data_array->size[ cache_entry_index ];
		}
		if( data_array->data_cache[ 0 ] == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid data cache.",
			 function );

			return( -1 );
		}
		if( entry_data_offset >= data_array->data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: invalid entry data offset value out of range.",
			 function );

			return( -1 );
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: cached data array: %" PRIu32 " entry: %" PRIu32 " in data cache entry: %" PRIu8 "\n",
			 function,
			 data_array->descriptor_identifier,
			 data_array_entry,
			 cache_entry_index );
		}
#endif
		*entry_data = &( ( data_array->data_cache[ 0 ] )[ entry_data_offset ] );

		return( 1 );
	}
	/* Check if the entry is already cached
	 */
	for( cache_entry_index = 0;
	     cache_entry_index < (uint32_t) data_array->available_cache_entries;
	     cache_entry_index++ )
	{
		if( data_array->cached_entry[ cache_entry_index ] == data_array_entry )
		{
			if( data_array->data_cache[ cache_entry_index ] == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid data cache entry: %" PRIu8 ".",
				 function,
				 cache_entry_index );

				return( -1 );
			}
			if( ( flags & LIBPFF_GET_ENTRY_FLAG_IGNORE_CACHE ) == 0 )
			{
				/* Make sure this cache entry is kept a little longer available
				 */
				if( cache_entry_index == (uint32_t) data_array->next_cache_entry )
				{
					data_array->next_cache_entry++;

					if( data_array->next_cache_entry >= LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES )
					{
						data_array->next_cache_entry = 0;
					}
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: cached data array: %" PRIu32 " entry: %" PRIu32 " in data cache entry: %" PRIu8 "\n",
					 function,
					 data_array->descriptor_identifier,
					 data_array_entry,
					 cache_entry_index );
				}
#endif
				*entry_data = data_array->data_cache[ cache_entry_index ];

				return( 1 );
			}
			entry_is_cached = 1;

			break;
		}
	}
	if( entry_is_cached == 0 )
	{
		cache_entry_index = data_array->next_cache_entry;

		/* Allocate the data cache entry if necessary
		 */
		if( data_array->data_cache[ cache_entry_index ] == NULL )
		{
			if( data_array->available_cache_entries != cache_entry_index )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid number of available cache entries.",
				 function );

				return( -1 );
			}
			data_array->data_cache[ cache_entry_index ] = (uint8_t *) memory_allocate(
										   sizeof( uint8_t ) * data_array->data_cache_size );

			if( data_array->data_cache[ cache_entry_index ] == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create data cache entry: %" PRIu8 ".",
				 function,
				 cache_entry_index );

				return( -1 );
			}
			data_array->available_cache_entries++;
		}
		data_array->next_cache_entry++;

		if( data_array->next_cache_entry >= LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES )
		{
			data_array->next_cache_entry = 0;
		}
	}
	/* Read the entry from file into the cache
	 */
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading data array: %" PRIu32 " entry: %" PRIu32 " in data cache entry: %" PRIu8 "\n",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry,
		 cache_entry_index );
	}
#endif

	if( libpff_data_array_read_entry(
	     data_array,
	     file_io_handle,
	     data_array_entry,
	     data_array->data_cache[ cache_entry_index ],
	     data_array->data_cache_size,
	     flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read data array: %" PRIu32 " entry: %" PRIu32 ".",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry );

		return( -1 );
	}
	data_array->cached_entry[ cache_entry_index ] = data_array_entry;

	*entry_data = data_array->data_cache[ cache_entry_index ];

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: data array: %" PRIu32 " entry: %" PRIu32 ":\n",
		 function,
		 data_array->descriptor_identifier,
		 data_array_entry );
		libnotify_print_data(
		 *entry_data,
		 *entry_data_size );
	}
#endif

	return( 1 );
}

/* Retrieves the entry for a certain data offset
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_get_entry_at_offset(
     libpff_data_array_t *data_array,
     off64_t data_offset,
     uint32_t *data_array_entry,
     size_t *entry_data_offset,
     liberror_error_t **error )
{
	static char *function = "libpff_data_array_get_entry_at_offset";

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( ( data_offset < 0 )
	 || ( data_offset >= (off64_t) data_array->data_size ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid data offset value out of range.",
		 function );

		return( -1 );
	}
	if( data_array_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array entry.",
		 function );

		return( -1 );
	}
	if( entry_data_offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry data offset.",
		 function );

		return( -1 );
	}
	for( *data_array_entry = 0;
	     *data_array_entry < data_array->number_of_entries;
	     *data_array_entry += 1 )
	{
		if( data_offset < (off64_t) data_array->size[ *data_array_entry ] )
		{
			break;
		}
		data_offset -= data_array->size[ *data_array_entry ];
	}
	if( *data_array_entry >= data_array->number_of_entries )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid data array entry value out of range.",
		 function );

		return( -1 );
	}
	*entry_data_offset = (size_t) data_offset;

	return( 1 );
}

/* Retrieves the entry of an entry aligned index
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_get_entry_at_aligned_index(
     libpff_data_array_t *data_array,
     int value_index,
     size_t value_size,
     uint32_t *data_array_entry,
     size_t *entry_data_offset,
     liberror_error_t **error )
{
	static char *function = "libpff_data_array_get_entry_at_aligned_index";
	int number_of_values  = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( value_index < 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid value index value out of range.",
		 function );

		return( -1 );
	}
	if( value_size == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid value size value zero or less.",
		 function );

		return( -1 );
	}
	if( value_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid value size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( data_array_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array entry.",
		 function );

		return( -1 );
	}
	if( entry_data_offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry data offset.",
		 function );

		return( -1 );
	}
	for( *data_array_entry = 0;
	     *data_array_entry < data_array->number_of_entries;
	     *data_array_entry += 1 )
	{
		number_of_values = data_array->size[ *data_array_entry ] / value_size;

		if( value_index < number_of_values )
		{
			break;
		}
		value_index -= number_of_values;
	}
	if( *data_array_entry >= data_array->number_of_entries )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid data array entry value out of range.",
		 function );

		return( -1 );
	}
	*entry_data_offset = (size_t) value_index * value_size;

	return( 1 );
}

/* Retrieves the data size of the data array
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_get_data_size(
     libpff_data_array_t *data_array,
     size64_t *data_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_array_get_data_size";

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data size.",
		 function );

		return( -1 );
	}
	*data_size = data_array->data_size;

	return( 1 );
}

/* Retrieves the data of the data array
 * Returns 1 if successful or -1 on error
 */
int libpff_data_array_get_data(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint8_t **data,
     size_t *data_size,
     liberror_error_t **error )
{
	static char *function     = "libpff_data_array_get_data";
	size_t data_offset        = 0;
	uint32_t data_array_entry = 0;

	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array->size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data array - missing entry sizes.",
		 function );

		return( -1 );
	}
	if( data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data.",
		 function );

		return( -1 );
	}
	if( data_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data size.",
		 function );

		return( -1 );
	}
	if( ( data_array->data_size == 0 )
	 || ( data_array->data_size > (size64_t) SSIZE_MAX ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid data size value out of range.",
		 function );

		return( -1 );
	}
	if( data_array->cached_entry[ 0 ] != LIBPFF_DATA_ARRAY_CACHED_ENTRY_ALL )
	{
		/* Make sure the data cache is large enough to hold all the data
		 * in a single data cache entry
		 */
		if( ( data_array->data_cache[ 0 ] == NULL )
		 || ( data_array->data_size > data_array->data_cache_size ) )
		{
			if( libpff_data_array_resize_data_cache(
			     data_array,
			     (size_t) data_array->data_size,
			     1,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
				 "%s: unable to resize data cache.",
				 function );

				return( -1 );
			}
		}
		/* Read all the entries into the first (and only) data cache entry
		 */
		for( data_array_entry = 0;
		     data_array_entry < data_array->number_of_entries;
		     data_array_entry++ )
		{
			if( data_offset >= data_array->data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid data offset value out of range.",
				 function );

				return( -1 );
			}
			if( libpff_data_array_read_entry(
			     data_array,
			     file_io_handle,
			     data_array_entry,
			     &( ( data_array->data_cache[ 0 ] )[ data_offset ] ),
			     data_array->size[ data_array_entry ],
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read data array: %" PRIu32 " entry: %" PRIu32 ".",
				 function,
				 data_array->descriptor_identifier,
				 data_array_entry );

				return( -1 );

			}
			data_offset += data_array->size[ data_array_entry ];
		}
		data_array->cached_entry[ 0 ] = LIBPFF_DATA_ARRAY_CACHED_ENTRY_ALL;
	}
	*data      = data_array->data_cache[ 0 ];
	*data_size = (size_t) data_array->data_size;

	return( 1 );
}

