/*
 * Data array functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#if !defined( _LIBPFF_DATA_ARRAY_H )
#define _LIBPFF_DATA_ARRAY_H

#include <common.h>
#include <types.h>

#include <liberror.h>

#include "libpff_io_handle.h"
#include "libpff_libbfio.h"

#if defined( __cplusplus )
extern "C" {
#endif

#define LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES	8

#define LIBPFF_DATA_ARRAY_CACHED_ENTRY_FREE		(uint32_t) -1
#define LIBPFF_DATA_ARRAY_CACHED_ENTRY_ALL		(uint32_t) -2

typedef struct libpff_data_array libpff_data_array_t;

struct libpff_data_array
{
	/* Value to indicate if the data array descriptor identifier
	 */
	uint32_t descriptor_identifier;

	/* A reference to the io handle
	 */
	libpff_io_handle_t *io_handle;

	/* The total size of the data in the array
	 */
	size64_t data_size;

	/* The number of array entries
	 */
	uint32_t number_of_entries;

	/* The offset index identifiers
	 */
	uint64_t *identifier;

	/* The entry data (file) offsets
	 */
	off64_t *offset;

	/* The entry data sizes
	 */
	size_t *size;

	/* The data cache entry sizes
	 */
	size_t data_cache_size;

	/* The data cache entry
	 */
	uint8_t *data_cache[ LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES ];

	/* The cached entries
	 */
	uint32_t cached_entry[ LIBPFF_DATA_ARRAY_NUMBER_OF_CACHE_ENTRIES ];

	/* The number of available cached entries
	 */
	uint8_t available_cache_entries;

	/* The next cache entry
	 */
	uint8_t next_cache_entry;

	/* Value to indicate if decryption was forced
	 * on one or more of the array entries
	 */
	uint8_t decryption_forced;
};

int libpff_data_array_initialize(
     libpff_data_array_t **data_array,
     uint32_t descriptor_identifier,
     libpff_io_handle_t *io_handle,
     size64_t data_size,
     uint32_t number_of_entries,
     liberror_error_t **error );

int libpff_data_array_free(
     libpff_data_array_t **data_array,
     liberror_error_t **error );

int libpff_data_array_resize(
     libpff_data_array_t *data_array,
     uint32_t number_of_entries,
     liberror_error_t **error );

int libpff_data_array_resize_data_cache(
     libpff_data_array_t *data_array,
     size_t data_cache_size,
     uint8_t required_cached_entries,
     liberror_error_t **error );

int libpff_data_array_set_entry(
     libpff_data_array_t *data_array,
     uint32_t data_array_entry,
     uint64_t entry_identifier,
     off64_t entry_data_offset,
     size_t entry_data_size,
     liberror_error_t **error );

int libpff_data_array_read_entry(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint32_t data_array_entry,
     uint8_t *buffer,
     size_t buffer_size,
     uint8_t flags,
     liberror_error_t **error );

int libpff_data_array_get_entry_data(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint32_t data_array_entry,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error );

int libpff_data_array_get_entry_at_offset(
     libpff_data_array_t *data_array,
     off64_t data_offset,
     uint32_t *data_array_entry,
     size_t *entry_data_offset,
     liberror_error_t **error );

int libpff_data_array_get_entry_at_aligned_index(
     libpff_data_array_t *data_array,
     int value_index,
     size_t value_size,
     uint32_t *data_array_entry,
     size_t *entry_data_offset,
     liberror_error_t **error );

int libpff_data_array_get_data_size(
     libpff_data_array_t *data_array,
     size64_t *data_size,
     liberror_error_t **error );

int libpff_data_array_get_data(
     libpff_data_array_t *data_array,
     libbfio_handle_t *file_io_handle,
     uint8_t **data,
     size_t *data_size,
     liberror_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif

