/*
 * Dat block functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_definitions.h"
#include "libpff_data_block.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"
#include "libpff_libfmapi.h"

#include "pff_block.h"

/* Reads the data block
 * Returns 1 if successful or -1 on error
 */
int libpff_data_block_read(
     libbfio_handle_t *file_io_handle,
     off64_t data_file_offset,
     size32_t data_size,
     int file_type,
     liberror_error_t **error )
{
	uint8_t *data_block            = NULL;
	uint8_t *data_block_footer     = NULL;
	static char *function          = "libpff_data_block_read";
	size_t data_block_footer_size  = 0;
	size_t data_block_padding_size = 0;
	size_t data_block_size         = 0;
	ssize_t read_count             = 0;
	uint64_t block_back_pointer    = 0;
	uint32_t calculated_crc        = 0;
	uint32_t stored_crc            = 0;
	uint16_t block_data_size       = 0;
	uint16_t block_signature       = 0;

	if( ( file_type != LIBPFF_FILE_TYPE_32BIT )
	 && ( file_type != LIBPFF_FILE_TYPE_64BIT ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: invalid file type.",
		 function );

		return( -1 );
	}
	if( data_file_offset < 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid data file offset value out of range.",
		 function );

		return( -1 );
	}
	if( data_size > (size32_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: data size value exceeds maximum.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading data block at offset: %" PRIu64 " (0x%08" PRIx64 ")\n",
		 function,
		 data_file_offset,
		 data_file_offset );
	}
#endif

	/* Read the data block
	 */
	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     data_file_offset,
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek value offset: %" PRIu64 ".",
		 function,
		 data_file_offset );

		return( -1 );
	}
	data_block_size = data_size;

	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		data_block_footer_size = sizeof( pff_block_footer_32bit_t );
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		data_block_footer_size = sizeof( pff_block_footer_64bit_t );
	}
	data_block_size = data_size / 64;

	if( ( data_size % 64 ) != 0 )
	{
		data_block_size += 1;
	}
	data_block_size *= 64;

	if( ( data_block_size - data_size ) < data_block_footer_size )
	{
		data_block_size += 64;
	}
	data_block = (uint8_t *) memory_allocate(
	                          sizeof( uint8_t ) * data_block_size );

	if( data_block == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create data block.",
		 function );

		return( -1 );
	}
	read_count = libbfio_handle_read(
	              file_io_handle,
	              data_block,
	              data_block_size,
	              error );

	if( read_count != (ssize_t) data_block_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read data block.",
		 function );

		memory_free(
		 data_block );

		return( -1 );
	}
	data_block_padding_size = data_block_size - ( data_size + data_block_footer_size );

	data_block_footer = &( data_block[ data_block_size - data_block_footer_size ] );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: data block padding size\t: %" PRIzd "\n",
		 function,
		 data_block_padding_size );
		libnotify_printf(
		 "%s: data block padding:\n",
		 function );
		libnotify_print_data(
		 &( data_block[ data_size ] ),
		 data_block_padding_size );

		libnotify_printf(
		 "%s: data block footer:\n",
		 function );
		libnotify_print_data(
		 data_block_footer,
		 data_block_footer_size );
	}
#endif

	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_block_footer_32bit_t *) data_block_footer )->size,
		 block_data_size );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_block_footer_32bit_t *) data_block_footer )->signature,
		 block_signature );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_block_footer_32bit_t *) data_block_footer )->back_pointer,
		 block_back_pointer );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_block_footer_32bit_t *) data_block_footer )->crc,
		 stored_crc );
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_block_footer_64bit_t *) data_block_footer )->size,
		 block_data_size );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_block_footer_64bit_t *) data_block_footer )->signature,
		 block_signature );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_block_footer_64bit_t *) data_block_footer )->crc,
		 stored_crc );

		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_block_footer_64bit_t *) data_block_footer )->back_pointer,
		 block_back_pointer );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: data block data size\t: %" PRIu16 "\n",
		 function,
		 block_data_size );
		libnotify_printf(
		 "%s: data block signature\t: 0x%04" PRIx16 "\n",
		 function,
		 block_signature );

		if( file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			libnotify_printf(
			 "%s: data block back pointer\t: %" PRIu64 "\n",
			 function,
			 block_back_pointer );
			libnotify_printf(
			 "%s: data block data crc\t: 0x%08" PRIx32 "\n",
			 function,
			 stored_crc );
		}
		else if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: data block data crc\t: 0x%08" PRIx32 "\n",
			 function,
			 stored_crc );
			libnotify_printf(
			 "%s: data block back pointer\t: %" PRIu64 "\n",
			 function,
			 block_back_pointer );
		}
		libnotify_printf(
		 "\n" );
	}
#endif
	if( data_size != (size_t) block_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_INPUT,
		 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
		 "%s: mismatch in data size ( %" PRIzd " != %" PRIu16 " ).",
		 function,
		 data_size,
		 block_data_size );

		memory_free(
		 data_block );

		return( -1 );
	}
	if( stored_crc != 0 )
	{
		if( libfmapi_crc_weak_calculate(
		     0,
		     data_block,
		     data_size,
		     &calculated_crc,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unable to calculate crc.",
			 function );

			memory_free(
			 data_block );

			return( -1 );
		}
		if( stored_crc != calculated_crc )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_INPUT,
			 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
			 "%s: mismatch in data block crc ( %" PRIu32 " != %" PRIu32 " ).",
			 function,
			 stored_crc,
			 calculated_crc );

			memory_free(
			 data_block );


			return( -1 );
		}
	}
	memory_free(
	 data_block );

	return( 1 );
}

