/*
 * Descriptor data functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_data_array.h"
#include "libpff_data_reference.h"
#include "libpff_libbfio.h"

/* Initializes the data reference and its values
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_initialize(
     libpff_data_reference_t **data_reference,
     uint8_t type,
     uint8_t flags,
     intptr_t *data,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_initialize";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 type );

		return( -1 );
	}
	if( *data_reference == NULL )
	{
		*data_reference = (libpff_data_reference_t *) memory_allocate(
		                                               sizeof( libpff_data_reference_t ) );

		if( *data_reference == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create data reference.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *data_reference,
		     0,
		     sizeof( libpff_data_reference_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear data reference.",
			 function );

			memory_free(
			 *data_reference );

			return( -1 );
		}
		( *data_reference )->type  = type;
		( *data_reference )->flags = flags;
		( *data_reference )->data  = data;
	}
	return( 1 );
}

/* Initializes a buffer data reference and its values
 * The data is copied into the data reference as a buffer
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_initialize_buffer(
     libpff_data_reference_t **data_reference,
     uint8_t *buffer,
     size_t buffer_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_initialize";

	if( libpff_data_reference_initialize(
	     data_reference,
	     LIBPFF_DATA_REFERENCE_TYPE_BUFFER,
	     LIBPFF_DATA_REFERENCE_FLAG_MANAGED,
	     NULL,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create data reference.",
		 function );

		return( -1 );
	}
	if( buffer_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid data size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( ( buffer != NULL )
	 && ( buffer_size != 0 ) )
	{
		( *data_reference )->data = (intptr_t *) memory_allocate(
							  sizeof( uint8_t ) * buffer_size );

		if( ( *data_reference )->data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create data buffer.",
			 function );

			libpff_data_reference_free(
			 data_reference,
			 NULL );

			return( -1 );
		}
		if( memory_copy(
		     ( *data_reference )->data,
		     buffer,
		     buffer_size ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy data buffer.",
			 function );

			libpff_data_reference_free(
			 data_reference,
			 NULL );

			return( -1 );
		}
		( *data_reference )->buffer_size = buffer_size;
	}
	return( 1 );
}

/* Frees the array and its values
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_free(
     libpff_data_reference_t **data_reference,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_free";
	int result            = 1;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( *data_reference != NULL )
	{
		if( ( ( *data_reference )->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
		 && ( ( *data_reference )->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
			 function,
			 ( *data_reference )->type );

			return( -1 );
		}
		if( ( ( ( *data_reference )->flags & LIBPFF_DATA_REFERENCE_FLAG_MANAGED ) != 0 )
		 && ( ( *data_reference )->data != NULL ) )
		{
			if( ( *data_reference )->type == LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
			{
				memory_free(
				 ( *data_reference )->data );
			}
			else if( ( *data_reference )->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
			{
				if( libpff_data_array_free(
				     (libpff_data_array_t **) &( ( *data_reference )->data ),
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
					 "%s: unable to free data array.",
					 function );

					result = -1;
				}
			}
		}
		memory_free(
		 *data_reference );

		*data_reference = NULL;
	}
	return( result );
}

/* Frees the array and its values
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_free_as_value(
     intptr_t *data_reference,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_free_as_value";
	int result            = 1;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( ( (libpff_data_reference_t *) data_reference )->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( ( (libpff_data_reference_t *) data_reference )->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 ( (libpff_data_reference_t *) data_reference )->type );

		return( -1 );
	}
	if( ( ( ( (libpff_data_reference_t *) data_reference )->flags & LIBPFF_DATA_REFERENCE_FLAG_MANAGED ) != 0 )
	 && ( ( (libpff_data_reference_t *) data_reference )->data != NULL ) )
	{
		if( ( (libpff_data_reference_t *) data_reference )->type == LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
		{
			memory_free(
			 ( (libpff_data_reference_t *) data_reference )->data );
		}
		else if( ( (libpff_data_reference_t *) data_reference )->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
		{
			if( libpff_data_array_free(
			     (libpff_data_array_t **) &( ( (libpff_data_reference_t *) data_reference )->data ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free data array.",
				 function );

				result = -1;
			}
		}
	}
	memory_free(
	 data_reference );

	return( result );
}

/* Clones the source data reference to the destination
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_clone(
     libpff_data_reference_t **destination,
     libpff_data_reference_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_clone";

	if( source == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid source data reference.",
		 function );

		return( -1 );
	}
	if( libpff_data_reference_initialize(
	     destination,
	     source->type,
	     LIBPFF_DATA_REFERENCE_FLAG_NON_MANAGED,
	     source->data,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create destination data reference.",
		 function );

		return( -1 );
	}
	if( ( *destination )->type == LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	{
		( *destination )->buffer_size = source->buffer_size;
	}
	return( 1 );
}

/* Retrieves the flags
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_flags(
     libpff_data_reference_t *data_reference,
     uint8_t *flags,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_get_flags";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( flags == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid flags.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	 && ( ( data_reference->flags & LIBPFF_DATA_REFERENCE_FLAG_DECRYPTION_FORCED ) == 0 ) )
	{
		if( ( (libpff_data_array_t *) data_reference->data )->decryption_forced != 0 )
		{
			data_reference->flags |= LIBPFF_DATA_REFERENCE_FLAG_DECRYPTION_FORCED;
		}
	}
	*flags = data_reference->flags;

	return( 1 );
}

/* Retrieves the number of entries
 * The number of entries is 0 when there is no data
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_number_of_entries(
     libpff_data_reference_t *data_reference,
     uint32_t *number_of_entries,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_get_number_of_entries";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( number_of_entries == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid number of entries.",
		 function );

		return( -1 );
	}
	if( data_reference->data == NULL )
	{
		*number_of_entries = 0;
	}
	else if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		*number_of_entries = ( (libpff_data_array_t *) data_reference->data )->number_of_entries;
	}
	else
	{
		*number_of_entries = 1;
	}
	return( 1 );
}

/* Retrieves the entry data of a specific entry in the data reference
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_entry_data(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     uint32_t entry,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_get_entry_data";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( data_reference->data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data reference - missing data.",
		 function );

		return( -1 );
	}
	if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		if( libpff_data_array_get_entry_data(
		     (libpff_data_array_t *) data_reference->data,
		     file_io_handle,
		     entry,
		     entry_data,
		     entry_data_size,
		     flags,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve data of entry: %" PRIu32 ".",
			 function,
			 entry );

			return( -1 );
		}
	}
	else
	{
		if( entry != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: entry value out of range.",
			 function );

			return( -1 );
		}
		if( entry_data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid entry data.",
			 function );

			return( -1 );
		}
		if( entry_data_size == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid entry data size.",
			 function );

			return( -1 );
		}
		*entry_data      = (uint8_t *) data_reference->data;
		*entry_data_size = data_reference->buffer_size;
	}
	return( 1 );
}

/* Retrieves the entry data of the data reference at a certain offset
 * The data size set to the remaining size in the corresponding entry
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_entry_data_at_offset(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     size_t data_offset,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error )
{
	static char *function    = "libpff_data_reference_get_entry_data_at_offset";
	size_t entry_data_offset = 0;
	uint32_t entry           = 0;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		if( libpff_data_array_get_entry_at_offset(
		     (libpff_data_array_t *) data_reference->data,
		     data_offset,
		     &entry,
		     &entry_data_offset,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve data array entry.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( data_offset > (size_t) SSIZE_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: invalid data offset value exceeds maximum.",
			 function );

			return( -1 );
		}
		entry_data_offset = (size_t) data_offset;
	}
	if( libpff_data_reference_get_entry_data(
	     data_reference,
	     file_io_handle,
	     entry,
	     entry_data,
	     entry_data_size,
	     flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve data of entry: %" PRIu32 ".",
		 function,
		 entry );

		return( -1 );
	}
	if( *entry_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid entry data.",
		 function );

		return( -1 );
	}
	if( entry_data_offset >= *entry_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid entry data offset value exceeds entry data size.",
		 function );

		return( -1 );
	}
	*entry_data      = &( ( *entry_data )[ entry_data_offset ] );
	*entry_data_size = *entry_data_size - entry_data_offset;

	return( 1 );
}

/* Retrieves the entry data of the data reference at an entry aligned index
 * The data size set to the remaining size in the corresponding entry
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_entry_data_at_aligned_index(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     int value_index,
     size_t value_size,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error )
{
	static char *function    = "libpff_data_reference_get_entry_data_at_aligned_index";
	size_t entry_data_offset = 0;
	uint32_t entry           = 0;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		if( libpff_data_array_get_entry_at_aligned_index(
		     (libpff_data_array_t *) data_reference->data,
		     value_index,
		     value_size,
		     &entry,
		     &entry_data_offset,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve data array entry.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( value_index < 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: invalid value index value out of range.",
			 function );

			return( -1 );
		}
		if( value_size > (size_t) SSIZE_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: invalid value size value exceeds maximum.",
			 function );

			return( -1 );
		}
		entry_data_offset = value_size * (size_t) value_index;
	}
	if( libpff_data_reference_get_entry_data(
	     data_reference,
	     file_io_handle,
	     entry,
	     entry_data,
	     entry_data_size,
	     flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve data of entry: %" PRIu32 ".",
		 function,
		 entry );

		return( -1 );
	}
	if( entry_data_offset >= *entry_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid entry data offset value exceeds entry data size.",
		 function );

		return( -1 );
	}
	*entry_data      = &( ( *entry_data )[ entry_data_offset ] );
	*entry_data_size = *entry_data_size - entry_data_offset;

	return( 1 );
}

/* Retrieves the data size of the data reference
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_data_size(
     libpff_data_reference_t *data_reference,
     size64_t *data_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_get_data_size";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		if( libpff_data_array_get_data_size(
		     (libpff_data_array_t *) data_reference->data,
		     data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve data size.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( data_size == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid data size.",
			 function );

			return( -1 );
		}
		*data_size = (size64_t) data_reference->buffer_size;
	}
	return( 1 );
}

/* Sets the buffer size of the data reference
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_set_buffer_size(
     libpff_data_reference_t *data_reference,
     size_t buffer_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_set_buffer_size";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( buffer_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid buffer size value exceeds maximum.",
		 function );

		return( -1 );
	}
	data_reference->buffer_size = buffer_size;

	return( 1 );
}

/* Retrieves the data of the data reference
 * Returns 1 if successful or -1 on error
 */
int libpff_data_reference_get_data(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     uint8_t **data,
     size_t *data_size,
     liberror_error_t **error )
{
	static char *function = "libpff_data_reference_get_data";

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( data_reference->type == LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY )
	{
		if( libpff_data_array_get_data(
		     (libpff_data_array_t *) data_reference->data,
		     file_io_handle,
		     data,
		     data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve data.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid data.",
			 function );

			return( -1 );
		}
		if( data_size == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid data size.",
			 function );

			return( -1 );
		}
		*data      = (uint8_t *) data_reference->data;
		*data_size = data_reference->buffer_size;
	}
	return( 1 );
}

/* Reads data from the curent offset into a buffer
 * Returns the number of bytes read or -1 on error
 */
ssize_t libpff_data_reference_read_buffer(
         libpff_data_reference_t *data_reference,
         libbfio_handle_t *file_io_handle,
         uint8_t *buffer,
         size_t buffer_size,
         liberror_error_t **error )
{
	uint8_t *entry_data    = NULL;
	static char *function  = "libpff_data_reference_read_buffer";
	size64_t data_size     = 0;
	size_t entry_data_size = 0;
	size_t buffer_offset   = 0;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( buffer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid buffer.",
		 function );

		return( -1 );
	}
	if( buffer_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid buffer size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( libpff_data_reference_get_data_size(
	     data_reference,
	     &data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve data size.",
		 function );

		return( -1 );
	}
	if( (off64_t) ( data_reference->data_offset + buffer_size ) > (off64_t) data_size )
	{
		buffer_size = (size_t) ( data_size - data_reference->data_offset );
	}
	while( buffer_size > 0 )
	{
		if( libpff_data_reference_get_entry_data_at_offset(
		     data_reference,
		     file_io_handle,
		     (size_t) data_reference->data_offset,
		     &entry_data,
		     &entry_data_size,
		     0,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve entry data at offset: %" PRIi64 ".",
			 function,
			 data_reference->data_offset );

			return( -1 );
		}
		if( entry_data_size > buffer_size )
		{
			entry_data_size = buffer_size;
		}
		if( entry_data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid entry data.",
			 function );

			return( -1 );
		}
		if( memory_copy(
		     &( buffer[ buffer_offset ] ),
		     entry_data,
		     entry_data_size ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy entry data to buffer.",
			 function );

			return( -1 );
		}
		data_reference->data_offset += entry_data_size;
		buffer_size                 -= entry_data_size;
		buffer_offset               += entry_data_size;
	}
	return( (size_t) buffer_offset );
}

/* Seeks a certain offset of the data
 * Returns the offset if seek is successful or -1 on error
 */
off64_t libpff_data_reference_seek_offset(
         libpff_data_reference_t *data_reference,
         off64_t offset,
         int whence,
         liberror_error_t **error )
{
	static char *function = "libpff_data_reference_seek_offset";
	size64_t data_size    = 0;

	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference.",
		 function );

		return( -1 );
	}
	if( ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_BUFFER )
	 && ( data_reference->type != LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported data descriptor type: 0x%02" PRIx8 ".",
		 function,
		 data_reference->type );

		return( -1 );
	}
	if( ( whence != SEEK_CUR )
	 && ( whence != SEEK_END )
	 && ( whence != SEEK_SET ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported whence.",
		 function );

		return( -1 );
	}
	if( libpff_data_reference_get_data_size(
	     data_reference,
	     &data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve data size.",
		 function );

		return( -1 );
	}
	if( whence == SEEK_CUR )
	{	
		offset += data_reference->data_offset;
	}
	else if( whence == SEEK_END )
	{	
		offset += (off64_t) data_size;
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: seeking offset: %" PRIu64 ".\n",
		 function,
		 offset );
	}
#endif

	if( ( offset < 0 )
	 || ( offset > (off64_t) data_size ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: offset out of range.",
		 function );

		return( -1 );
	}
	data_reference->data_offset = offset;

	return( offset );
}

