/*
 * Descriptor data functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#if !defined( _LIBPFF_DATA_REFERENCE_H )
#define _LIBPFF_DATA_REFERENCE_H

#include <common.h>
#include <types.h>

#include <liberror.h>

#include "libpff_libbfio.h"

#if defined( __cplusplus )
extern "C" {
#endif

enum LIBPFF_DATA_REFERENCE_TYPES
{
	/* The data is a buffer
	 */
	LIBPFF_DATA_REFERENCE_TYPE_BUFFER				= (int) 'b',

	/* The data is a data array
	 */
	LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY				= (int) 'a'
};

enum LIBPFF_DATA_REFERENCE_FLAGS
{
	/* The data is not managed by the data reference
	 */
	LIBPFF_DATA_REFERENCE_FLAG_NON_MANAGED				= 0x00,

	/* The data is managed by the data reference
	 */
	LIBPFF_DATA_REFERENCE_FLAG_MANAGED				= 0x01,

	/* Decryption of the data was forced
	 */
	LIBPFF_DATA_REFERENCE_FLAG_DECRYPTION_FORCED			= 0x02
};

typedef struct libpff_data_reference libpff_data_reference_t;

struct libpff_data_reference
{
	/* The type
	 */
	uint8_t type;

	/* The flags
	 */
	uint8_t flags;

	/* The data
	 */
	intptr_t *data;

	/* The current data offset
	 */
	off64_t data_offset;

	/* The buffer size
	 */
	size_t buffer_size;
};

int libpff_data_reference_initialize(
     libpff_data_reference_t **data_reference,
     uint8_t type,
     uint8_t flags,
     intptr_t *data,
     liberror_error_t **error );

int libpff_data_reference_initialize_buffer(
     libpff_data_reference_t **data_reference,
     uint8_t *buffer,
     size_t buffer_size,
     liberror_error_t **error );

int libpff_data_reference_free(
     libpff_data_reference_t **data_reference,
     liberror_error_t **error );

int libpff_data_reference_free_as_value(
     intptr_t *data_reference,
     liberror_error_t **error );

int libpff_data_reference_clone(
     libpff_data_reference_t **destination,
     libpff_data_reference_t *source,
     liberror_error_t **error );

int libpff_data_reference_get_flags(
     libpff_data_reference_t *data_reference,
     uint8_t *flags,
     liberror_error_t **error );

int libpff_data_reference_get_number_of_entries(
     libpff_data_reference_t *data_reference,
     uint32_t *number_of_entries,
     liberror_error_t **error );

int libpff_data_reference_get_entry_data(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     uint32_t entry,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error );

int libpff_data_reference_get_entry_data_at_offset(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     size_t data_offset,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error );

int libpff_data_reference_get_entry_data_at_aligned_index(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     int value_index,
     size_t value_size,
     uint8_t **entry_data,
     size_t *entry_data_size,
     uint8_t flags,
     liberror_error_t **error );

int libpff_data_reference_get_data_size(
     libpff_data_reference_t *data_reference,
     size64_t *data_size,
     liberror_error_t **error );

int libpff_data_reference_set_buffer_size(
     libpff_data_reference_t *data_reference,
     size_t buffer_size,
     liberror_error_t **error );

int libpff_data_reference_get_data(
     libpff_data_reference_t *data_reference,
     libbfio_handle_t *file_io_handle,
     uint8_t **data,
     size_t *data_size,
     liberror_error_t **error );

ssize_t libpff_data_reference_read_buffer(
         libpff_data_reference_t *data_reference,
         libbfio_handle_t *file_io_handle,
         uint8_t *buffer,
         size_t buffer_size,
         liberror_error_t **error );

off64_t libpff_data_reference_seek_offset(
         libpff_data_reference_t *data_reference,
         off64_t offset,
         int whence,
         liberror_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif

