/*
 * libpff descriptor
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_data_array.h"
#include "libpff_data_block.h"
#include "libpff_data_reference.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_encryption.h"
#include "libpff_libbfio.h"
#include "libpff_index.h"
#include "libpff_io_handle.h"
#include "libpff_local_descriptors.h"
#include "libpff_tree_type.h"

#include "pff_array.h"

/* Reads a descriptor
 * Returns 1 if successful, 0 if not available or -1 on error
 */
int libpff_descriptor_read(
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t **local_descriptors,
     libpff_data_reference_t **descriptor_data_reference,
     liberror_error_t **error )
{
	libpff_index_values_t *descriptor_index_values = NULL;
	libpff_tree_node_t *descriptor_index_root_node = NULL;
	static char *function                          = "libpff_descriptor_read";
	int result                                     = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( recovered == 0 )
	{
		descriptor_index_root_node = io_handle->descriptor_index_root_node;
	}
	else
	{
		descriptor_index_root_node = io_handle->recovered_descriptor_index_root_node;
	}
	if( descriptor_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid descriptor index root node.",
		 function );

		return( -1 );
	}
	result = libpff_index_get_index_values(
	          descriptor_index_root_node,
	          descriptor_identifier,
	          &descriptor_index_values,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: error while searching to find descriptor identifier: %" PRIu32 ".",
		 function,
		 descriptor_identifier );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( descriptor_index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid descriptor index values.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: identifier: %" PRIu64 " ",
		 function,
		 descriptor_index_values->identifier );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( descriptor_index_values->identifier & 0x0000001fUL ) );
		libnotify_printf(
		 ", data: %" PRIu64 ", local descriptors: %" PRIu64 ", parent: %" PRIu32 "\n",
		 descriptor_index_values->data_identifier,
		 descriptor_index_values->local_descriptors_identifier,
		 descriptor_index_values->parent_identifier );
	}
#endif

	if( descriptor_index_values->data_identifier == 0 )
	{
		return( 0 );
	}
	/* Read the local descriptors
	 */
	if( descriptor_index_values->local_descriptors_identifier != 0 )
	{
		if( local_descriptors == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid local descriptors.",
			 function );

			return( -1 );
		}
		if( libpff_descriptor_read_local_descriptors(
        	     descriptor_index_values->local_descriptors_identifier,
		     recovered,
		     io_handle,
		     file_io_handle,
	             local_descriptors,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read local descriptors.",
			 function );

			return( -1 );
		}
	}
	/* Read the data
	 */
	if( libpff_descriptor_read_data(
             descriptor_index_values->data_identifier,
	     descriptor_identifier,
	     recovered,
	     io_handle,
	     file_io_handle,
             descriptor_data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read descriptor data.",
		 function );

		if( ( local_descriptors != NULL )
		 && ( *local_descriptors != NULL ) )
		{
			libpff_tree_node_free(
			 local_descriptors,
			 &libpff_local_descriptor_values_free,
		         NULL );

			*local_descriptors = NULL;
		}
		return( -1 );
	}
	return( 1 );
}

/* Reads a local descriptor
 * Returns 1 if successful, 0 if not available or -1 on error
 */
int libpff_descriptor_local_read(
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_tree_node_t *parent_local_descriptors,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t **local_descriptors,
     libpff_data_reference_t **descriptor_data_reference,
     liberror_error_t **error )
{
	libpff_local_descriptor_values_t *local_descriptor_values = NULL;
	static char *function                                     = "libpff_descriptor_local_read";
	int result                                                = 0;

	if( parent_local_descriptors == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid parent local descriptors.",
		 function );

		return( -1 );
	}
	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	result = libpff_local_descriptors_get_values(
		  parent_local_descriptors,
		  descriptor_identifier,
		  &local_descriptor_values,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: error while searching for descriptor identifier: %" PRIu32 " in local descriptors.",
		 function,
		 descriptor_identifier );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( local_descriptor_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid local descriptor values.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: identifier: %" PRIu64 " ",
		 function,
		 local_descriptor_values->identifier );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( local_descriptor_values->identifier & 0x0000001fUL ) );
		libnotify_printf(
		 ", data: %" PRIu64 ", sub node: %" PRIu64 "\n",
		 local_descriptor_values->data_identifier,
		 local_descriptor_values->sub_node_identifier );
	}
#endif

	/* Read the local descriptors hierarchy
	 */
	if( local_descriptor_values->sub_node_identifier != 0 )
	{
		if( local_descriptors == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid local descriptors sub node.",
			 function );

			return( -1 );
		}
		if( libpff_descriptor_read_local_descriptors(
        	     local_descriptor_values->sub_node_identifier,
		     recovered,
		     io_handle,
		     file_io_handle,
	             local_descriptors,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read local descriptors sub node.",
			 function );

			return( -1 );
		}
	}
	if( local_descriptor_values->data_identifier == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid descriptor index values - missing data identifier.",
		 function );

		return( -1 );
	}
	/* Read the data
	 */
	if( libpff_descriptor_read_data(
	     local_descriptor_values->data_identifier,
	     descriptor_identifier,
	     recovered,
	     io_handle,
	     file_io_handle,
	     descriptor_data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read descriptor data.",
		 function );

		if( ( local_descriptors != NULL )
		 && ( *local_descriptors != NULL ) )
		{
			libpff_tree_node_free(
			 local_descriptors,
			 &libpff_local_descriptor_values_free,
			 NULL );

			*local_descriptors = NULL;
		}
		return( -1 );
	}
	return( 1 );
}

/* Reads the local descriptors
 * Returns 1 if successful or -1 on error
 */
int libpff_descriptor_read_local_descriptors(
     uint64_t local_descriptors_identifier,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t **local_descriptors,
     liberror_error_t **error )
{
	static char *function = "libpff_descriptor_read_local_descriptors";

	if( local_descriptors == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors.",
		 function );

		return( -1 );
	}
	if( *local_descriptors != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: local descriptors already set.",
		 function );

		return( -1 );
	}
	/* Create the local descriptors root node
	 */
	if( libpff_local_descriptors_node_initialize(
	     local_descriptors,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create local descriptors root node.",
		 function );

		return( -1 );
	}
	if( libpff_local_descriptors_read_tree(
	     *local_descriptors,
	     local_descriptors_identifier,
	     recovered,
	     io_handle,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read local descriptors with identifier: %" PRIu64 ".",
		 function,
		 local_descriptors_identifier );

		libpff_tree_node_free(
		 local_descriptors,
		 &libpff_local_descriptor_values_free,
		 NULL );

		*local_descriptors = NULL;

		return( -1 );
	}
	return( 1 );
}

/* Reads a descriptor data
 * Returns 1 if successful or -1 on error
 */
int libpff_descriptor_read_data(
     uint64_t data_identifier,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_data_reference_t **descriptor_data_reference,
     liberror_error_t **error )
{
	libpff_data_array_t *data_array            = NULL;
	libpff_index_values_t *offset_index_values = NULL;
	libpff_tree_node_t *offset_index_root_node = NULL;
	uint8_t *data                              = NULL;
	static char *function                      = "libpff_descriptor_read_data";
	size_t data_array_size                     = 0;
	size_t data_size                           = 0;
	ssize_t read_count                         = 0;
	int encryption_type                        = 0;
	uint8_t data_reference_flags               = LIBPFF_DATA_REFERENCE_FLAG_MANAGED;
	uint8_t decrypt_data                       = 0;
	uint8_t node_contains_table                = 0;
	uint8_t node_identifier_type               = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( descriptor_data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid descriptor data reference.",
		 function );

		return( -1 );
	}
	if( *descriptor_data_reference != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: descriptor data already set.",
		 function );

		return( -1 );
	}
	if( recovered == 0 )
	{
		offset_index_root_node = io_handle->offset_index_root_node;
	}
	else
	{
		offset_index_root_node = io_handle->recovered_offset_index_root_node;
	}
	if( offset_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid offset index root node.",
		 function );

		return( -1 );
	}
	if( libpff_index_get_index_values(
	     offset_index_root_node,
	     data_identifier & (uint64_t) LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK,
	     &offset_index_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to find data identifier: %" PRIu64 ".",
		 function,
		 data_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK );

		return( -1 );
	}
	if( offset_index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid offset index values.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: identifier: %" PRIu64 " (%s) at offset: %" PRIi64 " of size: %" PRIu32 "\n",
		 function,
		 offset_index_values->identifier,
		 ( ( offset_index_values->identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) ? "internal" : "external" ),
		 offset_index_values->file_offset,
		 offset_index_values->data_size );
	}
#endif

	if( offset_index_values->file_offset < 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid file offset value out of range.",
		 function );

		return( -1 );
	}
	if( offset_index_values->data_size > (size32_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: data size value exceeds maximum.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading descriptor data at offset: %" PRIu64 " (0x%08" PRIx64 ")\n",
		 function,
		 offset_index_values->file_offset,
		 offset_index_values->file_offset );
	}
#endif

#if defined( HAVE_DEBUG_OUTPUT )
	if( libpff_data_block_read(
	     file_io_handle,
	     offset_index_values->file_offset,
	     offset_index_values->data_size,
	     io_handle->file_type,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read data block.",
		 function );

		return( -1 );
	}
#endif
	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     offset_index_values->file_offset,
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek value offset: %" PRIu64 ".",
		 function,
		 offset_index_values->file_offset );

		return( -1 );
	}
	data = (uint8_t *) memory_allocate(
	                    sizeof( uint8_t ) * offset_index_values->data_size );

	if( data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create data.",
		 function );

		return( -1 );
	}
	read_count = libbfio_handle_read(
	              file_io_handle,
	              data,
	              offset_index_values->data_size,
	              error );

	if( read_count != (ssize_t) offset_index_values->data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read data.",
		 function );

		memory_free(
		 data );

		return( -1 );
	}
	data_size = (size_t) offset_index_values->data_size;

	/* Check if the descriptor data contains an array
	 * The array should have the internal flag set in the (data) offset index identifier
	 * The array starts with 0x01 followed by 0x01 or 0x02
	 */
	/* TODO: Can the flag be trusted in corrupt files ? */
	if( ( ( data_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) != 0 )
	 && ( ( data[ 0 ] == 0x01 )
	  && ( ( data[ 1 ] == 0x01 )
	   || ( data[ 1 ] == 0x02 ) ) ) )
	{
		if( libpff_descriptor_read_data_array(
		     descriptor_identifier,
		     recovered,
		     io_handle,
		     file_io_handle,
		     data,
		     data_size,
		     &data_array,
		     &data_array_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read array descriptor data.",
			 function );

			memory_free(
			 data );

			return( -1 );
		}
		memory_free(
		 data );

		if( data_array == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid data array.",
			 function );

			return( -1 );
		}
		/* Embed the data array in data reference structure
		 */
		if( libpff_data_reference_initialize(
		     descriptor_data_reference,
		     LIBPFF_DATA_REFERENCE_TYPE_DATA_ARRAY,
		     data_reference_flags,
		     (intptr_t *) data_array,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create data array.",
			 function );

			libpff_data_array_free(
			 &data_array,
			 NULL );

			return( -1 );
		}
		if( *descriptor_data_reference == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid descriptor data.",
			 function );

			libpff_data_array_free(
			 &data_array,
			 NULL );

			return( -1 );
		}
	}
	else
	{
		encryption_type = (int) io_handle->encryption_type;

		/* Check if internal (unencrypted) flag in the (data) offset index identifier is not set
		 */
		if( ( data_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) == 0 )
		{
			decrypt_data = 1;
		}
		/* Check if data is encrypted
		 * some 'invalid' PFF's have an encryption type of none
		 * but do actually contain encrypted data
		 * although they are considered invalid by Outlook
		 * it is still useful to be able to read them
		 */
		if( ( io_handle->encryption_type == LIBPFF_ENCRYPTION_TYPE_NONE )
		 && ( data_size > 4 ) )
		{
			node_identifier_type = (uint8_t) ( descriptor_identifier & 0x0000001fUL );

			/* Check if the node identifier type consists of a table
			 */
			if( ( ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_INTERNAL )
			  && ( ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_MESSAGE_STORE )
			   || ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_NAME_TO_ID_MAP )
			   || ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2049 )
			   || ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2081 )
			   || ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_2113 )
			   || ( descriptor_identifier == LIBPFF_DESCRIPTOR_IDENTIFIER_UNKNOWN_3073 ) ) )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_FOLDER )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SEARCH_FOLDER )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_MESSAGE )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_ASSOCIATED_CONTENT )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_FOLDERS )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_MESSAGES )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SUB_ASSOCIATED_CONTENTS )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_SEARCH_CONTENTS_TABLE )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_ATTACHMENTS )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_RECIPIENTS )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1718 )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1751 )
			 || ( node_identifier_type == LIBPFF_NODE_IDENTIFIER_TYPE_UNKNOWN_1784 ) )
			{
				node_contains_table = 1;
			}
			if( node_contains_table != 0 )
			{
				/* Test if the data contains an unencrypted table
				 * a table consists of 0xec in the third byte
				 * and 0x6c, 0x7c, 0x8c, 0x9c, 0xa5, 0xac, 0xbc, 0xcc in the fourth
				 */
				if( ( data[ 2 ] != 0xec )
				 || ( ( data[ 3 ] != 0x6c )
				  && ( data[ 3 ] != 0x7c )
				  && ( data[ 3 ] != 0x8c )
				  && ( data[ 3 ] != 0x9c )
				  && ( data[ 3 ] != 0xa5 )
				  && ( data[ 3 ] != 0xac )
				  && ( data[ 3 ] != 0xbc )
				  && ( data[ 3 ] != 0xcc ) ) )
				{
#if defined( HAVE_VERBOSE_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: compressible encrypted data detected while encryption type is none - decryption forced.\n",
						 function );
					}
#endif
					encryption_type             = LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE;
					decrypt_data                = 1;
					io_handle->force_decryption = 1;
					data_reference_flags       |= LIBPFF_DATA_REFERENCE_FLAG_DECRYPTION_FORCED;
				}
			}
			else if( io_handle->force_decryption == 1 )
			{
				/* Check if the internal (unencrypted) flag in (data) offset index identifier is not set
				 */
				if( ( data_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) == 0 )
				{
#if defined( HAVE_VERBOSE_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: decryption forced.\n",
						 function );
					}
#endif

					encryption_type       = LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE;
					decrypt_data          = 1;
					data_reference_flags |= LIBPFF_DATA_REFERENCE_FLAG_DECRYPTION_FORCED;
				}
			}
		}
		if( decrypt_data != 0 )
		{
			if( libpff_encryption_decrypt(
			     encryption_type,
			     (uint32_t) data_identifier,
			     data,
			     data_size,
			     error ) != (ssize_t) data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_ENCRYPTION,
				 LIBERROR_ENCRYPTION_ERROR_DECRYPT_FAILED,
				 "%s: unable to decrypt entry value data.",
				 function );

				memory_free(
				 data );

				return( -1 );
			}
		}
		/* Embed the data buffer in data reference structure
		 */
		if( libpff_data_reference_initialize(
		     descriptor_data_reference,
		     LIBPFF_DATA_REFERENCE_TYPE_BUFFER,
		     data_reference_flags,
		     (intptr_t *) data,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create descriptor data reference.",
			 function );

			memory_free(
			 data );

			return( -1 );
		}
		if( libpff_data_reference_set_buffer_size(
		     *descriptor_data_reference,
		     data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to set size of descriptor data reference.",
			 function );

			libpff_data_reference_free(
			 descriptor_data_reference,
			 NULL );

			return( -1 );
		}
	}
	return( 1 );
}

/* Reads descriptor data array
 * Returns 1 if successful or -1 on error
 */
int libpff_descriptor_read_data_array(
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     uint8_t *array_data,
     size_t array_data_size,
     libpff_data_array_t **data_array,
     size_t *data_array_size,
     liberror_error_t **error )
{
	libpff_index_values_t *offset_index_values = NULL;
	libpff_tree_node_t *offset_index_root_node = NULL;
	uint8_t *sub_array_data                    = NULL;
	static char *function                      = "libpff_descriptor_read_data_array";
	size_t calculated_data_array_size          = 0;
	size_t sub_array_data_size                 = 0;
	size_t sub_data_array_size                 = 0;
	ssize_t read_count                         = 0;
	uint64_t entry_identifier                  = 0;
	uint32_t previous_number_of_array_entries  = 0;
	uint16_t number_of_array_entries           = 0;
	uint16_t entry_iterator                    = 0;
	uint8_t array_entries_level                = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( ( io_handle->file_type != LIBPFF_FILE_TYPE_32BIT )
	 && ( io_handle->file_type != LIBPFF_FILE_TYPE_64BIT ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: invalid file type.",
		 function );

		return( -1 );
	}
	if( array_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid array data.",
		 function );

		return( -1 );
	}
	if( array_data_size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid array data size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( data_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array.",
		 function );

		return( -1 );
	}
	if( data_array_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data array size.",
		 function );

		return( -1 );
	}
	if( recovered == 0 )
	{
		offset_index_root_node = io_handle->offset_index_root_node;
	}
	else
	{
		offset_index_root_node = io_handle->recovered_offset_index_root_node;
	}
	if( offset_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid offset index root node.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: array:\n",
		 function );
		libnotify_print_data(
		 array_data,
		 array_data_size );
	}
#endif

	if( ( (pff_array_t *) array_data )->signature != 0x01 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported array signature.",
		 function );

		return( -1 );
	}
	array_entries_level = ( (pff_array_t *) array_data )->array_entries_level;

	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_array_t *) array_data )->number_of_entries,
	 number_of_array_entries );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_array_t *) array_data )->total_data_size,
	 *data_array_size );

	array_data      += sizeof( pff_array_t );
	array_data_size -= sizeof( pff_array_t );

#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: array entries level\t\t: %" PRIu8 "\n",
		 function,
		 array_entries_level );
		libnotify_printf(
		 "%s: array number of entries\t: %" PRIu16 "\n",
		 function,
		 number_of_array_entries );
		libnotify_printf(
		 "%s: array total data size\t: %" PRIu32 "\n",
		 function,
		 *data_array_size );
		libnotify_printf(
		 "\n" );
	}
#endif

	if( array_entries_level == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported array entries level: %" PRIu8 ".",
		 function,
		 array_entries_level );

		return( -1 );
	}
	if( *data_array == NULL )
	{
		/* Create the data array
		 */
		if( libpff_data_array_initialize(
		     data_array,
		     descriptor_identifier,
		     io_handle,
		     *data_array_size,
		     0,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create data array.",
			 function );

			return( -1 );
		}
	}
	previous_number_of_array_entries = ( *data_array )->number_of_entries;

	if( array_entries_level == 1 )
	{
		/* Resize the data array
		 */
		if( libpff_data_array_resize(
		     *data_array,
		     previous_number_of_array_entries + number_of_array_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize data array.",
			 function );

			libpff_data_array_free(
			 data_array,
			 NULL );

			return( -1 );
		}
	}
	for( entry_iterator = 0;
	     entry_iterator < number_of_array_entries;
	     entry_iterator++ )
	{
		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 array_data,
			 entry_identifier );

			array_data += 4;
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			byte_stream_copy_to_uint64_little_endian(
			 array_data,
			 entry_identifier );

			array_data += 8;
		}
		offset_index_values = NULL;

		/* Get the offset index entry values for the identifier
		 */
		if( libpff_index_get_index_values(
		     offset_index_root_node,
		     entry_identifier & (uint64_t) LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK,
		     &offset_index_values,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to find data identifier: %" PRIu64 ".",
			 function,
			 entry_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK );

			libpff_data_array_free(
			 data_array,
			 NULL );

			return( -1 );
		}
		if( offset_index_values == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid offset index values.",
			 function );

			libpff_data_array_free(
			 data_array,
			 NULL );

			return( -1 );
		}
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: array entry: %03" PRIu16 " at level: %" PRIu8 " identifier: %" PRIu64 " (%s) at offset: %" PRIi64 " of size: %" PRIu32 "\n",
			 function,
			 entry_iterator,
			 array_entries_level,
			 offset_index_values->identifier,
			 ( ( offset_index_values->identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) ? "internal" : "external" ),
			 offset_index_values->file_offset,
			 offset_index_values->data_size );
		}
#endif

		if( offset_index_values->file_offset < 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: invalid file offset value out of range.",
			 function );

			libpff_data_array_free(
			 data_array,
			 NULL );

			return( -1 );
		}
		if( offset_index_values->data_size > (size32_t) SSIZE_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: data size value exceeds maximum.",
			 function );

			libpff_data_array_free(
			 data_array,
			 NULL );

			return( -1 );
		}
		if( array_entries_level == 1 )
		{
			if( ( calculated_data_array_size + offset_index_values->data_size ) > ( *data_array )->data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: data size: %" PRIzd " exceeds total data size: %" PRIzd ".",
				 function,
				 calculated_data_array_size + offset_index_values->data_size,
				 ( *data_array )->data_size );

				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			if( libpff_data_array_set_entry(
			     *data_array,
			     previous_number_of_array_entries + entry_iterator,
			     offset_index_values->identifier,
			     (off64_t) offset_index_values->file_offset,
			     (size_t) offset_index_values->data_size,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_SET_FAILED,
				 "%s: unable to set data array entry: %" PRIu16 ".",
				 function,
				 entry_iterator );

				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			calculated_data_array_size += offset_index_values->data_size;
		}
		else
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: reading array descriptor data at offset: %" PRIu64 " (0x%08" PRIx64 ")\n",
				 function,
				 offset_index_values->file_offset,
				 offset_index_values->file_offset );
			}
#endif

			if( libbfio_handle_seek_offset(
			     file_io_handle,
			     offset_index_values->file_offset,
			     SEEK_SET,
			     error ) == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_SEEK_FAILED,
				 "%s: unable to seek value offset: %" PRIu64 ".",
				 function,
				 offset_index_values->file_offset );

				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			sub_array_data = (uint8_t *) memory_allocate(
			                              (size_t) offset_index_values->data_size );

			if( sub_array_data == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create sub array data.",
				 function );

				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			read_count = libbfio_handle_read(
				      file_io_handle,
				      sub_array_data,
				      offset_index_values->data_size,
				      error );

			if( read_count != (ssize_t) offset_index_values->data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to data.",
				 function );

				memory_free(
				 sub_array_data );
				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			sub_array_data_size = (size_t) offset_index_values->data_size;

			if( libpff_descriptor_read_data_array(
			     descriptor_identifier,
			     recovered,
			     io_handle,
			     file_io_handle,
			     sub_array_data,
			     sub_array_data_size,
			     data_array,
			     &sub_data_array_size,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read sub data array.",
				 function );

				memory_free(
				 sub_array_data );
				libpff_data_array_free(
				 data_array,
				 NULL );

				return( -1 );
			}
			memory_free(
			 sub_array_data );

			calculated_data_array_size += sub_data_array_size;
		}
	}
	if( calculated_data_array_size != (size_t) *data_array_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: mismatch in data size (%" PRIu32 " != %" PRIzd ").",
		 function,
		 *data_array_size,
		 calculated_data_array_size );

		libpff_data_array_free(
		 data_array,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

