/*
 * File functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <libcstring.h>
#include <liberror.h>
#include <libnotify.h>

#include "libpff_codepage.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_io_handle.h"
#include "libpff_item.h"
#include "libpff_item_descriptor.h"
#include "libpff_file.h"
#include "libpff_folder.h"
#include "libpff_index.h"
#include "libpff_item_tree.h"
#include "libpff_libbfio.h"
#include "libpff_name_to_id_map.h"
#include "libpff_offset_list.h"
#include "libpff_types.h"

/* Initialize a file
 * Make sure the value file is pointing to is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libpff_file_initialize(
     libpff_file_t **file,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_initialize";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	if( *file == NULL )
	{
		internal_file = (libpff_internal_file_t *) memory_allocate(
		                                            sizeof( libpff_internal_file_t ) );

		if( internal_file == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create file.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     internal_file,
		     0,
		     sizeof( libpff_internal_file_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear file.",
			 function );

			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_list_initialize(
		     &( internal_file->orphan_item_list ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create orphan item list.",
			 function );

			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_list_initialize(
		     &( internal_file->recovered_item_list ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create recovered item list.",
			 function );

			libpff_list_free(
			 &( internal_file->orphan_item_list ),
			 NULL,
			 NULL );
			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_list_initialize(
		     &( internal_file->unallocated_data_block_list ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create unallocated data block list.",
			 function );

			libpff_list_free(
			 &( internal_file->recovered_item_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->orphan_item_list ),
			 NULL,
			 NULL );
			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_list_initialize(
		     &( internal_file->unallocated_page_block_list ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create unallocated page block list.",
			 function );

			libpff_list_free(
			 &( internal_file->unallocated_data_block_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->recovered_item_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->orphan_item_list ),
			 NULL,
			 NULL );
			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_list_initialize(
		     &( internal_file->name_to_id_map_list ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create name to id map list.",
			 function );

			libpff_list_free(
			 &( internal_file->unallocated_page_block_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->unallocated_data_block_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->recovered_item_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->orphan_item_list ),
			 NULL,
			 NULL );
			memory_free(
			 internal_file );

			return( -1 );
		}
		if( libpff_io_handle_initialize(
		     &( internal_file->io_handle ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create io handle.",
			 function );

			libpff_list_free(
			 &( internal_file->name_to_id_map_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->unallocated_page_block_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->unallocated_data_block_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->recovered_item_list ),
			 NULL,
			 NULL );
			libpff_list_free(
			 &( internal_file->orphan_item_list ),
			 NULL,
			 NULL );
			memory_free(
			 internal_file );

			return( -1 );
		}
		internal_file->ascii_codepage = LIBPFF_CODEPAGE_WINDOWS_1252;

		*file = (libpff_file_t *) internal_file;
	}
	return( 1 );
}

/* Frees an exisisting file
 * Returns 1 if successful or -1 on error
 */
int libpff_file_free(
     libpff_file_t **file,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_free";
	int result                            = 1;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	if( *file != NULL )
	{
		internal_file = (libpff_internal_file_t *) *file;

		if( internal_file->item_tree_root_node != NULL )
		{
			if( libpff_tree_node_free(
			     &( internal_file->item_tree_root_node ),
			     &libpff_item_descriptor_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free item tree.",
				 function );

				result = -1;
			}
		}
		if( libpff_list_free(
		     &( internal_file->orphan_item_list ),
		     &libpff_item_tree_node_free_recovered,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free orphan item list.",
			 function );

			result = -1;
		}
		if( libpff_list_free(
		     &( internal_file->recovered_item_list ),
		     &libpff_item_tree_node_free_recovered,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free recovered item list.",
			 function );

			result = -1;
		}
		if( libpff_list_free(
		     &( internal_file->unallocated_data_block_list ),
		     &libpff_offset_list_values_free,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free unallocated data block list.",
			 function );

			result = -1;
		}
		if( libpff_list_free(
		     &( internal_file->unallocated_page_block_list ),
		     &libpff_offset_list_values_free,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free unallocated page block list.",
			 function );

			result = -1;
		}
		if( libpff_list_free(
		     &( internal_file->name_to_id_map_list ),
		     &libpff_name_to_id_map_entry_free,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free name to id map list.",
			 function );

			result = -1;
		}
		if( libpff_io_handle_free(
		     &( internal_file->io_handle ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free io handle.",
			 function );

			result = -1;
		}
		if( ( internal_file->file_io_handle_created_in_library != 0 )
		 && ( internal_file->file_io_handle != NULL ) )
		{
			if( libbfio_handle_free(
			     &( internal_file->file_io_handle ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free file io handle.",
				 function );

				result = -1;
			}
		}
		memory_free(
		 internal_file );

		*file = NULL;
	}
	return( result );
}

/* Signals the libpff file to abort its current activity
 * Returns 1 if successful or -1 on error
 */
int libpff_file_signal_abort(
     libpff_file_t *file,
     liberror_error_t **error )
{
	static char *function = "libpff_file_signal_abort";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	( (libpff_internal_file_t *) file )->abort = 1;

	return( 1 );
}

/* Opens a Personal Folder File
 * Returns 1 if successful or -1 on error
 */
int libpff_file_open(
     libpff_file_t *file,
     const char *filename,
     int flags,
     liberror_error_t **error )
{
	libbfio_handle_t *file_io_handle      = NULL;
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_open";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( filename == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid filename.",
		 function );

		return( -1 );
	}
	if( libbfio_file_initialize(
	     &file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create file io handle.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libbfio_handle_set_track_offsets_read(
	     file_io_handle,
	     1,
	     error ) != 1 )
	{
                liberror_error_set(
                 error,
                 LIBERROR_ERROR_DOMAIN_RUNTIME,
                 LIBERROR_RUNTIME_ERROR_SET_FAILED,
                 "%s: unable to set track offsets read in file io handle.",
                 function );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

                return( -1 );
	}
#endif
	if( libbfio_file_set_name(
	     file_io_handle,
	     filename,
	     libcstring_narrow_string_length(
	      filename ) + 1,
	     error ) != 1 )
	{
                liberror_error_set(
                 error,
                 LIBERROR_ERROR_DOMAIN_RUNTIME,
                 LIBERROR_RUNTIME_ERROR_SET_FAILED,
                 "%s: unable to set filename in file io handle.",
                 function );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

                return( -1 );
	}
	if( libpff_file_open_file_io_handle(
	     file,
	     file_io_handle,
	     flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_OPEN_FAILED,
		 "%s: unable to open file: %s.",
		 function,
		 filename );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

		return( -1 );
	}
	internal_file->file_io_handle_created_in_library = 1;

	return( 1 );
}

#if defined( HAVE_WIDE_CHARACTER_TYPE )

/* Opens a Personal Folder File
 * Returns 1 if successful or -1 on error
 */
int libpff_file_open_wide(
     libpff_file_t *file,
     const wchar_t *filename,
     int flags,
     liberror_error_t **error )
{
	libbfio_handle_t *file_io_handle      = NULL;
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_open_wide";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( filename == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid filename.",
		 function );

		return( -1 );
	}
	if( libbfio_file_initialize(
	     &file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create file io handle.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libbfio_handle_set_track_offsets_read(
	     file_io_handle,
	     1,
	     error ) != 1 )
	{
                liberror_error_set(
                 error,
                 LIBERROR_ERROR_DOMAIN_RUNTIME,
                 LIBERROR_RUNTIME_ERROR_SET_FAILED,
                 "%s: unable to set track offsets read in file io handle.",
                 function );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

                return( -1 );
	}
#endif
	if( libbfio_file_set_name_wide(
	     file_io_handle,
	     filename,
	     libcstring_wide_string_length(
	      filename ) + 1,
	     error ) != 1 )
	{
                liberror_error_set(
                 error,
                 LIBERROR_ERROR_DOMAIN_RUNTIME,
                 LIBERROR_RUNTIME_ERROR_SET_FAILED,
                 "%s: unable to set filename in file io handle.",
                 function );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

                return( -1 );
	}
	if( libpff_file_open_file_io_handle(
	     file,
	     file_io_handle,
	     flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_OPEN_FAILED,
		 "%s: unable to open file: %ls.",
		 function,
		 filename );

		libbfio_handle_free(
		 &file_io_handle,
		 NULL );

		return( -1 );
	}
	internal_file->file_io_handle_created_in_library = 1;

	return( 1 );
}

#endif

/* Opens a Personal Folder File using a Basic File IO (bfio) handle
 * Returns 1 if successful or -1 on error
 */
int libpff_file_open_file_io_handle(
     libpff_file_t *file,
     libbfio_handle_t *file_io_handle,
     int flags,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_open_file_io_handle";
	int file_io_flags                     = 0;
	int file_io_handle_is_open            = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->file_io_handle != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid internal file - file io handle already set.",
		 function );

		return( -1 );
	}
	if( file_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file io handle.",
		 function );

		return( -1 );
	}
	if( ( ( flags & LIBPFF_FLAG_READ ) != LIBPFF_FLAG_READ )
	 && ( ( flags & LIBPFF_FLAG_WRITE ) != LIBPFF_FLAG_WRITE ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported flags.",
		 function );

		return( -1 );
	}
	if( ( flags & LIBPFF_FLAG_WRITE ) == LIBPFF_FLAG_WRITE )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: write access to personal folder files currently not supported.",
		 function );

		return( -1 );
	}
	if( ( flags & LIBPFF_FLAG_READ ) == LIBPFF_FLAG_READ )
	{
		file_io_flags = LIBBFIO_FLAG_READ;
	}
	internal_file->file_io_handle = file_io_handle;

	file_io_handle_is_open = libbfio_handle_is_open(
	                          internal_file->file_io_handle,
	                          error );

	if( file_io_handle_is_open == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_OPEN_FAILED,
		 "%s: unable to open file.",
		 function );

		return( -1 );
	}
	else if( file_io_handle_is_open == 0 )
	{
		if( libbfio_handle_open(
		     internal_file->file_io_handle,
		     flags,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_OPEN_FAILED,
			 "%s: unable to open file io handle.",
			 function );

			return( -1 );
		}
	}
	if( libpff_file_open_read(
	     internal_file,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read from file handle.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Closes a Personal Folder File
 * Returns 0 if successful or -1 on error
 */
int libpff_file_close(
     libpff_file_t *file,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_close";
	int result                            = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->file_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing file IO handle.",
		 function );

		return( -1 );
	}
	if( internal_file->file_io_handle_created_in_library != 0 )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			if( libpff_debug_print_read_offsets(
			     internal_file->file_io_handle,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_PRINT_FAILED,
				 "%s: unable to print the read offsets.",
				 function );

				result = -1;
			}
		}
#endif
		if( libbfio_handle_close(
		     internal_file->file_io_handle,
		     error ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_CLOSE_FAILED,
			 "%s: unable to close file io handle.",
			 function );

			return( -1 );
		}
	}
	return( result );
}

/* Opens a Personal Folder File for reading
 * Returns 1 if successful or -1 on error
 */
int libpff_file_open_read(
     libpff_internal_file_t *internal_file,
     liberror_error_t **error )
{
	static char *function                     = "libpff_file_open_read";
	off64_t offset_index_root_node_offset     = 0;
	off64_t descriptor_index_root_node_offset = 0;
	uint64_t offset_index_back_pointer        = 0;
	uint64_t descriptor_index_back_pointer    = 0;
	int result_read_offset_index              = 0;
	int result_read_descriptor_index          = 0;

	if( internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid internal file.",
		 function );

		return( -1 );
	}
	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal file - missing io handle.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "Reading file header:\n" );
	}
#endif
	if( libpff_io_handle_read_file_header(
	     internal_file->io_handle,
	     internal_file->file_io_handle,
	     &offset_index_root_node_offset,
	     &offset_index_back_pointer,
	     &descriptor_index_root_node_offset,
	     &descriptor_index_back_pointer,
	     &( internal_file->content_type ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read file header.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "Reading offset index:\n" );
	}
#endif
	result_read_offset_index = libpff_io_handle_read_offset_index(
	                            internal_file->io_handle,
	                            internal_file->file_io_handle,
	                            offset_index_root_node_offset,
	                            offset_index_back_pointer,
	                            error );

	if( result_read_offset_index != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read offset index.",
		 function );

#if defined( HAVE_VERBOSE_OUTPUT )
		if( ( error != NULL )
		 && ( *error != NULL ) )
		{
			libnotify_print_error_backtrace(
			 *error );
		}
#endif
		liberror_error_free(
		 error );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "Reading descriptor index:\n" );
	}
#endif
	result_read_descriptor_index = libpff_io_handle_read_descriptor_index(
	                                internal_file->io_handle,
	                                internal_file->file_io_handle,
	                                descriptor_index_root_node_offset,
	                                descriptor_index_back_pointer,
	                                error );

	if( result_read_descriptor_index != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read descriptor index.",
		 function );

#if defined( HAVE_VERBOSE_OUTPUT )
		if( ( error != NULL )
		 && ( *error != NULL ) )
		{
			libnotify_print_error_backtrace(
			 *error );
		}
#endif
		liberror_error_free(
		 error );
	}
	if( ( result_read_offset_index == 1 )
	 && ( result_read_descriptor_index == 1 ) )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "Creating item tree:\n" );
		}
#endif
		if( libpff_item_tree_create(
		     internal_file->io_handle,
		     &( internal_file->item_tree_root_node ),
		     &( internal_file->root_folder_item_tree_node ),
		     internal_file->orphan_item_list,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create item tree.",
			 function );

			return( -1 );
		}
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "Name to ID map:\n" );
		}
#endif
		if( libpff_name_to_id_map_read(
		     internal_file->io_handle,
		     internal_file->file_io_handle,
		     internal_file->name_to_id_map_list,
		     error ) == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read name to id map.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Reads the allocation tables
 * Returns 1 if successful or -1 on error
 */
int libpff_file_read_allocation_tables(
     libpff_internal_file_t *internal_file,
     liberror_error_t **error )
{
	static char *function = "libpff_file_read_allocation_tables";

	if( internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid internal file.",
		 function );

		return( -1 );
	}
	if( internal_file->read_allocation_tables != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid internal file - allocation tables already set.",
		 function );

		return( -1 );
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "Reading the unallocated data blocks:\n" );
	}
#endif
	if( libpff_io_handle_read_unallocated_data_blocks(
	     internal_file->io_handle,
	     internal_file->file_io_handle,
	     internal_file->unallocated_data_block_list,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read unallocated data blocks.",
		 function );

		return( -1 );
	}

#if defined( HAVE_VERBOSE_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "Reading the unallocated page blocks:\n" );
	}
#endif
	if( libpff_io_handle_read_unallocated_page_blocks(
	     internal_file->io_handle,
	     internal_file->file_io_handle,
	     internal_file->unallocated_page_block_list,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read unallocated page blocks.",
		 function );

		return( -1 );
	}
	internal_file->read_allocation_tables = 1;

	return( 1 );
}

/* Recovers deleted items within a Personal Folder File
 * By default only the unallocated space is checked for recoverable items
 * Returns 1 if successful or -1 on error
 */
int libpff_file_recover_items(
     libpff_file_t *file,
     uint8_t recovery_flags,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_recover_items";
	int result                            = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->read_allocation_tables == 0 )
	{
		result = libpff_file_read_allocation_tables(
		          internal_file,
		          error );

		if( result != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read allocation tables.",
			 function );

#if defined( HAVE_VERBOSE_OUTPUT )
			if( ( error != NULL )
			 && ( *error != NULL ) )
			{
				libnotify_print_error_backtrace(
				 *error );
			}
#endif
			liberror_error_free(
			 error );
		}
	}
	if( libpff_io_handle_recover_items(
	     internal_file->io_handle,
	     internal_file->file_io_handle,
	     internal_file->unallocated_data_block_list,
	     internal_file->unallocated_page_block_list,
	     internal_file->recovered_item_list,
	     recovery_flags,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to recover items.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the file size
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_size(
     libpff_file_t *file,
     size64_t *size,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_size";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing io handle.",
		 function );

		return( -1 );
	}
	if( size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid size.",
		 function );

		return( -1 );
	}
	*size = internal_file->io_handle->file_size;

	return( 1 );
}

/* Retrieves the file content type
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_content_type(
     libpff_file_t *file,
     uint8_t *content_type,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_type";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( content_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid content type.",
		 function );

		return( -1 );
	}
	*content_type = (uint8_t) internal_file->content_type;

	return( 1 );
}

/* Retrieves the file type
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_type(
     libpff_file_t *file,
     uint8_t *type,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_type";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing io handle.",
		 function );

		return( -1 );
	}
	if( type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid type.",
		 function );

		return( -1 );
	}
	*type = internal_file->io_handle->file_type;

	return( 1 );
}

/* Retrieves the file encryption type
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_encryption_type(
     libpff_file_t *file,
     uint8_t *encryption_type,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_encryption_type";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing io handle.",
		 function );

		return( -1 );
	}
	if( encryption_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid encryption type.",
		 function );

		return( -1 );
	}
	*encryption_type = internal_file->io_handle->encryption_type;

	return( 1 );
}

/* Retrieves the file ASCII codepage
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_ascii_codepage(
     libpff_file_t *file,
     int *ascii_codepage,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_ascii_codepage";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( ascii_codepage == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid ASCII codepage.",
		 function );

		return( -1 );
	}
	*ascii_codepage = internal_file->ascii_codepage;

	return( 1 );
}

/* Sets the file ASCII codepage
 * Returns 1 if successful or -1 on error
 */
int libpff_file_set_ascii_codepage(
     libpff_file_t *file,
     int ascii_codepage,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_set_ascii_codepage";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( ( ascii_codepage != LIBPFF_CODEPAGE_ASCII )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_874 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1250 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1251 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1252 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1253 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1254 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1256 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1257 )
	 && ( ascii_codepage != LIBPFF_CODEPAGE_WINDOWS_1258 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported ASCII codepage.",
		 function );

		return( -1 );
	}
	internal_file->ascii_codepage = ascii_codepage;

	return( 1 );
}

/* Retrieves the number of unallocated blocks
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_number_of_unallocated_blocks(
     libpff_file_t *file,
     int unallocated_block_type,
     int *number_of_unallocated_blocks,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	libpff_list_t *unallocated_block_list = NULL;
	static char *function                 = "libpff_file_get_number_of_unallocated_blocks";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	if( ( unallocated_block_type != LIBPFF_UNALLOCATED_BLOCK_TYPE_DATA )
	 && ( unallocated_block_type != LIBPFF_UNALLOCATED_BLOCK_TYPE_PAGE ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported unallocated block type.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->read_allocation_tables == 0 )
	{
		if( libpff_file_read_allocation_tables(
		     internal_file,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read allocation tables.",
			 function );

			return( -1 );
		}
	}
	if( unallocated_block_type == LIBPFF_UNALLOCATED_BLOCK_TYPE_DATA )
	{
		unallocated_block_list = internal_file->unallocated_data_block_list;
	}
	else if( unallocated_block_type == LIBPFF_UNALLOCATED_BLOCK_TYPE_PAGE )
	{
		unallocated_block_list = internal_file->unallocated_page_block_list;
	}
	if( libpff_list_get_number_of_elements(
	     unallocated_block_list,
	     number_of_unallocated_blocks,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of unallocated blocks.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves a specific unallocated block
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_unallocated_block(
     libpff_file_t *file,
     int unallocated_block_type,
     int unallocated_block_index,
     off64_t *offset,
     size64_t *size,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	libpff_list_t *unallocated_block_list = NULL;
	static char *function                 = "libpff_file_get_unallocated_block";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	if( ( unallocated_block_type != LIBPFF_UNALLOCATED_BLOCK_TYPE_DATA )
	 && ( unallocated_block_type != LIBPFF_UNALLOCATED_BLOCK_TYPE_PAGE ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported unallocated block type.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( unallocated_block_type == LIBPFF_UNALLOCATED_BLOCK_TYPE_DATA )
	{
		unallocated_block_list = internal_file->unallocated_data_block_list;
	}
	else if( unallocated_block_type == LIBPFF_UNALLOCATED_BLOCK_TYPE_PAGE )
	{
		unallocated_block_list = internal_file->unallocated_page_block_list;
	}
	if( unallocated_block_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing unallocated block list.",
		 function );

		return( -1 );
	}
	if( ( internal_file->read_allocation_tables == 0 )
	 && ( libpff_file_read_allocation_tables(
	       internal_file,
	       error ) != 1 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read allocation tables.",
		 function );

		return( -1 );
	}
	if( libpff_offset_list_get_offset(
	     unallocated_block_list,
	     unallocated_block_index,
	     offset,
	     size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve offset list element.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the root item from the Personal Folder File
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_root_item(
     libpff_file_t *file,
     libpff_item_t **root_item,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_root_item";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->item_tree_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing item tree root node.",
		 function );

		return( -1 );
	}
	if( root_item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid root item.",
		 function );

		return( -1 );
	}
	if( *root_item != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: root item already set.",
		 function );

		return( -1 );
	}
	if( libpff_item_initialize(
	     root_item,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create root item.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *root_item,
	     internal_file->file_io_handle,
	     internal_file,
	     internal_file->item_tree_root_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach root item.",
		 function );

		libpff_item_free(
		 root_item,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the message store from the Personal Folder File
 * Returns 1 if successful, 0 if not present or -1 on error
 */
int libpff_file_get_message_store(
     libpff_file_t *file,
     libpff_item_t **message_store,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file            = NULL;
	libpff_tree_node_t *message_store_item_tree_node = NULL;
	static char *function                            = "libpff_file_get_message_store";
	int result                                       = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing io handle.",
		 function );

		return( -1 );
	}
	if( internal_file->io_handle->descriptor_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - invalid io handle - missing descriptor index root node.",
		 function );

		return( -1 );
	}
	if( message_store == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message store.",
		 function );

		return( -1 );
	}
	if( *message_store != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: message store already set.",
		 function );

		return( -1 );
	}
	result = libpff_index_get_tree_node(
	          internal_file->io_handle->descriptor_index_root_node,
	          LIBPFF_DESCRIPTOR_IDENTIFIER_MESSAGE_STORE,
	          &message_store_item_tree_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: error while searching to find descriptor identifier: %" PRIu32 ".",
		 function,
		 LIBPFF_DESCRIPTOR_IDENTIFIER_MESSAGE_STORE );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( libpff_item_initialize(
	     message_store,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create message store.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *message_store,
	     internal_file->file_io_handle,
	     internal_file,
	     message_store_item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach message store.",
		 function );

		libpff_item_free(
		 message_store,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the name to id map from the Personal Folder File
 * Returns 1 if successful, 0 if not present or -1 on error
 */
int libpff_file_get_name_to_id_map(
     libpff_file_t *file,
     libpff_item_t **name_to_id_map,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file             = NULL;
	libpff_tree_node_t *name_to_id_map_item_tree_node = NULL;
	static char *function                             = "libpff_file_get_name_to_id_map";
	int result                                        = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing io handle.",
		 function );

		return( -1 );
	}
	if( internal_file->io_handle->descriptor_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - invalid io handle - missing descriptor index root node.",
		 function );

		return( -1 );
	}
	if( name_to_id_map == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map.",
		 function );

		return( -1 );
	}
	if( *name_to_id_map != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: name to id map already set.",
		 function );

		return( -1 );
	}
	result = libpff_index_get_tree_node(
	          internal_file->io_handle->descriptor_index_root_node,
	          LIBPFF_DESCRIPTOR_IDENTIFIER_NAME_TO_ID_MAP,
	          &name_to_id_map_item_tree_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: error while searching to find descriptor identifier: %" PRIu32 ".",
		 function,
		 LIBPFF_DESCRIPTOR_IDENTIFIER_NAME_TO_ID_MAP );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( libpff_item_initialize(
	     name_to_id_map,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create name to id map.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *name_to_id_map,
	     internal_file->file_io_handle,
	     internal_file,
	     name_to_id_map_item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach name to id map.",
		 function );

		libpff_item_free(
		 name_to_id_map,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the root folder from the Personal Folder File
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_root_folder(
     libpff_file_t *file,
     libpff_item_t **root_folder,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_root_folder";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->root_folder_item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing root folder item tree node.",
		 function );

		return( -1 );
	}
	if( root_folder == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid root folder.",
		 function );

		return( -1 );
	}
	if( *root_folder != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: root folder already set.",
		 function );

		return( -1 );
	}
	if( libpff_item_initialize(
	     root_folder,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create root folder.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *root_folder,
	     internal_file->file_io_handle,
	     internal_file,
	     internal_file->root_folder_item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach root folder.",
		 function );

		libpff_item_free(
		 root_folder,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the item for the specific identifier from the Personal Folder File
 * Returns 1 if successful, 0 if no such item, or -1 on error
 */
int libpff_file_get_item_by_identifier(
     libpff_file_t *file,
     uint32_t item_identifier,
     libpff_item_t **item,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	libpff_tree_node_t *item_tree_node    = NULL;
	static char *function                 = "libpff_file_get_item_by_identifier";
	int result                            = 0;

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->item_tree_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid file - missing item tree root node.",
		 function );

		return( -1 );
	}
	if( item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *item != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: item already set.",
		 function );

		return( -1 );
	}
	result = libpff_item_tree_get_tree_node_by_identifier(
	          internal_file->item_tree_root_node,
                  item_identifier,
                  &item_tree_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve item tree node: %" PRIu32 ".",
		 function,
		 item_identifier );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item tree node.",
		 function );

		return( -1 );
	}
	if( libpff_item_initialize(
	     item,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to initialize item.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *item,
	     internal_file->file_io_handle,
	     internal_file,
	     item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach item.",
		 function );

		libpff_item_free(
		 item,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the number of orphan items
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_number_of_orphan_items(
     libpff_file_t *file,
     int *number_of_orphan_items,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_number_of_orphan_items";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( libpff_list_get_number_of_elements(
	     internal_file->orphan_item_list,
	     number_of_orphan_items,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of orphan items.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves a specific orphan item
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_orphan_item(
     libpff_file_t *file,
     int orphan_item_index,
     libpff_item_t **orphan_item,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file     = NULL;
	libpff_tree_node_t *orphan_item_tree_node = NULL;
	static char *function                     = "libpff_file_get_orphan_item";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->orphan_item_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal file - missing orphan item list.",
		 function );

		return( -1 );
	}
	if( orphan_item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid orphan item.",
		 function );

		return( -1 );
	}
	if( libpff_list_get_value(
	     internal_file->orphan_item_list,
	     orphan_item_index,
	     (intptr_t **) &orphan_item_tree_node,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve orphan item tree node.",
		 function );

		return( -1 );
	}
	if( libpff_item_initialize(
	     orphan_item,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create orphan item.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *orphan_item,
	     internal_file->file_io_handle,
	     internal_file,
	     orphan_item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach orphan item.",
		 function );

		libpff_item_free(
		 orphan_item,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the number of recovered items
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_number_of_recovered_items(
     libpff_file_t *file,
     int *number_of_recovered_items,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file = NULL;
	static char *function                 = "libpff_file_get_number_of_recovered_items";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( libpff_list_get_number_of_elements(
	     internal_file->recovered_item_list,
	     number_of_recovered_items,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of recovered items.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves a specific recovered item
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_recovered_item(
     libpff_file_t *file,
     int recovered_item_index,
     libpff_item_t **recovered_item,
     liberror_error_t **error )
{
	libpff_internal_file_t *internal_file        = NULL;
	libpff_tree_node_t *recovered_item_tree_node = NULL;
	static char *function                        = "libpff_file_get_recovered_item";

	if( file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file.",
		 function );

		return( -1 );
	}
	internal_file = (libpff_internal_file_t *) file;

	if( internal_file->recovered_item_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal file - missing recovered item list.",
		 function );

		return( -1 );
	}
	if( recovered_item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid recovered item.",
		 function );

		return( -1 );
	}
	if( libpff_list_get_value(
	     internal_file->recovered_item_list,
	     recovered_item_index,
	     (intptr_t **) &recovered_item_tree_node,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve recovered item tree node.",
		 function );

		return( -1 );
	}
	if( libpff_item_initialize(
	     recovered_item,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create recovered item.",
		 function );

		return( -1 );
	}
	if( libpff_item_attach(
	     (libpff_internal_item_t *) *recovered_item,
	     internal_file->file_io_handle,
	     internal_file,
	     recovered_item_tree_node,
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach recovered item.",
		 function );

		libpff_item_free(
		 recovered_item,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

