/*
 * Index functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_definitions.h"
#include "libpff_index.h"
#include "libpff_libbfio.h"
#include "libpff_libfmapi.h"
#include "libpff_list_type.h"
#include "libpff_tree_type.h"
#include "libpff_offset_list.h"

#include "pff_index_node.h"

/* Frees index values
 */
int libpff_index_values_free(
     intptr_t *index_values,
     liberror_error_t **error )
{
	if( index_values != NULL )
	{
		memory_free(
		 index_values );
	}
	return( 1 );
}

/* Compares two index values
 * Returns LIBPFF_TREE_NODE_COMPARE_LESS, LIBPFF_TREE_NODE_COMPARE_EQUAL, LIBPFF_TREE_NODE_COMPARE_GREATER if successful or -1 on error
 */
int libpff_index_values_compare(
     intptr_t *first,
     intptr_t *second,
     liberror_error_t **error )
{
	static char *function = "libpff_index_values_compare";

	if( first == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid first index values.",
		 function );

		return( -1 );
	}
	if( second == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid second index values.",
		 function );

		return( 1 );
	}
	if( ( (libpff_index_values_t *) first )->identifier < ( (libpff_index_values_t *) second )->identifier )
	{
		return( LIBPFF_TREE_NODE_COMPARE_LESS );
	}
	else if( ( (libpff_index_values_t *) first )->identifier > ( (libpff_index_values_t *) second )->identifier )
	{
		return( LIBPFF_TREE_NODE_COMPARE_GREATER );
	}
	return( LIBPFF_TREE_NODE_COMPARE_EQUAL );
}

/* Compares two index values but allows duplicate values
 * The second duplicate is considered greater than the first
 * Returns LIBPFF_TREE_NODE_COMPARE_LESS, LIBPFF_TREE_NODE_COMPARE_GREATER if successful or -1 on error
 */
int libpff_index_values_compare_allow_duplicate(
     intptr_t *first,
     intptr_t *second,
     liberror_error_t **error )
{
	static char *function = "libpff_index_values_compare_allow_duplicate";

	if( first == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid first index values.",
		 function );

		return( -1 );
	}
	if( second == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid second index values.",
		 function );

		return( 1 );
	}
	if( ( (libpff_index_values_t *) first )->identifier < ( (libpff_index_values_t *) second )->identifier )
	{
		return( LIBPFF_TREE_NODE_COMPARE_LESS );
	}
	return( LIBPFF_TREE_NODE_COMPARE_GREATER );
}

/* Creates an index tree node with item values
 * Returns 1 if successful or -1 on error
 */
int libpff_index_node_initialize(
     libpff_tree_node_t **index_tree_node,
     liberror_error_t **error )
{
	static char *function = "libpff_index_node_initialize";

	if( index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index tree node.",
		 function );

		return( -1 );
	}
	if( *index_tree_node == NULL )
	{
		if( libpff_tree_node_initialize(
		     index_tree_node,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create index tree node.",
			 function );

			return( -1 );
		}
		( *index_tree_node )->value = (intptr_t *) memory_allocate(
		                                            sizeof( libpff_index_values_t ) );

		if( ( *index_tree_node )->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create index values.",
			 function );

			libpff_tree_node_free(
			 index_tree_node,
			 NULL,
			 NULL );

			*index_tree_node = NULL;

			return( -1 );
		}
		if( memory_set(
		     ( *index_tree_node )->value,
		     0,
		     sizeof( libpff_index_values_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear index values.",
			 function );

			libpff_tree_node_free(
			 index_tree_node,
			 &libpff_index_values_free,
			 NULL );

			*index_tree_node = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Reads an index tree node and its elements
 * Returns 1 if successful or -1 on error
 */
int libpff_index_read_node(
     libpff_tree_node_t *index_tree_node,
     libbfio_handle_t *file_io_handle,
     off64_t node_file_offset,
     int file_type,
     uint8_t *node_level, 
     uint8_t *index_type, 
     uint64_t *back_pointer,
     liberror_error_t **error )
{
	libpff_tree_node_t *child_node  = NULL;
	uint8_t *index_node_data        = NULL;
	uint8_t *item_data              = NULL;
	static char *function           = "libpff_index_read_node";
	size_t read_size                = 0;
	ssize_t read_count              = 0;
        uint32_t stored_crc             = 0;
        uint32_t calculated_crc         = 0;
	uint8_t number_of_items         = 0;
	uint8_t expected_item_size      = 0;
	uint8_t index_type_copy         = 0;
	uint8_t item_size               = 0;
	uint8_t maximum_number_of_items = 0;
	int number_of_child_items       = 0;
	int child_node_iterator         = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint32_t value_32bit            = 0;
	uint16_t value_16bit            = 0;
#endif

	if( index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index tree node.",
		 function );

		return( -1 );
	}
	if( file_io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file io handle.",
		 function );

		return( -1 );
	}
	if( ( file_type != LIBPFF_FILE_TYPE_32BIT )
	 && ( file_type != LIBPFF_FILE_TYPE_64BIT ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: invalid file type.",
		 function );

		return( -1 );
	}
	if( node_level == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node level.",
		 function );

		return( -1 );
	}
	if( index_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index type.",
		 function );

		return( -1 );
	}
	if( back_pointer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid back pointer.",
		 function );

		return( -1 );
	}
	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     node_file_offset,
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek node offset: %" PRIu64 ".",
		 function,
		 node_file_offset );

		return( -1 );
	}
	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		read_size = sizeof( pff_index_node_32bit_t );
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		read_size = sizeof( pff_index_node_64bit_t );
	}
	index_node_data = (uint8_t *) memory_allocate(
	                               read_size );

	if( index_node_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create index node data.",
		 function );

		return( -1 );
	}
	read_count = libbfio_handle_read(
	              file_io_handle,
	              index_node_data,
	              read_size,
	              error );

	if( read_count != (ssize_t) read_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read index node.",
		 function );

		memory_free(
		 index_node_data );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: index node:\n",
		 function );
		libnotify_print_data(
		 index_node_data,
		 read_size );
	}
#endif

	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		*index_type     = ( (pff_index_node_32bit_t *) index_node_data )->type;
		index_type_copy = ( (pff_index_node_32bit_t *) index_node_data )->type_copy;

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_index_node_32bit_t *) index_node_data )->back_pointer,
		 *back_pointer );
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_index_node_32bit_t *) index_node_data )->crc,
		 stored_crc );

		number_of_items         = ( (pff_index_node_32bit_t *) index_node_data )->number_of_items ;
		maximum_number_of_items = ( (pff_index_node_32bit_t *) index_node_data )->maximum_number_of_items;
		item_size               = ( (pff_index_node_32bit_t *) index_node_data )->item_size;
		*node_level             = ( (pff_index_node_32bit_t *) index_node_data )->level;

		if( ( *index_type == LIBPFF_INDEX_TYPE_DESCRIPTOR )
		 && ( *node_level == PFF_INDEX_NODE_LEVEL_LEAF ) )
		{
			expected_item_size = 16;
		}
		else
		{
			expected_item_size = 12;
		}
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		*index_type     = ( (pff_index_node_64bit_t *) index_node_data )->type;
		index_type_copy = ( (pff_index_node_64bit_t *) index_node_data )->type_copy;

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_index_node_64bit_t *) index_node_data )->crc,
		 stored_crc );
		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_index_node_64bit_t *) index_node_data )->back_pointer,
		 *back_pointer );

		number_of_items         = ( (pff_index_node_64bit_t *) index_node_data )->number_of_items ;
		maximum_number_of_items = ( (pff_index_node_64bit_t *) index_node_data )->maximum_number_of_items;
		item_size               = ( (pff_index_node_64bit_t *) index_node_data )->item_size;
		*node_level             = ( (pff_index_node_64bit_t *) index_node_data )->level;

		if( ( *index_type == LIBPFF_INDEX_TYPE_DESCRIPTOR )
		 && ( *node_level == PFF_INDEX_NODE_LEVEL_LEAF ) )
		{
			expected_item_size = 32;
		}
		else
		{
			expected_item_size = 24;
		}
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: number of items\t\t\t\t: %" PRIu8 "\n",
		 function,
		 number_of_items );
		libnotify_printf(
		 "%s: maximum number of items\t\t\t: %" PRIu8 "\n",
		 function,
		 maximum_number_of_items );
		libnotify_printf(
		 "%s: item size\t\t\t\t: %" PRIu8 "\n",
		 function,
		 item_size );
		libnotify_printf(
		 "%s: node level\t\t\t\t: %" PRIu8 "\n",
		 function,
		 *node_level );

		if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_index_node_64bit_t *) index_node_data )->padding1,
			 value_32bit );
			libnotify_printf(
			 "%s: padding1\t\t\t\t: 0x%08" PRIx32 "\n",
			 function,
			 value_32bit );
		}
		libnotify_printf(
		 "%s: index type\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 *index_type );
		libnotify_printf(
		 "%s: index type copy\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 index_type_copy );

		if( file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_index_node_32bit_t *) index_node_data )->signature,
			 value_16bit );
			libnotify_printf(
			 "%s: signature\t\t\t\t: 0x%04" PRIx16 "\n",
			 function,
			 value_16bit );

			libnotify_printf(
			 "%s: back pointer\t\t\t\t: %" PRIu64 "\n",
			 function,
			 *back_pointer );
			libnotify_printf(
			 "%s: crc\t\t\t\t\t: 0x%08" PRIx32 "\n",
			 function,
			 stored_crc );
		}
		else if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_index_node_64bit_t *) index_node_data )->signature,
			 value_16bit );
			libnotify_printf(
			 "%s: signature\t\t\t\t: 0x%04" PRIx16 "\n",
			 function,
			 value_16bit );

			libnotify_printf(
			 "%s: crc\t\t\t\t\t: 0x%08" PRIx32 "\n",
			 function,
			 stored_crc );
			libnotify_printf(
			 "%s: back pointer\t\t\t\t: %" PRIu64 "\n",
			 function,
			 *back_pointer );
		}
	}
#endif

	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		if( libfmapi_crc_weak_calculate(
		     0,
		     index_node_data,
		     500,
		     &calculated_crc,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to calculate weak CRC.",
			 function );

			memory_free(
			 index_node_data );

			return( -1 );
		}
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		if( libfmapi_crc_weak_calculate(
		     0,
		     index_node_data,
		     496,
		     &calculated_crc,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to calculate weak CRC.",
			 function );

			memory_free(
			 index_node_data );

			return( -1 );
		}
	}
	if( stored_crc != calculated_crc )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_INPUT,
		 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
		 "%s: mismatch in crc ( 0x%08" PRIx32 " != 0x%08" PRIx32 " ).",
		 function,
		 stored_crc,
		 calculated_crc );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	if( *index_type != index_type_copy )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_INPUT,
		 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
		 "%s: mismatch in index type ( 0x%02" PRIx8 " != 0x%02" PRIx8 " ).",
		 function,
		 *index_type,
		 index_type_copy );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	if( ( *index_type != LIBPFF_INDEX_TYPE_DESCRIPTOR )
	 && ( *index_type != LIBPFF_INDEX_TYPE_OFFSET ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported index node type: 0x%02" PRIx8 "",
		 function,
		 *index_type );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	if( expected_item_size != item_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: item size mismatch (expected: %" PRIu8 ", read: %" PRIu8 ").",
		 function,
		 expected_item_size,
		 item_size );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	if( number_of_items > maximum_number_of_items )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: number of items: %" PRIu8 ", exceeds maximum: %" PRIu8 ".",
		 function,
		 number_of_items,
		 maximum_number_of_items );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	if( ( number_of_items * item_size ) > PFF_INDEX_NODE_MAXIMUM_ITEMS_DATA_SIZE )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: size of items: %" PRIu16 ", exceeds maximum: %" PRIu8 ".",
		 function,
		 number_of_items * item_size,
		 PFF_INDEX_NODE_MAXIMUM_ITEMS_DATA_SIZE );

		/* TODO implement error tollerance
		 */
		memory_free(
		 index_node_data );

		return( -1 );
	}
	number_of_child_items = (int) number_of_items;

#if defined( HAVE_DEBUG_OUTPUT )
	/* In debug mode print all the entries
	 */
	if( *index_type == LIBPFF_INDEX_TYPE_DESCRIPTOR )
	{
		number_of_child_items = (int) maximum_number_of_items;
	}
#endif

	for( child_node_iterator = 0;
	     child_node_iterator < number_of_child_items;
	     child_node_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( ( libnotify_verbose != 0 )
		 && ( child_node_iterator == (int) number_of_items ) )
		{
			libnotify_printf(
			 "%s: remaining node items\n",
			 function );
		}
#endif
		child_node = (libpff_tree_node_t *) memory_allocate(
		                                     sizeof( libpff_tree_node_t ) );

		if( child_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create tree node.",
			 function );

			memory_free(
			 index_node_data );

			return( -1 );
		}
		if( memory_set(
		     child_node,
		     0,
		     sizeof( libpff_tree_node_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear tree node.",
			 function );

			memory_free(
			 child_node );
			memory_free(
			 index_node_data );

			return( -1 );
		}
		child_node->value = (intptr_t *) memory_allocate(
		                                  sizeof( libpff_index_values_t ) );

		if( child_node->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create index values.",
			 function );

			memory_free(
			 child_node );
			memory_free(
			 index_node_data );

			return( -1 );
		}
		if( file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			item_data = (uint8_t *) &( ( ( (pff_index_node_32bit_t *) index_node_data )->items_data )[ ( child_node_iterator * item_size ) ] );

			byte_stream_copy_to_uint32_little_endian(
			 item_data,
			 ( (libpff_index_values_t *) child_node->value )->identifier );
		}
		else if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			item_data = (uint8_t *) &( ( ( (pff_index_node_64bit_t *) index_node_data )->items_data )[ ( child_node_iterator * item_size ) ] );

			byte_stream_copy_to_uint32_little_endian(
			 item_data,
			 ( (libpff_index_values_t *) child_node->value )->identifier );
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: item: %02d index identifier\t\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
			 function,
			 child_node_iterator,
			 ( (libpff_index_values_t *) child_node->value )->identifier,
			 ( (libpff_index_values_t *) child_node->value )->identifier );
		}
#endif

		/* Handle descriptor index leaf nodes
		 */
		if( ( *index_type == LIBPFF_INDEX_TYPE_DESCRIPTOR )
		 && ( *node_level == PFF_INDEX_NODE_LEVEL_LEAF ) )
		{
			if( file_type == LIBPFF_FILE_TYPE_32BIT )
			{
				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_index_node_descriptor_item_32bit_t*) item_data )->data_identifier,
				 ( (libpff_index_values_t *) child_node->value )->data_identifier );
				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_index_node_descriptor_item_32bit_t*) item_data )->local_descriptors_identifier,
				 ( (libpff_index_values_t *) child_node->value )->local_descriptors_identifier );
				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_index_node_descriptor_item_32bit_t*) item_data )->parent_identifier,
				 ( (libpff_index_values_t *) child_node->value )->parent_identifier );
			}
			else if( file_type == LIBPFF_FILE_TYPE_64BIT )
			{
				byte_stream_copy_to_uint64_little_endian(
				 ( (pff_index_node_descriptor_item_64bit_t*) item_data )->data_identifier,
				 ( (libpff_index_values_t *) child_node->value )->data_identifier );
				byte_stream_copy_to_uint64_little_endian(
				 ( (pff_index_node_descriptor_item_64bit_t*) item_data )->local_descriptors_identifier,
				 ( (libpff_index_values_t *) child_node->value )->local_descriptors_identifier );
				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_index_node_descriptor_item_64bit_t*) item_data )->parent_identifier,
				 ( (libpff_index_values_t *) child_node->value )->parent_identifier );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: item: %02d data identifier\t\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
				 function,
				 child_node_iterator,
				 ( (libpff_index_values_t *) child_node->value )->data_identifier,
				 ( (libpff_index_values_t *) child_node->value )->data_identifier );
				libnotify_printf(
				 "%s: item: %02d local descriptors identifier\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
				 function,
				 child_node_iterator,
				 ( (libpff_index_values_t *) child_node->value )->local_descriptors_identifier,
				 ( (libpff_index_values_t *) child_node->value )->local_descriptors_identifier );
				libnotify_printf(
				 "%s: item: %02d parent identifier\t\t: 0x%08" PRIx32 " (%" PRIu32 ")\n",
				 function,
				 child_node_iterator,
				 ( (libpff_index_values_t *) child_node->value )->parent_identifier,
				 ( (libpff_index_values_t *) child_node->value )->parent_identifier );

				if( file_type == LIBPFF_FILE_TYPE_64BIT )
				{
					byte_stream_copy_to_uint32_little_endian(
					 ( (pff_index_node_descriptor_item_64bit_t*) item_data )->unknown1,
					 value_32bit );
					libnotify_printf(
					 "%s: item: %02d unknown1\t\t\t: 0x%08" PRIx32 " (%" PRIu32 ")\n",
					 function,
					 child_node_iterator,
					 value_32bit,
					 value_32bit );
				}
			}
#endif
		}
		/* Handle offset and descriptor index branch nodes
		 * and offset index leaf nodes
		 */
		else
		{
			if( file_type == LIBPFF_FILE_TYPE_32BIT )
			{
				if( *node_level != PFF_INDEX_NODE_LEVEL_LEAF )
				{
					byte_stream_copy_to_uint32_little_endian(
					 ( (pff_index_node_branch_item_32bit_t *) item_data )->back_pointer,
					 ( (libpff_index_values_t *) child_node->value )->back_pointer );
					byte_stream_copy_to_uint32_little_endian(
					 ( (pff_index_node_branch_item_32bit_t *) item_data )->file_offset,
					 ( (libpff_index_values_t *) child_node->value )->file_offset );
				}
				else if( *node_level == PFF_INDEX_NODE_LEVEL_LEAF )
				{
					byte_stream_copy_to_uint32_little_endian(
					 ( (pff_index_node_offset_item_32bit_t *) item_data )->file_offset,
					 ( (libpff_index_values_t *) child_node->value )->file_offset );
					byte_stream_copy_to_uint16_little_endian(
					 ( (pff_index_node_offset_item_32bit_t *) item_data )->data_size,
					 ( (libpff_index_values_t *) child_node->value )->data_size );
					byte_stream_copy_to_uint16_little_endian(
					 ( (pff_index_node_offset_item_32bit_t *) item_data )->reference_count,
					 ( (libpff_index_values_t *) child_node->value )->reference_count );
				}
			}
			else if( file_type == LIBPFF_FILE_TYPE_64BIT )
			{
				if( *node_level != PFF_INDEX_NODE_LEVEL_LEAF )
				{
					byte_stream_copy_to_uint64_little_endian(
					 ( (pff_index_node_branch_item_64bit_t *) item_data )->back_pointer,
					 ( (libpff_index_values_t *) child_node->value )->back_pointer );
					byte_stream_copy_to_uint64_little_endian(
					 ( (pff_index_node_branch_item_64bit_t *) item_data )->file_offset,
					 ( (libpff_index_values_t *) child_node->value )->file_offset );
				}
				else if( *node_level == PFF_INDEX_NODE_LEVEL_LEAF )
				{
					byte_stream_copy_to_uint64_little_endian(
					 ( (pff_index_node_offset_item_64bit_t *) item_data )->file_offset,
					 ( (libpff_index_values_t *) child_node->value )->file_offset );
					byte_stream_copy_to_uint16_little_endian(
					 ( (pff_index_node_offset_item_64bit_t *) item_data )->data_size,
					 ( (libpff_index_values_t *) child_node->value )->data_size );
					byte_stream_copy_to_uint16_little_endian(
					 ( (pff_index_node_offset_item_64bit_t *) item_data )->reference_count,
					 ( (libpff_index_values_t *) child_node->value )->reference_count );
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( *node_level != PFF_INDEX_NODE_LEVEL_LEAF )
				{
					libnotify_printf(
					 "%s: item: %02d back pointer\t\t\t: %" PRIu64 "\n",
					 function,
					 child_node_iterator,
					 ( (libpff_index_values_t *) child_node->value )->back_pointer );
				}
				libnotify_printf(
				 "%s: item: %02d file offset\t\t\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
				 function,
				 child_node_iterator,
				 ( (libpff_index_values_t *) child_node->value )->file_offset,
				 ( (libpff_index_values_t *) child_node->value )->file_offset );

				if( *node_level == PFF_INDEX_NODE_LEVEL_LEAF )
				{
					libnotify_printf(
					 "%s: item: %02d data size\t\t\t: %" PRIu32 "\n",
					 function,
					 child_node_iterator,
					 ( (libpff_index_values_t *) child_node->value )->data_size );
					libnotify_printf(
					 "%s: item: %02d reference count\t\t: %" PRIu32 "\n",
					 function,
					 child_node_iterator,
					 ( (libpff_index_values_t *) child_node->value )->reference_count );

					if( file_type == LIBPFF_FILE_TYPE_64BIT )
					{
						byte_stream_copy_to_uint32_little_endian(
						 ( (pff_index_node_offset_item_64bit_t*) item_data )->data_allocation_table_file_offset,
						 value_32bit );
						libnotify_printf(
						 "%s: item: %02d data allocation table offset\t: 0x%08" PRIx32 " (%" PRIu32 ")\n",
						 function,
						 child_node_iterator,
						 value_32bit,
						 value_32bit );
					}
				}
			}
#endif
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( child_node_iterator >= (int) number_of_items )
		{
			memory_free(
			 child_node->value );
			memory_free(
			 child_node );

			continue;
		}
#endif
		if( libpff_tree_node_insert_node(
		     index_tree_node,
		     child_node,
		     &libpff_index_values_compare,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to insert tree node.",
			 function );

			memory_free(
			 child_node->value );
			memory_free(
			 child_node );
			memory_free(
			 index_node_data );

			return( -1 );
		}
	}
	memory_free(
	 index_node_data );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );
}

/* Reads the index tree node and its child nodes
 * Returns 1 if successful or -1 on error
 */
int libpff_index_read_tree(
     libpff_tree_node_t *index_tree_node,
     libbfio_handle_t *file_io_handle,
     off64_t node_file_offset,
     int file_type,
     uint8_t expected_index_type,
     uint64_t expected_back_pointer,
     liberror_error_t **error )
{
	libpff_tree_node_t *child_node = NULL;
	static char *function          = "libpff_index_read_tree";
	uint64_t back_pointer          = 0;
	uint8_t index_type             = 0;
	uint8_t node_level             = 0;
	int child_node_iterator        = 0;

	if( libpff_index_read_node(
	     index_tree_node,
	     file_io_handle,
	     node_file_offset,
	     file_type,
	     &node_level,
	     &index_type,
	     &back_pointer,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
 		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read index root node.",
		 function );

		return( -1 );
	}
	if( back_pointer != expected_back_pointer )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: back pointer mismatch (expected: %" PRIu64 ", read: %" PRIu64 ").",
		 function,
		 expected_back_pointer,
		 back_pointer );

		/* TODO implement error tollerance
		 */
		return( -1 );
	}
	if( index_type != expected_index_type )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: index type mismatch (expected: 0x%02" PRIx8 ", read: 0x%02" PRIx8 ").",
		 function,
		 expected_index_type,
		 index_type );

		/* TODO implement error tollerance
		 */
		return( -1 );
	}
	if( node_level != PFF_INDEX_NODE_LEVEL_LEAF )
	{
		child_node = index_tree_node->first_child;

		for( child_node_iterator = 0;
		     child_node_iterator < index_tree_node->number_of_child_nodes;
		     child_node_iterator++ )
		{
			if( child_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for index node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			if( libpff_index_read_tree(
			     child_node,
			     file_io_handle,
			     ( (libpff_index_values_t *) child_node->value )->file_offset,
			     file_type,
			     expected_index_type,
			     ( (libpff_index_values_t *) child_node->value )->back_pointer,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
		 		 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read index node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			child_node = child_node->next;
		}
	}
	return( 1 );
}

/* Retrieves the tree node of an index node
 * Returns 1 if successful, 0 if the index node was not found or -1 on error
 */
int libpff_index_get_tree_node(
     libpff_tree_node_t *index_tree_node,
     uint64_t index_identifier,
     libpff_tree_node_t **result_index_tree_node,
     liberror_error_t **error )
{
	libpff_tree_node_t *child_index_tree_node = NULL;
	static char *function                     = "libpff_index_get_tree_node";
	int child_node_iterator                   = 0;
	int result                                = 0;

	if( index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index tree node.",
		 function );

		return( -1 );
	}
	if( result_index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid result index tree node.",
		 function );

		return( -1 );
	}
	if( *result_index_tree_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: result index tree node already set.",
		 function );

		return( -1 );
	}
	/* A branch node contains the index identifier of its first leaf node
	 */
	if( index_tree_node->number_of_child_nodes > 0 )
	{
		child_index_tree_node = index_tree_node->first_child;

		for( child_node_iterator = 0;
		     child_node_iterator < index_tree_node->number_of_child_nodes;
		     child_node_iterator++ )
		{
			if( child_index_tree_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			/* Check if the identifier is in the current child node
			 */
			if( ( ( child_index_tree_node->next != NULL )
			  && ( child_index_tree_node->next->value != NULL )
			  && ( index_identifier < ( (libpff_index_values_t *) child_index_tree_node->next->value )->identifier ) )
			 || ( index_tree_node->last_child == child_index_tree_node ) )
			{
				result = libpff_index_get_tree_node(
				          child_index_tree_node,
				          index_identifier,
				          result_index_tree_node,
				          error );

				if( result == -1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to traverse child node: %d.",
					 function,	
					 child_node_iterator );

					return( -1 );
				}
				else if( result == 1 )
				{
					break;
				}
			}
			child_index_tree_node = child_index_tree_node->next;
		}
	}
	else if( index_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid index tree node - missing value.",
		 function );

		return( -1 );
	}
	else if( ( (libpff_index_values_t *) index_tree_node->value )->identifier == index_identifier )
	{
		*result_index_tree_node = index_tree_node;

		return( 1 );
	}
	return( result );
}

/* Retrieves the reference of the index values of an index node
 * Returns 1 if successful, 0 if the index node was not found or -1 on error
 */
int libpff_index_get_index_values(
     libpff_tree_node_t *index_tree_node,
     uint64_t index_identifier,
     libpff_index_values_t **index_values,
     liberror_error_t **error )
{
	libpff_tree_node_t *result_index_tree_node = NULL;
	static char *function                      = "libpff_index_get_index_values";
	int result                                 = 0;

	if( index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index tree node.",
		 function );

		return( -1 );
	}
	if( index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid index values.",
		 function );

		return( -1 );
	}
	if( *index_values != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: index values already set.",
		 function );

		return( -1 );
	}
	result = libpff_index_get_tree_node(
	          index_tree_node,
	          index_identifier,
	          &result_index_tree_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to find index tree node: %" PRIu64 ".",
		 function,
		 index_identifier );

		return( -1 );
	}
	else if( result == 1 )
	{
		if( result_index_tree_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid result index tree node.",
			 function );

			return( -1 );
		}
		*index_values = (libpff_index_values_t *) result_index_tree_node->value;
	}
	return( result );
}

