/*
 * libpff Input/Output (IO) handle
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_allocation_table.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_index.h"
#include "libpff_item_descriptor.h"
#include "libpff_item_tree.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"
#include "libpff_libfmapi.h"
#include "libpff_local_descriptors.h"
#include "libpff_offset_list.h"
#include "libpff_tree_type.h"

#include "pff_file_header.h"

const uint8_t pff_file_signature[ 4 ] = { 0x21, 0x42, 0x44, 0x4e };

/* Initialize an io handle
 * Make sure the value io_handle is pointing to is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_initialize(
     libpff_io_handle_t **io_handle,
     liberror_error_t **error )
{
	static char *function = "libpff_io_handle_initialize";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( *io_handle == NULL )
	{
		*io_handle = (libpff_io_handle_t *) memory_allocate(
		                                     sizeof( libpff_io_handle_t ) );

		if( *io_handle == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create io handle.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *io_handle,
		     0,
		     sizeof( libpff_io_handle_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear file.",
			 function );

			memory_free(
			 *io_handle );

			*io_handle = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees an exisisting io handle
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_free(
     libpff_io_handle_t **io_handle,
     liberror_error_t **error )
{
	static char *function = "libpff_io_handle_free";
	int result            = 1;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( *io_handle != NULL )
	{
		if( ( *io_handle )->descriptor_index_root_node != NULL )
		{
			if( libpff_tree_node_free(
			     &( ( *io_handle )->descriptor_index_root_node ),
			     &libpff_index_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free descriptor index .",
				 function );

				result = -1;
			}
		}
		if( ( *io_handle )->offset_index_root_node != NULL )
		{
			if( libpff_tree_node_free(
			     &( ( *io_handle )->offset_index_root_node ),
			     &libpff_index_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free offset index .",
				 function );

				result = -1;
			}
		}
		if( ( *io_handle )->recovered_descriptor_index_root_node != NULL )
		{
			if( libpff_tree_node_free(
			     &( ( *io_handle )->recovered_descriptor_index_root_node ),
			     &libpff_index_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free recovered descriptor index .",
				 function );

				result = -1;
			}
		}
		if( ( *io_handle )->recovered_offset_index_root_node != NULL )
		{
			if( libpff_tree_node_free(
			     &( ( *io_handle )->recovered_offset_index_root_node ),
			     &libpff_index_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free recovered offset index .",
				 function );

				result = -1;
			}
		}
		memory_free(
		 *io_handle );

		*io_handle = NULL;
	}
	return( result );
}

/* Reads the file header
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_read_file_header(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t *offset_index_root_node_offset,
     uint64_t *offset_index_back_pointer,
     off64_t *descriptor_index_root_node_offset,
     uint64_t *descriptor_index_back_pointer,
     int *file_content_type,
     liberror_error_t **error )
{
	uint8_t *file_header      = NULL;
	uint8_t *file_header_data = NULL;
	static char *function     = "libpff_io_handle_read_file_header";
	size_t data_size          = 564;
	size_t read_size          = 564;
	ssize_t read_count        = 0;
	uint32_t calculated_crc   = 0;
	uint32_t stored_crc       = 0;
	uint16_t content_type     = 0;
	uint16_t data_version     = 0;
	uint8_t sentinal          = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint8_t *value_data       = 0;
	uint64_t value_64bit      = 0;
	uint32_t value_32bit      = 0;
	uint16_t value_16bit      = 0;
	int value_iterator        = 0;
#endif

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( offset_index_root_node_offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid offset index root node offset.",
		 function );

		return( -1 );
	}
	if( offset_index_back_pointer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid offset index back pointer.",
		 function );

		return( -1 );
	}
	if( descriptor_index_root_node_offset == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid descriptor index root node offset.",
		 function );

		return( -1 );
	}
	if( descriptor_index_back_pointer == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid descriptor index back pointer.",
		 function );

		return( -1 );
	}
	if( file_content_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid file content type.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading file header at offset: 0 (0x00000000)\n",
		 function );
	}
#endif

	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     0,
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek file header offset: 0.",
		 function );

		return( -1 );
	}
	file_header = (uint8_t *) memory_allocate(
	                           read_size );

	if( file_header == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create file header.",
		 function );

		return( -1 );
	}
	read_count = libbfio_handle_read(
	              file_io_handle,
	              file_header,
	              read_size,
	              error );

	if( read_count != (ssize_t) read_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read file header data.",
		 function );

		memory_free(
		 file_header );

		return( -1 );
	}
	file_header_data = &( file_header[ sizeof( pff_file_header_t ) ] );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: file header:\n",
		 function );
		libnotify_print_data(
		 file_header,
		 sizeof( pff_file_header_t ) );
	}
#endif

	if( memory_compare(
	     ( (pff_file_header_t *) file_header )->signature,
	     pff_file_signature,
	     4 ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: invalid file signature.",
		 function );

		memory_free(
		 file_header );

		return( -1 );
	}
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_file_header_t *) file_header )->crc,
	 stored_crc );

	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_file_header_t *) file_header )->content_type,
	 content_type );

	if( content_type == PFF_FILE_HEADER_CONTENT_TYPE_PAB )
	{
		*file_content_type = LIBPFF_FILE_CONTENT_TYPE_PAB;
	}
	else if( content_type == PFF_FILE_HEADER_CONTENT_TYPE_PST )
	{
		*file_content_type = LIBPFF_FILE_CONTENT_TYPE_PST;
	}
	else if( content_type == PFF_FILE_HEADER_CONTENT_TYPE_OST )
	{
		*file_content_type = LIBPFF_FILE_CONTENT_TYPE_OST;
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsuported content type: 0x%04" PRIx16 ".",
		 function,
		 content_type );

		return( -1 );
	}
	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_file_header_t *) file_header )->data_version,
	 data_version );

	if( data_version <= 0x000f )
	{
		data_size            = sizeof( pff_file_header_data_32bit_t );
		io_handle->file_type = LIBPFF_FILE_TYPE_32BIT;
	}
	else if( data_version >= 0x0015 )
	{
		data_size            = sizeof( pff_file_header_data_64bit_t );
		io_handle->file_type = LIBPFF_FILE_TYPE_64BIT;
	}
	else
	{
		if( ( ( (pff_file_header_data_32bit_t *) file_header_data )->sentinal == 0x80 )
		 && ( ( (pff_file_header_data_64bit_t *) file_header_data )->sentinal != 0x80 ) )
		{
#if defined( HAVE_VERBOSE_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: detected 32-bit file type for unsuported data version: 0x%04" PRIx16 ".\n",
				 function,
				 data_version );
			}
#endif
			data_size            = sizeof( pff_file_header_data_32bit_t );
			io_handle->file_type = LIBPFF_FILE_TYPE_32BIT;
		}
		else if( ( ( (pff_file_header_data_32bit_t *) file_header_data )->sentinal != 0x80 )
		      && ( ( (pff_file_header_data_64bit_t *) file_header_data )->sentinal == 0x80 ) )
		{
#if defined( HAVE_VERBOSE_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: detected 64-bit file type for unsuported data version: 0x%04" PRIx16 ".\n",
				 function,
				 data_version );
			}
#endif
			data_size            = sizeof( pff_file_header_data_64bit_t );
			io_handle->file_type = LIBPFF_FILE_TYPE_64BIT;
		}
#if defined( HAVE_VERBOSE_OUTPUT )
		else if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: unsuported data version: 0x%04" PRIx16 ".\n",
			 function,
			 data_version );
		}
#endif
	}
	if( ( io_handle->file_type != LIBPFF_FILE_TYPE_32BIT )
	 && ( io_handle->file_type != LIBPFF_FILE_TYPE_64BIT ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unable to determine file type.",
		 function );

		memory_free(
		 file_header );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: signature\t\t: %c%c%c%c\n",
		 function,
		 ( (pff_file_header_t *) file_header )->signature[ 0 ],
		 ( (pff_file_header_t *) file_header )->signature[ 1 ],
		 ( (pff_file_header_t *) file_header )->signature[ 2 ],
		 ( (pff_file_header_t *) file_header )->signature[ 3 ] );

		libnotify_printf(
		 "%s: crc\t\t\t: 0x%08" PRIx64 "\n",
		 function,
		 stored_crc );

		libnotify_printf(
		 "%s: content type\t\t: %c%c\n",
		 function,
		 ( (pff_file_header_t *) file_header )->content_type[ 0 ],
		 ( (pff_file_header_t *) file_header )->content_type[ 1 ] );

		libnotify_printf(
		 "%s: data version\t\t: %" PRIu16 "\n",
		 function,
		 data_version );
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_file_header_t *) file_header )->content_version,
		 value_16bit );

		libnotify_printf(
		 "%s: content version\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		libnotify_printf(
		 "%s: creation platform\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_file_header_t *) file_header )->creation_platform );

		libnotify_printf(
		 "%s: access platform\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_file_header_t *) file_header )->access_platform );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_t *) file_header )->unknown1,
		 value_32bit );
		libnotify_printf(
		 "%s: unknown1\t\t: 0x%08" PRIx32 " (%" PRIu32 ")\n",
		 function,
		 value_32bit,
		 value_32bit );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_t *) file_header )->unknown2,
		 value_32bit );
		libnotify_printf(
		 "%s: unknown2\t\t: 0x%08" PRIx32 " (%" PRIu32 ")\n",
		 function,
		 value_32bit,
		 value_32bit );

		libnotify_printf(
		 "\n" );
	}
#endif

	if( libfmapi_crc_weak_calculate(
	     0,
	     &( file_header[ 8 ] ),
	     471,
	     &calculated_crc,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unable to calculate crc.",
		 function );

		memory_free(
		 file_header );

		return( -1 );
	}
	if( stored_crc != calculated_crc )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_INPUT,
		 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
		 "%s: mismatch in file header crc ( %" PRIu32 " != %" PRIu32 " ).",
		 function,
		 stored_crc,
		 calculated_crc );

		memory_free(
		 file_header );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: file header data:\n",
		 function );
		libnotify_print_data(
		 file_header_data,
		 data_size );
	}
#endif

	if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_32bit_t *) file_header_data )->file_size,
		 io_handle->file_size );

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_32bit_t *) file_header_data )->descriptor_index_back_pointer,
		 *descriptor_index_back_pointer );
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_32bit_t *) file_header_data )->descriptor_index_root_node_offset,
		 *descriptor_index_root_node_offset );
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_32bit_t *) file_header_data )->offset_index_back_pointer,
		 *offset_index_back_pointer );
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_32bit_t *) file_header_data )->offset_index_root_node_offset,
		 *offset_index_root_node_offset );

		sentinal                   = ( (pff_file_header_data_32bit_t *) file_header_data )->sentinal;
		io_handle->encryption_type = ( (pff_file_header_data_32bit_t *) file_header_data )->encryption_type;
	}
	else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->file_size,
		 io_handle->file_size );

		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->descriptor_index_back_pointer,
		 *descriptor_index_back_pointer );
		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->descriptor_index_root_node_offset,
		 *descriptor_index_root_node_offset );
		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->offset_index_back_pointer,
		 *offset_index_back_pointer );
		byte_stream_copy_to_uint64_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->offset_index_root_node_offset,
		 *offset_index_root_node_offset );

		sentinal                   = ( (pff_file_header_data_64bit_t *) file_header_data )->sentinal;
		io_handle->encryption_type = ( (pff_file_header_data_64bit_t *) file_header_data )->encryption_type;

		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_file_header_data_64bit_t *) file_header_data )->crc,
		 stored_crc );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->next_index_pointer,
			 value_32bit );
			libnotify_printf(
			 "%s: next index pointer\t\t\t: %" PRIu32 "\n",
			 function,
			 value_32bit );

			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->next_index_back_pointer,
			 value_32bit );
			libnotify_printf(
			 "%s: next index back pointer\t\t: %" PRIu32 "\n",
			 function,
			 value_32bit );

			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->seed_value,
			 value_32bit );
			libnotify_printf(
			 "%s: seed value\t\t\t\t: 0x%08" PRIx32 "\n",
			 function,
			 value_32bit );

			value_data = ( (pff_file_header_data_32bit_t *) file_header_data )->descriptor_index_high_water_marks;

			for( value_iterator = 0;
			     value_iterator < 32;
			     value_iterator++ )
			{
				byte_stream_copy_to_uint32_little_endian(
				 value_data,
				 value_32bit );
				libnotify_printf(
				 "%s: high water mark: %02d value\t\t: %" PRIu32 "\n",
				 function,
				 value_iterator,
				 value_32bit );

				value_data += 4;
			}
			libnotify_printf(
			 "\n" );
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: unknown3:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown3,
			 8 );

			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->next_index_back_pointer,
			 value_64bit );
			libnotify_printf(
			 "%s: next index back pointer\t\t: %" PRIu64 "\n",
			 function,
			 value_64bit );

			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->seed_value,
			 value_32bit );
			libnotify_printf(
			 "%s: seed value\t\t\t\t: 0x%08" PRIx32 "\n",
			 function,
			 value_32bit );

			value_data = ( (pff_file_header_data_64bit_t *) file_header_data )->descriptor_index_high_water_marks;

			for( value_iterator = 0;
			     value_iterator < 32;
			     value_iterator++ )
			{
				byte_stream_copy_to_uint32_little_endian(
				 value_data,
				 value_32bit );
				libnotify_printf(
				 "%s: high water mark: %02d value\t\t: %" PRIu32 "\n",
				 function,
				 value_iterator,
				 value_32bit );

				value_data += 4;
			}
			libnotify_printf(
			 "\n" );
			libnotify_printf(
			 "%s: unknown4:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown4,
			 8 );
		}
		libnotify_printf(
		 "%s: file header data root:\n",
		 function );

		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			libnotify_printf(
			 "%s: unknown5:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown5,
			 4 );
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: unknown5:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown5,
			 4 );
		}
		libnotify_printf(
		 "%s: file size\t\t\t\t: %" PRIu64 "\n",
		 function,
		 io_handle->file_size );

		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->last_data_allocation_table_offset,
			 value_32bit );
			libnotify_printf(
			 "%s: last data allocation table offset\t: %" PRIu32 "\n",
			 function,
			 value_32bit );

			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->total_available_data_size,
			 value_32bit );
			libnotify_printf(
			 "%s: total available data size\t\t: %" PRIu32 "\n",
			 function,
			 value_32bit );

			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->total_available_page_size,
			 value_32bit );
			libnotify_printf(
			 "%s: total available page size\t\t: %" PRIu32 "\n",
			 function,
			 value_32bit );
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->last_data_allocation_table_offset,
			 value_64bit );
			libnotify_printf(
			 "%s: last data allocation table offset\t: %" PRIu64 "\n",
			 function,
			 value_64bit );

			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->total_available_data_size,
			 value_64bit );
			libnotify_printf(
			 "%s: total available data size\t\t: %" PRIu64 "\n",
			 function,
			 value_64bit );

			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->total_available_page_size,
			 value_64bit );
			libnotify_printf(
			 "%s: total available page size\t\t: %" PRIu64 "\n",
			 function,
			 value_64bit );
		}
		libnotify_printf(
		 "%s: descriptor index root node offset\t: %" PRIu64 "\n",
		 function,
		 *descriptor_index_root_node_offset );

		libnotify_printf(
		 "%s: descriptor index back pointer\t: %" PRIu64 "\n",
		 function,
		 *descriptor_index_back_pointer );

		libnotify_printf(
		 "%s: offset index root node offset\t: %" PRIu64 "\n",
		 function,
		 *offset_index_root_node_offset );

		libnotify_printf(
		 "%s: offset index back pointer\t\t: %" PRIu64 "\n",
		 function,
		 *offset_index_back_pointer );

		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			libnotify_printf(
			 "%s: allocation table validation type\t: 0x%02" PRIx8 "\n",
			 function,
			 ( (pff_file_header_data_32bit_t *) file_header_data )->allocation_table_validation_type );

			libnotify_printf(
			 "%s: unknown6:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown6,
			 3 );

			libnotify_printf(
			 "%s: initial data free map:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->initial_data_free_map,
			 128 );

			libnotify_printf(
			 "%s: initial page free map:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->initial_page_free_map,
			 128 );
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: allocation table validation type\t: 0x%02" PRIx8 "\n",
			 function,
			 ( (pff_file_header_data_64bit_t *) file_header_data )->allocation_table_validation_type );

			libnotify_printf(
			 "%s: unknown6:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown6,
			 3 );

			libnotify_printf(
			 "%s: unknown7:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown7,
			 3 );

			libnotify_printf(
			 "%s: initial data free map:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->initial_data_free_map,
			 128 );

			libnotify_printf(
			 "%s: initial page free map:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->initial_page_free_map,
			 128 );
		}
		libnotify_printf(
		 "%s: sentinal\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 sentinal );

		libnotify_printf(
		 "%s: encryption type\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 io_handle->encryption_type );

		if( io_handle->file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			libnotify_printf(
			 "%s: unknown8:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown8,
			 2 );

			libnotify_printf(
			 "%s: unknown9:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown9,
			 8 );

			libnotify_printf(
			 "%s: unknown10:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown10,
			 4 );

			libnotify_printf(
			 "%s: unknown11:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown11,
			 3 );

			libnotify_printf(
			 "%s: unknown12:\n",
			 function );
			libnotify_print_data(
			 &( ( (pff_file_header_data_32bit_t *) file_header_data )->unknown12 ),
			 1 );

			libnotify_printf(
			 "%s: unknown13:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_32bit_t *) file_header_data )->unknown13,
			 32 );
		}
		else if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: unknown8:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown8,
			 2 );

			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->next_index_pointer,
			 value_32bit );
			libnotify_printf(
			 "%s: next index pointer\t\t\t: %" PRIu64 "\n",
			 function,
			 value_64bit );

			libnotify_printf(
			 "%s: crc\t\t\t\t\t: 0x%08" PRIx64 "\n",
			 function,
			 stored_crc );

			libnotify_printf(
			 "%s: unknown11:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown11,
			 3 );

			libnotify_printf(
			 "%s: unknown12:\n",
			 function );
			libnotify_print_data(
			 &( ( (pff_file_header_data_32bit_t *) file_header_data )->unknown12 ),
			 1 );

			libnotify_printf(
			 "%s: unknown13:\n",
			 function );
			libnotify_print_data(
			 ( (pff_file_header_data_64bit_t *) file_header_data )->unknown13,
			 32 );
		}
	}
#endif
	if( io_handle->file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		if( libfmapi_crc_weak_calculate(
		     0,
		     &( file_header[ 8 ] ),
		     516,
		     &calculated_crc,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unable to calculate crc.",
			 function );

			memory_free(
			 file_header );

			return( -1 );
		}
		if( stored_crc != calculated_crc )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_INPUT,
			 LIBERROR_INPUT_ERROR_CRC_MISMATCH,
			 "%s: mismatch in file header crc ( %" PRIu32 " != %" PRIu32 " ).",
			 function,
			 stored_crc,
			 calculated_crc );

			memory_free(
			 file_header );

			return( -1 );
		}
	}
	memory_free(
	 file_header );

	if( ( io_handle->encryption_type != LIBPFF_ENCRYPTION_TYPE_NONE )
	 && ( io_handle->encryption_type != LIBPFF_ENCRYPTION_TYPE_COMPRESSIBLE )
	 && ( io_handle->encryption_type != LIBPFF_ENCRYPTION_TYPE_HIGH ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported encryption type: 0x%02x",
		 function,
		 io_handle->encryption_type );

		return( -1 );
	}
	return( 1 );
}

/* Reads the unallocated data blocks
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_read_unallocated_data_blocks(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     liberror_error_t **error )
{
	static char *function           = "libpff_io_handle_read_unallocated_data_blocks";
	off64_t allocation_table_offset = 0;
	size64_t allocation_block_size  = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	allocation_table_offset = 0x4400;
	allocation_block_size   = 496 * 512;

	while( allocation_table_offset < (off64_t) io_handle->file_size )
	{
		if( libpff_allocation_table_read(
		     unallocated_data_block_list,
		     file_io_handle,
		     allocation_table_offset,
		     (int) io_handle->file_type,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read allocation table at offset: %" PRIu64 ".",
			 function,
			 allocation_table_offset );

			return( -1 );
		}
		allocation_table_offset += allocation_block_size;
	}
	return( 1 );
}

/* Reads the unallocated page blocks
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_read_unallocated_page_blocks(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_page_block_list,
     liberror_error_t **error )
{
	static char *function           = "libpff_io_handle_read_unallocated_page_blocks";
	off64_t allocation_table_offset = 0;
	size64_t allocation_block_size  = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	allocation_table_offset = 0x4600;
	allocation_block_size   = 496 * 4096;

	while( allocation_table_offset < (off64_t) io_handle->file_size )
	{
		if( libpff_allocation_table_read(
		     unallocated_page_block_list,
		     file_io_handle,
		     allocation_table_offset,
		     (int) io_handle->file_type,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read allocation table at offset: %" PRIu64 ".",
			 function,
			 allocation_table_offset );

			return( -1 );
		}
		allocation_table_offset += allocation_block_size;
	}
	return( 1 );
}

/* Reads the descriptor index
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_read_descriptor_index(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t descriptor_index_root_node_offset,
     uint64_t descriptor_index_back_pointer,
     liberror_error_t **error )
{
	static char *function = "libpff_io_handle_read_descriptor_index";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( io_handle->descriptor_index_root_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid io handle - descriptor index root node already set.",
		 function );

		return( -1 );
	}
	if( libpff_tree_node_initialize(
	     &( io_handle->descriptor_index_root_node ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create descriptor index root node.",
		 function );

		return( -1 );
	}
	if( libpff_index_read_tree(
	     io_handle->descriptor_index_root_node,
	     file_io_handle,
	     descriptor_index_root_node_offset,
	     (int) io_handle->file_type,
	     LIBPFF_INDEX_TYPE_DESCRIPTOR,
	     descriptor_index_back_pointer,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read descriptor index.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the (file) offset index
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_read_offset_index(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t offset_index_root_node_offset,
     uint64_t offset_index_back_pointer,
     liberror_error_t **error )
{
	static char *function = "libpff_io_handle_read_offset_index";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( io_handle->offset_index_root_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid io handle - offset index root node already set.",
		 function );

		return( -1 );
	}
	if( libpff_tree_node_initialize(
	     &( io_handle->offset_index_root_node ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create offset index root node.",
		 function );

		return( -1 );
	}
	if( libpff_index_read_tree(
	     io_handle->offset_index_root_node,
	     file_io_handle,
	     offset_index_root_node_offset,
	     (int) io_handle->file_type,
	     LIBPFF_INDEX_TYPE_OFFSET,
	     offset_index_back_pointer,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read offset index.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Scans for recoverable items
 * By default only the unallocated space is checked for recoverable items
 * Returns 1 if successful or -1 on error
 */
int libpff_io_handle_recover_items(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     libpff_list_t *unallocated_page_block_list,
     libpff_list_t *recovered_item_list,
     uint8_t recovery_flags,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor     = NULL;
	libpff_index_values_t *child_index_values     = NULL;
	libpff_index_values_t *recovered_index_values = NULL;
	libpff_list_element_t *list_element           = NULL;
	libpff_tree_node_t *child_node                = NULL;
	libpff_tree_node_t *item_tree_node            = NULL;
	libpff_tree_node_t *local_descriptors         = NULL;
	libpff_tree_node_t *recovered_index_tree_node = NULL;
	libpff_tree_node_t *result_node               = NULL;
	static char *function                         = "libpff_io_handle_recover_items";
	off64_t block_offset                          = 0;
	size64_t block_size                           = 0;
	uint8_t supported_recovery_flags              = 0;
	int number_of_child_nodes                     = 0;
	int number_of_list_elements                   = 0;
	int child_iterator                            = 0;
	int recoverable                               = 0;
	int result                                    = 0;
	int unallocated_block_iterator                = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( recovered_item_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid recovered item list.",
		 function );

		return( -1 );
	}
	supported_recovery_flags = LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA;

	if( ( recovery_flags & ~( supported_recovery_flags ) ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported recovery flags.",
		 function );

		return( -1 );
	}
	if( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) == 0 )
	{
		if( unallocated_page_block_list == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid unallocated page block list.",
			 function );

			return( -1 );
		}
	}
	/* Create the recovered descriptor index root node
	 */
	if( io_handle->recovered_descriptor_index_root_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid io handle - recovered descriptor index root node already set.",
		 function );

		return( -1 );
	}
	if( libpff_index_node_initialize(
	     &( io_handle->recovered_descriptor_index_root_node ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create recovered descriptor index root node.",
		 function );

		return( -1 );
	}
	/* Create the recovered offset index root node
	 */
	if( io_handle->recovered_offset_index_root_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid io handle - recovered offset index root node already set.",
		 function );

		return( -1 );
	}
	if( libpff_index_node_initialize(
	     &( io_handle->recovered_offset_index_root_node ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create recovered offset index root node.",
		 function );

		return( -1 );
	}
	if( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) == 0 )
	{
		if( libpff_list_get_number_of_elements(
		     unallocated_page_block_list,
		     &number_of_list_elements,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of list elements.",
			 function );

			return( -1 );
		}
		list_element = unallocated_page_block_list->first;
	}
	if( ( number_of_list_elements > 0 )
	 || ( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) != 0 ) )
	{
		/* Scan the unallocated page block list or all blocks
		 */
		while( block_offset < (off64_t) io_handle->file_size )
		{
			if( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) == 0 )
			{
				if( list_element == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
					 "%s: corruption detected for list element: %d.",
					 function,
					 unallocated_block_iterator );

					return( -1 );
				}
				if( list_element->value == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
					 "%s: missing offset values for list element: %d.",
					 function,
					 unallocated_block_iterator );

					return( -1 );
				}
				block_offset = ( (libpff_offset_list_values_t *) list_element->value )->offset;
				block_size   = ( (libpff_offset_list_values_t *) list_element->value )->size;
				list_element = list_element->next;

				unallocated_block_iterator++;
			}
			else
			{
				block_size = 512;
			}
			/* The index nodes are 512 bytes of size
			 */
			while( block_size >= 512 )
			{
				/* The index nodes are 512 bytes aligned
				 */
				if( ( block_offset % 512 ) == 0 )
				{
					/* Scan the data in the index block
					 */
					result = libpff_io_handle_recover_index_nodes(
						  io_handle,
						  file_io_handle,
						  unallocated_data_block_list,
						  recovery_flags,
						  block_offset,
						  error );

					if( result == -1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
						 "%s: error while recovering index nodes at offset: %" PRIu64 ".\n",
						 function,
						 block_offset );

						return( -1 );
					}
					block_size   -= 512;
					block_offset += 512;
				}
				else
				{
					block_size   -= block_offset % 512;
					block_offset  = ( ( block_offset / 512 ) + 1 ) * 512;
				}
			}
			if( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) == 0 )
			{
				if( unallocated_block_iterator >= unallocated_page_block_list->number_of_elements )
				{
					break;
				}
			}
		}
	}
	/* TODO
	 * handle scanning without index data
	 * what about 'encryption' ?
	 */

	/* Check if the list and data offset index values exists for the recovered descriptor index nodes
	 */
	number_of_child_nodes = io_handle->recovered_descriptor_index_root_node->number_of_child_nodes;
	child_node            = io_handle->recovered_descriptor_index_root_node->first_child;

	for( child_iterator = 0;
	     child_iterator < number_of_child_nodes;
	     child_iterator++ )
	{
		if( child_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: corruption detected for index node: %d.",
			 function,
			 child_iterator );

			libpff_tree_node_free(
			 &recovered_index_tree_node,
			 &libpff_index_values_free,
			 NULL );

			return( -1 );
		}
		if( child_node->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid index node: %d - missing value.",
			 function,
			 child_iterator );

			libpff_tree_node_free(
			 &recovered_index_tree_node,
			 &libpff_index_values_free,
			 NULL );

			return( -1 );
		}
		child_index_values = (libpff_index_values_t *) child_node->value;

#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: identifier: %" PRIu64 ", data: %" PRIu64 ", local descriptors: %" PRIu64 ", parent: %" PRIu32 "\n",
			 function,
			 child_index_values->identifier,
			 child_index_values->data_identifier,
			 child_index_values->local_descriptors_identifier,
			 child_index_values->parent_identifier );
		}
#endif

		recoverable = 1;

		/* Check if the local descriptors index identifier is also recoverable
		 */
		result_node = NULL;

		result = libpff_index_get_tree_node(
			  io_handle->recovered_offset_index_root_node,
			  child_index_values->local_descriptors_identifier,
			  &result_node,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: error retrieving recovered local descriptors index identifier: %" PRIu64 ".",
			 function,
			 child_index_values->local_descriptors_identifier );

			libpff_tree_node_free(
			 &recovered_index_tree_node,
			 &libpff_index_values_free,
			 NULL );

			return( -1 );
		}
		else if( result == 0 )
		{
#if defined( HAVE_VERBOSE_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: missing recovered local descriptors index identifier: %" PRIu64 ".\n",
				 function,
				 child_index_values->local_descriptors_identifier );
			}
#endif

			recoverable = 0;
		}
		if( recoverable != 0 )
		{
			/* Check if the data identifier is also recoverable
			 */
			result_node = NULL;

			result = libpff_index_get_tree_node(
				  io_handle->recovered_offset_index_root_node,
				  child_index_values->data_identifier,
				  &result_node,
			          error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: error retrieving recovered data index identifier: %" PRIu64 ".",
				 function,
				 child_index_values->data_identifier );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			else if( result == 0 )
			{
#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: missing recovered data index identifier: %" PRIu64 ".\n",
					 function,
					 child_index_values->data_identifier );
				}
#endif

				recoverable = 0;
			}
		}
		if( recoverable != 0 )
		{
			/* Check if local descriptors are readable
			 */
			result = libpff_descriptor_read_local_descriptors(
			          child_index_values->local_descriptors_identifier,
			          1,
			          io_handle,
			          file_io_handle,
			          &local_descriptors,
			          error );

			if( result != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read local descriptors index identifier: %" PRIu64 ".",
				 function,
				 child_index_values->local_descriptors_identifier );

#if defined( HAVE_DEBUG_OUTPUT )
				if( ( libnotify_verbose != 0 )
				 && ( error != NULL )
				 && ( *error != NULL ) )
				{
					libnotify_print_error_backtrace(
					 *error );
				}
#endif
				liberror_error_free(
				 error );

				recoverable = 0;
			}
			else
			{
				if( libpff_tree_node_free(
				     &local_descriptors,
				     &libpff_local_descriptor_values_free,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
					 "%s: unable to free local descriptors.",
					 function );

					libpff_tree_node_free(
					 &recovered_index_tree_node,
					 &libpff_index_values_free,
					 NULL );

					return( -1 );
				}
			}
		}
		if( recoverable == 0 )
		{
			recovered_index_tree_node = child_node;
		}
		else
		{
			if( child_index_values->identifier > (uint64_t) UINT32_MAX )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
				 "%s: invalid descriptor index identifier value exceeds maximum.",
				 function );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			/* Create a new item descriptor
			 */
			if( libpff_item_descriptor_initialize(
			     &item_descriptor,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create item descriptor.",
				 function );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			item_descriptor->descriptor_identifier = (uint32_t) child_index_values->identifier;
			item_descriptor->recovered             = 1;

			/* Create a new tree node with item tree values
			 */
			if( libpff_tree_node_initialize(
			     &item_tree_node,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create item tree node.",
				 function );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );
				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			item_tree_node->value = (intptr_t *) item_descriptor;

			item_descriptor = NULL;

			result = libpff_list_append_value(
			          recovered_item_list,
			          (intptr_t *) item_tree_node,
			          error );

			/* Free the node if it could not be appended
			 */
			if( result != 1 )
			{
				libpff_tree_node_free(
				 &item_tree_node,
				 &libpff_item_descriptor_free,
				 NULL );
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to append tree node in recovered item list.",
				 function );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			item_tree_node = NULL;
		}
		child_node = child_node->next;

		/* Remove incomplete recoverable descriptors
		 */
		if( recovered_index_tree_node != NULL )
		{
			if( recovered_index_tree_node->value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid recovered index tree node - missing value.",
				 function );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			recovered_index_values = (libpff_index_values_t *) recovered_index_tree_node->value;

#if defined( HAVE_VERBOSE_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: incomplete recoverable descriptor index identifier: %" PRIu64 ".\n",
				 function,
				 recovered_index_values->identifier );
			}
#endif

			if( libpff_tree_node_remove_node(
			     io_handle->recovered_descriptor_index_root_node,
			     recovered_index_tree_node,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_REMOVE_FAILED,
				 "%s: unable to remove descriptor index identifier: %" PRIu64 ".",
				 function,
				 recovered_index_values->identifier );

				libpff_tree_node_free(
				 &recovered_index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			if( libpff_tree_node_free(
			     &recovered_index_tree_node,
			     &libpff_index_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free recovered index node.",
				 function );

				return( -1 );
			}
			recovered_index_tree_node = NULL;
		}
	}

	/* TODO
	 * link recovered descriptors to parent? and add to item hierarchy?
	 * what about remaining deleted offset index identifiers?
	 * what about remaining unallocated data offsets?
	 */

	return( 1 );
}

/* Scans for recoverable descriptor index nodes
 * Returns 1 if successful, 0 if no index nodes could be found or -1 on error
 */
int libpff_io_handle_recover_index_nodes(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     uint8_t recovery_flags,
     size64_t node_offset,
     liberror_error_t **error )
{
	libpff_index_values_t *child_index_values     = NULL;
	libpff_index_values_t *recovered_index_values = NULL;
	libpff_tree_node_t *child_node                = NULL;
	libpff_tree_node_t *index_root_node           = NULL;
	libpff_tree_node_t *index_tree_node           = NULL;
	libpff_tree_node_t *recovered_index_root_node = NULL;
	libpff_tree_node_t *recovered_index_tree_node = NULL;
	const char *index_string                      = NULL;
	static char *function                         = "libpff_io_handle_recover_index_nodes";
	uint64_t back_pointer                         = 0;
	uint8_t index_type                            = 0;
	uint8_t node_level                            = 0;
	uint8_t node_is_recoverable                   = 0;
	int number_of_child_nodes                     = 0;
	int child_iterator                            = 0;
	int result                                    = 0;
	int index_nodes_recovered                     = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( libpff_tree_node_initialize(
	     &index_tree_node,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create index tree node.",
		 function );

		return( -1 );
	}
	if( libpff_index_read_node(
	     index_tree_node,
	     file_io_handle,
	     node_offset,
	     (int) io_handle->file_type,
	     &node_level,
	     &index_type,
	     &back_pointer,
	     error ) != 1 )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( ( libnotify_verbose != 0 )
		 && ( error != NULL )
		 && ( *error != NULL ) )
		{
			libnotify_print_error_backtrace(
			 *error );
		}
#endif
		liberror_error_free(
		 error );

		libpff_tree_node_free(
		 &index_tree_node,
		 &libpff_index_values_free,
		 NULL );

		return( 0 );
	}
	if( ( index_type != LIBPFF_INDEX_TYPE_DESCRIPTOR )
	 && ( index_type != LIBPFF_INDEX_TYPE_OFFSET ) )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: unsupported index type: 0x%02" PRIx8 ".\n",
			 function,
			 index_type );
		}
#endif
		index_nodes_recovered = 0;
	}
	else if( node_level != 0 )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: unsupported node level: %" PRIu8 ".\n",
			 function,
			 node_level );
		}
#endif
		index_nodes_recovered = 0;
	}
	else
	{
		if( libpff_tree_node_get_number_of_child_nodes(
		     index_tree_node,
		     &number_of_child_nodes,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of child nodes.",
			 function );

			libpff_tree_node_free(
			 &index_tree_node,
			 &libpff_index_values_free,
			 NULL );

			return( -1 );
		}
		/* Check if the index leaf nodes are recoverable
		 */
		child_node = index_tree_node->first_child;

		for( child_iterator = 0;
		     child_iterator < number_of_child_nodes;
		     child_iterator++ )
		{
			if( child_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for index node: %d.",
				 function,
				 child_iterator );

				libpff_tree_node_free(
				 &index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			if( child_node->value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid index node: %d - missing value.",
				 function,
				 child_iterator );

				libpff_tree_node_free(
				 &index_tree_node,
				 &libpff_index_values_free,
				 NULL );

				return( -1 );
			}
			child_index_values = (libpff_index_values_t *) child_node->value;

			if( index_type == LIBPFF_INDEX_TYPE_DESCRIPTOR )
			{
				index_string              = "descriptor";
				index_root_node           = io_handle->descriptor_index_root_node;
				recovered_index_root_node = io_handle->recovered_descriptor_index_root_node;
			}
			else if( index_type == LIBPFF_INDEX_TYPE_OFFSET )
			{
				index_string              = "offset";
				index_root_node           = io_handle->offset_index_root_node;
				recovered_index_root_node = io_handle->recovered_offset_index_root_node;
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: %s index identifier: %" PRIu64 " does not exist.\n",
				 function,
				 index_string,
				 child_index_values->identifier );
			}
#endif
			node_is_recoverable = 1;

			if( ( ( recovery_flags & LIBPFF_RECOVERY_FLAG_IGNORE_ALLOCATION_DATA ) == 0 )
			 && ( index_type == LIBPFF_INDEX_TYPE_OFFSET ) )
			{
				/* Check if the offset index value is unallocated according to the
				 * unallocated data block list
				 */
				result = libpff_offset_list_range_is_present(
					  unallocated_data_block_list,
					  child_index_values->file_offset,
					  child_index_values->data_size,
					  error );

				if( result == -1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: error determining if offset range is unallocated.",
					 function );

					libpff_tree_node_free(
					 &index_tree_node,
					 &libpff_index_values_free,
					 NULL );

					return( -1 );
				}
				else if( result == 0 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: %s index identifier: %" PRIu64 " refers to allocated range: %" PRIu64 " - %" PRIu64 ".\n",
						 function,
						 index_string,
						 child_index_values->identifier,
						 child_index_values->file_offset,
						 child_index_values->file_offset + child_index_values->data_size );
					}
#endif
					node_is_recoverable = 0;
				}
#if defined( HAVE_DEBUG_OUTPUT )
				else if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: %s index identifier: %" PRIu64 " refers to unallocated range: %" PRIu64 " - %" PRIu64 ".\n",
					 function,
					 index_string,
					 child_index_values->identifier,
					 child_index_values->file_offset,
					 child_index_values->file_offset + child_index_values->data_size );
				}
#endif
			}
			if( node_is_recoverable == 1 )
			{
				recovered_index_tree_node = child_node;
			}
			child_node = child_node->next;

			/* Move the recovered node to the index
			 */
			if( recovered_index_tree_node != NULL )
			{
				if( recovered_index_tree_node->value == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
					 "%s: invalid recovered index tree node - missing value.",
					 function );

					libpff_tree_node_free(
					 &index_tree_node,
					 &libpff_index_values_free,
					 NULL );

					return( -1 );
				}
				recovered_index_values = (libpff_index_values_t *) recovered_index_tree_node->value;

#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: found recoverable %s index identifier: %" PRIu64 ".\n",
					 function,
					 index_string,
					 recovered_index_values->identifier );
				}
#endif
				if( libpff_tree_node_remove_node(
				     index_tree_node,
				     recovered_index_tree_node,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_REMOVE_FAILED,
					 "%s: unable to remove %s index identifier: %" PRIu64 ".",
					 function,
					 index_string,
					 recovered_index_values->identifier );

					libpff_tree_node_free(
					 &index_tree_node,
					 &libpff_index_values_free,
					 NULL );

					return( -1 );
				}
				if( libpff_tree_node_insert_node(
				     recovered_index_root_node,
				     recovered_index_tree_node,
				     &libpff_index_values_compare_allow_duplicate,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
					 "%s: unable to insert %s index identifier: %" PRIu64 ".",
					 function,
					 index_string,
					 recovered_index_values->identifier );

					libpff_tree_node_free(
					 &recovered_index_tree_node,
					 &libpff_index_values_free,
					 NULL );
					libpff_tree_node_free(
					 &index_tree_node,
					 &libpff_index_values_free,
					 NULL );

					return( -1 );
				}
				recovered_index_tree_node = NULL;

				index_nodes_recovered = 1;
			}
		}
	}
	if( libpff_tree_node_free(
	     &index_tree_node,
	     &libpff_index_values_free,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free %s index node.",
		 function,
		 index_string );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif
	return( index_nodes_recovered );
}

