/*
 * libpff Input/Output (IO) handle
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#if !defined( _LIBPFF_IO_HANDLE_H )
#define _LIBPFF_IO_HANDLE_H

#include <common.h>
#include <types.h>

#include <liberror.h>

#include "libpff_libbfio.h"
#include "libpff_list_type.h"
#include "libpff_tree_type.h"

#if defined( __cplusplus )
extern "C" {
#endif

extern const uint8_t pff_file_signature[ 4 ];

typedef struct libpff_io_handle libpff_io_handle_t;

struct libpff_io_handle
{
	/* The descriptor index root tree node
	 */
	libpff_tree_node_t *descriptor_index_root_node;

	/* The offset index root tree node
	 */
	libpff_tree_node_t *offset_index_root_node;

	/* The recovered descriptor index root tree node
	 */
	libpff_tree_node_t *recovered_descriptor_index_root_node;

	/* The recovered offset index root tree node
	 */
	libpff_tree_node_t *recovered_offset_index_root_node;

	/* The file size
	 */
	size64_t file_size;

	/* The encryption type
	 */
	uint8_t encryption_type;

	/* Value to indicate decryption should be forced
	 */
	uint8_t force_decryption;

	/* The file type
	 */
	uint8_t file_type;
};

int libpff_io_handle_initialize(
     libpff_io_handle_t **io_handle,
     liberror_error_t **error );

int libpff_io_handle_free(
     libpff_io_handle_t **io_handle,
     liberror_error_t **error );

int libpff_io_handle_read_file_header(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t *offset_index_root_node_offset,
     uint64_t *offset_index_back_pointer,
     off64_t *descriptor_index_root_node_offset,
     uint64_t *descriptor_index_back_pointer,
     int *file_content_type,
     liberror_error_t **error );

int libpff_io_handle_read_unallocated_data_blocks(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     liberror_error_t **error );

int libpff_io_handle_read_unallocated_page_blocks(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_page_block_list,
     liberror_error_t **error );

int libpff_io_handle_read_descriptor_index(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t descriptor_index_root_node_offset,
     uint64_t descriptor_index_back_pointer,
     liberror_error_t **error );

int libpff_io_handle_read_offset_index(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     off64_t offset_index_root_node_offset,
     uint64_t offset_index_back_pointer,
     liberror_error_t **error );

int libpff_io_handle_recover_items(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     libpff_list_t *unallocated_page_block_list,
     libpff_list_t *recovered_item_list,
     uint8_t recovery_flags,
     liberror_error_t **error );

int libpff_io_handle_recover_index_nodes(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *unallocated_data_block_list,
     uint8_t recovery_flags,
     size64_t node_offset,
     liberror_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif

