/*
 * Item descriptor functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>

#include "libpff_item_descriptor.h"
#include "libpff_tree_type.h"

/* Creates an item descriptor
 * Returns 1 if successful or -1 on error
 */
int libpff_item_descriptor_initialize(
     libpff_item_descriptor_t **item_descriptor,
     liberror_error_t **error )
{
	static char *function = "libpff_item_descriptor_initialize";

	if( item_descriptor == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item descriptor.",
		 function );

		return( -1 );
	}
	if( *item_descriptor == NULL )
	{
		*item_descriptor = (libpff_item_descriptor_t *) memory_allocate(
		                                                 sizeof( libpff_item_descriptor_t ) );

		if( *item_descriptor == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create item descriptor.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *item_descriptor,
		     0,
		     sizeof( libpff_item_descriptor_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear item descriptor.",
			 function );

			memory_free(
			 *item_descriptor );

			*item_descriptor = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees item descriptor
 * Returns 1 if successful or -1 on error
 */
int libpff_item_descriptor_free(
     intptr_t *item_descriptor,
     liberror_error_t **error )
{
	if( item_descriptor != NULL )
	{
		memory_free(
		 item_descriptor );
	}
	return( 1 );
}

/* Clones the existing item descriptor
 * Returns 1 if successful or -1 on error
 */
int libpff_item_descriptor_clone(
     intptr_t **destination,
     intptr_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_item_descriptor_clone";

	if( destination == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination item descriptor.",
		 function );

		return( -1 );
	}
	if( *destination != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination item descriptor already set.",
		 function );

		return( -1 );
	}
	if( source == NULL )
	{
		*destination = NULL;
	}
	else
	{
		*destination = (intptr_t *) memory_allocate(
		                             sizeof( libpff_item_descriptor_t ) );

		if( *destination == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create item descriptor.",
			 function );

			return( -1 );
		}
		if( memory_copy(
		     *destination,
		     source,
		     sizeof( libpff_item_descriptor_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy item descriptor.",
			 function );

			memory_free(
			 *destination );

			*destination = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Compares two item descriptors
 * Returns LIBPFF_TREE_NODE_COMPARE_LESS, LIBPFF_TREE_NODE_COMPARE_EQUAL, LIBPFF_TREE_NODE_COMPARE_GREATER if successful or -1 on error
 */
int libpff_item_descriptor_compare(
     intptr_t *first,
     intptr_t *second,
     liberror_error_t **error )
{
	static char *function = "libpff_item_descriptor_compare";

	if( first == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid first item descriptor.",
		 function );

		return( -1 );
	}
	if( second == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid second item descriptor.",
		 function );

		return( 1 );
	}
	if( ( (libpff_item_descriptor_t *) first )->descriptor_identifier < ( (libpff_item_descriptor_t *) second )->descriptor_identifier )
	{
		return( LIBPFF_TREE_NODE_COMPARE_LESS );
	}
	else if( ( (libpff_item_descriptor_t *) first )->descriptor_identifier > ( (libpff_item_descriptor_t *) second )->descriptor_identifier )
	{
		return( LIBPFF_TREE_NODE_COMPARE_GREATER );
	}
	return( LIBPFF_TREE_NODE_COMPARE_EQUAL );
}

