/*
 * Item tree functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_index.h"
#include "libpff_io_handle.h"
#include "libpff_item_descriptor.h"
#include "libpff_item_tree.h"
#include "libpff_tree_type.h"

/* Frees a recovered item tree node
 * Return 1 if successful or -1 on error
 */
int libpff_item_tree_node_free_recovered(
     intptr_t *item_tree_node,
     liberror_error_t **error )
{
	static char *function = "libpff_item_tree_node_free_recovered";

	if( ( item_tree_node != NULL )
	 && ( libpff_tree_node_free(
	       (libpff_tree_node_t **) &item_tree_node,
	       &libpff_item_descriptor_free,
	       error ) != 1 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free tree node.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the tree node of an item node
 * Returns 1 if successful, 0 if the item node was not found or -1 on error
 */
int libpff_item_tree_get_tree_node_by_identifier(
     libpff_tree_node_t *item_tree_node,
     uint32_t item_identifier,
     libpff_tree_node_t **result_item_tree_node,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor = NULL;
	libpff_tree_node_t *child_tree_node       = NULL;
	static char *function                     = "libpff_item_tree_get_tree_node_by_identifier";
	int child_node_iterator                   = 0;
	int result                                = 0;

	if( item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree node.",
		 function );

		return( -1 );
	}
	if( item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) item_tree_node->value;
	
	if( result_item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid result item tree node.",
		 function );

		return( -1 );
	}
	if( *result_item_tree_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: result item tree node already set.",
		 function );

		return( -1 );
	}
	if( item_descriptor->descriptor_identifier == item_identifier )
	{
		*result_item_tree_node = item_tree_node;

		return( 1 );
	}
	else if( item_tree_node->number_of_child_nodes > 0 )
	{
		child_tree_node = item_tree_node->first_child;

		for( child_node_iterator = 0;
		     child_node_iterator < item_tree_node->number_of_child_nodes;
		     child_node_iterator++ )
		{
			if( child_tree_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			result = libpff_item_tree_get_tree_node_by_identifier(
			          child_tree_node,
			          item_identifier,
			          result_item_tree_node,
			          error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to traverse child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			else if( result == 1 )
			{
				break;
			}
			child_tree_node = child_tree_node->next;
		}
	}
	return( result );
}

/* Retrieves the child node of an item node
 * Returns 1 if successful, 0 if the item node was not found or -1 on error
 */
int libpff_item_tree_get_child_node_by_identifier(
     libpff_tree_node_t *item_tree_node,
     uint32_t child_node_identifier,
     libpff_tree_node_t **child_node,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor = NULL;
	libpff_tree_node_t *child_tree_node       = NULL;
	static char *function                     = "libpff_item_tree_get_child_node_by_identifier";
	int child_node_iterator                   = 0;
	int result                                = 0;

	if( item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree node.",
		 function );

		return( -1 );
	}
	if( item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	if( child_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid child node.",
		 function );

		return( -1 );
	}
	if( *child_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: child node already set.",
		 function );

		return( -1 );
	}
	if( item_tree_node->number_of_child_nodes > 0 )
	{
		child_tree_node = item_tree_node->first_child;

		for( child_node_iterator = 0;
		     child_node_iterator < item_tree_node->number_of_child_nodes;
		     child_node_iterator++ )
		{
			if( child_tree_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			if( child_tree_node->value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: missing value for child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			item_descriptor = (libpff_item_descriptor_t *) child_tree_node->value;

			if( item_descriptor->descriptor_identifier == child_node_identifier )
			{
				*child_node = child_tree_node;

				return( 1 );
			}
			child_tree_node = child_tree_node->next;
		}
	}
	return( result );
}

/* Creates an item tree from the descriptor index
 * Returns 1 if successful or -1 on error
 */
int libpff_item_tree_create(
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t **item_tree_root_node,
     libpff_tree_node_t **root_folder_item_tree_node,
     libpff_list_t *orphan_node_list,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor = NULL;
	static char *function                     = "libpff_item_tree_create";

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	if( item_tree_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree root node.",
		 function );

		return( -1 );
	}
	if( *item_tree_root_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: item tree root node already set.",
		 function );

		return( -1 );
	}
	if( orphan_node_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid orphan node list.",
		 function );

		return( -1 );
	}
	/* Create a new item descriptor
	 */
	item_descriptor = (libpff_item_descriptor_t *) memory_allocate(
							sizeof( libpff_item_descriptor_t ) );

	if( item_descriptor == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create item descriptor.",
		 function );

		return( -1 );
	}
	if( memory_set(
	     item_descriptor,
	     0,
	     sizeof( libpff_item_descriptor_t ) ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear item descriptor.",
		 function );

		memory_free(
		 item_descriptor );

		return( -1 );
	}
	/* Create the item tree root node
	 */
	*item_tree_root_node = (libpff_tree_node_t *) memory_allocate(
						       sizeof( libpff_tree_node_t ) );

	if( *item_tree_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create item tree root node.",
		 function );

		memory_free(
		 item_descriptor );

		return( -1 );
	}
	if( memory_set(
	     *item_tree_root_node,
	     0,
	     sizeof( libpff_tree_node_t ) ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear item tree root node.",
		 function );

		memory_free(
		 item_descriptor );
		memory_free(
		 *item_tree_root_node );

		*item_tree_root_node = NULL;

		return( -1 );
	}
	( *item_tree_root_node )->value = (intptr_t *) item_descriptor;

	/* Create the item tree
	 */
	if( libpff_item_tree_create_node(
	     io_handle->descriptor_index_root_node,
	     io_handle->descriptor_index_root_node,
	     *item_tree_root_node,
	     orphan_node_list,
	     root_folder_item_tree_node,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create item tree node.",
		 function );

		libpff_tree_node_free(
		 item_tree_root_node,
		 &libpff_item_descriptor_free,
		 NULL );

		*item_tree_root_node        = NULL;
		*root_folder_item_tree_node = NULL;

		return( -1 );
	}
	return( 1 );
}

/* Creates an item tree node from the descriptor index
 * Returns 1 if successful, 0 if the node has no parent and was added to the orphan node list or -1 on error
 */
int libpff_item_tree_create_node(
     libpff_tree_node_t *descriptor_index_root_node,
     libpff_tree_node_t *descriptor_index_tree_node,
     libpff_tree_node_t *item_tree_root_node,
     libpff_list_t *orphan_node_list,
     libpff_tree_node_t **root_folder_item_tree_node,
     liberror_error_t **error )
{
	libpff_index_values_t *descriptor_index_values = NULL;
	libpff_item_descriptor_t *item_descriptor      = NULL;
	libpff_tree_node_t *child_node                 = NULL;
	libpff_tree_node_t *item_tree_node             = NULL;
	libpff_tree_node_t *parent_node                = NULL;
	static char *function                          = "libpff_item_tree_create_node";
	int child_node_iterator                        = 0;
	int result                                     = 0;

	if( descriptor_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid descriptor index root node.",
		 function );

		return( -1 );
	}
	if( descriptor_index_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid descriptor index tree node.",
		 function );

		return( -1 );
	}
	if( item_tree_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree root node.",
		 function );

		return( -1 );
	}
	if( orphan_node_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid orphan node list.",
		 function );

		return( -1 );
	}
	if( root_folder_item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid root folder item tree node.",
		 function );

		return( -1 );
	}

	/* Skip the branch nodes in the comparison
	 * The branch node can contain index identifiers that also are assigned to leaf nodes
	 */
	if( descriptor_index_tree_node->number_of_child_nodes > 0 )
	{
		child_node = descriptor_index_tree_node->first_child;

		for( child_node_iterator = 0;
		     child_node_iterator < descriptor_index_tree_node->number_of_child_nodes;
		     child_node_iterator++ )
		{
			if( child_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			result = libpff_item_tree_create_node(
			          descriptor_index_root_node,
			          child_node,
			          item_tree_root_node,
			          orphan_node_list,
			          root_folder_item_tree_node,
			          error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to traverse child node: %d.",
				 function,
				 child_node_iterator );

				return( -1 );
			}
			child_node = child_node->next;
		}
	}
	else
	{
		if( descriptor_index_tree_node->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid descriptor index tree node - missing value.",
			 function );

			return( -1 );
		}
		descriptor_index_values = (libpff_index_values_t *) descriptor_index_tree_node->value;

		if( descriptor_index_values->identifier > (uint64_t) UINT32_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: descriptor index identifier value exceeds maximum.",
			 function );

			return( -1 );
		}
		/* Create a new item descriptor
		 */
		if( libpff_item_descriptor_initialize(
		     &item_descriptor,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create item descriptor.",
			 function );

			return( -1 );
		}
		item_descriptor->descriptor_identifier = (uint32_t) descriptor_index_values->identifier;

		/* The root folder index descriptor points to itself as its parent
		 */
		if( descriptor_index_values->identifier == descriptor_index_values->parent_identifier )
		{
			if( *root_folder_item_tree_node != NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
				 "%s: root folder item tree node already set.",
				 function );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );

				return( -1 );
			}
			*root_folder_item_tree_node = (libpff_tree_node_t *) memory_allocate(
			                                                      sizeof( libpff_tree_node_t ) );

			if( *root_folder_item_tree_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create root folder item tree node.",
				 function );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );

				return( -1 );
			}
			if( memory_set(
			     *root_folder_item_tree_node,
			     0,
			     sizeof( libpff_tree_node_t ) ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear root folder item tree node.",
				 function );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );
				memory_free(
				 *root_folder_item_tree_node );

				*root_folder_item_tree_node = NULL;

				return( -1 );
			}
			( *root_folder_item_tree_node )->value = (intptr_t *) item_descriptor;

			result = libpff_tree_node_insert_node(
				  item_tree_root_node,
				  *root_folder_item_tree_node,
				  &libpff_item_descriptor_compare,
			          NULL );

			/* Free the node if it could not be inserted
			 */
			if( result != 1 )
			{
				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );
				memory_free(
				 *root_folder_item_tree_node );

				*root_folder_item_tree_node = NULL;
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to insert item descriptor in item tree node.",
				 function );

				return( -1 );
			}
		}
		else
		{
			result = libpff_item_tree_get_tree_node_by_identifier(
				  item_tree_root_node,
				  descriptor_index_values->parent_identifier,
				  &parent_node,
			          error );

			if( result == 0 )
			{
#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: reading ahead for descriptor: %" PRIu64 " parent %" PRIu32 ".\n",
					 function,
					 descriptor_index_values->identifier,
					 descriptor_index_values->parent_identifier );
				}
#endif

				result = libpff_index_get_tree_node(
					  descriptor_index_root_node,
					  descriptor_index_values->parent_identifier,
					  &parent_node,
				          error );

				if( result == 1 )
				{
					if( parent_node == NULL )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
						 "%s: invalid parent node.",
						 function );

						libpff_item_descriptor_free(
						 (intptr_t *) item_descriptor,
						 NULL );

						return( -1 );
					}
					result = libpff_item_tree_create_node(
					          descriptor_index_root_node,
					          parent_node,
					          item_tree_root_node,
					          orphan_node_list,
					          root_folder_item_tree_node,
					          error );

					if( result == -1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
						 "%s: unable to traverse parent node: %d.",
						 function,
						 child_node_iterator );

						libpff_item_descriptor_free(
						 (intptr_t *) item_descriptor,
						 NULL );

						return( -1 );
					}
					parent_node = NULL;

					result = libpff_item_tree_get_tree_node_by_identifier(
						  item_tree_root_node,
						  descriptor_index_values->parent_identifier,
						  &parent_node,
					          error );
				}
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to find parent node: %" PRIu32 ".",
				 function,
				 descriptor_index_values->parent_identifier );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );

				return( -1 );
			}
			else if( result == 0 )
			{
#if defined( HAVE_VERBOSE_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: parent node: %" PRIu32 " missing found orphan node: %" PRIu64 ".\n",
					 function,
					 descriptor_index_values->parent_identifier,
					 descriptor_index_values->identifier );
				}
#endif
				if( libpff_tree_node_initialize(
				     &item_tree_node,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
					 "%s: unable to create item tree node.",
					 function );

					libpff_item_descriptor_free(
					 (intptr_t *) item_descriptor,
					 NULL );

					return( -1 );
				}
				item_tree_node->value = (intptr_t *) item_descriptor;

				item_descriptor = NULL;

				if( libpff_list_append_value(
				     orphan_node_list,
				     (intptr_t *) item_tree_node,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
					 "%s: unable to append orphan node in orphan node list.",
					 function );

					libpff_tree_node_free(
					 &item_tree_node,
					 &libpff_item_descriptor_free,
					 NULL );

					return( -1 );
				}
				return( 0 );
			}
			if( parent_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid parent node.",
				 function );

				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );

				return( -1 );
			}
			result = libpff_tree_node_insert_value(
				  parent_node,
				  (intptr_t *) item_descriptor,
				  &libpff_item_descriptor_compare,
			          error );

			/* Free the node if it could not be inserted
			 */
			if( result != 1 )
			{
				libpff_item_descriptor_free(
				 (intptr_t *) item_descriptor,
				 NULL );
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to insert item descriptor in item tree node.",
				 function );

				return( -1 );
			}
		}
	}
	return( 1 );
}

/* Retrieves the identifier of the item
 * Returns 1 if successful or -1 on error
 */
int libpff_item_tree_get_identifier(
     libpff_tree_node_t *item_tree_node,
     uint32_t *identifier,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor = NULL;
	static char *function                     = "libpff_item_tree_get_identifier";

	if( item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree node.",
		 function );

		return( -1 );
	}
	if( item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) item_tree_node->value;

	if( identifier == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid identifier.",
		 function );

		return( -1 );
	}
	*identifier = item_descriptor->descriptor_identifier;

	return( 1 );
}

/* Appends the identifier of the item
 * Returns 1 if successful or -1 on error
 */
int libpff_item_tree_append_identifier(
     libpff_tree_node_t *item_tree_node,
     uint32_t identifier,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *child_item_descriptor = NULL;
	libpff_item_descriptor_t *item_descriptor       = NULL;
	static char *function                           = "libpff_item_tree_append_identifier";

	if( item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item tree node.",
		 function );

		return( -1 );
	}
	if( item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) item_tree_node->value;

	child_item_descriptor = (libpff_item_descriptor_t *) memory_allocate(
	                                                      sizeof( libpff_item_descriptor_t ) );

	if( child_item_descriptor == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create child item descriptor.",
		 function );

		return( -1 );
	}
	if( memory_set(
	     child_item_descriptor,
	     0,
	     sizeof( libpff_item_descriptor_t ) ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear item descriptor.",
		 function );

		memory_free(
		 child_item_descriptor );

		return( -1 );
	}
	child_item_descriptor->descriptor_identifier = identifier;
	child_item_descriptor->recovered             = item_descriptor->recovered;

	if( libpff_tree_node_append_value(
	     item_tree_node,
	     (intptr_t *) child_item_descriptor,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to append item descriptor to item tree node.",
		 function );

		memory_free(
		 child_item_descriptor );

		return( -1 );
	}
	return( 1 );
}

