/*
 * Item values functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_data_reference.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_item_values.h"
#include "libpff_libuna.h"
#include "libpff_list_type.h"
#include "libpff_local_descriptors.h"
#include "libpff_multi_value.h"
#include "libpff_table.h"
#include "libpff_tree_type.h"
#include "libpff_types.h"
#include "libpff_value_type.h"

/* Creates the item values
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_initialize(
     libpff_item_values_t **item_values,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_initialize";

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( *item_values == NULL )
	{
		*item_values = (libpff_item_values_t *) memory_allocate(
		                                         sizeof( libpff_item_values_t ) );

		if( *item_values == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create item values.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *item_values,
		     0,
		     sizeof( libpff_item_values_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear item values.",
			 function );

			memory_free(
			 *item_values );

			*item_values = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees item values
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_free(
     libpff_item_values_t **item_values,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_free";
	int result            = 1;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( *item_values != NULL )
	{
		if( ( *item_values )->local_descriptors != NULL )
		{
			if( libpff_tree_node_free(
			     &( ( *item_values )->local_descriptors ),
			     &libpff_local_descriptor_values_free,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free item local descriptors.",
				 function );

				result = -1;
			}
		}
		if( ( *item_values )->table != NULL )
		{
			if( libpff_table_free(
			     ( *item_values )->table,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free item values table.",
				 function );

				result = -1;
			}
		}
		memory_free(
		 *item_values );

		*item_values = NULL;
	}
	return( result );
}

/* Clone copies the existing item values
 * Copies the values and creates sub elements if necessary
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_clone_copy(
     libpff_item_values_t *destination,
     libpff_item_values_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_clone_copy";

	if( destination == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination item values.",
		 function );

		return( -1 );
	}
	if( destination->local_descriptors != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination item values - local descriptors already set.",
		 function );

		return( -1 );
	}
	if( destination->table != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination item values - table already set.",
		 function );

		return( -1 );
	}
	if( source == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid source item values.",
		 function );

		return( -1 );
	}
	if( libpff_tree_node_clone(
	     &( destination->local_descriptors ),
	     source->local_descriptors,
	     &libpff_local_descriptor_values_clone,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to clone local descriptors.",
		 function );

		/* libpff_tree_node_clone can return a partial tree node
		 */
		libpff_tree_node_free(
		 &destination->local_descriptors,
		 &libpff_local_descriptor_values_free,
		 NULL );

		destination->local_descriptors = NULL;

		return( -1 );
	}
	if( libpff_table_clone(
	     &( destination->table ),
	     source->table,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to clone table.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the item values for a specific descriptor
 * Returns 1 if successful, 0 if not available or -1 on error
 */
int libpff_item_values_read(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *data_reference = NULL;
	static char *function                   = "libpff_item_values_read";
	int result                              = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->local_descriptors != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item values - local descriptors already set.",
		 function );

		return( -1 );
	}
	if( item_values->table != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item values - table already set.",
		 function );

		return( -1 );
	}
	result = libpff_descriptor_read(
	          descriptor_identifier,
	          recovered,
	          io_handle,
	          file_io_handle,
	          &( item_values->local_descriptors ),
	          &data_reference,
	          error );

	/* Both the local_descriptors and data_reference are freed and set to NULL on error
	 */
	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read descriptor for item identifier: %" PRIu32 ".",
		 function,
		 descriptor_identifier );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data reference.",
		 function  );

		if( item_values->local_descriptors != NULL )
		{
			libpff_tree_node_free(
			 &item_values->local_descriptors,
			 &libpff_local_descriptor_values_free,
			 NULL );

			item_values->local_descriptors = NULL;
		}
		return( -1 );
	}
	if( libpff_table_initialize(
	     &( item_values->table ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table.",
		 function );

		libpff_data_reference_free(
		 &data_reference,
		 NULL );

		if( item_values->local_descriptors != NULL )
		{
			libpff_tree_node_free(
			 &item_values->local_descriptors,
			 &libpff_local_descriptor_values_free,
			 NULL );

			item_values->local_descriptors = NULL;
		}
		return( -1 );
	}
	if( libpff_table_read(
	     item_values->table,
	     recovered,
	     io_handle,
	     item_values->local_descriptors,
	     name_to_id_map_list,
	     data_reference,
	     file_io_handle,
	     debug_item_type,
	     debug_ascii_codepage,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table.",
		 function );

		libpff_data_reference_free(
		 &data_reference,
		 NULL );
		libpff_table_free(
		 item_values->table,
		 NULL );

		item_values->table = NULL;

		if( item_values->local_descriptors != NULL )
		{
			libpff_tree_node_free(
			 &item_values->local_descriptors,
			 &libpff_local_descriptor_values_free,
			 NULL );

			item_values->local_descriptors = NULL;
		}
		return( -1 );
	}
	if( libpff_data_reference_free(
	     &data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free data reference.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the item values for a specific local descriptor
 * Returns 1 if successful, 0 if not available or -1 on error
 */
int libpff_item_values_local_read(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_tree_node_t *parent_local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *data_reference = NULL;
	static char *function                   = "libpff_item_values_local_read";
	int result                              = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->local_descriptors != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item values - local descriptors already set.",
		 function );

		return( -1 );
	}
	if( item_values->table != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item values - table already set.",
		 function );

		return( -1 );
	}
	result = libpff_descriptor_local_read(
	          descriptor_identifier,
	          recovered,
	          parent_local_descriptors,
	          io_handle,
	          file_io_handle,
	          &( item_values->local_descriptors ),
	          &data_reference,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read local descriptor for identifier: %" PRIu32 ".",
		 function,
		 descriptor_identifier );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid data reference.",
		 function  );

		return( -1 );
	}
	if( libpff_table_initialize(
	     &( item_values->table ),
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table.",
		 function );

		libpff_data_reference_free(
		 &data_reference,
		 NULL );
		libpff_tree_node_free(
		 &item_values->local_descriptors,
		 &libpff_local_descriptor_values_free,
		 NULL );

		item_values->local_descriptors = NULL;

		return( -1 );
	}
	if( libpff_table_read(
	     item_values->table,
	     recovered,
	     io_handle,
	     item_values->local_descriptors,
	     name_to_id_map_list,
	     data_reference,
	     file_io_handle,
	     debug_item_type,
	     debug_ascii_codepage,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table.",
		 function );

		libpff_data_reference_free(
		 &data_reference,
		 NULL );
		libpff_tree_node_free(
		 &item_values->local_descriptors,
		 &libpff_local_descriptor_values_free,
		 NULL );
		libpff_table_free(
		 item_values->table,
		 NULL );

		item_values->local_descriptors = NULL;
		item_values->table             = NULL;

		return( -1 );
	}
	if( libpff_data_reference_free(
	     &data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free data reference.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the sub node of item values
 * Returns 1 if successful, 0 if the item node was not found or -1 on error
 */
int libpff_item_values_get_sub_node_by_identifier(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     uint32_t sub_node_identifier,
     libpff_tree_node_t **sub_node,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_sub_node_by_identifier";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	if( item_values->local_descriptors == NULL )
	{
		return( 0 );
	}
	result = libpff_local_descriptors_get_tree_node_by_identifier(
	          item_values->local_descriptors,
	          sub_node_identifier,
	          sub_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve local descriptors tree node.",
		 function );

		return( -1 );
	}
	return( result );
}

/* Retrieves the number of item value sets
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_get_number_of_sets(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     uint32_t *number_of_sets,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_number_of_sets";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( number_of_sets == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid number of sets.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			*number_of_sets = 0;

			return( 1 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	*number_of_sets = item_values->table->number_of_sets;

	return( 1 );
}

/* Retrieves the number of item entries
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_get_number_of_entries(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     uint32_t *number_of_entries,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_number_of_entries";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( number_of_entries == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid number of entries.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			*number_of_entries = 0;

			return( 1 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	*number_of_entries = item_values->table->number_of_entries;

	return( 1 );
}

/* Retrieves the entry type of a specific entry from the referenced item values
 * Returns 1 if successful or -1 on error
 */
int libpff_item_values_get_entry_type(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int table_set_index,
     int table_entry_index,
     uint32_t *entry_type,
     uint32_t *value_type,
     libpff_name_to_id_map_entry_t **name_to_id_map_entry,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_entry_type";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	if( libpff_table_get_entry_type_by_index(
	     item_values->table,
	     table_set_index,
	     table_entry_index,
	     entry_type,
	     value_type,
	     name_to_id_map_entry,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entry type.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the value of a specific entry from the referenced item values
 *
 * When the LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE is set
 * the value type is ignored and set. The default behavior is a strict
 * matching of the value type. In this case the value type must be filled
 * with the corresponding value type
 *
 * When the LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP is set
 * the name to identifier mapping is ignored. The default behavior is
 * the use the mapped entry value. In this case named properties are not
 * retrieved.
 *
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_item_values_get_entry_value(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int table_set_index,
     uint32_t entry_type,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     uint8_t flags,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_entry_value";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			return( 0 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	result = libpff_table_get_entry_value_by_entry_type(
	          item_values->table,
	          table_set_index,
	          entry_type,
	          value_type,
	          value_data_reference, 
	          flags,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entry value.",
		 function );

		return( -1 );
	}
	return( result );
}

/* Retrieves the value of a specific entry from the referenced item values by an UTF-8 formatted name
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_item_values_get_entry_value_by_utf8_name(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int table_set_index,
     uint8_t *utf8_entry_name,
     size_t utf8_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_entry_value_by_utf8_name";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			return( 0 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	result = libpff_table_get_entry_value_by_utf8_entry_name(
	          item_values->table,
	          table_set_index,
	          utf8_entry_name,
	          utf8_entry_name_length,
	          value_type,
	          value_data_reference, 
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entry value.",
		 function );

		return( -1 );
	}
	return( result );
}

/* Retrieves the value of a specific entry from the referenced item values by an UTF-16 formatted name
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_item_values_get_entry_value_by_utf16_name(
     libpff_item_values_t *item_values,
     uint32_t descriptor_identifier,
     uint8_t recovered,
     libpff_list_t *name_to_id_map_list,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     int table_set_index,
     uint16_t *utf16_entry_name,
     size_t utf16_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_item_values_get_entry_value_by_utf16_name";
	int result            = 0;

	if( item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item values.",
		 function );

		return( -1 );
	}
	if( item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          item_values,
		          descriptor_identifier,
		          recovered,
		          name_to_id_map_list,
		          io_handle,
		          file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          debug_ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			return( 0 );
		}
		if( item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	result = libpff_table_get_entry_value_by_utf16_entry_name(
	          item_values->table,
	          table_set_index,
	          utf16_entry_name,
	          utf16_entry_name_length,
	          value_type,
	          value_data_reference, 
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entry value.",
		 function );

		return( -1 );
	}
	return( result );
}

