/*
 * Legacy functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>

#include "libpff_attachment.h"
#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_file.h"
#include "libpff_folder.h"
#include "libpff_item.h"
#include "libpff_item_descriptor.h"
#include "libpff_legacy.h"
#include "libpff_mapi.h"
#include "libpff_message.h"
#include "libpff_multi_value.h"
#include "libpff_types.h"

#if !defined( HAVE_LOCAL_LIBPFF )

/* Retrieves the attachment data
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_attachment_get_data(
     libpff_item_t *attachment,
     uint8_t *data,
     size_t size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item         = NULL;
	libpff_item_descriptor_t *item_descriptor     = NULL;
	libpff_data_reference_t *value_data_reference = NULL;
	uint8_t *value_data                           = NULL;
	static char *function                         = "libpff_attachment_get_data";
	size_t value_data_size                        = 0;
	uint32_t embedded_object_data_identifier      = 0;
	uint32_t value_type                           = 0;
	int result                                    = 0;

	if( attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid attachment.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) attachment;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid attachment - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid attachment - missing item values.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) internal_item->item_tree_node->value;

	if( data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data.",
		 function );

		return( -1 );
	}
	if( size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid size value exceeds maximum.",
		 function );

		return( -1 );
	}
	result = libpff_item_values_get_entry_value(
	          internal_item->item_values,
	          item_descriptor->descriptor_identifier,
	          item_descriptor->recovered,
	          internal_item->internal_file->name_to_id_map_list,
	          internal_item->internal_file->io_handle,
	          internal_item->file_io_handle,
	          0,
	          LIBPFF_ENTRY_TYPE_ATTACHMENT_DATA_OBJECT,
	          &value_type,
	          &value_data_reference,
	          LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE,
	          internal_item->internal_file->ascii_codepage,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	/* The OLE attachment method could refer to an OLE embedded object
	 */
	if( value_type == LIBPFF_VALUE_TYPE_OBJECT )
	{
		if( internal_item->embedded_object == NULL )
		{
			if( libpff_data_reference_get_data(
			     value_data_reference,
			     internal_item->file_io_handle,
			     &value_data,
			     &value_data_size,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve value data reference data.",
				 function );

				return( -1 );
			}
			if( value_data == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid entry embedded object data.",
				 function );

				return( -1 );
			}
			/* The descriptor identifier is located in the local descriptors
			 */
			byte_stream_copy_to_uint32_little_endian(
			 value_data,
			 embedded_object_data_identifier );

			if( libpff_descriptor_local_read(
			     embedded_object_data_identifier,
			     item_descriptor->recovered,
			     internal_item->item_values->local_descriptors,
			     internal_item->internal_file->io_handle,
			     internal_item->file_io_handle,
			     NULL,
			     &( internal_item->embedded_object ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to read embedded object: %" PRIu32 ".",
				 function,
				 embedded_object_data_identifier );

				return( -1 );
			}
		}
		value_data_reference = internal_item->embedded_object;
	}
	if( libpff_data_reference_get_data(
	     value_data_reference,
	     internal_item->file_io_handle,
	     &value_data,
	     &value_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve value data reference data.",
		 function );

		return( -1 );
	}
	if( value_data == NULL )
	{
		return( 0 );
	}
	if( size < value_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: data size too small.",
		 function );

		return( -1 );
	}
	if( memory_copy(
	     data,
	     value_data,
	     value_data_size ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_COPY_FAILED,
		 "%s: unable to set data.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the amount of unallocated blocks
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_amount_of_unallocated_blocks(
     libpff_file_t *file,
     int unallocated_block_type,
     int *amount_of_unallocated_blocks,
     liberror_error_t **error )
{
	return( libpff_file_get_number_of_unallocated_blocks(
	         file,
	         unallocated_block_type,
	         amount_of_unallocated_blocks,
	         error ) );
}

/* Retrieves the amount of orphan items
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_amount_of_orphan_items(
     libpff_file_t *file,
     int *amount_of_orphan_items,
     liberror_error_t **error )
{
	return( libpff_file_get_number_of_orphan_items(
	         file,
	         amount_of_orphan_items,
	         error ) );
}

/* Retrieves the amount of recovered items
 * Returns 1 if successful or -1 on error
 */
int libpff_file_get_amount_of_recovered_items(
     libpff_file_t *file,
     int *amount_of_recovered_items,
     liberror_error_t **error )
{
	return( libpff_file_get_number_of_recovered_items(
	         file,
	         amount_of_recovered_items,
	         error ) );
}

/* Retrieves the amount of sub folders from a folder
 * Returns 1 if successful or -1 on error
 */
int libpff_folder_get_amount_of_sub_folders(
     libpff_item_t *folder,
     int *amount_of_sub_folders,
     liberror_error_t **error )
{
	return( libpff_folder_get_number_of_sub_folders(
	         folder,
	         amount_of_sub_folders,
	         error ) );
}

/* Retrieves the amount of sub messages from a folder
 * Returns 1 if successful or -1 on error
 */
int libpff_folder_get_amount_of_sub_messages(
     libpff_item_t *folder,
     int *amount_of_sub_messages,
     liberror_error_t **error )
{
	return( libpff_folder_get_number_of_sub_messages(
	         folder,
	         amount_of_sub_messages,
	         error ) );
}

/* Retrieves the amount of sub associated contents from a folder
 * Returns 1 if successful or -1 on error
 */
int libpff_folder_get_amount_of_sub_associated_contents(
     libpff_item_t *folder,
     int *amount_of_sub_associated_contents,
     liberror_error_t **error )
{
	return( libpff_folder_get_number_of_sub_associated_contents(
	         folder,
	         amount_of_sub_associated_contents,
	         error ) );
}

/* Retrieves the amount of sets values of the referenced item
 * Returns 1 if successful or -1 on error
 */
int libpff_item_get_amount_of_sets(
     libpff_item_t *item,
     uint32_t *amount_of_sets,
     liberror_error_t **error )
{
	return( libpff_item_get_number_of_sets(
	         item,
	         amount_of_sets,
	         error ) );
}

/* Retrieves the amount of entries (of a set) of the referenced item
 * All sets in an item contain the same amount of entries
 * Returns 1 if successful or -1 on error
 */
int libpff_item_get_amount_of_entries(
     libpff_item_t *item,
     uint32_t *amount_of_entries,
     liberror_error_t **error )
{
	return( libpff_item_get_number_of_entries(
	         item,
	         amount_of_entries,
	         error ) );
}

/* Retrieves the amount of sub items from a item
 * Returns 1 if successful or -1 on error
 */
int libpff_item_get_amount_of_sub_items(
     libpff_item_t *item,
     int *amount_of_sub_items,
     liberror_error_t **error )
{
	return( libpff_item_get_number_of_sub_items(
	         item,
	         amount_of_sub_items,
	         error ) );
}

/* Retrieves the amount of attachments from a message item
 * Returns 1 if successful or -1 on error
 */
int libpff_message_get_amount_of_attachments(
     libpff_item_t *message,
     int *amount_of_attachments,
     liberror_error_t **error )
{
	return( libpff_message_get_number_of_attachments(
	         message,
	         amount_of_attachments,
	         error ) );
}

/* Retrieves the amount of values of the multi value
 * Returns 1 if successful or -1 on error
 */
int libpff_multi_value_get_amount_of_values(
     libpff_multi_value_t *multi_value,
     int *amount_of_values,
     liberror_error_t **error )
{
	return( libpff_multi_value_get_number_of_values(
	         multi_value,
	         amount_of_values,
	         error ) );
}

#endif /* !defined( HAVE_LOCAL_LIBPFF ) */

