/*
 * Local descriptors functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_index.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"
#include "libpff_local_descriptors.h"
#include "libpff_tree_type.h"

#include "pff_local_descriptors_node.h"

/* Frees local descriptor values
 * Returns 1 if successful or -1 on error
 */
int libpff_local_descriptor_values_free(
     intptr_t *local_descriptor_values,
     liberror_error_t **error )
{
	if( local_descriptor_values != NULL )
	{
		memory_free(
		 local_descriptor_values );
	}
	return( 1 );
}

/* Clones the existing local descriptor values
 * Returns 1 if successful or -1 on error
 */
int libpff_local_descriptor_values_clone(
     intptr_t **destination,
     intptr_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_local_descriptor_values_clone";

	if( destination == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination local descriptor values.",
		 function );

		return( -1 );
	}
	if( *destination != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination local descriptor values already set.",
		 function );

		return( -1 );
	}
	if( source == NULL )
	{
		*destination = NULL;
	}
	else
	{
		*destination = (intptr_t *) memory_allocate(
		                             sizeof( libpff_local_descriptor_values_t ) );

		if( *destination == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create local descriptor values.",
			 function );

			return( -1 );
		}
		if( memory_copy(
		     *destination,
		     source,
		     sizeof( libpff_local_descriptor_values_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to copy local descriptor values.",
			 function );

			memory_free(
			 *destination );

			*destination = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Creates a local descriptors tree node with item values
 * Returns 1 if successful or -1 on error
 */
int libpff_local_descriptors_node_initialize(
     libpff_tree_node_t **local_descriptors_tree_node,
     liberror_error_t **error )
{
	static char *function = "libpff_local_descriptors_node_initialize";

	if( local_descriptors_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors tree node.",
		 function );

		return( -1 );
	}
	if( *local_descriptors_tree_node == NULL )
	{
		*local_descriptors_tree_node = (libpff_tree_node_t *) memory_allocate(
		                                                       sizeof( libpff_tree_node_t ) );

		if( *local_descriptors_tree_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create local descriptors tree node.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *local_descriptors_tree_node,
		     0,
		     sizeof( libpff_tree_node_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear local descriptors tree node.",
			 function );

			memory_free(
			 *local_descriptors_tree_node );

			*local_descriptors_tree_node = NULL;

			return( -1 );
		}
		( *local_descriptors_tree_node )->value = (intptr_t *) memory_allocate(
		                                                        sizeof( libpff_local_descriptor_values_t ) );

		if( ( *local_descriptors_tree_node )->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create local descriptor values.",
			 function );

			memory_free(
			 *local_descriptors_tree_node );

			*local_descriptors_tree_node = NULL;

			return( -1 );
		}
		if( memory_set(
		     ( *local_descriptors_tree_node )->value,
		     0,
		     sizeof( libpff_local_descriptor_values_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear local descriptors values.",
			 function );

			memory_free(
			 ( *local_descriptors_tree_node )->value );
			memory_free(
			 *local_descriptors_tree_node );

			*local_descriptors_tree_node = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Reads the local descriptors tree node
 * Returns 1 if successful or -1 on error
 */
int libpff_local_descriptors_read_node(
     libpff_tree_node_t *local_descriptors_tree_node,
     libbfio_handle_t *file_io_handle,
     off64_t node_file_offset,
     size32_t node_data_size,
     int file_type,
     uint8_t *node_level,
     liberror_error_t **error )
{
	libpff_local_descriptor_values_t *values = NULL;
	uint8_t *element_data                    = NULL;
	uint8_t *node_data                       = NULL;
	static char *function                    = "libpff_local_descriptors_read_node";
	size_t element_size                      = 0;
	size_t header_size                       = 0;
	ssize_t read_count                       = 0;
	uint32_t unknown1                        = 0;
	uint16_t number_of_elements              = 0;
	uint16_t element_iterator                = 0;
	uint8_t node_signature                   = 0;

	if( local_descriptors_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors tree node.",
		 function );

		return( -1 );
	}
	if( ( file_type != LIBPFF_FILE_TYPE_32BIT )
	 && ( file_type != LIBPFF_FILE_TYPE_64BIT ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: invalid file type.",
		 function );

		return( -1 );
	}
	if( node_file_offset < 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid node file offset value out of range.",
		 function );

		return( -1 );
	}
	if( node_data_size > (size32_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: node data size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( node_level == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node level.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading local descriptors node data at offset: %" PRIu64 " (0x%08" PRIx64 ")\n",
		 function,
		 node_file_offset,
		 node_file_offset );
	}
#endif

	/* Read the local descriptors node data
	 */
	if( libbfio_handle_seek_offset(
	     file_io_handle,
	     node_file_offset,
	     SEEK_SET,
	     error ) == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_SEEK_FAILED,
		 "%s: unable to seek value offset: %" PRIu64 ".",
		 function,
		 node_file_offset );

		return( -1 );
	}
	node_data = (uint8_t *) memory_allocate(
	                         sizeof( uint8_t ) * node_data_size );

	if( node_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create node data.",
		 function );

		return( -1 );
	}
	read_count = libbfio_handle_read(
	              file_io_handle,
	              node_data,
	              node_data_size,
	              error );

	if( read_count != (ssize_t) node_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read node data.",
		 function );

		memory_free(
		 node_data );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: local descriptors node:\n",
		 function );
		libnotify_print_data(
		 node_data,
		 node_data_size );
	}
#endif

	/* Parse the local descriptors node data
	 */
	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		node_signature = ( (pff_local_descriptors_node_32bit_t *) node_data )->signature;
		*node_level    = ( (pff_local_descriptors_node_32bit_t *) node_data )->level;

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_local_descriptors_node_32bit_t *) node_data )->number_of_elements,
		 number_of_elements );

		element_data = &( node_data[ sizeof( pff_local_descriptors_node_32bit_t ) ] );
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		node_signature = ( (pff_local_descriptors_node_64bit_t *) node_data )->signature;
		*node_level    = ( (pff_local_descriptors_node_64bit_t *) node_data )->level;

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_local_descriptors_node_64bit_t *) node_data )->number_of_elements,
		 number_of_elements );

		element_data = &( node_data[ sizeof( pff_local_descriptors_node_64bit_t ) ] );
	}
	header_size = 4;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: local descriptors node signature\t: 0x%02" PRIx8 "\n",
		 function,
		 node_signature );
		libnotify_printf(
		 "%s: local descriptors node level\t: %" PRIu8 "\n",
		 function,
		 *node_level );
		libnotify_printf(
		 "%s: number of elements\t\t\t: %" PRIu16 "\n",
		 function,
		 number_of_elements );

		if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			libnotify_printf(
			 "%s: padding1:\n",
			 function );
			libnotify_print_data(
			 ( (pff_local_descriptors_node_64bit_t *) node_data )->padding1,
			 4 );
		}
	}
#endif
	if( node_signature != 0x02 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: unsupported local descriptors node signature: 0x%02" PRIx8 ".",
		 function,
		 node_signature );

		memory_free(
		 node_data );

		return( -1 );
	}
	if( file_type == LIBPFF_FILE_TYPE_32BIT )
	{
		if( *node_level == 0 )
		{
			element_size = 12;
		}
		else
		{
			element_size = 8;
		}
	}
	else if( file_type == LIBPFF_FILE_TYPE_64BIT )
	{
		header_size += 4;

		if( *node_level == 0 )
		{
			element_size = 24;
		}
		else
		{
			element_size = 16;
		}
	}
	if( number_of_elements == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid number of local descriptor elements.",
		 function );

		/* TODO implement error tollerance
		 */
		memory_free(
		 node_data );

		return( -1 );
	}
	if( ( number_of_elements * element_size ) != ( node_data_size - header_size ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: mismatch in total size of the elements data and the size of the elements.",
		 function );

		/* TODO implement error tollerance
		 */
		memory_free(
		 node_data );

		return( -1 );
	}
	for( element_iterator = 0;
	     element_iterator < number_of_elements;
	     element_iterator++ )
	{
		values = (libpff_local_descriptor_values_t *) memory_allocate(
		                                               sizeof( libpff_local_descriptor_values_t ) );

		if( values == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create values.",
			 function );

			memory_free(
			 node_data );

			return( -1 );
		}
		values->level = *node_level;

		if( file_type == LIBPFF_FILE_TYPE_32BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 element_data,
			 values->identifier );

			element_data += 4;

			if( *node_level == 0 )
			{
				byte_stream_copy_to_uint32_little_endian(
				 element_data,
				 values->data_identifier );

				element_data += 4;
			}
			else
			{
				values->data_identifier = 0;
			}
			byte_stream_copy_to_uint32_little_endian(
			 element_data,
			 values->sub_node_identifier );

			element_data += 4;
		}
		else if( file_type == LIBPFF_FILE_TYPE_64BIT )
		{
			byte_stream_copy_to_uint32_little_endian(
			 element_data,
			 values->identifier );

			element_data += 4;

			byte_stream_copy_to_uint32_little_endian(
			 element_data,
			 unknown1 );

			element_data += 4;

			if( *node_level == 0 )
			{
				byte_stream_copy_to_uint64_little_endian(
				 element_data,
				 values->data_identifier );

				element_data += 8;
			}
			byte_stream_copy_to_uint64_little_endian(
			 element_data,
			 values->sub_node_identifier );

			element_data += 8;
		}
		
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: element: %03d identifier\t\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
			 function,
			 element_iterator,
			 values->identifier,
			 values->identifier );

			if( file_type == LIBPFF_FILE_TYPE_64BIT )
			{
				libnotify_printf(
				 "%s: element: %03d unknown1\t\t: %" PRIu16 "\n",
				 function,
				 element_iterator,
				 unknown1 );
			}
			if( *node_level == 0 )
			{
				libnotify_printf(
				 "%s: element: %03d data identifier\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
				 function,
				 element_iterator,
				 values->data_identifier,
				 values->data_identifier );
			}
			libnotify_printf(
			 "%s: element: %03d sub node identifier\t: 0x%08" PRIx64 " (%" PRIu64 ")\n",
			 function,
			 element_iterator,
			 values->sub_node_identifier,
			 values->sub_node_identifier );
		}
#endif

		if( libpff_tree_node_append_value(
		     local_descriptors_tree_node,
		     (intptr_t *) values,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to append local descriptor values.",
			 function );

			memory_free(
			 values );
			memory_free(
			 node_data );

			return( -1 );
		}
	}
	memory_free(
	 node_data );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );
}

/* Reads the local descriptors tree node and its elements
 * Returns 1 if successful or -1 on error
 */
int libpff_local_descriptors_read_tree(
     libpff_tree_node_t *local_descriptors_tree_node,
     uint64_t local_descriptors_node_identifier,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_index_values_t *offset_index_values = NULL;
	libpff_tree_node_t *child_node             = NULL;
	libpff_tree_node_t *offset_index_root_node = NULL;
	static char *function                      = "libpff_local_descriptors_read_tree";
	int child_iterator                         = 0;
	uint8_t node_level                         = 0;

	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	/* Determine the file offset of the local descriptors node
	 */
	if( recovered == 0 )
	{
		offset_index_root_node = io_handle->offset_index_root_node;
	}
	else
	{
		offset_index_root_node = io_handle->recovered_offset_index_root_node;
	}
	if( offset_index_root_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid offset index root node.",
		 function );

		return( -1 );
	}
	if( libpff_index_get_index_values(
	     offset_index_root_node,
	     local_descriptors_node_identifier & (uint64_t) LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK,
	     &offset_index_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to find local descriptors identifier: %" PRIu64 ".",
		 function,
		 local_descriptors_node_identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_MASK );

		return( -1 );
	}
	if( offset_index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid offset index values.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: identifier: %" PRIu64 " (%s) at offset: %" PRIi64 " of size: %" PRIu32 "\n",
		 function,
		 offset_index_values->identifier,
		 ( ( offset_index_values->identifier & LIBPFF_OFFSET_INDEX_IDENTIFIER_FLAG_INTERNAL ) ? "internal" : "external" ),
		 offset_index_values->file_offset,
		 offset_index_values->data_size );
	}
#endif

	/* Read the local descriptors node data
	 */
	if( libpff_local_descriptors_read_node(
	     local_descriptors_tree_node,
	     file_io_handle,
	     offset_index_values->file_offset,
	     offset_index_values->data_size,
	     io_handle->file_type,
	     &node_level,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read local descriptors node.",
		 function );

		return( -1 );
	}
	child_node = local_descriptors_tree_node->first_child;

	for( child_iterator = 0;
	     child_iterator < local_descriptors_tree_node->number_of_child_nodes;
	     child_iterator++ )
	{
		if( child_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: corruption detected for local descriptors node: %d.",
			 function,
			 child_iterator );

			return( -1 );
		}
		if( ( (libpff_local_descriptor_values_t *) child_node->value )->sub_node_identifier != 0 )
		{
			if( libpff_local_descriptors_read_tree(
			     child_node,
			     ( (libpff_local_descriptor_values_t *) child_node->value )->sub_node_identifier,
			     recovered,
			     io_handle,
			     file_io_handle,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read local descriptors node: %d.",
				 function,
				 child_iterator );

				return( -1 );
			}
		}
		child_node = child_node->next;
	}
	return( 1 );
}

/* Retrieves the tree node of a local descriptors node
 * Returns 1 if successful, 0 if the local descriptors node was not found or -1 on error
 */
int libpff_local_descriptors_get_tree_node_by_identifier(
     libpff_tree_node_t *local_descriptors_tree_node,
     uint64_t local_descriptor_identifier,
     libpff_tree_node_t **result_local_descriptors_tree_node,
     liberror_error_t **error )
{
	libpff_tree_node_t *child_local_descriptors_tree_node = NULL;
	libpff_local_descriptor_values_t *values              = NULL;
	static char *function                                 = "libpff_local_descriptors_get_tree_node_by_identifier";
	int iterator                                          = 0;
	int result                                            = 0;

	if( local_descriptors_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors tree node.",
		 function );

		return( -1 );
	}
	if( result_local_descriptors_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid result local descriptors tree node.",
		 function );

		return( -1 );
	}
	if( *result_local_descriptors_tree_node != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: result local descriptors tree node already set.",
		 function );

		return( -1 );
	}
	if( local_descriptors_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid local descriptors tree node - missing value.",
		 function );

		return( -1 );
	}
	values = (libpff_local_descriptor_values_t *) local_descriptors_tree_node->value;

	/* Make sure to find a matching local descriptor identifier at level 0
	 */
	if( ( values->level == 0 )
	 && ( values->identifier == local_descriptor_identifier ) )
	{
		*result_local_descriptors_tree_node = local_descriptors_tree_node;

		return( 1 );
	}
	else if( local_descriptors_tree_node->number_of_child_nodes > 0 )
	{
		child_local_descriptors_tree_node = local_descriptors_tree_node->first_child;

		for( iterator = 0;
		     iterator < local_descriptors_tree_node->number_of_child_nodes;
		     iterator++ )
		{
			if( child_local_descriptors_tree_node == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: corruption detected for child node: %d.",
				 function,
				 iterator );

				return( -1 );
			}
			result = libpff_local_descriptors_get_tree_node_by_identifier(
				  child_local_descriptors_tree_node,
				  local_descriptor_identifier,
				  result_local_descriptors_tree_node,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to traverse child node: %d.",
				 function,	
				 iterator );

				return( -1 );
			}
			else if( result == 1 )
			{
				break;
			}
			child_local_descriptors_tree_node = child_local_descriptors_tree_node->next;
		}
	}
	return( result );
}

/* Retrieves the reference of the values of a local descriptor
 * Returns 1 if successful, 0 if the local descriptor was not found or -1 on error
 */
int libpff_local_descriptors_get_values(
     libpff_tree_node_t *local_descriptors_tree_node,
     uint64_t local_descriptor_identifier,
     libpff_local_descriptor_values_t **local_descriptors_values,
     liberror_error_t **error )
{
	libpff_tree_node_t *result_local_descriptors_tree_node = NULL;
	static char *function                                  = "libpff_local_descriptors_get_values";
	int result                                             = 0;

	if( local_descriptors_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors tree node.",
		 function );

		return( -1 );
	}
	if( local_descriptors_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid local descriptors values.",
		 function );

		return( -1 );
	}
	if( *local_descriptors_values != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: local descriptor values already set.",
		 function );

		return( -1 );
	}
	result = libpff_local_descriptors_get_tree_node_by_identifier(
	          local_descriptors_tree_node,
	          local_descriptor_identifier,
	          &result_local_descriptors_tree_node,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to find local descriptors tree node: %" PRIu64 ".",
		 function,
		 local_descriptor_identifier );

		return( -1 );
	}
	else if( result == 1 )
	{
		if( result_local_descriptors_tree_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid result local descriptors tree node.",
			 function );

			return( -1 );
		}
		*local_descriptors_values = (libpff_local_descriptor_values_t *) result_local_descriptors_tree_node->value;
	}
	return( result );
}

