/*
 * Message functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_data_reference.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_item.h"
#include "libpff_item_descriptor.h"
#include "libpff_item_values.h"
#include "libpff_libfmapi.h"
#include "libpff_mapi.h"
#include "libpff_message.h"
#include "libpff_item.h"
#include "libpff_libuna.h"
#include "libpff_value_type.h"

#define LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS	0
#define LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS	1

/* Retrieves the UTF-8 string size of a specific entry from the referenced message item
 * Length includes the end of string character
 * Returns 1 if successful, 0 if the message does not contain such value or -1 on error
 */
int libpff_message_get_entry_value_utf8_string_size(
     libpff_item_t *message,
     uint32_t entry_type,
     size_t *utf8_string_size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_entry_value_utf8_string_size";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( utf8_string_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-8 string size.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  entry_type,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			result = libpff_item_get_entry_value_32bit(
				  message,
				  0,
				  LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
				  &ascii_codepage,
				  0,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			else if( result == 0 )
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_get_utf8_string_size(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     utf8_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine UTF-8 string size.",
			 function );

			return( -1 );
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the UTF-8 string value of a specific entry from the referenced message item
 * The function uses a codepage if necessary, it uses the codepage set for the library
 * Returns 1 if successful, 0 if the message does not contain such value or -1 on error
 */
int libpff_message_get_entry_value_utf8_string(
     libpff_item_t *message,
     uint32_t entry_type,
     uint8_t *utf8_string,
     size_t utf8_string_size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_entry_value_utf8_string";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( utf8_string == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-8 string.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  entry_type,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			result = libpff_item_get_entry_value_32bit(
				  message,
				  0,
				  LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
				  &ascii_codepage,
				  0,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			else if( result == 0 )
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_copy_to_utf8_string(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     utf8_string,
		     utf8_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-8 string.",
			 function );

			return( -1 );
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the UTF-16 string size of a specific entry from the referenced message item
 * Length includes the end of string character
 * Returns 1 if successful, 0 if the message does not contain such value or -1 on error
 */
int libpff_message_get_entry_value_utf16_string_size(
     libpff_item_t *message,
     uint32_t entry_type,
     size_t *utf16_string_size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_entry_value_utf16_string_size";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( utf16_string_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-16 string size.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  entry_type,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			result = libpff_item_get_entry_value_32bit(
				  message,
				  0,
				  LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
				  &ascii_codepage,
				  0,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			else if( result == 0 )
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_get_utf16_string_size(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     utf16_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine UTF-16 string size.",
			 function );

			return( -1 );
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the UTF-16 string value of a specific entry from the referenced message item
 * The function uses a codepage if necessary, it uses the codepage set for the library
 * Returns 1 if successful, 0 if the message does not contain such value or -1 on error
 */
int libpff_message_get_entry_value_utf16_string(
     libpff_item_t *message,
     uint32_t entry_type,
     uint16_t *utf16_string,
     size_t utf16_string_size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_entry_value_utf16_string";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( utf16_string == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-16 string.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  entry_type,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			result = libpff_item_get_entry_value_32bit(
				  message,
				  0,
				  LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
				  &ascii_codepage,
				  0,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			else if( result == 0 )
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_copy_to_utf16_string(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     utf16_string,
		     utf16_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-16 string.",
			 function );

			return( -1 );
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Determine if the message item has attachments
 * Returns 1 if successful or -1 on error
 */
int libpff_message_determine_attachments(
     libpff_internal_item_t *internal_item,
     liberror_error_t **error )
{
	libpff_data_reference_t *value_data_reference         = NULL;
	libpff_item_descriptor_t *item_descriptor             = NULL;
	libpff_item_descriptor_t *attachments_item_descriptor = NULL;
	libpff_item_descriptor_t *attachment_item_descriptor  = NULL;
	uint8_t *value_data                                   = NULL;
	static char *function                                 = "libpff_message_determine_attachments";
	size_t value_data_size                                = 0;
	uint32_t attachment_descriptor_identifier             = 0;
	uint32_t value_type                                   = LIBPFF_VALUE_TYPE_INTEGER_32BIT_SIGNED;
	int attachment_iterator                               = 0;
	int result                                            = 0;

	if( internal_item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid internal item.",
		 function );

		return( -1 );
	}
	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) internal_item->item_tree_node->value;

	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing item values.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: attachments sub item tree node already set.",
		 function );

		return( -1 );
	}
	/* Make sure the item values have been read
	 */
	if( internal_item->item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 item_descriptor->descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          internal_item->item_values,
		          item_descriptor->descriptor_identifier,
		          item_descriptor->recovered,
		          internal_item->internal_file->name_to_id_map_list,
		          internal_item->internal_file->io_handle,
		          internal_item->file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          internal_item->internal_file->ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
		if( internal_item->item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	/* Determine if the item has attachments
	 */
	if( internal_item->item_values->local_descriptors != NULL )
	{
		/* Cache the attachments items values for successive usage
		 */
		if( libpff_item_values_initialize(
		     &( internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create attachments item values.",
			 function );

			return( -1 );
		}
		result = libpff_item_values_local_read(
			  internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
			  LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_ATTACHMENTS,
			  item_descriptor->recovered,
			  internal_item->item_values->local_descriptors,
			  internal_item->internal_file->name_to_id_map_list,
			  internal_item->internal_file->io_handle,
			  internal_item->file_io_handle,
			  LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
			  internal_item->internal_file->ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: error while read local descriptor identifier: %" PRIu32 ".",
			 function,
			 LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_ATTACHMENTS );

			return( -1 );
		}
		else if( result == 1 )
		{
			if( internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ]->table == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid attachments item values - missing table.",
				 function );

				return( -1 );
			}
			/* Create an attachments item descriptor
			 */
			if( libpff_item_descriptor_initialize(
			     &attachments_item_descriptor,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create attachments item descriptor.",
				 function );

				return( -1 );
			}
			attachments_item_descriptor->descriptor_identifier = LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_ATTACHMENTS;
			attachments_item_descriptor->recovered             = item_descriptor->recovered;

			if( libpff_tree_node_initialize(
			     &( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create attachments sub item tree node.",
				 function );

				memory_free(
				 attachments_item_descriptor );

				return( -1 );
			}
			if( libpff_tree_node_set_value(
			     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
			     (intptr_t *) attachments_item_descriptor,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_SET_FAILED,
				 "%s: unable to set attachments item descriptor in attachments sub item tree node.",
				 function );

				libpff_tree_node_free(
				 &( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] ),
				 NULL,
				 NULL );
				memory_free(
				 attachments_item_descriptor );

				return( -1 );
			}
			for( attachment_iterator = 0;
			     attachment_iterator < internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ]->table->number_of_sets;
			     attachment_iterator++ )
			{
				if( libpff_item_values_get_entry_value(
				     internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
				     attachments_item_descriptor->descriptor_identifier,
				     attachments_item_descriptor->recovered,
				     internal_item->internal_file->name_to_id_map_list,
				     internal_item->internal_file->io_handle,
				     internal_item->file_io_handle,
				     attachment_iterator,
				     LIBPFF_ENTRY_TYPE_SUB_ITEM_IDENTIFIER,
				     &value_type,
				     &value_data_reference,
				     0,
				     internal_item->internal_file->ascii_codepage,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve entry value.",
					 function );

					return( -1 );
				}
				if( libpff_data_reference_get_data(
				     value_data_reference,
				     internal_item->file_io_handle,
				     &value_data,
				     &value_data_size,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve value data reference data.",
					 function );

					return( -1 );
				}
				if( libpff_value_type_copy_to_32bit(
				     value_data,
				     value_data_size,
				     &attachment_descriptor_identifier,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_CONVERSION,
					 LIBERROR_CONVERSION_ERROR_GENERIC,
					 "%s: unable to set 32-bit entry value.",
					 function );

					return( -1 );
				}
				/* Create an attachment item descriptor
				 */
				attachment_item_descriptor = NULL;

				if( libpff_item_descriptor_initialize(
				     &attachment_item_descriptor,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
					 "%s: unable to create attachment item descriptor.",
					 function );

					return( -1 );
				}
				attachment_item_descriptor->descriptor_identifier = attachment_descriptor_identifier;
				attachment_item_descriptor->recovered             = item_descriptor->recovered;

				if( libpff_tree_node_append_value(
				     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
				     (intptr_t *) attachment_item_descriptor,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
					 "%s: unable to append attachment item descriptor to item tree node.",
					 function );

					memory_free(
					 attachment_item_descriptor );

					return( -1 );
				}
			}
		}
	}
	return( 1 );
}

/* Determine if the message item has recipients
 * Returns 1 if successful or -1 on error
 */
int libpff_message_determine_recipients(
     libpff_internal_item_t *internal_item,
     liberror_error_t **error )
{
	libpff_item_descriptor_t *item_descriptor            = NULL;
	libpff_item_descriptor_t *recipients_item_descriptor = NULL;
	static char *function                                = "libpff_message_determine_recipients";
	int result                                           = 0;

	if( internal_item == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid internal item.",
		 function );

		return( -1 );
	}
	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->item_tree_node->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - invalid item tree node - missing value.",
		 function );

		return( -1 );
	}
	item_descriptor = (libpff_item_descriptor_t *) internal_item->item_tree_node->value;

	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid internal item - missing item values.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: recipients sub item tree node already set.",
		 function );

		return( -1 );
	}
	/* Make sure the item values have been read
	 */
	if( internal_item->item_values->table == NULL )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: reading item values of descriptor: %" PRIu32 "\n",
			 function,
			 item_descriptor->descriptor_identifier );
		}
#endif

		result = libpff_item_values_read(
		          internal_item->item_values,
		          item_descriptor->descriptor_identifier,
		          item_descriptor->recovered,
		          internal_item->internal_file->name_to_id_map_list,
		          internal_item->internal_file->io_handle,
		          internal_item->file_io_handle,
		          LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
		          internal_item->internal_file->ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read item values.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
		if( internal_item->item_values->table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid item values - missing table.",
			 function );

			return( -1 );
		}
	}
	/* Determine if the item has recipients
	 */
	if( internal_item->item_values->local_descriptors != NULL )
	{
		/* Cache the recipients items values for successive usage
		 */
		if( libpff_item_values_initialize(
		     &( internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] ),
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create recipients item values.",
			 function );

			return( -1 );
		}
		result = libpff_item_values_local_read(
			  internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ],
			  LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_RECIPIENTS,
			  item_descriptor->recovered,
			  internal_item->item_values->local_descriptors,
			  internal_item->internal_file->name_to_id_map_list,
			  internal_item->internal_file->io_handle,
			  internal_item->file_io_handle,
			  LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
			  internal_item->internal_file->ascii_codepage,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: error while read local descriptor identifier: %" PRIu32 ".",
			 function,
			 LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_RECIPIENTS );

			return( -1 );
		}
		else if( result == 1 )
		{
			if( internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ]->table == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid recipients item values - missing table.",
				 function );

				return( -1 );
			}
			/* Create an recipients item descriptor
			 */
			if( libpff_item_descriptor_initialize(
			     &recipients_item_descriptor,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create recipients item descriptor.",
				 function );

				return( -1 );
			}
			recipients_item_descriptor->descriptor_identifier = LIBPFF_LOCAL_DESCRIPTOR_IDENTIFIER_RECIPIENTS;
			recipients_item_descriptor->recovered             = item_descriptor->recovered;

			if( libpff_tree_node_initialize(
			     &( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create recipients sub item tree node.",
				 function );

				memory_free(
				 recipients_item_descriptor );

				return( -1 );
			}
			if( libpff_tree_node_set_value(
			     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ],
			     (intptr_t *) recipients_item_descriptor,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_SET_FAILED,
				 "%s: unable to set recipients item descriptor in recipients sub item tree node.",
				 function );

				libpff_tree_node_free(
				 &( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] ),
				 NULL,
				 NULL );
				memory_free(
				 recipients_item_descriptor );

				return( -1 );
			}
			/* All the recipient data is in the recipients item
			 * there are no sub items like for the attachments item
			 */
		}
	}
	return( 1 );
}

/* Retrieves the number of attachments from a message item
 * Returns 1 if successful or -1 on error
 */
int libpff_message_get_number_of_attachments(
     libpff_item_t *message,
     int *number_of_attachments,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item  = NULL;
	static char *function                  = "libpff_message_get_number_of_attachments";

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item tree node.",
		 function );

		return( -1 );
	}
	/* TODO add item type check */

	if( number_of_attachments == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid number of attachments.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] == NULL )
	{
		if( libpff_message_determine_attachments(
		     internal_item,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine attachments.",
			 function );

			return( -1 );
		}
	}
	*number_of_attachments = 0;

	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] != NULL )
	{
		*number_of_attachments = internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ]->number_of_child_nodes;
	}
	return( 1 );
}

/* Retrieves the attachment for the specific index from a message item
 * Returns 1 if successful or -1 on error
 */
int libpff_message_get_attachment(
     libpff_item_t *message,
     int attachment_index,
     libpff_item_t **attachment,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item    = NULL;
	libpff_tree_node_t *attachment_tree_node = NULL;
	static char *function                    = "libpff_message_get_attachment";
	int result                               = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item values.",
		 function );

		return( -1 );
	}
	/* TODO add item type check */

	if( attachment == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid attachment.",
		 function );

		return( -1 );
	}
	if( *attachment != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: attachment already set.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] == NULL )
	{
		if( libpff_message_determine_attachments(
		     internal_item,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine attachments.",
			 function );

			return( -1 );
		}
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] != NULL )    
	{
		if( ( attachment_index < 0 )
		 || ( (uint64_t) attachment_index > (uint64_t) UINT32_MAX )
		 || ( attachment_index >= internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ]->number_of_child_nodes ) )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: invalid attachment index value out of range.",
			 function );

			return( -1 );
		}
		if( libpff_tree_node_get_child_node(
		     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
		     attachment_index,
		     &attachment_tree_node,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve attachment item tree node.",
			 function );

			return( -1 );
		}
		if( attachment_tree_node == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid attachment item tree node.",
			 function );

			return( -1 );
		}
		if( libpff_item_initialize(
		     attachment,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create attachment.",
			 function );

			return( -1 );
		}
		if( *attachment == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid attachment.",
			 function );

			return( -1 );
		}
		( (libpff_internal_item_t *) *attachment )->type = LIBPFF_ITEM_TYPE_ATTACHMENT;

		if( libpff_item_attach(
		     (libpff_internal_item_t *) *attachment,
		     internal_item->file_io_handle,
		     internal_item->internal_file,
		     attachment_tree_node,
		     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE | LIBPFF_ITEM_FLAG_MANAGED_ITEM_TREE_NODE,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to attach attachment.",
			 function );

			libpff_item_free(
			 attachment,
			 NULL );

			return( -1 );
		}
		/* This is necessary otherwise the library looks in the
		 * the (global) descriptor list for the item descriptor
		 */
		/* TODO is there a better way to solve this?
		 */
		result = libpff_item_values_local_read(
			  ( (libpff_internal_item_t *) *attachment )->item_values,
			  ( (libpff_item_descriptor_t *) attachment_tree_node->value )->descriptor_identifier,
		          ( (libpff_item_descriptor_t *) attachment_tree_node->value )->recovered,
			  internal_item->item_values->local_descriptors,
			  internal_item->internal_file->name_to_id_map_list,
			  internal_item->internal_file->io_handle,
			  internal_item->file_io_handle,
			  LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
			  internal_item->internal_file->ascii_codepage,
		          error );

		if( result != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: error while read local descriptor identifier: %" PRIu32 ".",
			 function,
			 ( (libpff_item_descriptor_t *) attachment_tree_node->value )->descriptor_identifier );

			libpff_item_free(
			 attachment,
			 NULL );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the attachments from a message item
 * Returns 1 if successful, 0 if the message item does not contain such value or -1 on error
 */
int libpff_message_get_attachments(
     libpff_item_t *message,
     libpff_item_t **attachments,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	static char *function                 = "libpff_message_get_attachments";

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item values.",
		 function );

		return( -1 );
	}
	/* TODO add item type check */

	if( attachments == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid attachments.",
		 function );

		return( -1 );
	}
	if( *attachments != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: attachments already set.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] == NULL )
	{
		if( libpff_message_determine_attachments(
		     internal_item,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine attachments.",
			 function );

			return( -1 );
		}
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ] == NULL )
	{
		return( 0 );
	}
	if( libpff_item_initialize(
	     attachments,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create attachments.",
		 function );

		return( -1 );
	}
	if( *attachments == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid attachments.",
		 function );

		return( -1 );
	}
	( (libpff_internal_item_t *) *attachments )->type = LIBPFF_ITEM_TYPE_ATTACHMENTS;

	if( libpff_item_attach(
	     (libpff_internal_item_t *) *attachments,
	     internal_item->file_io_handle,
	     internal_item->internal_file,
	     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE | LIBPFF_ITEM_FLAG_MANAGED_ITEM_TREE_NODE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach attachments.",
		 function );

		libpff_item_free(
		 attachments,
		 NULL );

		return( -1 );
	}
	/* Clones the item values sub elements from the cached sub item values
	 */
	if( libpff_item_values_clone_copy(
	     ( (libpff_internal_item_t *) *attachments )->item_values,
	     internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_ATTACHMENTS ],
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_COPY_FAILED,
		 "%s: unable to copy attachments item values.",
		 function );

		libpff_item_free(
		 attachments,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the recipients from a message item
 * Returns 1 if successful, 0 if the message item does not contain such value or -1 on error
 */
int libpff_message_get_recipients(
     libpff_item_t *message,
     libpff_item_t **recipients,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	static char *function                 = "libpff_message_get_recipients";

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->item_tree_node == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item tree node.",
		 function );

		return( -1 );
	}
	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing internal file.",
		 function );

		return( -1 );
	}
	if( internal_item->item_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid message - missing item values.",
		 function );

		return( -1 );
	}
	/* TODO add item type check */

	if( recipients == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid recipients.",
		 function );

		return( -1 );
	}
	if( *recipients != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: recipients already set.",
		 function );

		return( -1 );
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] == NULL )
	{
		if( libpff_message_determine_recipients(
		     internal_item,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine recipients.",
			 function );

			return( -1 );
		}
	}
	if( internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ] == NULL )
	{
		return( 0 );
	}
	if( libpff_item_initialize(
	     recipients,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create recipients.",
		 function );

		return( -1 );
	}
	if( *recipients == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid recipients.",
		 function );

		return( -1 );
	}
	( (libpff_internal_item_t *) *recipients )->type = LIBPFF_ITEM_TYPE_RECIPIENTS;

	if( libpff_item_attach(
	     (libpff_internal_item_t *) *recipients,
	     internal_item->file_io_handle,
	     internal_item->internal_file,
	     internal_item->sub_item_tree_node[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ],
	     LIBPFF_ITEM_FLAG_MANAGED_FILE_IO_HANDLE | LIBPFF_ITEM_FLAG_MANAGED_ITEM_TREE_NODE,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to attach recipients.",
		 function );

		libpff_item_free(
		 recipients,
		 NULL );

		return( -1 );
	}
	/* Clones the item values sub elements from the cached sub item values
	 */
	if( libpff_item_values_clone_copy(
	     ( (libpff_internal_item_t *) *recipients )->item_values,
	     internal_item->sub_item_values[ LIBPFF_MESSAGE_SUB_ITEM_RECIPIENTS ],
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_COPY_FAILED,
		 "%s: unable to copy recipients item values.",
		 function );

		libpff_item_free(
		 recipients,
		 NULL );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the plain text message body size
 * Size includes the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_plain_text_body_size(
     libpff_item_t *message,
     size_t *size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_plain_text_body_size";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t message_codepage             = 0;
	uint32_t message_body_codepage        = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid size.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_PLAIN_TEXT,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			if( libpff_item_get_entry_value_32bit(
			     message,
			     0,
			     LIBPFF_ENTRY_TYPE_MESSAGE_BODY_CODEPAGE,
			     &message_body_codepage,
			     0,
			     error ) == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message body codepage.",
				 function );

				return( -1 );
			}
			if( libpff_item_get_entry_value_32bit(
			     message,
			     0,
			     LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
			     &message_codepage,
			     0,
			     error ) == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			if( message_body_codepage != 0 )
			{
				ascii_codepage = message_body_codepage;
			}
			else if( message_codepage != 0 )
			{
				ascii_codepage = message_codepage;
			}
			else
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_get_utf8_string_size(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     size,
		     error ) != 1 )
		{
			/* Sometimes the message codepage is not available
			 */
			if( ( message_codepage == message_body_codepage )
			 || ( message_codepage == 0 ) )
			{
				message_codepage = internal_item->internal_file->ascii_codepage;
			}
			/* Sometimes the message body codepade is not the one used to encode
			 * the message body, so try the message codepage as well
			 */
			if( ( ascii_codepage == message_codepage )
			 || ( libpff_value_type_get_utf8_string_size(
			       value_data,
			       value_data_size,
			       is_ascii_string,
			       message_codepage,
			       size,
			       error ) != 1 ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to determine message body size.",
				 function );

				return( -1 );
			}
			liberror_error_free(
			 error );
		}
	}
	else if( value_type == LIBPFF_VALUE_TYPE_BINARY_DATA )
	{
		/* Add the end of string byte
		 */
		*size = value_data_size + 1;
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the plain text message body
 * The body is formatted in UTF-8
 * The function uses a codepage if necessary, it tries to read PidTagInternetCodepage first,
 * after that PidTagMessageCodepage and if not available it uses the codepage set for the library
 * Size should include the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_plain_text_body(
     libpff_item_t *message,
     uint8_t *message_body,
     size_t size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_plain_text_body";
	size_t value_data_size                = 0;
	uint32_t ascii_codepage               = 0;
	uint32_t message_codepage             = 0;
	uint32_t message_body_codepage        = 0;
	uint32_t value_type                   = 0;
	uint8_t is_ascii_string               = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( message_body == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message body.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_PLAIN_TEXT,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE ) )
	{
		if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		{
			if( libpff_item_get_entry_value_32bit(
			     message,
			     0,
			     LIBPFF_ENTRY_TYPE_MESSAGE_BODY_CODEPAGE,
			     &message_body_codepage,
			     0,
			     error ) == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message body codepage.",
				 function );

				return( -1 );
			}
			if( libpff_item_get_entry_value_32bit(
			     message,
			     0,
			     LIBPFF_ENTRY_TYPE_MESSAGE_CODEPAGE,
			     &message_codepage,
			     0,
			     error ) == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve the message codepage.",
				 function );

				return( -1 );
			}
			if( message_body_codepage != 0 )
			{
				ascii_codepage = message_body_codepage;
			}
			else if( message_codepage != 0 )
			{
				ascii_codepage = message_codepage;
			}
			else
			{
				ascii_codepage = (uint32_t) internal_item->internal_file->ascii_codepage;
			}
			is_ascii_string = 1;
		}
		if( libpff_value_type_copy_to_utf8_string(
		     value_data,
		     value_data_size,
		     is_ascii_string,
		     ascii_codepage,
		     message_body,
		     size,
		     error ) != 1 )
		{
			/* Sometimes the message codepage is not available
			 */
			if( ( message_codepage == message_body_codepage )
			 || ( message_codepage == 0 ) )
			{
				message_codepage = internal_item->internal_file->ascii_codepage;
			}
			/* Sometimes the message body codepade is not the one used to encode
			 * the message body, so try the message codepage as well
			 */
			if( ( ascii_codepage == message_codepage )
			 || ( libpff_value_type_copy_to_utf8_string(
			       value_data,
			       value_data_size,
			       is_ascii_string,
			       message_codepage,
			       message_body,
			       size,
			       error ) != 1 ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_CONVERSION,
				 LIBERROR_CONVERSION_ERROR_GENERIC,
				 "%s: unable to set message body.",
				 function );

				return( -1 );
			}
			liberror_error_free(
			 error );
		}
	}
	else if( value_type == LIBPFF_VALUE_TYPE_BINARY_DATA )
	{
		if( size < value_data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
			 "%s: message body size too small.",
			 function );

			return( -1 );
		}
		if( memory_copy(
		     message_body,
		     value_data,
		     value_data_size ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to set message body .",
			 function );

			return( -1 );
		}
		/* Add the end of string byte
		 */
		message_body[ value_data_size ] = 0;
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the RTF message body size
 * Size includes the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_rtf_body_size(
     libpff_item_t *message,
     size_t *size,
     liberror_error_t **error )
{
	uint8_t *value_data   = NULL;
	static char *function = "libpff_message_get_rtf_body_size";
	uint32_t value_type   = LIBPFF_VALUE_TYPE_BINARY_DATA;
	int result            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	if( size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid size.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_COMPRESSED_RTF,
		  &value_type,
		  &value_data,
		  size,
		  0,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( value_data == NULL )
	{
		*size = 0;
	}
	else if( libfmapi_lzfu_get_uncompressed_data_size(
	          value_data,
	          *size,
	          size,
	          error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_COMPRESSION,
		 LIBERROR_COMPRESSION_ERROR_UNCOMPRESS_FAILED,
		 "%s: unable to determine uncompressed RTF body size.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the RTF message body
 * The body is formatted in UTF-8
 * Size should include the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_rtf_body(
     libpff_item_t *message,
     uint8_t *message_body,
     size_t size,
     liberror_error_t **error )
{
	uint8_t *value_data    = NULL;
	static char *function  = "libpff_message_get_rtf_body";
	size_t value_data_size = 0;
	uint32_t value_type    = LIBPFF_VALUE_TYPE_BINARY_DATA;
	int result             = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	if( message_body == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message body.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_COMPRESSED_RTF,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  0,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( libfmapi_lzfu_decompress(
	     message_body,
	     &size,
	     value_data,
	     value_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_COMPRESSION,
		 LIBERROR_COMPRESSION_ERROR_UNCOMPRESS_FAILED,
		 "%s: unable to decompress message body.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the HTML message body size
 * Size includes the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_html_body_size(
     libpff_item_t *message,
     size_t *size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_html_body_size";
	size_t value_data_size                = 0;
	uint32_t value_type                   = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid size.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_HTML,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( value_data == NULL )
	{
		return( 0 );
	}
	if( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	{
		*size = value_data_size;

		if( value_data[ value_data_size - 1 ] != 0 )
		{
			/* Add the end of string byte
			 */
			*size += 1;
		}
	}
	else if( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE )
	{
		*size = value_data_size;

		if( ( value_data[ value_data_size - 1 ] != 0 )
		 || ( value_data[ value_data_size - 2 ] != 0 ) )
		{
			/* Add the end of string byte
			 */
			*size += 2;
		}
	}
	else if( value_type == LIBPFF_VALUE_TYPE_BINARY_DATA )
	{
		*size = value_data_size;

		if( value_data[ value_data_size - 1 ] != 0 )
		{
			/* Add the end of string byte
			 */
			*size += 1;
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the HTML message body
 * Size should include the end of string character
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_message_get_html_body(
     libpff_item_t *message,
     uint8_t *message_body,
     size_t size,
     liberror_error_t **error )
{
	libpff_internal_item_t *internal_item = NULL;
	uint8_t *value_data                   = NULL;
	static char *function                 = "libpff_message_get_html_body";
	size_t value_data_size                = 0;
	uint32_t value_type                   = 0;
	int result                            = 0;

	if( message == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message.",
		 function );

		return( -1 );
	}
	internal_item = (libpff_internal_item_t *) message;

	if( internal_item->internal_file == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item - missing internal file.",
		 function );

		return( -1 );
	}
	if( message_body == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid message body.",
		 function );

		return( -1 );
	}
	result = libpff_item_get_entry_value(
		  message,
		  0,
		  LIBPFF_ENTRY_TYPE_MESSAGE_BODY_HTML,
		  &value_type,
		  &value_data,
		  &value_data_size,
		  1,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve entry value.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( value_data == NULL )
	{
		return( 0 );
	}
	if( ( value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
	 || ( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE )
	 || ( value_type == LIBPFF_VALUE_TYPE_BINARY_DATA ) )
	{
		if( size < value_data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
			 "%s: message body size too small.",
			 function );

			return( -1 );
		}
		if( memory_copy(
		     message_body,
		     value_data,
		     value_data_size ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_COPY_FAILED,
			 "%s: unable to set message body .",
			 function );

			return( -1 );
		}
		/* Add the end of string byte
		 */
		message_body[ value_data_size - 1 ] = 0;

		if( value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE )
		{
			message_body[ value_data_size - 2 ] = 0;
		}
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported value type: 0x%04" PRIx32 ".",
		 function,
		 value_type );

		return( -1 );
	}
	return( 1 );
}

