/*
 * Name to ID map functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_codepage.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_mapi.h"
#include "libpff_name_to_id_map.h"
#include "libpff_index.h"
#include "libpff_io_handle.h"
#include "libpff_item_values.h"
#include "libpff_libbfio.h"
#include "libpff_libfmapi.h"
#include "libpff_libuna.h"
#include "libpff_list_type.h"
#include "libpff_types.h"
#include "libpff_value_type.h"

#include "pff_value_data.h"

/* Creates a name to id map entry
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_initialize(
     libpff_internal_name_to_id_map_entry_t **name_to_id_map_entry,
     liberror_error_t **error )
{
	static char *function = "libpff_name_to_id_map_entry_initialize";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	if( *name_to_id_map_entry == NULL )
	{
		*name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) memory_allocate(
		                                                                    sizeof( libpff_internal_name_to_id_map_entry_t ) );

		if( *name_to_id_map_entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create name to id map entry.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *name_to_id_map_entry,
		     0,
		     sizeof( libpff_internal_name_to_id_map_entry_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear name to id map entry.",
			 function );

			memory_free(
			 *name_to_id_map_entry );

			*name_to_id_map_entry = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees a name to id map entry
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_free(
     intptr_t *name_to_id_map_entry,
     liberror_error_t **error )
{
	if( name_to_id_map_entry != NULL )
	{
		if( ( ( (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry )->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
		 && ( ( (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry )->string_value != NULL ) )
		{
			memory_free(
			 ( (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry )->string_value );
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( ( (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry )->debug_string != NULL )
		{
			memory_free(
			 ( (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry )->debug_string );
		}
#endif
		memory_free(
		 name_to_id_map_entry );
	}
	return( 1 );
}

/* Read the name to id map
 * Return 1 if successful, 0 if not available or -1 on error
 */
int libpff_name_to_id_map_read(
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *name_to_id_map_list,
     liberror_error_t **error )
{
	libpff_data_reference_t *name_to_id_map_class_identifiers_data_reference = NULL;
	libpff_data_reference_t *name_to_id_map_entries_data_reference           = NULL;
	libpff_data_reference_t *name_to_id_map_strings_data_reference           = NULL;
	libpff_item_values_t *item_values                                        = NULL;
	libpff_internal_name_to_id_map_entry_t *name_to_id_map_entry             = NULL;
	uint8_t *name_to_id_map_class_identifiers_data                           = NULL;
	uint8_t *name_to_id_map_entries_data                                     = NULL;
	uint8_t *name_to_id_map_entry_data                                       = NULL;
	uint8_t *name_to_id_map_string_data                                      = NULL;
	uint8_t *name_to_id_map_strings_data                                     = NULL;
	static char *function                                                    = "libpff_name_to_id_map_read";
	size_t name_to_id_map_class_identifiers_data_size                        = 0;
	size_t name_to_id_map_entries_data_size                                  = 0;
	size_t name_to_id_map_strings_data_size                                  = 0;
	uint32_t number_of_name_to_id_map_entries                                = 0;
	uint32_t name_to_id_map_string_size                                      = 0;
	uint32_t name_to_id_map_entry_iterator                                   = 0;
	uint32_t name_to_id_map_entry_value                                      = 0;
	uint32_t value_type                                                      = LIBPFF_VALUE_TYPE_BINARY_DATA;
	uint16_t name_to_id_map_class_identifier_index                           = 0;
	uint16_t name_to_id_map_entry_number                                     = 0;
	uint16_t name_to_id_map_entry_type                                       = 0;
	int result                                                               = 0;

	if( name_to_id_map_list == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map list.",
		 function );

		return( -1 );
	}
	if( libpff_item_values_initialize(
	     &item_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create item values.",
		 function );

		return( -1 );
	}
	result = libpff_item_values_read(
	          item_values,
	          LIBPFF_DESCRIPTOR_IDENTIFIER_NAME_TO_ID_MAP,
	          0,
	          NULL,
	          io_handle,
	          file_io_handle,
	          LIBPFF_DEBUG_ITEM_TYPE_NAME_TO_ID_MAP,
	          LIBPFF_CODEPAGE_ASCII,
	          error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read name to id map item values.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	else if( result == 0 )
	{
		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( 0 );
	}
	if( item_values->table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid item values - missing table.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	if( libpff_table_get_entry_value_by_entry_type(
	     item_values->table,
	     0,
	     LIBPFF_ENTRY_TYPE_NAME_TO_ID_MAP_ENTRIES,
	     &value_type,
	     &name_to_id_map_entries_data_reference,
	     LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map entries data reference.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	/* Reading the data at once reduces the amount of reads back and forth
	 * also it prevents having to deal with entry boundary in the middle
	 * of entry values
	 */
	if( libpff_data_reference_get_data(
	     name_to_id_map_entries_data_reference,
	     file_io_handle,
	     &name_to_id_map_entries_data,
	     &name_to_id_map_entries_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map entries data.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	if( ( name_to_id_map_entries_data_size % 8 ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entries size.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	number_of_name_to_id_map_entries = (uint32_t) ( name_to_id_map_entries_data_size / 8 );

	if( libpff_table_get_entry_value_by_entry_type(
	     item_values->table,
	     0,
	     LIBPFF_ENTRY_TYPE_NAME_TO_ID_MAP_STRINGS,
	     &value_type,
	     &name_to_id_map_strings_data_reference,
	     LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map strings data reference.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	/* Reading the data at once reduces the amount of reads back and forth
	 * also it prevents having to deal with entry boundary in the middle
	 * of string values
	 */
	if( libpff_data_reference_get_data(
	     name_to_id_map_strings_data_reference,
	     file_io_handle,
	     &name_to_id_map_strings_data,
	     &name_to_id_map_strings_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map strings data.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	if( libpff_table_get_entry_value_by_entry_type(
	     item_values->table,
	     0,
	     LIBPFF_ENTRY_TYPE_NAME_TO_ID_MAP_CLASS_IDENTIFIERS,
	     &value_type,
	     &name_to_id_map_class_identifiers_data_reference,
	     LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map class identifiers data reference.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	/* Reading the data at once reduces the amount of reads back and forth
	 * also it prevents having to deal with entry boundary in the middle
	 * of class identifier values
	 */
	if( libpff_data_reference_get_data(
	     name_to_id_map_class_identifiers_data_reference,
	     file_io_handle,
	     &name_to_id_map_class_identifiers_data,
	     &name_to_id_map_class_identifiers_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve name to id map class idendentifiers data.",
		 function );

		libpff_item_values_free(
		 &item_values,
		 NULL );

		return( -1 );
	}
	name_to_id_map_entry_data = name_to_id_map_entries_data;

	for( name_to_id_map_entry_iterator = 0;
	     name_to_id_map_entry_iterator < number_of_name_to_id_map_entries;
	     name_to_id_map_entry_iterator++ )
	{
		name_to_id_map_entry = NULL;

		if( libpff_name_to_id_map_entry_initialize(
		     &name_to_id_map_entry,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create name to id map entry.",
			 function );

			libpff_item_values_free(
			 &item_values,
			 NULL );

			return( -1 );
		}
		byte_stream_copy_to_uint32_little_endian(
		 ( (pff_name_to_id_map_entry_t *) name_to_id_map_entry_data )->entry_value,
		 name_to_id_map_entry_value );
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_name_to_id_map_entry_t *) name_to_id_map_entry_data )->entry_type,
		 name_to_id_map_entry_type );
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_name_to_id_map_entry_t *) name_to_id_map_entry_data )->entry_number,
		 name_to_id_map_entry_number );

		name_to_id_map_entry->identifier = name_to_id_map_entry_number + 0x8000;

		if( name_to_id_map_entry_type > 5 )
		{
			name_to_id_map_class_identifier_index = (uint16_t) ( ( name_to_id_map_entry_type / 2 ) - 3 );

			if( (size_t) ( name_to_id_map_class_identifier_index * 16 ) > name_to_id_map_class_identifiers_data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid name to id map class identifier index value exceeds class identifiers data size.",
				 function );

				return( -1 );
			}
			if( memory_copy(
			     name_to_id_map_entry->guid,
			     &( name_to_id_map_class_identifiers_data[ name_to_id_map_class_identifier_index * 16 ] ),
			     16 ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to copy name to id map class identifier.",
				 function );

				 return( -1 );
			}
		}
		else if( name_to_id_map_entry_type == 5 )
		{
			if( memory_copy(
			     name_to_id_map_entry->guid,
			     libfmapi_guid_public_strings,
			     16 ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to set GUID.",
				 function );

				 return( -1 );
			}
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: entry: %03d name to id map entry value\t: 0x%08" PRIx32 "\n",
			 function,
			 name_to_id_map_entry_iterator,
			 name_to_id_map_entry_value );

			if( name_to_id_map_entry_type > 5 )
			{
				libnotify_printf(
				 "%s: entry: %03d name to id map entry type\t: 0x%04" PRIx16 " (class identifier: %02" PRIu16 ", class: %s)\n",
				 function,
				 name_to_id_map_entry_iterator,
				 name_to_id_map_entry_type,
				 name_to_id_map_class_identifier_index,
				 libfmapi_class_identifier_get_name(
				  name_to_id_map_entry->guid ) );
			}
			else if( name_to_id_map_entry_type == 5 )
			{
				libnotify_printf(
				 "%s: entry: %03d name to id map entry type\t: 0x%04" PRIx16 " (class: %s)\n",
				 function,
				 name_to_id_map_entry_iterator,
				 name_to_id_map_entry_type,
				 libfmapi_class_identifier_get_name(
				  name_to_id_map_entry->guid ) );
			}
			else
			{
				libnotify_printf(
				 "%s: entry: %03d name to id map entry type\t: 0x%04" PRIx16 "\n",
				 function,
				 name_to_id_map_entry_iterator,
				 name_to_id_map_entry_type );
			}
			libnotify_printf(
			 "%s: entry: %03d name to id map entry number\t: 0x%04" PRIx16 " (0x%04" PRIx32 ")\n",
			 function,
			 name_to_id_map_entry_iterator,
			 name_to_id_map_entry_number,
			 name_to_id_map_entry->identifier );
		}
#endif

		/* The lowest bit of the name to id map entry type signifies
		 * that the name to id map entry value refers to the name to id map string table or the item values
		 */
		if( ( name_to_id_map_entry_type & 0x0001 ) != 0 )
		{
			if( name_to_id_map_entry_value > name_to_id_map_strings_data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid name to id map entry value exceeds strings data size.",
				 function );

				return( -1 );
			}
			name_to_id_map_string_data = &( name_to_id_map_strings_data[ name_to_id_map_entry_value ] );

			byte_stream_copy_to_uint32_little_endian(
			 name_to_id_map_string_data,
			 name_to_id_map_string_size );

			name_to_id_map_string_data += 4;

			name_to_id_map_entry->type = LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING;

			result = libpff_value_type_string_contains_zero_bytes(
			          name_to_id_map_string_data,
			          name_to_id_map_string_size,
			          error ) ;

			if( result == -1 )
			{
                                liberror_error_set(
                                 error,
                                 LIBERROR_ERROR_DOMAIN_RUNTIME,
                                 LIBERROR_RUNTIME_ERROR_GET_FAILED,
                                 "%s: unable to determine name to id map entry string contains zero bytes.",
                                 function );

				memory_free(
				 name_to_id_map_entry );
				libpff_item_values_free(
				 &item_values,
				 NULL );

                                return( -1 );
			}
			else if( result == 0 )
			{
				name_to_id_map_entry->is_ascii_string = 1;
			}
			name_to_id_map_entry->value_size = name_to_id_map_string_size;

			name_to_id_map_entry->string_value = (uint8_t *) memory_allocate(
			                                                  sizeof( uint8_t ) * name_to_id_map_entry->value_size );

			if( name_to_id_map_entry->string_value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create name to id map entry string.",
				 function );

				memory_free(
				 name_to_id_map_entry );
				libpff_item_values_free(
				 &item_values,
				 NULL );

				return( -1 );
			}
			if( memory_copy(
			     name_to_id_map_entry->string_value,
			     name_to_id_map_string_data,
			     name_to_id_map_entry->value_size ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to set name to id map entry string.",
				 function );

				memory_free(
				 name_to_id_map_entry );
				libpff_item_values_free(
				 &item_values,
				 NULL );

				return( -1 );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( name_to_id_map_entry->is_ascii_string == 0 )
				{
					/* TODO handle different system string sizes */
					if( libuna_utf8_string_size_from_utf16_stream(
					     name_to_id_map_entry->string_value,
					     name_to_id_map_entry->value_size,
					     LIBPFF_ENDIAN_LITTLE,
					     &( name_to_id_map_entry->debug_string_size ),
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine name to id map entry UTF-8 string size.",
						 function );

						memory_free(
						 name_to_id_map_entry );
						libpff_item_values_free(
						 &item_values,
						 NULL );

						return( -1 );
					}
				}
				else
				{
					/* TODO handle different system string sizes */
					if( libuna_utf8_string_size_from_byte_stream(
					     name_to_id_map_entry->string_value,
					     name_to_id_map_entry->value_size,
					     LIBUNA_CODEPAGE_ASCII,
					     &( name_to_id_map_entry->debug_string_size ),
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine name to id map entry UTF-8 string size.",
						 function );

						memory_free(
						 name_to_id_map_entry );
						libpff_item_values_free(
						 &item_values,
						 NULL );

						return( -1 );
					}
				}
				name_to_id_map_entry->debug_string = (libcstring_system_character_t *) memory_allocate(
										                        sizeof( libcstring_system_character_t ) * name_to_id_map_entry->debug_string_size );

				if( name_to_id_map_entry->debug_string == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_MEMORY,
					 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to create UTF-8 name to id map entry string.",
					 function );

					memory_free(
					 name_to_id_map_entry );
					libpff_item_values_free(
					 &item_values,
					 NULL );

					return( -1 );
				}
				if( name_to_id_map_entry->is_ascii_string == 0 )
				{
					/* TODO handle different system string sizes */
					if( libuna_utf8_string_copy_from_utf16_stream(
					     (libuna_utf8_character_t *) name_to_id_map_entry->debug_string,
					     name_to_id_map_entry->debug_string_size,
					     name_to_id_map_entry->string_value,
					     name_to_id_map_entry->value_size,
					     LIBPFF_ENDIAN_LITTLE,
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_CONVERSION,
						 LIBERROR_CONVERSION_ERROR_GENERIC,
						 "%s: unable to set UTF-8 name to id map entry string.",
						 function );

						memory_free(
						 name_to_id_map_entry->debug_string );
						memory_free(
						 name_to_id_map_entry );
						libpff_item_values_free(
						 &item_values,
						 NULL );

						return( -1 );
					}
				}
				else
				{
					/* TODO handle different system string sizes */
					if( libuna_utf8_string_copy_from_byte_stream(
					     (libuna_utf8_character_t *) name_to_id_map_entry->debug_string,
					     name_to_id_map_entry->debug_string_size,
					     name_to_id_map_entry->string_value,
					     name_to_id_map_entry->value_size,
					     LIBUNA_CODEPAGE_ASCII,
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_CONVERSION,
						 LIBERROR_CONVERSION_ERROR_GENERIC,
						 "%s: unable to set UTF-8 name to id map entry string.",
						 function );

						memory_free(
						 name_to_id_map_entry->debug_string );
						memory_free(
						 name_to_id_map_entry );
						libpff_item_values_free(
						 &item_values,
						 NULL );

						return( -1 );
					}
				}
				libnotify_printf(
				 "%s: entry: %03d name to id map entry string\t: %s\n",
				 function,
				 name_to_id_map_entry_iterator,
				 name_to_id_map_entry->debug_string );
			}
#endif
		}
		else
		{
			name_to_id_map_entry->type          = LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_NUMERIC;
			name_to_id_map_entry->numeric_value = name_to_id_map_entry_value;
			name_to_id_map_entry->value_size    = 4;
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "\n" );
		}
#endif

		if( libpff_list_append_value(
		     name_to_id_map_list,
		     (intptr_t *) name_to_id_map_entry,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to append name to id map entry to list.",
			 function );

			memory_free(
			 name_to_id_map_entry->string_value );
			memory_free(
			 name_to_id_map_entry );
			libpff_item_values_free(
			 &item_values,
			 NULL );

			return( -1 );
		}
		name_to_id_map_entry_data += sizeof( pff_name_to_id_map_entry_t );
	}
	if( libpff_item_values_free(
	     &item_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free item values.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the type of the name to id map entry
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_type(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     uint8_t *entry_type,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_type";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( entry_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry type.",
		 function );

		return( -1 );
	}
	*entry_type = internal_name_to_id_map_entry->type;

	return( 1 );
}

/* Retrieves the number of the name to id map entry
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_number(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     uint32_t *number,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_number";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( internal_name_to_id_map_entry->type != LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_NUMERIC )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entry type.",
		 function );

		return( -1 );
	}
	if( number == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid number.",
		 function );

		return( -1 );
	}
	*number = internal_name_to_id_map_entry->numeric_value;

	return( 1 );
}

/* Retrieves the UTF-8 string size of the name to id map entry
 * The returned size includes the end of string character
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_utf8_string_size(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     size_t *utf8_string_size,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_utf8_string_size";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( internal_name_to_id_map_entry->type != LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entry type.",
		 function );

		return( -1 );
	}
	if( internal_name_to_id_map_entry->is_ascii_string == 0 )
	{
		if( libuna_utf8_string_size_from_utf16_stream(
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBPFF_ENDIAN_LITTLE,
		     utf8_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine name to id map entry UTF-8 string size.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( libuna_utf8_string_size_from_byte_stream(
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBUNA_CODEPAGE_ASCII,
		     utf8_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine name to id map entry UTF-8 string size.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the UTF-8 string of the name to id map entry
 * The size should include the end of string character
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_utf8_string(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     uint8_t *utf8_string,
     size_t utf8_string_size,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_utf8_string";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( internal_name_to_id_map_entry->type != LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entry type.",
		 function );

		return( -1 );
	}
	if( internal_name_to_id_map_entry->is_ascii_string == 0 )
	{
		if( libuna_utf8_string_copy_from_utf16_stream(
		     utf8_string,
		     utf8_string_size,
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBPFF_ENDIAN_LITTLE,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-8 name to id map entry string.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( libuna_utf8_string_copy_from_byte_stream(
		     utf8_string,
		     utf8_string_size,
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBUNA_CODEPAGE_ASCII,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-8 name to id map entry string.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the UTF-16 string size of the name to id map entry
 * The returned size includes the end of string character
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_utf16_string_size(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     size_t *utf16_string_size,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_utf16_string_size";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( internal_name_to_id_map_entry->type != LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entry type.",
		 function );

		return( -1 );
	}
	if( internal_name_to_id_map_entry->is_ascii_string == 0 )
	{
		if( libuna_utf16_string_size_from_utf16_stream(
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBPFF_ENDIAN_LITTLE,
		     utf16_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine name to id map entry UTF-16 string size.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( libuna_utf16_string_size_from_byte_stream(
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBUNA_CODEPAGE_ASCII,
		     utf16_string_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to determine name to id map entry UTF-16 string size.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the UTF-16 string of the name to id map entry
 * The size should include the end of string character
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_utf16_string(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     uint16_t *utf16_string,
     size_t utf16_string_size,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_utf16_string";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( internal_name_to_id_map_entry->type != LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported name to id map entry type.",
		 function );

		return( -1 );
	}
	if( internal_name_to_id_map_entry->is_ascii_string == 0 )
	{
		if( libuna_utf16_string_copy_from_utf16_stream(
		     utf16_string,
		     utf16_string_size,
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBPFF_ENDIAN_LITTLE,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-16 name to id map entry string.",
			 function );

			return( -1 );
		}
	}
	else
	{
		if( libuna_utf16_string_copy_from_byte_stream(
		     utf16_string,
		     utf16_string_size,
		     internal_name_to_id_map_entry->string_value,
		     internal_name_to_id_map_entry->value_size,
		     LIBUNA_CODEPAGE_ASCII,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_CONVERSION,
			 LIBERROR_CONVERSION_ERROR_GENERIC,
			 "%s: unable to set UTF-16 name to id map entry string.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the GUID of the name to id map entry
 * Returns 1 if successful or -1 on error
 */
int libpff_name_to_id_map_entry_get_guid(
     libpff_name_to_id_map_entry_t *name_to_id_map_entry,
     uint8_t *guid,
     size_t size,
     liberror_error_t **error )
{
	libpff_internal_name_to_id_map_entry_t *internal_name_to_id_map_entry = NULL;
	static char *function                                                 = "libpff_name_to_id_map_entry_get_guid";

	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	internal_name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) name_to_id_map_entry;

	if( guid == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid GUID.",
		 function );

		return( -1 );
	}
	if( size > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( size < 16 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: size is too small.",
		 function );

		return( -1 );
	}
	if( memory_copy(
	     guid,
	     internal_name_to_id_map_entry->guid,
	     16 ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_COPY_FAILED,
		 "%s: unable to set GUID.",
		 function );

		return( -1 );
	}
	return( 1 );
}

