/*
 * Table functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#if !defined( _LIBPFF_TABLE_H )
#define _LIBPFF_TABLE_H

#include <common.h>
#include <types.h>

#include <liberror.h>

#include "libpff_array_type.h"
#include "libpff_data_reference.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"
#include "libpff_list_type.h"
#include "libpff_name_to_id_map.h"
#include "libpff_tree_type.h"
#include "libpff_types.h"

#if defined( __cplusplus )
extern "C" {
#endif

enum LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMATS
{
	/* The entry identifier consists of a MAPI property type
	 * (entry type, value type)
	 */
	LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_MAPI_PROPERTY	= 1,

	/* The entry identifier consists of a GUID
	 */
	LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_GUID		= 2,

	/* The entry identifier consists of a PRQ_ID_SECURE4 value
	 * the PRQ_ID_SECURE4 value is a yet unknown 10-byte value
	 */
	LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_SECURE4		= 3
};

enum LIBPFF_TABLE_ENTRY_FLAGS
{
	/* The data descriptor could not be read
	 */
	LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR		= 0x01,
};

enum LIBPFF_TABLE_FLAGS
{
	/* The data of one or more table entries could not be read
	 */
	LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA			= 0x01,
};

typedef struct libpff_table_index_value libpff_table_index_value_t;

struct libpff_table_index_value
{
	/* The table data array entry of the value
	 */
	uint32_t array_entry;

	/* The offset of the value
	 */
	uint16_t offset;

	/* The size of the value
	 */
	uint16_t size;
};

typedef struct libpff_table_index_values libpff_table_index_values_t;

struct libpff_table_index_values
{
	/* The number of values
	 */
	uint16_t number_of_values;

	/* The table index values
	 */
	libpff_table_index_value_t *value;
};

typedef struct libpff_table_entry_identifier libpff_table_entry_identifier_t;

struct libpff_table_entry_identifier
{
	/* The table entry identifier format
	 */
	uint8_t format;

	/* The entry identifier
	 */
	union
	{
		/* The MAPI property based entry identifier
		 */
		struct
		{
			/* The entry type
			 */
			uint32_t entry_type;

			/* The value type
			 */
			uint32_t value_type;
		};
		/* The GUID based entry identifier
		 */
		uint8_t guid[ 16 ];

		/* The PRQ_ID_SECURE4 based entry identifier
		 */
		uint64_t secure4;
	};
};

typedef struct libpff_table_entry libpff_table_entry_t;

/* Different types of tables have different entry identifier formats
 * Some table entry are mapped to a name or number in the name-to-id map
 */
struct libpff_table_entry
{
	/* The entry identifier
	 */
	libpff_table_entry_identifier_t identifier;

	/* The value data reference
	 */
	libpff_data_reference_t *value_data_reference;

	/* The name to id map entry
	 */
	libpff_internal_name_to_id_map_entry_t *name_to_id_map_entry;

	/* The flags
	 */
	uint8_t flags;
};

typedef struct libpff_table libpff_table_t;

struct libpff_table
{
	/* The table entries
	 */
	libpff_table_entry_t **entry;

	/* The number of table entry sets
	 */
	int number_of_sets;

	/* The number of table entries
	 */
	int number_of_entries;

	/* The flags
	 */
	uint8_t flags;
};

typedef struct libpff_table_entry_definition libpff_table_entry_definition_t;

struct libpff_table_entry_definition
{
	/* The item entry type
	 */
	uint16_t entry_type;

	/* The item entry value type
	 */
	uint16_t value_type;

	/* The values array entry offset
	 */
	uint16_t entry_offset;

	/* The value array entry size
	 */
	uint16_t entry_size;

	/* The item entry values table
	 */
	libpff_table_t *item_entry_values_table;

	/* The name to id map entry
	 */
	libpff_internal_name_to_id_map_entry_t *name_to_id_map_entry;
};

typedef struct libpff_table_values_array_entry libpff_table_values_array_entry_t;

struct libpff_table_values_array_entry
{
	/* The identifier
	 */
	uint32_t identifier;

	/* The entry number
	 */
	uint32_t entry_number;
};

int libpff_table_index_values_free(
     intptr_t *table_index_values,
     liberror_error_t **error );

int libpff_table_index_array_get_value(
     libpff_array_t *table_index_array,
     uint32_t table_index_reference,
     libpff_table_index_value_t **table_index_value,
     liberror_error_t **error );

int libpff_table_entry_definition_free(
     intptr_t *table_entry_definition,
     liberror_error_t **error );

int libpff_table_entries_resize(
     libpff_table_t *table,
     int number_of_sets,
     int number_of_entries,
     liberror_error_t **error );

int libpff_table_entries_clone(
     libpff_table_t *destination,
     libpff_table_t *source,
     liberror_error_t **error );

int libpff_table_initialize(
     libpff_table_t **table,
     liberror_error_t **error );

int libpff_table_free(
     libpff_table_t *table,
     liberror_error_t **error );

int libpff_table_clone(
     libpff_table_t **destination,
     libpff_table_t *source,
     liberror_error_t **error );

int libpff_table_get_entry_type_by_index(
     libpff_table_t *table,
     int set_index,
     int entry_index,
     uint32_t *entry_type,
     uint32_t *value_type,
     libpff_name_to_id_map_entry_t **name_to_id_map_entry,
     liberror_error_t **error );

int libpff_table_get_entry_value_by_index(
     libpff_table_t *table,
     int set_index,
     int entry_type,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error );

int libpff_table_get_entry_value_by_entry_type(
     libpff_table_t *table,
     int set_index,
     uint32_t entry_type,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     uint8_t flags,
     liberror_error_t **error );

int libpff_table_get_entry_value_by_utf8_entry_name(
     libpff_table_t *table,
     int set_index,
     uint8_t *utf8_entry_name,
     size_t utf8_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error );

int libpff_table_get_entry_value_by_utf16_entry_name(
     libpff_table_t *table,
     int set_index,
     uint16_t *utf16_entry_name,
     size_t utf16_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error );

int libpff_table_read(
     libpff_table_t *table,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_index(
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_entries(
     libpff_array_t *table_entries_array,
     uint8_t table_entries_level,
     uint8_t table_entry_identifier_size,
     uint32_t table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_values(
     libpff_table_t *table,
     uint8_t table_type,
     uint32_t table_value_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_6c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_7c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_8c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_9c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_a5_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_ac_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_bc_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_b5_header(
     uint32_t table_header_reference,
     uint8_t *table_entry_identifier_size,
     uint8_t *table_entry_value_size,
     uint8_t *table_entries_level,
     uint32_t *table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_6c_item_entries(
     libpff_table_t *table,
     uint32_t values_array_reference,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entries_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_7c_item_entry_definitions(
     libpff_array_t *table_entry_definitions_array,
     libpff_data_reference_t *table_entry_definition_data_reference,
     int number_of_table_entry_definitions,
     libpff_array_t *table_index_array,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *name_to_id_map_list,
     liberror_error_t **error );

int libpff_table_read_8c_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_entries_array,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_9c_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_entries_array,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_a5_item_entries(
     libpff_table_t *table,
     uint32_t table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error );

int libpff_table_read_ac_item_entry_definitions(
     libpff_array_t *table_entry_definitions_array,
     uint32_t table_entry_definitions_reference,
     int number_of_table_entry_definitions,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_bc_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entries_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_values_array(
     libpff_table_t *table,
     uint32_t table_entries_reference,
     uint32_t values_array_reference,
     uint8_t table_entry_identifier_size,
     uint8_t table_entry_value_size,
     uint16_t values_array_entry_size,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entry_definitions_array,
     libpff_array_t *table_entries_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error );

int libpff_table_read_entry_value(
     libpff_table_t *table,
     int set_index,
     int entry_index,
     uint32_t table_entry_type,
     uint32_t table_entry_value_type,
     uint8_t *table_entry_value,
     size_t table_entry_value_size,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_internal_name_to_id_map_entry_t *name_to_id_map_entry,
     libpff_table_t *item_entry_values_table,
     libpff_data_reference_t *table_data_reference,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif

