/*
 * Table functions
 *
 * Copyright (c) 2010, Joachim Metz <jbmetz@users.sourceforge.net>
 * Copyright (c) 2008-2010, Joachim Metz <forensics@hoffmannbv.nl>,
 * Hoffmann Investigations.
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include <liberror.h>
#include <libnotify.h>

#include "libpff_array_type.h"
#include "libpff_data_array.h"
#include "libpff_data_reference.h"
#include "libpff_debug.h"
#include "libpff_definitions.h"
#include "libpff_descriptor.h"
#include "libpff_index.h"
#include "libpff_io_handle.h"
#include "libpff_libbfio.h"
#include "libpff_libfmapi.h"
#include "libpff_libuna.h"
#include "libpff_list_type.h"
#include "libpff_local_descriptors.h"
#include "libpff_mapi.h"
#include "libpff_name_to_id_map.h"
#include "libpff_table.h"
#include "libpff_tree_type.h"
#include "libpff_types.h"

#include "pff_table.h"

/* Frees table index values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_index_values_free(
     intptr_t *table_index_values,
     liberror_error_t **error )
{
	if( table_index_values != NULL )
	{
		if( ( (libpff_table_index_values_t *) table_index_values )->value != NULL )
		{
			memory_free(
			 ( (libpff_table_index_values_t *) table_index_values )->value );
		}
		memory_free(
		 table_index_values );
	}
	return( 1 );
}

/* Retrieves a specific table index value from the table index array
 * Returns 1 if successful or -1 on error
 */
int libpff_table_index_array_get_value(
     libpff_array_t *table_index_array,
     uint32_t table_index_reference,
     libpff_table_index_value_t **table_index_value,
     liberror_error_t **error )
{
	libpff_table_index_values_t *table_index_values = NULL;
	static char *function                           = "libpff_table_index_array_get_value";
	uint16_t table_index_array_reference            = 0;
	uint16_t table_index_value_reference            = 0;

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	if( ( table_index_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table index reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_index_reference & 0x0000001fUL,
		 table_index_reference );

		return( -1 );
	}
	/* Determine the index array reference
	 */
	table_index_array_reference = (uint16_t) ( table_index_reference >> 16 );

	if( libpff_array_get_entry(
	     table_index_array,
	     (int) table_index_array_reference,
	     (intptr_t **) &table_index_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table index array entry: %" PRIu16 ".",
		 function,
		 table_index_array_reference );

		return( -1 );
	}
	if( table_index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index values.",
		 function );

		return( -1 );
	}
	table_index_value_reference = (uint16_t) ( ( table_index_reference & 0x0000ffe0 ) >> 5 ) - 1;

	if( table_index_value_reference >= table_index_values->number_of_values )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid table index value reference value out of range.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: retrieving table index array entry: %" PRIu16 " value: %" PRIu16 ".\n",
		 function,
		 table_index_array_reference,
		 table_index_value_reference );
	}
#endif
	*table_index_value = &( table_index_values->value[ table_index_value_reference ] );

	return( 1 );
}

/* Frees a table entry definition
 * Returns 1 if successful or -1 on error
 */
int libpff_table_entry_definition_free(
     intptr_t *table_entry_definition,
     liberror_error_t **error )
{
	static char *function = "libpff_table_entry_definition_free";
	int result            = 1;

	if( table_entry_definition != NULL )
	{
		if( ( ( libpff_table_entry_definition_t *) table_entry_definition )->item_entry_values_table != NULL )
		{
			if( libpff_table_free(
			     ( ( ( libpff_table_entry_definition_t *) table_entry_definition )->item_entry_values_table ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
				 "%s: unable to free item entry values table.",
				 function );

				result = -1;
			}
		}
		memory_free(
		 table_entry_definition );
	}
	return( result );
}

/* Resizes the table entries
 * Returns 1 if successful or -1 on error
 */
int libpff_table_entries_resize(
     libpff_table_t *table,
     int number_of_sets,
     int number_of_entries,
     liberror_error_t **error )
{
	static char *function = "libpff_table_entries_resize";
	void *reallocation    = NULL;
	int set_iterator      = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( number_of_sets <= 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid number of sets value zero or less.",
		 function );

		return( -1 );
	}
	if( number_of_entries <= 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid number of entries value zero or less.",
		 function );

		return( -1 );
	}
	if( number_of_sets > table->number_of_sets )
	{
		reallocation = memory_reallocate(
		                table->entry,
		                sizeof( libpff_table_entry_t * ) * number_of_sets );

		if( reallocation == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to resize table entries.",
			 function );

			return( -1 );
		}
		table->entry = (libpff_table_entry_t **) reallocation;

		if( memory_set(
		     &( table->entry[ table->number_of_sets ] ),
		     0,
		     sizeof( libpff_table_entry_t * ) * ( number_of_sets - table->number_of_sets ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear table entries.",
			 function );

			return( -1 );
		}
		for( set_iterator = table->number_of_sets;
		     set_iterator < number_of_sets;
		     set_iterator++ )
		{
			table->entry[ set_iterator ] = (libpff_table_entry_t *) memory_allocate(
										 sizeof( libpff_table_entry_t ) * table->number_of_entries );

			if( table->entry[ set_iterator ] == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create table entries for set: %d.",
				 function,
				 set_iterator );

				return( -1 );
			}
			if( memory_set(
			     table->entry[ set_iterator ],
			     0,
			     sizeof( libpff_table_entry_t ) * table->number_of_entries ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear table entries for set: %d.",
				 function,
				 set_iterator );

				return( -1 );
			}
		}
		table->number_of_sets = number_of_sets;
	}
	if( ( table->number_of_sets > 0 )
	 && ( number_of_entries > table->number_of_entries ) )
	{
		for( set_iterator = 0;
		     set_iterator < table->number_of_sets;
		     set_iterator++ )
		{
			reallocation = memory_reallocate(
			                table->entry[ set_iterator ],
			                sizeof( libpff_table_entry_t ) * number_of_entries );

			if( reallocation == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to resize table entries for set: %d.",
				 function,
				 set_iterator );

				return( -1 );
			}
			table->entry[ set_iterator ] = (libpff_table_entry_t *) reallocation;

			if( memory_set(
			     &( ( table->entry[ set_iterator ] )[ table->number_of_entries ] ),
			     0,
			     sizeof( libpff_table_entry_t ) * ( number_of_entries - table->number_of_entries ) ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear table entries for set: %d.",
				 function,
				 set_iterator );

				return( -1 );
			}
		}
		table->number_of_entries = number_of_entries;
	}
	return( 1 );
}

/* Copies the values in the table entries
 * This function can return a partially cloned table entries on error
 * Returns 1 if successful or -1 on error
 */
int libpff_table_entries_clone(
     libpff_table_t *destination,
     libpff_table_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_table_entries_copy";
	int entry_iterator    = 0;
	int set_iterator      = 0;

	if( destination == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination table.",
		 function );

		return( -1 );
	}
	if( destination->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination table - entries already set.",
		 function );

		return( -1 );
	}
	if( source == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid source table.",
		 function );

		return( -1 );
	}
	if( source->entry == NULL )
	{
		destination->entry             = NULL;
		destination->number_of_sets    = 0;
		destination->number_of_entries = 0;

		return( 1 );
	}
	destination->entry = (libpff_table_entry_t **) memory_allocate(
	                                                sizeof( libpff_table_entry_t * ) * source->number_of_sets );

	if( destination->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create table entries.",
		 function );

		return( -1 );
	}
	if( memory_set(
	     destination->entry,
	     0,
	     sizeof( libpff_table_entry_t * ) * source->number_of_sets ) == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_MEMORY,
		 LIBERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear table entries.",
		 function );

		return( -1 );
	}
	for( set_iterator = 0;
	     set_iterator < source->number_of_sets;
	     set_iterator++ )
	{
		if( source->entry[ set_iterator ] == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing source table set: %d entries.",
			 function,
			 set_iterator );

			return( -1 );
		}
		destination->entry[ set_iterator ] = (libpff_table_entry_t *) memory_allocate(
		                                                               sizeof( libpff_table_entry_t ) * source->number_of_entries );

		if( destination->entry[ set_iterator ] == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create destination table set: %d entries.",
			 function,
			 set_iterator );

			return( -1 );
		}
		if( memory_set(
		     destination->entry[ set_iterator ],
		     0,
		     sizeof( libpff_table_entry_t ) * source->number_of_entries ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear destination table set: %d entries.",
			 function,
			 set_iterator );

			return( -1 );
		}
		for( entry_iterator = 0;
		     entry_iterator < source->number_of_entries;
		     entry_iterator++ )
		{
			if( memory_copy(
			     &( ( destination->entry[ set_iterator ] )[ entry_iterator ].identifier ),
			     &( ( source->entry[ set_iterator ] )[ entry_iterator ].identifier ),
			     sizeof( libpff_table_entry_identifier_t ) ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to copy table entry identifier of table set: %d entry: %d.",
				 function,
				 set_iterator,
				 entry_iterator );

				return( -1 );
			}
			if( libpff_data_reference_clone(
			     &( ( destination->entry[ set_iterator ] )[ entry_iterator ].value_data_reference ),
			     ( source->entry[ set_iterator ] )[ entry_iterator ].value_data_reference,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create destination value data reference of table set: %d entry: %d.",
				 function,
				 set_iterator,
				 entry_iterator );

				return( -1 );
			}
		}
	}
	destination->number_of_sets    = source->number_of_sets;
	destination->number_of_entries = source->number_of_entries;

	return( 1 );
}

/* Creates a table
 * Returns 1 if successful or -1 on error
 */
int libpff_table_initialize(
     libpff_table_t **table,
     liberror_error_t **error )
{
	static char *function = "libpff_table_initialize";

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( *table == NULL )
	{
		*table = (libpff_table_t *) memory_allocate(
		                             sizeof( libpff_table_t ) );

		if( *table == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create table.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *table,
		     0,
		     sizeof( libpff_table_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear table.",
			 function );

			memory_free(
			 *table );

			*table = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Frees a table
 * Returns 1 if successful or -1 on error
 */
int libpff_table_free(
     libpff_table_t *table,
     liberror_error_t **error )
{
	static char *function = "libpff_table_free";
	int entry_iterator    = 0;
	int set_iterator      = 0;
	int result            = 1;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		for( set_iterator = 0;
		     set_iterator < table->number_of_sets;
		     set_iterator++ )
		{
			for( entry_iterator = 0;
			     entry_iterator < table->number_of_entries;
			     entry_iterator++ )
			{
				if( ( table->entry[ set_iterator ] )[ entry_iterator ].value_data_reference != NULL )
				{
					if( libpff_data_reference_free(
					     &( ( table->entry[ set_iterator ] )[ entry_iterator ].value_data_reference ),
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
						 "%s: unable to free table set: %d entry: %d value data reference.",
						 function,
						 set_iterator,
						 entry_iterator );

						result = -1;
					}
				}
			}
			memory_free(
			 table->entry[ set_iterator ] );
		}
		memory_free(
		 table->entry );

		table->entry = NULL;
	}
	memory_free(
	 table );

	return( result );
}

/* Clones the existing table
 * Returns 1 if successful or -1 on error
 */
int libpff_table_clone(
     libpff_table_t **destination,
     libpff_table_t *source,
     liberror_error_t **error )
{
	static char *function = "libpff_table_clone";

	if( destination == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid destination table.",
		 function );

		return( -1 );
	}
	if( *destination != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid destination table already set.",
		 function );

		return( -1 );
	}
	if( source == NULL )
	{
		*destination = NULL;
	}
	else
	{
		*destination = (libpff_table_t *) memory_allocate(
		                                   sizeof( libpff_table_t ) );

		if( *destination == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create table.",
			 function );

			return( -1 );
		}
		if( memory_set(
		     *destination,
		     0,
		     sizeof( libpff_table_t ) ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear table.",
			 function );

			memory_free(
			 *destination );

			*destination = NULL;

			return( -1 );
		}
		if( libpff_table_entries_clone(
		     *destination,
		     source,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to clone table entries.",
			 function );

			libpff_table_free(
			 *destination,
			 NULL );

			*destination = NULL;

			return( -1 );
		}
	}
	return( 1 );
}

/* Retrieves the entry and value type of a the entry matching the index from a table
 * Returns 1 if successful or -1 on error
 */
int libpff_table_get_entry_type_by_index(
     libpff_table_t *table,
     int set_index,
     int entry_index,
     uint32_t *entry_type,
     uint32_t *value_type,
     libpff_name_to_id_map_entry_t **name_to_id_map_entry,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry = NULL;
	static char *function             = "libpff_table_get_entry_type_by_index";

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->number_of_sets == 0 )
	{
		return( 0 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( entry_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid entry type.",
		 function );

		return( -1 );
	}
	if( value_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value type.",
		 function );

		return( -1 );
	}
	if( name_to_id_map_entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid name to id map entry.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: retrieving table set: %d entry index: %d\n",
		 function,
		 set_index,
		 entry_index );
	}
#endif

	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

	if( ( entry_index < 0 )
	 || ( entry_index >= table->number_of_entries ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid entry index value out of range.",
		 function );

		return( -1 );
	}
	if( ( table_entry[ entry_index ] ).identifier.format != LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_MAPI_PROPERTY )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier format: %" PRIu8 ".",
		 function,
		 ( table_entry[ entry_index ] ).identifier.format );

		return( -1 );
	}
	*entry_type           = ( table_entry[ entry_index ] ).identifier.entry_type;
	*value_type           = ( table_entry[ entry_index ] ).identifier.value_type;
	*name_to_id_map_entry = (libpff_name_to_id_map_entry_t *) ( table_entry[ entry_index ] ).name_to_id_map_entry;

	return( 1 );
}

/* Retrieves the value of a the entry matching the index from a table
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_table_get_entry_value_by_index(
     libpff_table_t *table,
     int set_index,
     int entry_index,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry = NULL;
	static char *function             = "libpff_table_get_entry_value_by_index";

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->number_of_sets == 0 )
	{
		return( 0 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
	if( value_data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value data reference.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: retrieving entry index: %d from set: %d\n",
		 function,
		 entry_index,
		 set_index );
	}
#endif

	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

	if( ( entry_index < 0 )
	 || ( entry_index >= table->number_of_entries ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid entry index value out of range.",
		 function );

		return( -1 );
	}
	*value_data_reference = ( table_entry[ entry_index ] ).value_data_reference;

	return( 1 );
}

/* Retrieves the value of a the entry matching the entry type from a table
 *
 * When the LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE is set
 * the value type is ignored and set. The default behavior is a strict
 * matching of the value type. In this case the value type must be filled
 * with the corresponding value type
 *
 * When the LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP is set
 * the name to identifier mapping is ignored. The default behavior is
 * the use the mapped entry value. In this case named properties are not
 * retrieved.
 *
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_table_get_entry_value_by_entry_type(
     libpff_table_t *table,
     int set_index,
     uint32_t entry_type,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     uint8_t flags,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry = NULL;
	static char *function             = "libpff_table_get_entry_value_by_entry_type";
	int entry_type_match              = 0;
	int table_entry_iterator          = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->number_of_sets == 0 )
	{
		return( 0 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
	if( value_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value type.",
		 function );

		return( -1 );
	}
	if( value_data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value data reference.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: retrieving table set: %d entry type: 0x%04" PRIx32 "\n",
		 function,
		 set_index,
		 entry_type );
	}
#endif

	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

	for( table_entry_iterator = 0;
	     table_entry_iterator < table->number_of_entries;
	     table_entry_iterator++ )
	{
		/* Skip tables entries that do not contain a MAPI identifier
		 */
		if( ( table_entry[ table_entry_iterator ] ).identifier.format != LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_MAPI_PROPERTY )
		{
			continue;
		}
		/* Ignore the name to identifier map
		 */
		if( ( flags & LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP ) == LIBPFF_ENTRY_VALUE_FLAG_IGNORE_NAME_TO_ID_MAP )
		{
			if( ( table_entry[ table_entry_iterator ] ).identifier.entry_type == entry_type )
			{
				entry_type_match = 1;
			}
		}
		else
		{
			/* Mapped properties must be accessed through their mapped entry type value
			 */
			if( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry != NULL )
			{
				if( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_NUMERIC )
				{
					if( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->numeric_value == entry_type )
					{
						entry_type_match = 1;
					}
				}
			}
			/* Other properties are accessed through their entry type value
			 */
			else if( ( table_entry[ table_entry_iterator ] ).identifier.entry_type == entry_type )
			{
				entry_type_match = 1;
			}
		}
		if( entry_type_match != 0 )
		{
			if( ( ( flags & LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE ) == LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE )
			 || ( ( table_entry[ table_entry_iterator ] ).identifier.value_type == *value_type ) )
			{
				if( ( flags & LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE ) == LIBPFF_ENTRY_VALUE_FLAG_MATCH_ANY_VALUE_TYPE )
				{
					*value_type = ( table_entry[ table_entry_iterator ] ).identifier.value_type;
				}
				*value_data_reference = ( table_entry[ table_entry_iterator ] ).value_data_reference;

				return( 1 );
			}
		}
	}
	return( 0 );
}

/* Retrieves the value of an UTF-8 formatted entry name matching the entry name from a table
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_table_get_entry_value_by_utf8_entry_name(
     libpff_table_t *table,
     int set_index,
     uint8_t *utf8_entry_name,
     size_t utf8_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry = NULL;
	static char *function             = "libpff_table_get_entry_value_by_utf8_entry_name";
	int result                        = 0;
	int table_entry_iterator          = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->number_of_sets == 0 )
	{
		return( 0 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
	if( utf8_entry_name == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-8 entry name.",
		 function );

		return( -1 );
	}
	if( utf8_entry_name_length == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid UTF-8 entry name length value zero or less.",
		 function );

		return( -1 );
	}
	if( utf8_entry_name_length > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid UTF-8 entry name length value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( value_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value type.",
		 function );

		return( -1 );
	}
	if( value_data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value data reference.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: retrieving table set: %d entry name: %s\n",
		 function,
		 set_index,
		 utf8_entry_name );
	}
#endif

	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

	for( table_entry_iterator = 0;
	     table_entry_iterator < table->number_of_entries;
	     table_entry_iterator++ )
	{
		/* There is a match
		 * if there is a name-to-id map entry
		 *   and the name-to-id map entry is a string
		 *   and the entry name string matches the one in the name-to-id map
		 */
		if( ( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry != NULL )
		 && ( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING ) )
		{
			if( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->is_ascii_string == 0 )
			{
				result = libuna_utf8_string_compare_with_utf16_stream(
				          utf8_entry_name,
				          utf8_entry_name_length + 1,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->string_value,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->value_size,
				          LIBPFF_ENDIAN_LITTLE,
				          error );
			}
			else
			{
				result = libuna_utf8_string_compare_with_byte_stream(
				          utf8_entry_name,
				          utf8_entry_name_length + 1,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->string_value,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->value_size,
				          LIBUNA_CODEPAGE_ASCII,
				          error );
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GENERIC,
				 "%s: unable to compare entry name.",
				 function );

				return( -1 );
			}
			else if( result != 0 )
			{
				*value_type           = ( table_entry[ table_entry_iterator ] ).identifier.value_type;
				*value_data_reference = ( table_entry[ table_entry_iterator ] ).value_data_reference;

				return( 1 );
			}
		}
	}
	return( 0 );
}

/* Retrieves the value of an UTF-16 formatted entry name matching the entry name from a table
 * Returns 1 if successful, 0 if the item does not contain such value or -1 on error
 */
int libpff_table_get_entry_value_by_utf16_entry_name(
     libpff_table_t *table,
     int set_index,
     uint16_t *utf16_entry_name,
     size_t utf16_entry_name_length,
     uint32_t *value_type,
     libpff_data_reference_t **value_data_reference,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry = NULL;
	static char *function             = "libpff_table_get_entry_value_by_utf16_entry_name";
	int result                        = 0;
	int table_entry_iterator          = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->number_of_sets == 0 )
	{
		return( 0 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
	if( utf16_entry_name == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid UTF-16 entry name.",
		 function );

		return( -1 );
	}
	if( utf16_entry_name_length == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid UTF-16 entry name length value zero or less.",
		 function );

		return( -1 );
	}
	if( utf16_entry_name_length > (size_t) SSIZE_MAX )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid UTF-16 entry name length value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( value_type == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value type.",
		 function );

		return( -1 );
	}
	if( value_data_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value data reference.",
		 function );

		return( -1 );
	}
	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

	for( table_entry_iterator = 0;
	     table_entry_iterator < table->number_of_entries;
	     table_entry_iterator++ )
	{
		/* There is a match
		 * if there is a name-to-id map entry
		 *   and the name-to-id map entry is a string
		 *   and the entry name string matches the one in the name-to-id map
		 */
		if( ( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry != NULL )
		 && ( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING ) )
		{
			if( ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->is_ascii_string == 0 )
			{
				result = libuna_utf16_string_compare_with_utf16_stream(
				          utf16_entry_name,
				          utf16_entry_name_length + 1,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->string_value,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->value_size,
				          LIBPFF_ENDIAN_LITTLE,
				          error );
			}
			else
			{
				result = libuna_utf16_string_compare_with_byte_stream(
				          utf16_entry_name,
				          utf16_entry_name_length + 1,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->string_value,
				          ( table_entry[ table_entry_iterator ] ).name_to_id_map_entry->value_size,
				          LIBUNA_CODEPAGE_ASCII,
				          error );
			}
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GENERIC,
				 "%s: unable to compare entry name.",
				 function );

				return( -1 );
			}
			else if( result != 0 )
			{
				*value_type           = ( table_entry[ table_entry_iterator ] ).identifier.value_type;
				*value_data_reference = ( table_entry[ table_entry_iterator ] ).value_data_reference;

				return( 1 );
			}
		}
	}
	return( 0 );
}

/* Reads a table and its values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read(
     libpff_table_t *table,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_array_t *table_index_array               = NULL;
	uint8_t *table_data                             = NULL;
	static char *function                           = "libpff_table_read";
	size_t table_data_size                          = 0;
	uint32_t table_value_reference                  = 0;
	uint8_t table_signature                         = 0;
	uint8_t table_type                              = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	libpff_table_index_values_t *table_index_values = NULL;
	uint16_t table_index_values_iterator            = 0;
	int number_of_table_index_array_entries         = 0;
	int table_index_array_iterator                  = 0;
#endif

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	/* Retrieve the first data reference entry data
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     0,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve data reference entry: 0.",
		 function );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid data reference entry: 0.",
		 function );

		return( -1 );
	}
	if( table_data_size < sizeof( pff_table_t ) )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: descriptor data:\n",
			 function );
			libnotify_print_data(
			 table_data,
			 table_data_size );
		}
#endif

		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: data reference entry: 0 too small to be a table.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table start:\n",
		 function );
		libnotify_print_data(
		 table_data,
		 sizeof( pff_table_t ) );
	}
#endif

	table_signature = ( (pff_table_t *) table_data )->signature;
	table_type      = ( (pff_table_t *) table_data )->type;

	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_t *) table_data )->value_reference,
	 table_value_reference );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table signature\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 table_signature );
		libnotify_printf(
		 "%s: table type\t\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 table_type );
		libnotify_printf(
		 "%s: table value reference\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 table_value_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( table_value_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
	}
#endif

	if( table_signature != 0xec )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table signature: 0x%02" PRIx8 ".",
		 function,
		 table_signature );

		return( -1 );
	}
	if( ( table_type != 0x6c )
	 && ( table_type != 0x7c )
	 && ( table_type != 0x8c )
	 && ( table_type != 0x9c )
	 && ( table_type != 0xa5 )
	 && ( table_type != 0xac )
	 && ( table_type != 0xbc ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table type: 0x%02" PRIx8 ".",
		 function,
		 table_type );

		return( -1 );
	}
	/* Create the table index array
	 */
	if( libpff_array_initialize(
	     &table_index_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table index array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_index(
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table index.",
		 function );

		libpff_array_free(
		 &table_index_array,
		 &libpff_table_index_values_free,
		 NULL );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		if( libpff_array_get_number_of_entries(
		     table_index_array,
		     &number_of_table_index_array_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve the number of table index array entries.",
			 function );

			libpff_array_free(
			 &table_index_array,
			 &libpff_table_index_values_free,
			 NULL );

			return( -1 );
		}
		for( table_index_array_iterator = 0;
		     table_index_array_iterator < number_of_table_index_array_entries;
		     table_index_array_iterator++ )
		{
			libnotify_printf(
			 "%s: table index array entry: %d\n",
			 function,
			 table_index_array_iterator );

			if( libpff_array_get_entry(
			     table_index_array,
			     table_index_array_iterator,
			     (intptr_t **) &table_index_values,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table index array entry: %d.",
				 function,
				 table_index_array_iterator );

				libpff_array_free(
				 &table_index_array,
				 &libpff_table_index_values_free,
				 NULL );

				return( -1 );
			}
			if( table_index_values == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
				 "%s: invalid table index values.",
				 function );

				libpff_array_free(
				 &table_index_array,
				 &libpff_table_index_values_free,
				 NULL );

				return( -1 );
			}
			if( table_index_values->value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid table index values - missing values.",
				 function );

				libpff_array_free(
				 &table_index_array,
				 &libpff_table_index_values_free,
				 NULL );

				return( -1 );
			}
			for( table_index_values_iterator = 0;
			     table_index_values_iterator < table_index_values->number_of_values;
			     table_index_values_iterator++ )
			{
				if( libpff_data_reference_get_entry_data(
				     table_data_reference,
				     file_io_handle,
				     ( table_index_values->value[ table_index_values_iterator ] ).array_entry,
				     &table_data,
				     &table_data_size,
				     0,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve table data array entry: %" PRIu16 ".",
					 function,
					 ( table_index_values->value[ table_index_values_iterator ] ).array_entry );

					libpff_array_free(
					 &table_index_array,
					 &libpff_table_index_values_free,
					 NULL );

					return( -1 );
				}
				if( table_data == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
					 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
					 "%s: invalid table data array entry: %" PRIu16 ".",
					 function,
					 ( table_index_values->value[ table_index_values_iterator ] ).array_entry );

					libpff_array_free(
					 &table_index_array,
					 &libpff_table_index_values_free,
					 NULL );

					return( -1 );
				}
				libnotify_printf(
				 "%s: table value: %d at offset: %" PRIu32 " of size: %" PRIu16 "\n",
				 function,
				 table_index_values_iterator,
				 ( table_index_values->value[ table_index_values_iterator ] ).offset,
				 ( table_index_values->value[ table_index_values_iterator ] ).size );
				libnotify_print_data(
				 &( table_data[ ( table_index_values->value[ table_index_values_iterator ] ).offset ] ),
				 ( table_index_values->value[ table_index_values_iterator ] ).size );
			}
		}
		libnotify_printf(
		 "\n" );
	}
#endif

	/* Read the table values
	 */
	if( libpff_table_read_values(
	     table,
	     table_type,
	     table_value_reference,
	     table_index_array,
	     recovered,
	     io_handle,
	     local_descriptors,
	     name_to_id_map_list,
	     table_data_reference,
	     file_io_handle,
	     debug_item_type,
	     debug_ascii_codepage,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table values.",
		 function );

		libpff_array_free(
		 &table_index_array,
		 &libpff_table_index_values_free,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_index_array,
	     &libpff_table_index_values_free,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table index array.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif

	return( 1 );
}

/* Reads the table index
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_index(
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_table_index_values_t *table_index_values = NULL;
	uint8_t *table_data                             = NULL;
	uint8_t *table_index_data                       = NULL;
	static char *function                           = "libpff_table_read_index";
	size_t table_index_values_size                  = 0;
	size_t table_data_size                          = 0;
	ssize_t table_data_offset                       = 0;
	uint32_t number_of_table_data_array_entries     = 0;
	uint32_t table_data_array_entry_iterator        = 0;
	uint16_t table_index_offset                     = 0;
	uint16_t table_value_start_offset               = 0;
	uint16_t table_value_end_offset                 = 0;
	uint16_t table_number_of_index_offsets          = 0;
	uint16_t table_number_of_unused_index_offsets   = 0;
	uint16_t table_index_iterator                   = 0;

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( libpff_data_reference_get_number_of_entries(
	     table_data_reference,
	     &number_of_table_data_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table data array entries.",
		 function );

		return( -1 );
	}
	/* Resize the table index array
	 */
	if( libpff_array_resize(
	     table_index_array,
	     number_of_table_data_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
		 "%s: unable to resize table index array.",
		 function );

		return( -1 );
	}
	/* Iterate the table data array to fill the table index array
	 */
	for( table_data_array_entry_iterator = 0;
	     table_data_array_entry_iterator < number_of_table_data_array_entries;
	     table_data_array_entry_iterator++ )
	{
		if( libpff_data_reference_get_entry_data(
		     table_data_reference,
		     file_io_handle,
		     table_data_array_entry_iterator,
		     &table_data,
		     &table_data_size,
		     0,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table data array entry: %" PRIu16 ".",
			 function,
			 table_data_array_entry_iterator );

			return( -1 );
		}
		if( table_data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid table data array entry: %" PRIu16 ".",
			 function,
			 table_data_array_entry_iterator );

			return( -1 );
		}
		byte_stream_copy_to_uint16_little_endian(
		 table_data,
		 table_index_offset );

#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table data offset\t\t\t: %" PRIzd "\n",
			 function,
			 table_data_offset );
			libnotify_printf(
			 "%s: table size\t\t\t\t: %" PRIzd "\n",
			 function,
			 table_data_size );
			libnotify_printf(
			 "%s: table index offset\t\t\t: %" PRIu16 "\n",
			 function,
			 table_index_offset );
		}
#endif

		if( table_index_offset == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: invalid table index offset.",
			 function );

			return( -1 );
		}
		/* The table index offset should point to an offset within the table
		 */
		if( (size_t) table_index_offset >= table_data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: table index offset exceeds table data size.",
			 function );

			return( -1 );
		}
		/* Determine which values are in the table using the index
		 */
		table_index_data = &( table_data[ table_index_offset ] );

#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table index:\n",
			 function );
			libnotify_print_data(
			 table_index_data,
			 4 );
		}
#endif
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_index_t *) table_index_data )->number_of_offsets,
		 table_number_of_index_offsets );
		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_index_t *) table_index_data )->number_of_unused_offsets,
		 table_number_of_unused_index_offsets );

		table_index_data += 4;

#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table number of index offsets\t\t: %" PRIu16 "\n",
			 function,
			 table_number_of_index_offsets );

			libnotify_printf(
			 "%s: table number of unused index offsets\t: %" PRIu16 "\n",
			 function,
			 table_number_of_unused_index_offsets );
		}
#endif
		if( (size_t) ( table_index_offset + 4 + ( ( table_number_of_index_offsets + 1 ) * 2 ) ) > table_data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: number of index offsets do not fit in table.",
			 function );

			return( -1 );
		}
		/* Create the table index values structure
		 */
		table_index_values = (libpff_table_index_values_t *) memory_allocate(
		                                                      sizeof( libpff_table_index_values_t ) );

		if( table_index_values == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create table index values.",
			 function );

			return( -1 );
		}
		/* The table number of index items should be considered more of a last item number
		 * The table actually contains 1 additional table index value
		 */
		table_index_values->number_of_values = table_number_of_index_offsets;

		table_index_values_size = sizeof( libpff_table_index_value_t ) * table_index_values->number_of_values;

		table_index_values->value = (libpff_table_index_value_t *) memory_allocate(
		                                                            table_index_values_size );

		if( table_index_values->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
			 "%s: unable to create table index values.",
			 function );

			memory_free(
			 table_index_values );

			return( -1 );
		}
		if( memory_set(
		     table_index_values->value,
		     0,
		     table_index_values_size ) == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_MEMORY,
			 LIBERROR_MEMORY_ERROR_SET_FAILED,
			 "%s: unable to clear table index values.",
			 function );

			memory_free(
			 table_index_values->value );
			memory_free(
			 table_index_values );

			return( -1 );
		}
		if( libpff_array_set_entry(
		     table_index_array,
		     table_data_array_entry_iterator,
		     (intptr_t *) table_index_values,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to set table index array entry: %" PRIu16 ".",
			 function,
			 table_data_array_entry_iterator );

			memory_free(
			 table_index_values->value );
			memory_free(
			 table_index_values );

			return( -1 );
		}
		if( table_number_of_index_offsets > 0 )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "\n" );

				libnotify_printf(
				 "%s: table index offsets:\n",
				 function );
				libnotify_print_data(
				 table_index_data,
				 ( table_number_of_index_offsets + 1 ) * 2 );
			}
#endif
			/* Fill the table index values
			 */
			byte_stream_copy_to_uint16_little_endian(
			 table_index_data,
			 table_value_start_offset );

			table_index_data += 2;

			for( table_index_iterator = 0;
			     table_index_iterator < table_number_of_index_offsets;
			     table_index_iterator++ )
			{
				byte_stream_copy_to_uint16_little_endian(
				 table_index_data,
				 table_value_end_offset );

				table_index_data += 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: table index value: %03" PRIu16 " offset\t\t: %" PRIu16 " - %" PRIu16 "\n",
					 function,
					 table_index_iterator,
					 table_value_start_offset,
					 table_value_end_offset );
				}
#endif

				if( table_value_start_offset > table_value_end_offset )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_SET_FAILED,
					 "%s: table index start offset: %" PRIu16 " exceeds end offset: %" PRIu16 ".",
					 function,
					 table_value_start_offset,
					 table_value_end_offset );

					return( -1 );
				}
				( table_index_values->value[ table_index_iterator ] ).array_entry = table_data_array_entry_iterator;
				( table_index_values->value[ table_index_iterator ] ).offset      = table_value_start_offset;
				( table_index_values->value[ table_index_iterator ] ).size        = table_value_end_offset - table_value_start_offset;

				table_value_start_offset = table_value_end_offset;
			}
			if( table_value_end_offset > table_index_offset )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_SET_FAILED,
				 "%s: last table index value end offset: %" PRIu16 " exceeds table index offset: %" PRIu16 ".",
				 function,
				 table_value_end_offset,
				 table_index_offset );

				return( -1 );
			}
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( ( libnotify_verbose != 0 )
		 && ( table_value_end_offset < table_index_offset ) )
		{
			libnotify_printf(
			 "%s: last table index values end offset: %" PRIu16 " does not match table index offset: %" PRIu16 "\n",
			 function,
			 table_value_start_offset,
			 table_index_offset );

			libnotify_print_data(
			 &( table_data[ table_value_end_offset ] ),
			 ( table_index_offset - table_value_end_offset ) );
		}
		if( ( libnotify_verbose != 0 )
		 && ( (size_t) ( table_index_data - table_data ) != table_data_size ) )
		{
			libnotify_printf(
			 "\n" );
			libnotify_printf(
			 "%s: data after table index of size: %" PRIzd "\n",
			 function,
			 table_data_size - (size_t) ( table_index_data - table_data ) );
			libnotify_print_data(
			 table_index_data,
			 table_data_size - (size_t) ( table_index_data - table_data ) );
		}
#endif
		table_data_offset += table_data_size;
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif

	return( 1 );
}

/* Reads the table entries
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_entries(
     libpff_array_t *table_entries_array,
     uint8_t table_entries_level,
     uint8_t table_entry_identifier_size,
     uint32_t table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
#if defined( HAVE_DEBUG_OUTPUT )
	uint8_t guid_string[ LIBFMAPI_GUID_STRING_SIZE ];
#endif

	libpff_data_reference_t *table_entries_data_reference = NULL;
	libpff_table_index_value_t *table_index_value         = NULL;
	uint8_t *table_data                                   = NULL;
	uint8_t *table_entry_data                             = NULL;
	static char *function                                 = "libpff_table_read_entries";
	size_t number_of_table_entries                        = 0;
	size_t table_data_size                                = 0;
	size_t table_entry_data_size                          = 0;
	size_t table_entry_size                               = 0;
	int table_entry_iterator                              = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint64_t table_entry_identifier                       = 0;
#endif

	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 2 )
	 && ( table_entry_identifier_size != 4 )
	 && ( table_entry_identifier_size != 8 )
	 && ( table_entry_identifier_size != 16 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: %" PRIu8 ".",
		 function,
		 table_entry_identifier_size );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: reading table entries at level: %" PRIu8 " with reference: 0x%08" PRIx32 " ",
		 function,
		 table_entries_level,
		 table_entries_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( table_entries_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
		libnotify_printf(
		 "\n" );
	}
#endif

	/* Check if the table contains any entries
	 */
	if( table_entries_reference == 0 )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table contains no entries.\n",
			 function );
		}
#endif

		return( 1 );
	}
	if( ( table_entries_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entries reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_entries_reference & 0x0000001fUL,
		 table_entries_reference );

		return( -1 );
	}
	/* Fetch the table entries reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_entries_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entries reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	/* Create a data reference of the table entries data
	 */
	if( libpff_data_reference_initialize_buffer(
	     &table_entries_data_reference,
	     &( table_data[ table_index_value->offset ] ),
	     table_index_value->size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries data reference.",
		 function );

		return( -1 );
	}
	if( table_entries_level == 0 )
	{
		if( libpff_array_append_entry(
		     table_entries_array,
		     &table_entry_iterator,
		     (intptr_t *) table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_SET_FAILED,
			 "%s: unable to append table entries data reference.",
			 function );

			libpff_data_reference_free(
			 &table_entries_data_reference,
			 NULL );

			return( -1 );
		}
	}
	else
	{
		table_entry_size = table_entry_identifier_size + 4;

		if( ( table_index_value->size % table_entry_size ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries size.",
			 function );

			libpff_data_reference_free(
			 &table_entries_data_reference,
			 NULL );

			return( -1 );
		}
		number_of_table_entries = table_index_value->size / table_entry_size;

		for( table_entry_iterator = 0;
		     (size_t) table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     table_entry_iterator * table_entry_size,
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 table_entry_iterator * table_entry_size );

				libpff_data_reference_free(
				 &table_entries_data_reference,
				 NULL );

				return( -1 );
			}
			if( table_entry_data_size < table_entry_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				libpff_data_reference_free(
				 &table_entries_data_reference,
				 NULL );

				return( -1 );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( ( table_entry_identifier_size == 2 )
				 || ( table_entry_identifier_size == 4 )
				 || ( table_entry_identifier_size == 8 ) )
				{
					if( table_entry_identifier_size == 2 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 table_entry_data,
						 table_entry_identifier );
					}
					else if( table_entry_identifier_size == 4 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 table_entry_data,
						 table_entry_identifier );
					}
					else if( table_entry_identifier_size == 8 )
					{
						byte_stream_copy_to_uint64_little_endian(
						 table_entry_data,
						 table_entry_identifier );
					}
					libnotify_printf(
					 "%s: table entry: %03d at level: %" PRIu8 " identifier\t\t\t: 0x%08" PRIx64 "\n",
					 function,
					 table_entry_iterator,
					 table_entries_level,
					 table_entry_identifier );
				}
				else if( table_entry_identifier_size == 16 )
				{
					if( libfmapi_guid_to_string(
					     table_entry_data,
					     LIBFMAPI_ENDIAN_LITTLE,
					     guid_string,
					     LIBFMAPI_GUID_STRING_SIZE,
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_CONVERSION,
						 LIBERROR_CONVERSION_ERROR_GENERIC,
						 "%s: unable to create guid string.",
						 function );

						libpff_data_reference_free(
						 &table_entries_data_reference,
						 NULL );

						return( -1 );
					}
					libnotify_printf(
					 "%s: table entry: %03d at level: %" PRIu8 " identifier\t\t\t: %s\n",
					 function,
					 table_entry_iterator,
					 table_entries_level,
					 guid_string );
				}
			}
#endif

			/* TODO use the table entry identifier to validate sub level table entries
			 */
			table_entry_data += table_entry_identifier_size;

			byte_stream_copy_to_uint32_little_endian(
			 table_entry_data,
			 table_entries_reference );

			table_entry_data += 4;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table entry: %03d at level: %" PRIu8 " reference\t\t\t: 0x%08" PRIx32 " ",
				 function,
				 table_entry_iterator,
				 table_entries_level,
				 table_entries_reference );
				libpff_debug_print_node_identifier_type(
				 (uint8_t) ( table_entries_reference & 0x0000001fUL ) );
				libnotify_printf(
				 "\n" );
				libnotify_printf(
				 "\n" );
			}
#endif

			if( libpff_table_read_entries(
			     table_entries_array,
			     table_entries_level - 1,
			     table_entry_identifier_size,
			     table_entries_reference,
			     table_index_array,
			     table_data_reference,
			     file_io_handle,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read table entries.",
				 function );

				libpff_data_reference_free(
				 &table_entries_data_reference,
				 NULL );

				return( -1 );
			}
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "\n" );
		}
#endif

		if( libpff_data_reference_free(
		     &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free table entries data reference.",
			 function );

			return( -1 );
		}
	}
	return( 1 );
}

/* Reads the table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_values(
     libpff_table_t *table,
     uint8_t table_type,
     uint32_t table_value_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	static char *function = "libpff_table_read_values";
	int result            = 0;

	switch( table_type )
	{
		case 0x6c:
			result = libpff_table_read_6c_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  table_data_reference,
				  file_io_handle,
				  error );
			break;

		case 0x7c:
			result = libpff_table_read_7c_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  recovered,
				  io_handle,
				  local_descriptors,
				  name_to_id_map_list,
				  table_data_reference,
				  file_io_handle,
				  debug_ascii_codepage,
				  error );
			break;

		case 0x8c:
			result = libpff_table_read_8c_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  table_data_reference,
				  file_io_handle,
				  error );
			break;

		case 0x9c:
			result = libpff_table_read_9c_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  table_data_reference,
				  file_io_handle,
				  error );
			break;

		case 0xa5:
			result = libpff_table_read_a5_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  table_data_reference,
				  file_io_handle,
				  error );
			break;

		case 0xac:
			result = libpff_table_read_ac_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  recovered,
				  io_handle,
				  local_descriptors,
				  name_to_id_map_list,
				  table_data_reference,
				  file_io_handle,
				  debug_ascii_codepage,
				  error );
			break;

		case 0xbc:
			result = libpff_table_read_bc_values(
				  table,
				  table_value_reference,
				  table_index_array,
				  recovered,
				  io_handle,
				  local_descriptors,
				  name_to_id_map_list,
				  table_data_reference,
				  file_io_handle,
				  debug_item_type,
				  debug_ascii_codepage,
				  error );
			break;
	}
	if( result != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table values.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the 6c table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_6c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array           = NULL;
	libpff_table_index_value_t *table_index_value = NULL;
	uint8_t *table_data                           = NULL;
	uint8_t *table_value_data                     = NULL;
	static char *function                         = "libpff_table_read_6c_values";
	size_t table_data_size                        = 0;
	uint32_t b5_table_header_reference            = 0;
	uint32_t table_entries_reference              = 0;
	uint32_t values_array_reference               = 0;
	uint8_t table_entry_identifier_size           = 0;
	uint8_t table_entry_value_size                = 0;
	uint8_t table_entries_level                   = 0;

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( ( table_header_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_header_reference & 0x0000001fUL,
		 table_header_reference );

		return( -1 );
	}
	/* Fetch the 6c table header reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_header_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve 6c table header reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_value_data = &( table_data[ table_index_value->offset ] );

	/* Read the 6c table header
	 * This table header contains no type indicator
	 * to make sure the it is supported the size is checked
	 */
	if( table_index_value->size != 8 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: unsupported 6c table header of size: %" PRIu16 ".",
		 function,
		 table_index_value->size );

		return( -1 );
	}
	byte_stream_copy_to_uint32_little_endian(
	 table_value_data,
	 b5_table_header_reference );

	table_value_data += 4;

	byte_stream_copy_to_uint32_little_endian(
	 table_value_data,
	 values_array_reference );

	table_value_data += 4;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: b5 table header reference\t\t: 0x%08" PRIx32 " ",
		 function,
		 b5_table_header_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( b5_table_header_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
		libnotify_printf(
		 "%s: values array reference\t\t: 0x%08" PRIx32 " ",
		 function,
		 values_array_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( values_array_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
	}
#endif

	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 16 )
	 || ( table_entry_value_size != 2 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: %" PRIu8 " and table entry value size: %" PRIu8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	/* Check if the table contains any entries
	 */
	if( ( table_entries_reference == 0 )
	 && ( values_array_reference == 0 ) )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table contains no entries.\n",
			 function );
		}
#endif

		return( 1 );
	}
	if( table_entries_reference == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table contains value array but no table entries.",
		 function );

		return( -1 );
	}
	if( values_array_reference == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table contains table entries but no value array.",
		 function );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_6c_item_entries(
	     table,
	     values_array_reference,
	     table_index_array,
	     table_entries_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read 6c table item entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the 7c table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_7c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array                            = NULL;
	libpff_array_t *table_entry_definitions_array                  = NULL;
	libpff_data_reference_t *table_entry_definition_data_reference = NULL;
	libpff_table_index_value_t *table_index_value                  = NULL;
	uint8_t *table_data                                            = NULL;
	uint8_t *table_header_data                                     = NULL;
	static char *function                                          = "libpff_table_read_7c_values";
	size_t table_data_size                                         = 0;
	size_t table_header_data_size                                  = 0;
	uint32_t b5_table_header_reference                             = 0;
	uint32_t values_array_reference                                = 0;
	uint32_t table_entries_reference                               = 0;
	uint16_t values_array_entry_size                               = 0;
	uint8_t table_entry_identifier_size                            = 0;
	uint8_t table_entry_value_size                                 = 0;
	uint8_t table_entries_level                                    = 0;
	int number_of_table_entry_definitions                          = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint16_t value_16bit                                           = 0;
#endif

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( ( table_header_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_header_reference & 0x0000001fUL,
		 table_header_reference );

		return( -1 );
	}
	/* Fetch the 7c table header reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_header_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve 7c table header reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_header_data      = &( table_data[ table_index_value->offset ] );
	table_header_data_size = (size_t) table_index_value->size;

	/* Read the 7c table header
	 */
	if( ( (pff_table_header_7c_t *) table_header_data )->type != 0x7c )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header type: 0x%02x.",
		 function,
		 ( (pff_table_header_7c_t *) table_header_data )->type );

		return( -1 );
	}
	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_table_header_7c_t *) table_header_data )->values_array_entry_end_offset_cell_existence_block,
	 values_array_entry_size );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_7c_t *) table_header_data )->b5_table_header_reference,
	 b5_table_header_reference );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_7c_t *) table_header_data )->values_array_reference,
	 values_array_reference );

	number_of_table_entry_definitions = (int) ( (pff_table_header_7c_t *) table_header_data )->number_of_entry_definitions;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table header type\t\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_table_header_7c_t *) table_header_data )->type );

		libnotify_printf(
		 "%s: number of entry definitions\t\t\t: %d\n",
		 function,
		 number_of_table_entry_definitions );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_7c_t *) table_header_data )->values_array_entry_end_offset_32bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 32-bit values\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_7c_t *) table_header_data )->values_array_entry_end_offset_16bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 16-bit values\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_7c_t *) table_header_data )->values_array_entry_end_offset_8bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 8-bit values\t\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_7c_t *) table_header_data )->values_array_entry_end_offset_cell_existence_block,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset cell existence block\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		libnotify_printf(
		 "%s: b5 table header reference\t\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 b5_table_header_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( b5_table_header_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );

		libnotify_printf(
		 "%s: values array reference\t\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 values_array_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( values_array_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );

		libnotify_printf(
		 "%s: unknown1:\n",
		 function );
		libnotify_print_data(
		 ( (pff_table_header_7c_t *) table_header_data )->unknown1,
		 4 );
	}
#endif

	table_header_data      += sizeof( pff_table_header_7c_t );
	table_header_data_size -= sizeof( pff_table_header_7c_t );

	/* Read the entry definitions in the 7c table header
	 */
	if( ( (size_t) number_of_table_entry_definitions * sizeof( pff_table_entry_definition_7c_t ) ) != table_header_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: mismatch in number of the table entry definitions and data size.",
		 function );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: 7c table entry definitions data:\n",
		 function );
		libnotify_print_data(
		 table_header_data,
		 table_header_data_size );
	}
#endif

	/* Create a data reference of the table entry definition data
	 */
	if( libpff_data_reference_initialize_buffer(
	     &table_entry_definition_data_reference,
	     table_header_data,
	     table_header_data_size,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entry definitions data reference.",
		 function );

		return( -1 );
	}
	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 4 )
	 || ( ( table_entry_value_size != 2 )
	  && ( table_entry_value_size != 4 ) ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	/* Create the table entry definitions array
	 */
	if( libpff_array_initialize(
	     &table_entry_definitions_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entry definitions array.",
		 function );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_7c_item_entry_definitions(
	     table_entry_definitions_array,
	     table_entry_definition_data_reference,
	     (uint8_t) number_of_table_entry_definitions,
	     table_index_array,
	     file_io_handle,
	     name_to_id_map_list,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read 7c table item entry definitions.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );
		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );
		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );
		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );
		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( libpff_array_get_number_of_entries(
	     table_index_array,
	     &number_of_table_entry_definitions,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table entry definitions.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );
		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );
		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( number_of_table_entry_definitions > 0 )
	{
		if( libpff_table_read_values_array(
		     table,
		     table_entries_reference,
		     values_array_reference,
		     table_entry_identifier_size,
		     table_entry_value_size,
		     values_array_entry_size,
		     table_index_array,
		     table_entry_definitions_array,
		     table_entries_array,
		     recovered,
		     io_handle,
		     local_descriptors,
		     table_data_reference,
		     file_io_handle,
		     debug_ascii_codepage,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read values array.",
			 function );

			libpff_array_free(
			 &table_entries_array,
			 &libpff_data_reference_free_as_value,
			 NULL );
			libpff_array_free(
			 &table_entry_definitions_array,
			 &libpff_table_entry_definition_free,
			 NULL );
			libpff_data_reference_free(
			 &table_entry_definition_data_reference,
			 NULL );

			return( -1 );
		}
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );
		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entry_definitions_array,
	     &libpff_table_entry_definition_free,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entry definitions array.",
		 function );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( libpff_data_reference_free(
	     &table_entry_definition_data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free table entry definition data reference.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the 8c table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_8c_values(
     libpff_table_t *table,
     uint32_t b5_table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array = NULL;
	static char *function               = "libpff_table_read_8c_values";
	uint32_t table_entries_reference    = 0;
	uint8_t table_entry_identifier_size = 0;
	uint8_t table_entry_value_size      = 0;
	uint8_t table_entries_level         = 0;

	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 8 )
	 || ( table_entry_value_size != 4 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_8c_item_entries(
	     table,
	     table_entries_array,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table item entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the 9c table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_9c_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array           = NULL;
	libpff_table_index_value_t *table_index_value = NULL;
	uint8_t *table_data                           = NULL;
	uint8_t *table_value_data                     = NULL;
	static char *function                         = "libpff_table_read_9c_values";
	size_t table_data_size                        = 0;
	uint32_t b5_table_header_reference            = 0;
	uint32_t table_entries_reference              = 0;
	uint8_t table_entry_identifier_size           = 0;
	uint8_t table_entry_value_size                = 0;
	uint8_t table_entries_level                   = 0;

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( ( table_header_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_header_reference & 0x0000001fUL,
		 table_header_reference );

		return( -1 );
	}
	/* Fetch the 9c table header reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_header_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve 9c table header reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_value_data = &( table_data[ table_index_value->offset ] );

	/* Read the 9c table header
	 * This table header contains no type indicator
	 * to make sure the it is supported the size is checked
	 */
	if( table_index_value->size != 4 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: unsupported 9c table header of size: %" PRIu16 ".",
		 function,
		 table_index_value->size );

		return( -1 );
	}
	byte_stream_copy_to_uint32_little_endian(
	 table_value_data,
	 b5_table_header_reference );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: b5 table header reference\t\t: 0x%08" PRIx32 " ",
		 function,
		 b5_table_header_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( b5_table_header_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
	}
#endif

	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 16 )
	 || ( table_entry_value_size != 4 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_9c_item_entries(
	     table,
	     table_entries_array,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read 9c table item entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the a5 table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_a5_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_table_index_values_t *table_index_values = NULL;
	static char *function                           = "libpff_table_read_a5_values";

	/* The a5 table contains no b5 table header
	 */
	if( table_header_reference != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference: 0x%08" PRIx32 ".",
		 function,
		 table_header_reference );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( libpff_array_get_entry(
	     table_index_array,
	     0,
	     (intptr_t **) &table_index_values,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table index array entry: 0.",
		 function );

		return( -1 );
	}
	if( table_index_values == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index values.",
		 function );

		return( -1 );
	}
	if( table_index_values->value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index values - missing values.",
		 function );

		return( -1 );
	}
	if( table_index_values->number_of_values > 1 )
	{
		if( libpff_table_read_a5_item_entries(
		     table,
		     0x00000020,
		     table_index_array,
		     table_data_reference,
		     file_io_handle,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read a5 table item entries.",
			 function );

			return( -1 );
		}
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	else if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table contains no entries.\n",
		 function );
	}
#endif
	return( 1 );
}

/* Reads the ac table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_ac_values(
     libpff_table_t *table,
     uint32_t table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array           = NULL;
	libpff_array_t *table_entry_definitions_array = NULL;
	libpff_table_index_value_t *table_index_value = NULL;
	uint8_t *table_data                           = NULL;
	uint8_t *table_header_data                    = NULL;
	static char *function                         = "libpff_table_read_ac_values";
	size_t table_data_size                        = 0;
	size_t table_header_data_size                 = 0;
	uint32_t b5_table_header_reference            = 0;
	uint32_t table_entries_reference              = 0;
	uint32_t table_entry_definitions_reference    = 0;
	uint32_t values_array_reference               = 0;
	uint16_t values_array_entry_size              = 0;
	uint8_t table_entry_identifier_size           = 0;
	uint8_t table_entry_value_size                = 0;
	uint8_t table_entries_level                   = 0;
	int number_of_table_entry_definitions         = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint16_t value_16bit                          = 0;
#endif

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( ( table_header_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_header_reference & 0x0000001fUL,
		 table_header_reference );

		return( -1 );
	}
	/* Fetch the ac table header reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_header_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve ac table header reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_header_data      = &( table_data[ table_index_value->offset ] );
	table_header_data_size = (size_t) table_index_value->size;

	/* Read the ac table header
	 */
	if( ( (pff_table_header_ac_t *) table_header_data )->type != 0xac )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header type: 0x%02x.",
		 function,
		 ( (pff_table_header_ac_t *) table_header_data )->type );

		return( -1 );
	}
	/* TODO check ac header size */

	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_table_header_ac_t *) table_header_data )->values_array_entry_end_offset_cell_existence_block,
	 values_array_entry_size );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_ac_t *) table_header_data )->b5_table_header_reference,
	 b5_table_header_reference );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_ac_t *) table_header_data )->values_array_reference,
	 values_array_reference );
	byte_stream_copy_to_uint16_little_endian(
	 ( (pff_table_header_ac_t *) table_header_data )->number_of_entry_definitions,
	 number_of_table_entry_definitions );
	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_ac_t *) table_header_data )->table_entry_definitions_reference,
	 table_entry_definitions_reference );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table header type\t\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_table_header_ac_t *) table_header_data )->type );

		libnotify_printf(
		 "%s: padding1\t\t\t\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_table_header_ac_t *) table_header_data )->padding1 );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_ac_t *) table_header_data )->values_array_entry_end_offset_32bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 32-bit values\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_ac_t *) table_header_data )->values_array_entry_end_offset_16bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 16-bit values\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_ac_t *) table_header_data )->values_array_entry_end_offset_8bit_values,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset 8-bit values\t\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		byte_stream_copy_to_uint16_little_endian(
		 ( (pff_table_header_ac_t *) table_header_data )->values_array_entry_end_offset_cell_existence_block,
		 value_16bit );
		libnotify_printf(
		 "%s: values array entry end offset cell existence block\t: %" PRIu16 "\n",
		 function,
		 value_16bit );

		libnotify_printf(
		 "%s: b5 table header reference\t\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 b5_table_header_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( b5_table_header_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );

		libnotify_printf(
		 "%s: values array reference\t\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 values_array_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( values_array_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );

		libnotify_printf(
		 "%s: padding2:\n",
		 function );
		libnotify_print_data(
		 ( (pff_table_header_ac_t *) table_header_data )->padding2,
		 4 );

		libnotify_printf(
		 "%s: number of entry definitions\t\t\t: %d\n",
		 function,
		 number_of_table_entry_definitions );

		libnotify_printf(
		 "%s: table entry definitions reference\t\t\t: 0x%08" PRIx32 " ",
		 function,
		 table_entry_definitions_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( table_entry_definitions_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );

		libnotify_printf(
		 "%s: unknown2:\n",
		 function );
		libnotify_print_data(
		 ( (pff_table_header_ac_t *) table_header_data )->unknown2,
		 12 );
	}
#endif

	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 4 )
	 || ( table_entry_value_size != 4 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	/* Create the table entry definitions array
	 */
	if( libpff_array_initialize(
	     &table_entry_definitions_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entry definitions array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_ac_item_entry_definitions(
	     table_entry_definitions_array,
	     table_entry_definitions_reference,
	     number_of_table_entry_definitions,
	     recovered,
	     io_handle,
	     file_io_handle,
	     local_descriptors,
	     name_to_id_map_list,
	     debug_ascii_codepage,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read ac table item entry definitions.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );
		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );

		return( -1 );
	}
	if( libpff_array_get_number_of_entries(
	     table_index_array,
	     &number_of_table_entry_definitions,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table entry definitions.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );
		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );

		return( -1 );
	}
	if( number_of_table_entry_definitions > 0 )
	{
		if( libpff_table_read_values_array(
		     table,
		     table_entries_reference,
		     values_array_reference,
		     table_entry_identifier_size,
		     table_entry_value_size,
		     values_array_entry_size,
		     table_index_array,
		     table_entry_definitions_array,
		     table_entries_array,
		     recovered,
		     io_handle,
		     local_descriptors,
		     table_data_reference,
		     file_io_handle,
		     debug_ascii_codepage,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read values array.",
			 function );

			libpff_array_free(
			 &table_entries_array,
			 &libpff_data_reference_free_as_value,
			 NULL );
			libpff_array_free(
			 &table_entry_definitions_array,
			 &libpff_table_entry_definition_free,
			 NULL );

			return( -1 );
		}
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		libpff_array_free(
		 &table_entry_definitions_array,
		 &libpff_table_entry_definition_free,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entry_definitions_array,
	     &libpff_table_entry_definition_free,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entry definitions array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the bc table values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_bc_values(
     libpff_table_t *table,
     uint32_t b5_table_header_reference,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_array_t *table_entries_array = NULL;
	static char *function               = "libpff_table_read_bc_values";
	uint32_t table_entries_reference    = 0;
	uint8_t table_entry_identifier_size = 0;
	uint8_t table_entry_value_size      = 0;
	uint8_t table_entries_level         = 0;

	/* Read the b5 table header
	 */
	if( libpff_table_read_b5_header(
	     b5_table_header_reference,
	     &table_entry_identifier_size,
	     &table_entry_value_size,
	     &table_entries_level,
	     &table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read b5 table header.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 2 )
	 || ( table_entry_value_size != 6 ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	/* Create the table entries array
	 */
	if( libpff_array_initialize(
	     &table_entries_array,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_table_read_entries(
	     table_entries_array,
	     table_entries_level,
	     table_entry_identifier_size,
	     table_entries_reference,
	     table_index_array,
	     table_data_reference,
	     file_io_handle,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_table_read_bc_item_entries(
	     table,
	     table_index_array,
	     table_entries_array,
	     recovered,
	     io_handle,
	     file_io_handle,
	     local_descriptors,
	     name_to_id_map_list,
	     table_data_reference,
	     debug_item_type,
	     debug_ascii_codepage,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read table item entries.",
		 function );

		libpff_array_free(
		 &table_entries_array,
		 &libpff_data_reference_free_as_value,
		 NULL );

		return( -1 );
	}
	if( libpff_array_free(
	     &table_entries_array,
	     &libpff_data_reference_free_as_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free the table entries array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the b5 table header
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_b5_header(
     uint32_t table_header_reference,
     uint8_t *table_entry_identifier_size,
     uint8_t *table_entry_value_size,
     uint8_t *table_entries_level,
     uint32_t *table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_table_index_value_t *table_index_value = NULL;
	uint8_t *table_data                           = NULL;
	uint8_t *table_value_data                     = NULL;
	static char *function                         = "libpff_table_read_b5_header";
	size_t table_data_size                        = 0;

	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( table_entry_identifier_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry identifier size.",
		 function );

		return( -1 );
	}
	if( table_entry_value_size == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry value size.",
		 function );

		return( -1 );
	}
	if( table_entries_level == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries level.",
		 function );

		return( -1 );
	}
	if( table_entries_reference == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries reference.",
		 function );

		return( -1 );
	}
	if( ( table_header_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_header_reference & 0x0000001fUL,
		 table_header_reference );

		return( -1 );
	}
	/* Fetch the b5 table header reference
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     table_header_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve b5 table header reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_value_data = &( table_data[ table_index_value->offset ] );

	if( ( (pff_table_header_b5_t *) table_value_data )->type != 0xb5 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table header type: 0x%02x.",
		 function,
		 ( (pff_table_header_b5_t *) table_value_data )->type );

		return( -1 );
	}
	*table_entry_identifier_size = ( (pff_table_header_b5_t *) table_value_data )->table_entry_identifier_size;
	*table_entry_value_size      = ( (pff_table_header_b5_t *) table_value_data )->table_entry_value_size;
	*table_entries_level         = ( (pff_table_header_b5_t *) table_value_data )->table_entries_level;

	byte_stream_copy_to_uint32_little_endian(
	 ( (pff_table_header_b5_t *) table_value_data )->table_entries_reference,
	 *table_entries_reference );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table header type\t\t\t: 0x%02" PRIx8 "\n",
		 function,
		 ( (pff_table_header_b5_t *) table_value_data )->type );
		libnotify_printf(
		 "%s: table entry identifier size\t: %" PRIu8 "\n",
		 function,
		 *table_entry_identifier_size );
		libnotify_printf(
		 "%s: table entry value size\t\t: %" PRIu8 "\n",
		 function,
		 *table_entry_value_size );
		libnotify_printf(
		 "%s: table entries level\t\t: %" PRIu8 "\n",
		 function,
		 ( (pff_table_header_b5_t *) table_value_data )->table_entries_level );
		libnotify_printf(
		 "%s: table entries reference\t\t: 0x%08" PRIx32 " ",
		 function,
		 *table_entries_reference );
		libpff_debug_print_node_identifier_type(
		 (uint8_t) ( *table_entries_reference & 0x0000001fUL ) );
		libnotify_printf(
		 "\n" );
		libnotify_printf(
		 "\n" );
	}
#endif

	return( 1 );
}

/* Reads the 6c table item entries and their values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_6c_item_entries(
     libpff_table_t *table,
     uint32_t values_array_reference,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entries_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
#if defined( HAVE_DEBUG_OUTPUT )
	uint8_t guid_string[ LIBFMAPI_GUID_STRING_SIZE ];
#endif

	libpff_data_reference_t *table_entries_data_reference = NULL;
	libpff_table_entry_t *table_entry                     = NULL;
	libpff_table_index_value_t *table_index_value         = NULL;
	uint8_t *table_data                                   = NULL;
	uint8_t *table_entry_data                             = NULL;
	uint8_t *table_values_array_data                      = NULL;
	static char *function                                 = "libpff_table_read_6c_item_entries";
	size64_t table_entries_data_size                      = 0;
	size_t number_of_table_entries                        = 0;
	size_t table_data_size                                = 0;
	size_t table_entry_data_size                          = 0;
	size_t table_values_array_data_size                   = 0;
	uint32_t table_entry_iterator                         = 0;
	uint16_t values_array_number                          = 0;
	int number_of_table_entries_array_entries             = 0;
	int table_entries_array_iterator                      = 0;
	int table_entry_index                                 = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid table - entries already set.",
		 function );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	if( values_array_reference == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table contains no value array.",
		 function );

		return( -1 );
	}
	if( ( values_array_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported values array entries reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 values_array_reference & 0x0000001fUL,
		 values_array_reference );

		return( -1 );
	}
	/* Fetch the values array
	 */
	if( libpff_table_index_array_get_value(
	     table_index_array,
	     values_array_reference,
	     &table_index_value,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve values array entries reference.",
		 function );

		return( -1 );
	}
	if( table_index_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table index value.",
		 function );

		return( -1 );
	}
	/* Retrieve the corresponding table data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_data_reference,
	     file_io_handle,
	     table_index_value->array_entry,
	     &table_data,
	     &table_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( table_data == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table data array entry: %" PRIu32 ".",
		 function,
		 table_index_value->array_entry );

		return( -1 );
	}
	if( (size_t) table_index_value->offset >= table_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table value offset exceeds table data size.",
		 function );

		return( -1 );
	}
	table_values_array_data      = &( table_data[ table_index_value->offset ] );
	table_values_array_data_size = (size_t) table_index_value->size;

	if( libpff_array_get_number_of_entries(
	     table_entries_array,
	     &number_of_table_entries_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table entries array entries.",
		 function );

		return( -1 );
	}
	for( table_entries_array_iterator = 0;
	     table_entries_array_iterator < number_of_table_entries_array_entries;
	     table_entries_array_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table entries array entry: %d\n",
			 function,
			 table_entries_array_iterator );
		}
#endif

		if( libpff_array_get_entry(
		     table_entries_array,
		     table_entries_array_iterator,
		     (intptr_t **) &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d.",
			 function,
			 table_entries_array_iterator );

			return( -1 );
		}
		if( libpff_data_reference_get_data_size(
		     table_entries_data_reference,
		     &table_entries_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d data size.",
			 function,
			 table_entries_array_iterator );

			return( -1 );
		}
		if( ( table_entries_data_size % sizeof( pff_table_entry_6c_t ) ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries data size.",
			 function );

			return( -1 );
		}
		number_of_table_entries = (size_t) ( table_entries_data_size / sizeof( pff_table_entry_6c_t ) );

		if( number_of_table_entries > (size_t) INT_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: number of table entries value exceeds maximum.",
			 function );

			return( -1 );
		}
		if( libpff_table_entries_resize(
		     table,
		     1,
		     table->number_of_entries + (int) number_of_table_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entries.",
			 function );

			return( -1 );
		}
		if( table->entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table - missing table entries.",
			 function );

			return( -1 );
		}
		table_entry = table->entry[ 0 ];

		if( table_entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table entries.",
			 function );

			return( -1 );
		}
		for( table_entry_iterator = 0;
		     table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     (size_t) table_entry_iterator * sizeof( pff_table_entry_6c_t ),
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_iterator * sizeof( pff_table_entry_6c_t ) );

				return( -1 );
			}
			if( table_entry_data_size < sizeof( pff_table_entry_6c_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				return( -1 );
			}
			( table_entry[ table_entry_index ] ).identifier.format = LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_GUID;

			if( memory_copy(
			     ( table_entry[ table_entry_index ] ).identifier.guid,
			     ( (pff_table_entry_6c_t *) table_entry_data )->item_entry_guid,
			     16 ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to copy table entry identifier GUID.",
				 function );

				return( -1 );
			}
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_6c_t *) table_entry_data )->entry_number,
			 values_array_number );

			if( (size_t) ( 16 * values_array_number ) > table_values_array_data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: values array number exceeds table values array data size.",
				 function );

				return( -1 );
			}
			/* Create a data reference of the table entry value data
			 */
			if( libpff_data_reference_initialize_buffer(
			     &( ( table_entry[ table_entry_index ] ).value_data_reference ),
			     &( table_values_array_data[ 16 * values_array_number ] ),
			     16,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create table entry value data reference.",
				 function );

				return( -1 );
			}
			/* TODO do something with values_array_number ? */

#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( libfmapi_guid_to_string(
				     (libfmapi_guid_t *) ( table_entry[ table_entry_index ] ).identifier.guid,
				     LIBFMAPI_ENDIAN_LITTLE,
				     guid_string,
				     LIBFMAPI_GUID_STRING_SIZE,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_CONVERSION,
					 LIBERROR_CONVERSION_ERROR_GENERIC,
					 "%s: unable to create guid string.",
					 function );

					return( -1 );
				}
				libnotify_printf(
				 "%s: table set: %03d entry: %03d item entry guid\t\t\t: %s\n",
				 function,
				 0,
				 table_entry_index,
				 guid_string );

				if( libfmapi_guid_to_string(
				     (libfmapi_guid_t *) &( table_values_array_data[ 16 * values_array_number ] ),
				     LIBFMAPI_ENDIAN_LITTLE,
				     guid_string,
				     LIBFMAPI_GUID_STRING_SIZE,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_CONVERSION,
					 LIBERROR_CONVERSION_ERROR_GENERIC,
					 "%s: unable to create guid string.",
					 function );

					return( -1 );
				}
				libnotify_printf(
				 "%s: table set: %03d entry: %03d item entry value guid\t\t: %s\n",
				 function,
				 0,
				 table_entry_index,
				 guid_string );
				libnotify_printf(
				 "%s: table set: %03d entry: %03d item entry value identifier\t: 0x%04" PRIx16 "\n",
				 function,
				 0,
				 table_entry_index,
				 values_array_number );

				libnotify_printf(
				 "\n" );
			}
#endif
			table_entry_data += sizeof( pff_table_entry_6c_t );

			table_entry_index++;
		}
	}
	return( 1 );
}

/* Reads the 7c table item entry definitions
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_7c_item_entry_definitions(
     libpff_array_t *table_entry_definitions_array,
     libpff_data_reference_t *table_entry_definition_data_reference,
     int number_of_table_entry_definitions,
     libpff_array_t *table_index_array,
     libbfio_handle_t *file_io_handle,
     libpff_list_t *name_to_id_map_list,
     liberror_error_t **error )
{
	libpff_list_element_t *list_element               = NULL;
	libpff_table_entry_definition_t *entry_definition = NULL;
	uint8_t *table_entry_definition_data              = NULL;
	static char *function                             = "libpff_table_read_7c_item_entry_definitions";
	size_t table_entry_definition_data_size           = 0;
	uint8_t entry_definition_number                   = 0;
	int list_iterator                                 = 0;
	int table_entry_definition_iterator               = 0;

	if( table_entry_definitions_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry definitions array.",
		 function );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( number_of_table_entry_definitions > 0 )
	{
		if( libpff_array_resize(
		     table_entry_definitions_array,
		     table_entry_definitions_array->number_of_entries + number_of_table_entry_definitions,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entry definition array.",
			 function );

			return( -1 );
		}
		for( table_entry_definition_iterator = 0;
		     table_entry_definition_iterator < number_of_table_entry_definitions;
		     table_entry_definition_iterator++ )
		{
			entry_definition = (libpff_table_entry_definition_t *) memory_allocate(
			                                                        sizeof( libpff_table_entry_definition_t ) );

			if( entry_definition == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create table entry definition.",
				 function );

				return( -1 );
			}
			if( memory_set(
			     entry_definition,
			     0,
			     sizeof( libpff_table_entry_definition_t ) ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear table entry definition.",
				 function );

				memory_free(
				 entry_definition );

				return( -1 );
			}
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entry_definition_data_reference,
			     file_io_handle,
			     (size_t) table_entry_definition_iterator * sizeof( pff_table_entry_definition_7c_t ),
			     &table_entry_definition_data,
			     &table_entry_definition_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_definition_iterator * sizeof( pff_table_entry_definition_7c_t ) );

				memory_free(
				 entry_definition );

				return( -1 );
			}
			if( table_entry_definition_data_size < sizeof( pff_table_entry_definition_7c_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry definition data size value too small.",
				 function );

				memory_free(
				 entry_definition );

				return( -1 );
			}
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_7c_t *) table_entry_definition_data )->item_entry_type,
			 entry_definition->entry_type );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_7c_t *) table_entry_definition_data )->item_entry_value_type,
			 entry_definition->value_type );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_7c_t *) table_entry_definition_data )->entry_offset,
			 entry_definition->entry_offset );

			entry_definition->entry_size = ( (pff_table_entry_definition_7c_t *) table_entry_definition_data )->entry_size;
			entry_definition_number      = ( (pff_table_entry_definition_7c_t *) table_entry_definition_data )->entry_number;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table entry definition: %03d item entry type\t\t: 0x%04" PRIx16 "",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_type );
			}
#endif

			if( ( entry_definition->entry_type >= 0x8000 )
			 || ( entry_definition->entry_type <= 0xfffe ) )
			{
				if( name_to_id_map_list != NULL )
				{
					list_element = name_to_id_map_list->first;

					for( list_iterator = 0;
					     list_iterator < name_to_id_map_list->number_of_elements;
					     list_iterator++ )
					{
						if( list_element == NULL )
						{
							break;
						}
						if( list_element->value == NULL )
						{
							continue;
						}
						if( ( (libpff_internal_name_to_id_map_entry_t *) list_element->value )->identifier == (uint32_t) entry_definition->entry_type )
						{
							entry_definition->name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) list_element->value;
						}
						list_element = list_element->next;
					}
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( entry_definition->name_to_id_map_entry != NULL )
				{
					if( entry_definition->name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
					{
						libnotify_printf(
						 " maps to: %s (%s : %s)\n",
						 (char *) entry_definition->name_to_id_map_entry->debug_string,
						 libfmapi_named_property_type_get_identifier(
						  entry_definition->name_to_id_map_entry->guid,
						  (char *) entry_definition->name_to_id_map_entry->debug_string,
						  entry_definition->name_to_id_map_entry->value_size,
						  entry_definition->value_type ),
						 libfmapi_named_property_type_get_description(
						  entry_definition->name_to_id_map_entry->guid,
						  (char *) entry_definition->name_to_id_map_entry->debug_string,
						  entry_definition->name_to_id_map_entry->value_size,
						  entry_definition->value_type ) );
					}
					else
					{
						libnotify_printf(
						 " maps to: 0x%04" PRIx32 " (%s : %s)\n",
						 entry_definition->name_to_id_map_entry->numeric_value,
						 libfmapi_property_type_get_identifier(
						  entry_definition->name_to_id_map_entry->guid,
						  entry_definition->name_to_id_map_entry->numeric_value,
						  entry_definition->value_type ),
						 libfmapi_property_type_get_description(
						  entry_definition->name_to_id_map_entry->guid,
						  entry_definition->name_to_id_map_entry->numeric_value,
						  entry_definition->value_type ) );
					}
				}
				else
				{
					libnotify_printf(
					 " (%s : %s)\n",
					 libfmapi_property_type_get_identifier(
					  NULL,
					  entry_definition->entry_type,
					  entry_definition->value_type ),
					 libfmapi_property_type_get_description(
					  NULL,
					  entry_definition->entry_type,
					  entry_definition->value_type ) );
				}
				libnotify_printf(
				 "%s: table entry definition: %03d item entry value type\t\t: 0x%04" PRIx16 " (%s : %s)\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->value_type,
				 libfmapi_value_type_get_identifier(
				  entry_definition->value_type ),
				 libfmapi_value_type_get_description(
				  entry_definition->value_type ) );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry offset\t: %" PRIu16 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_offset );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry size\t: %" PRIu16 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_size );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry number\t: %" PRIu8 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition_number );
				libnotify_printf(
				 "\n" );
			}
#endif

			if( libpff_array_set_entry(
			     table_entry_definitions_array,
			     (int) entry_definition_number,
			     (intptr_t *) entry_definition,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_SET_FAILED,
				 "%s: unable to set table entry definitions array entry: %" PRIu8 ".",
				 function,
				 entry_definition_number );

				memory_free(
				 entry_definition );

				return( -1 );
			}
		}
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	else if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table contains no item entry definitions.\n",
		 function );
	}
#endif
	return( 1 );
}

/* Reads the 8c table item entries and their values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_8c_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_entries_array,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_data_reference_t *table_entries_data_reference = NULL;
	libpff_table_entry_t *table_entry                     = NULL;
	uint8_t *table_entry_data                             = NULL;
	static char *function                                 = "libpff_table_read_8c_item_entries";
	size64_t table_entries_data_size                      = 0;
	size_t number_of_table_entries                        = 0;
	size_t table_entry_data_size                          = 0;
	uint32_t table_entry_iterator                         = 0;
	int number_of_table_entries_array_entries             = 0;
	int table_entries_array_iterator                      = 0;
	int table_entry_index                                 = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint32_t test                                         = 0;
#endif

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid table - entries already set.",
		 function );

		return( -1 );
	}
	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_array_get_number_of_entries(
	     table_entries_array,
	     &number_of_table_entries_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table entries array entries.",
		 function );

		return( -1 );
	}
	for( table_entries_array_iterator = 0;
	     table_entries_array_iterator < number_of_table_entries_array_entries;
	     table_entries_array_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table entries array entry: %d\n",
			 function,
			 table_entries_array_iterator );
		}
#endif

		if( libpff_array_get_entry(
		     table_entries_array,
		     table_entries_array_iterator,
		     (intptr_t **) &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d.",
			 function,
			 table_entries_data_reference );

			return( -1 );
		}
		if( libpff_data_reference_get_data_size(
		     table_entries_data_reference,
		     &table_entries_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries data size.",
			 function );

			return( -1 );
		}
		if( ( table_entries_data_size % sizeof( pff_table_entry_8c_t ) ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries data size.",
			 function );

			return( -1 );
		}
		number_of_table_entries = (size_t) ( table_entries_data_size / sizeof( pff_table_entry_8c_t ) );

		if( number_of_table_entries > (size_t) INT_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: number of table entries value exceeds maximum.",
			 function );

			return( -1 );
		}
		if( libpff_table_entries_resize(
		     table,
		     1,
		     table->number_of_entries + (int) number_of_table_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entries.",
			 function );

			return( -1 );
		}
		if( table->entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table - missing table entries.",
			 function );

			return( -1 );
		}
		table_entry = table->entry[ 0 ];

		if( table_entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table entries.",
			 function );

			return( -1 );
		}
		for( table_entry_iterator = 0;
		     table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     (size_t) table_entry_iterator * sizeof( pff_table_entry_8c_t ),
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_iterator * sizeof( pff_table_entry_8c_t ) );

				return( -1 );
			}
			if( table_entry_data_size < sizeof( pff_table_entry_8c_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				return( -1 );
			}
			( table_entry[ table_entry_index ] ).identifier.format = LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_SECURE4;

			byte_stream_copy_to_uint64_little_endian(
			 ( (pff_table_entry_8c_t *) table_entry_data )->identifier,
			 ( table_entry[ table_entry_index ] ).identifier.secure4 );

			/* Create a data reference of the table entry value data
			 */
			if( libpff_data_reference_initialize_buffer(
			     &( ( table_entry[ table_entry_index ] ).value_data_reference ),
			     ( (pff_table_entry_8c_t *) table_entry_data )->descriptor_identifier,
			     sizeof( uint32_t ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create table entry value data reference.",
				 function );

				return( -1 );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03d entry: %03d identifier\t\t\t: 0x%" PRIx64 "\n",
				 function,
				 0,
				 table_entry_index,
				 ( table_entry[ table_entry_index ] ).identifier.secure4 );

				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_table_entry_8c_t *) table_entry_data )->descriptor_identifier,
				 test );

				libnotify_printf(
				 "%s: table set: %03d entry: %03d descriptor identifier\t: 0x%08" PRIx32 "\n",
				 function,
				 0,
				 table_entry_index,
				 test );

				libnotify_printf(
				 "\n" );
			}
#endif
			table_entry_data += sizeof( pff_table_entry_8c_t );

			table_entry_index++;
		}
	}
	return( 1 );
}

/* Reads the 9c table item entries and their values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_9c_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_entries_array,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
#if defined( HAVE_DEBUG_OUTPUT )
	uint8_t guid_string[ LIBFMAPI_GUID_STRING_SIZE ];
#endif

	libpff_data_reference_t *table_entries_data_reference = NULL;
	libpff_table_entry_t *table_entry                     = NULL;
	uint8_t *table_entry_data                             = NULL;
	static char *function                                 = "libpff_table_read_9c_item_entries";
	size64_t table_entries_data_size                      = 0;
	size_t number_of_table_entries                        = 0;
	size_t table_entry_data_size                          = 0;
	uint32_t table_entry_iterator                         = 0;
	int number_of_table_entries_array_entries             = 0;
	int table_entries_array_iterator                      = 0;
	int table_entry_index                                 = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	uint32_t test                                         = 0;
#endif

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid table - entries already set.",
		 function );

		return( -1 );
	}
	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	if( libpff_array_get_number_of_entries(
	     table_entries_array,
	     &number_of_table_entries_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table entries array entries.",
		 function );

		return( -1 );
	}
	for( table_entries_array_iterator = 0;
	     table_entries_array_iterator < number_of_table_entries_array_entries;
	     table_entries_array_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table entries array entry: %d\n",
			 function,
			 table_entries_array_iterator );
		}
#endif

		if( libpff_array_get_entry(
		     table_entries_array,
		     table_entries_array_iterator,
		     (intptr_t **) &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d.",
			 function,
			 table_entries_array_iterator );

			return( -1 );
		}
		if( libpff_data_reference_get_data_size(
		     table_entries_data_reference,
		     &table_entries_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d data size.",
			 function,
			 table_entries_array_iterator );

			return( -1 );
		}
		if( ( table_entries_data_size % sizeof( pff_table_entry_9c_t ) ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries data size.",
			 function );

			return( -1 );
		}
		number_of_table_entries = (size_t) ( table_entries_data_size / sizeof( pff_table_entry_9c_t ) );

		if( number_of_table_entries > (size_t) INT_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: number of table entries value exceeds maximum.",
			 function );

			return( -1 );
		}
		if( libpff_table_entries_resize(
		     table,
		     1,
		     table->number_of_entries + (int) number_of_table_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entries.",
			 function );

			return( -1 );
		}
		if( table->entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table - missing table entries.",
			 function );

			return( -1 );
		}
		table_entry = table->entry[ 0 ];

		if( table_entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table entries.",
			 function );

			return( -1 );
		}
		for( table_entry_iterator = 0;
		     table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     (size_t) table_entry_iterator * sizeof( pff_table_entry_9c_t ),
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_iterator * sizeof( pff_table_entry_9c_t ) );

				return( -1 );
			}
			if( table_entry_data_size < sizeof( pff_table_entry_9c_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				return( -1 );
			}
			( table_entry[ table_entry_index ] ).identifier.format = LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_GUID;

			if( memory_copy(
			     ( table_entry[ table_entry_index ] ).identifier.guid,
			     ( (pff_table_entry_9c_t *) table_entry_data )->item_entry_guid,
			     16 ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_COPY_FAILED,
				 "%s: unable to copy table entry identifier GUID.",
				 function );

				return( -1 );
			}
			/* Create a data reference of the table entry value data
			 */
			if( libpff_data_reference_initialize_buffer(
			     &( ( table_entry[ table_entry_index ] ).value_data_reference ),
			     ( (pff_table_entry_9c_t *) table_entry_data )->descriptor_identifier,
			     sizeof( uint32_t ),
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create table entry value data reference.",
				 function );

				return( -1 );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( libfmapi_guid_to_string(
				     (libfmapi_guid_t *) ( table_entry[ table_entry_index ] ).identifier.guid,
				     LIBFMAPI_ENDIAN_LITTLE,
				     guid_string,
				     LIBFMAPI_GUID_STRING_SIZE,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_CONVERSION,
					 LIBERROR_CONVERSION_ERROR_GENERIC,
					 "%s: unable to create guid string.",
					 function );

					return( -1 );
				}
				libnotify_printf(
				 "%s: table set: %03d entry: %03d item entry guid\t\t\t: %s\n",
				 function,
				 0,
				 table_entry_index,
				 guid_string );

				byte_stream_copy_to_uint32_little_endian(
				 ( (pff_table_entry_9c_t *) table_entry_data )->descriptor_identifier,
				 test );

				libnotify_printf(
				 "%s: table set: %03d entry: %03d item entry descriptor identifier\t: 0x%08" PRIx32 "\n",
				 function,
				 0,
				 table_entry_index,
				 test );

				libnotify_printf(
				 "\n" );
			}
#endif
			table_entry_data += sizeof( pff_table_entry_9c_t );

			table_entry_index++;
		}
	}
	return( 1 );
}

/* Reads the a5 table item entry values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_a5_item_entries(
     libpff_table_t *table,
     uint32_t table_entries_reference,
     libpff_array_t *table_index_array,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     liberror_error_t **error )
{
	libpff_table_entry_t *table_entry               = NULL;
	libpff_table_index_value_t *table_index_value   = NULL;
	libpff_table_index_values_t *table_index_values = NULL;
	uint8_t *table_data                             = NULL;
	static char *function                           = "libpff_table_read_a5_item_entries";
	size_t table_data_size                          = 0;
	uint16_t table_index_values_iterator            = 0;
	int number_of_table_index_array_entries         = 0;
	int table_index_array_entries_iterator          = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid table - entries already set.",
		 function );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	/* Check if the table contains any entries
	 */
	if( table_entries_reference == 0 )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table contains no entries.\n",
			 function );
		}
#endif

		return( 1 );
	}
	if( ( table_entries_reference & 0x0000001fUL ) != 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entries reference type: 0x%08" PRIx32 " (0x%08" PRIx32 ").",
		 function,
		 table_entries_reference & 0x0000001fUL,
		 table_entries_reference );

		return( -1 );
	}
	if( libpff_array_get_number_of_entries(
	     table_index_array,
	     &number_of_table_index_array_entries,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve the number of table index array entries.",
		 function );

		return( -1 );
	}
	for( table_index_array_entries_iterator = 0;
	     table_index_array_entries_iterator < number_of_table_index_array_entries;
	     table_index_array_entries_iterator++ )
	{
		if( libpff_array_get_entry(
		     table_index_array,
		     table_index_array_entries_iterator,
		     (intptr_t **) &table_index_values,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table index array entry: %" PRIu16 ".",
			 function,
			 table_index_array_entries_iterator );

			return( -1 );
		}
		if( table_index_values == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table index values.",
			 function );

			return( -1 );
		}
		if( table_index_values->value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table index values - missing value.",
			 function );

			return( -1 );
		}
		if( ( number_of_table_index_array_entries > table->number_of_sets )
		 || ( table_index_values->number_of_values > table->number_of_entries ) )
		{
			if( libpff_table_entries_resize(
			     table,
			     number_of_table_index_array_entries,
			     (int) table_index_values->number_of_values,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
				 "%s: unable to resize table entries.",
				 function );

				return( -1 );
			}
		}
		if( table->entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table - missing table entries.",
			 function );

			return( -1 );
		}
		table_entry = table->entry[ table_index_array_entries_iterator ];

		if( table_entry == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table entries.",
			 function );

			return( -1 );
		}
		for( table_index_values_iterator = 0;
		     table_index_values_iterator < table_index_values->number_of_values;
		     table_index_values_iterator++ )
		{
			table_index_value = &( table_index_values->value[ table_index_values_iterator ] );

			if( table_index_value == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: invalid table index value.",
				 function );

				return( -1 );
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03d value at offset: %" PRIu32 " with size: %" PRIu16 ".\n",
				 function,
				 table_index_array_entries_iterator,
				 table_index_values_iterator,
				 table_index_value->offset,
				 table_index_value->size );
			}
#endif

			/* Retrieve the corresponding table data array entry
			 */
			if( libpff_data_reference_get_entry_data(
			     table_data_reference,
			     file_io_handle,
			     table_index_value->array_entry,
			     &table_data,
			     &table_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
				 function,
				 table_index_value->array_entry );

				return( -1 );
			}
			if( table_data == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
				 "%s: invalid table data array entry: %" PRIu32 ".",
				 function,
				 table_index_value->array_entry );

				return( -1 );
			}
			if( (size_t) table_index_value->offset >= table_data_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: table value offset exceeds table data size.",
				 function );

				return( -1 );
			}
			/* Create a data reference of the table entry value data
			 */
			if( libpff_data_reference_initialize_buffer(
			     &( ( table_entry[ table_index_values_iterator ] ).value_data_reference ),
			     &( table_data[ table_index_value->offset ] ),
			     table_index_value->size,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create table entry value data reference.",
				 function );

				return( -1 );
			}
		}
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );
}

/* Reads the ac table item entry definitions
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_ac_item_entry_definitions(
     libpff_array_t *table_entry_definitions_array,
     uint32_t table_entry_definitions_reference,
     int number_of_table_entry_definitions,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *table_entry_definition_data_reference  = NULL;
	libpff_data_reference_t *table_item_entry_values_data_reference = NULL;
	libpff_tree_node_t *table_item_entry_values_local_descriptors   = NULL;
	libpff_list_element_t *list_element                             = NULL;
	libpff_table_entry_definition_t *entry_definition               = NULL;
	uint8_t *table_entry_definition_data                            = NULL;
	uint8_t *table_entry_definitions_data                           = NULL;
	static char *function                                           = "libpff_table_read_ac_item_entry_definitions";
	size_t table_entry_definition_data_size                         = 0;
	size_t table_entry_definitions_data_size                        = 0;
	uint32_t item_entry_values_table_descriptor                     = 0;
	uint16_t entry_definition_number                                = 0;
	int list_iterator                                               = 0;
	int result                                                      = 0;
	int table_entry_definition_iterator                             = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	libpff_table_index_value_t *table_index_value                   = NULL;
#endif

	if( table_entry_definitions_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry definitions array.",
		 function );

		return( -1 );
	}
	/* Read the entry definitions
	 */
	result = libpff_descriptor_local_read(
		  table_entry_definitions_reference,
		  recovered,
		  local_descriptors,
		  io_handle,
		  file_io_handle,
		  NULL,
		  &table_entry_definition_data_reference,
		  error );

	if( result == -1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_IO,
		 LIBERROR_IO_ERROR_READ_FAILED,
		 "%s: unable to read entry definitions data with descriptor: 0x%08" PRIx32 " (%" PRIu32 ").",
		 function,
		 table_entry_definitions_reference,
		 table_entry_definitions_reference );

		return( -1 );
	}
	else if( result == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: missing entry definitions data descriptor: 0x%08" PRIx32 " (%" PRIu32 ").",
		 function,
		 table_entry_definitions_reference,
		 table_entry_definitions_reference );

		return( -1 );
	}
	/* TODO what about entry definitons in a table array ?
	 * check for table array
	 */
	/* Retrieve the corresponding table entry definitions data array entry
	 */
	if( libpff_data_reference_get_entry_data(
	     table_entry_definition_data_reference,
	     file_io_handle,
	     0,
	     &table_entry_definitions_data,
	     &table_entry_definitions_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table entry definitions data array entry: 0.",
		 function );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
	if( ( (size_t) number_of_table_entry_definitions * sizeof( pff_table_entry_definition_ac_t ) ) != table_entry_definitions_data_size )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: mismatch in the number of table entry definitions and the data size.",
		 function );

		libpff_data_reference_free(
		 &table_entry_definition_data_reference,
		 NULL );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		if( table_entry_definitions_data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid table data array entry: %" PRIu32 ".",
			 function,
			 table_index_value->array_entry );

			libpff_data_reference_free(
			 &table_entry_definition_data_reference,
			 NULL );

			return( -1 );
		}
		libnotify_printf(
		 "%s: ac table entry definitions:\n",
		 function );
		libnotify_print_data(
		 table_entry_definitions_data,
		 table_entry_definitions_data_size );
	}
#endif

	if( number_of_table_entry_definitions > 0 )
	{
		if( libpff_array_resize(
		     table_entry_definitions_array,
		     table_entry_definitions_array->number_of_entries + number_of_table_entry_definitions,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entry definition array.",
			 function );

			libpff_data_reference_free(
			 &table_entry_definition_data_reference,
			 NULL );

			return( -1 );
		}
		for( table_entry_definition_iterator = 0;
		     table_entry_definition_iterator < number_of_table_entry_definitions;
		     table_entry_definition_iterator++ )
		{
			entry_definition = (libpff_table_entry_definition_t *) memory_allocate(
			                                                        sizeof( libpff_table_entry_definition_t ) );

			if( entry_definition == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_INSUFFICIENT,
				 "%s: unable to create table entry definition.",
				 function );

				libpff_data_reference_free(
				 &table_entry_definition_data_reference,
				 NULL );

				return( -1 );
			}
			if( memory_set(
			     entry_definition,
			     0,
			     sizeof( libpff_table_entry_definition_t ) ) == NULL )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_MEMORY,
				 LIBERROR_MEMORY_ERROR_SET_FAILED,
				 "%s: unable to clear table entry definition.",
				 function );

				memory_free(
				 entry_definition );
				libpff_data_reference_free(
				 &table_entry_definition_data_reference,
				 NULL );

				return( -1 );
			}
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entry_definition_data_reference,
			     file_io_handle,
			     table_entry_definition_iterator * sizeof( pff_table_entry_definition_ac_t ),
			     &table_entry_definition_data,
			     &table_entry_definition_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 table_entry_definition_iterator * sizeof( pff_table_entry_definition_ac_t ) );

				memory_free(
				 entry_definition );
				libpff_data_reference_free(
				 &table_entry_definition_data_reference,
				 NULL );

				return( -1 );
			}
			if( table_entry_definition_data_size < sizeof( pff_table_entry_definition_ac_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry definition data size value too small.",
				 function );

				memory_free(
				 entry_definition );
				libpff_data_reference_free(
				 &table_entry_definition_data_reference,
				 NULL );

				return( -1 );
			}
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->item_entry_type,
			 entry_definition->entry_type );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->item_entry_value_type,
			 entry_definition->value_type );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->entry_offset,
			 entry_definition->entry_offset );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->entry_size,
			 entry_definition->entry_size );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->entry_number,
			 entry_definition_number );
			byte_stream_copy_to_uint32_little_endian(
			 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->item_entry_values_table_descriptor,
			 item_entry_values_table_descriptor );

#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table entry definition: %03d item entry type\t\t\t: 0x%04" PRIx16 "",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_type );
			}
#endif

			if( ( entry_definition->entry_type >= 0x8000 )
			 || ( entry_definition->entry_type <= 0xfffe ) )
			{
				if( name_to_id_map_list != NULL )
				{
					list_element = name_to_id_map_list->first;

					for( list_iterator = 0;
					     list_iterator < name_to_id_map_list->number_of_elements;
					     list_iterator++ )
					{
						if( list_element == NULL )
						{
							break;
						}
						if( list_element->value == NULL )
						{
							continue;
						}
						if( ( (libpff_internal_name_to_id_map_entry_t *) list_element->value )->identifier == (uint32_t) entry_definition->entry_type )
						{
							entry_definition->name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) list_element->value;
						}
						list_element = list_element->next;
					}
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				if( entry_definition->name_to_id_map_entry != NULL )
				{
					if( entry_definition->name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
					{
						libnotify_printf(
						 " maps to: %s (%s : %s)\n",
						 (char *) entry_definition->name_to_id_map_entry->debug_string,
						 libfmapi_named_property_type_get_identifier(
						  entry_definition->name_to_id_map_entry->guid,
						  (char *) entry_definition->name_to_id_map_entry->debug_string,
						  entry_definition->name_to_id_map_entry->value_size,
						  entry_definition->value_type ),
						 libfmapi_named_property_type_get_description(
						  entry_definition->name_to_id_map_entry->guid,
						  (char *) entry_definition->name_to_id_map_entry->debug_string,
						  entry_definition->name_to_id_map_entry->value_size,
						  entry_definition->value_type ) );
					}
					else
					{
						libnotify_printf(
						 " maps to: 0x%04" PRIx32 " (%s : %s)\n",
						 entry_definition->name_to_id_map_entry->numeric_value,
						 libfmapi_property_type_get_identifier(
						  entry_definition->name_to_id_map_entry->guid,
						  entry_definition->name_to_id_map_entry->numeric_value,
						  entry_definition->value_type ),
						 libfmapi_property_type_get_description(
						  entry_definition->name_to_id_map_entry->guid,
						  entry_definition->name_to_id_map_entry->numeric_value,
						  entry_definition->value_type ) );
					}
				}
				else
				{
					libnotify_printf(
					 " (%s : %s)\n",
					 libfmapi_property_type_get_identifier(
					  NULL,
					  entry_definition->entry_type,
					  entry_definition->value_type ),
					 libfmapi_property_type_get_description(
					  NULL,
					  entry_definition->entry_type,
					  entry_definition->value_type ) );
				}
				libnotify_printf(
				 "%s: table entry definition: %03d item entry value type\t\t\t: 0x%04" PRIx16 " (%s : %s)\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->value_type,
				 libfmapi_value_type_get_identifier(
				  entry_definition->value_type ),
				 libfmapi_value_type_get_description(
				  entry_definition->value_type ) );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry offset\t\t: %" PRIu16 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_offset );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry size\t\t: %" PRIu16 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition->entry_size );
				libnotify_printf(
				 "%s: table entry definition: %03d values array entry number\t\t: %" PRIu16 "\n",
				 function,
				 table_entry_definition_iterator,
				 entry_definition_number );
				libnotify_printf(
				 "%s: padding1:\n",
				 function );
				libnotify_print_data(
				 ( (pff_table_entry_definition_ac_t *) table_entry_definition_data )->padding1,
				 2 );
				libnotify_printf(
				 "%s: table entry definition: %03d item entry values table descriptor\t: %" PRIu32 "\n",
				 function,
				 table_entry_definition_iterator,
				 item_entry_values_table_descriptor );
				libnotify_printf(
				 "\n" );
			}
#endif

			/* Read the item entry values table if necessary
			 */
			if( item_entry_values_table_descriptor > 0 )
			{
				if( libpff_table_initialize(
				     &( entry_definition->item_entry_values_table ),
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
					 "%s: unable to create item entry values table.",
					 function );

					memory_free(
					 entry_definition );
					libpff_data_reference_free(
					 &table_entry_definition_data_reference,
					 NULL );

					return( -1 );
				}
				if( libpff_descriptor_local_read(
				     item_entry_values_table_descriptor,
				     recovered,
				     local_descriptors,
				     io_handle,
				     file_io_handle,
				     &table_item_entry_values_local_descriptors,
				     &table_item_entry_values_data_reference,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_IO,
					 LIBERROR_IO_ERROR_READ_FAILED,
					 "%s: unable to read table item entry values descriptor.",
					 function );

					libpff_table_free(
					 entry_definition->item_entry_values_table,
					 NULL );
					memory_free(
					 entry_definition );
					libpff_data_reference_free(
					 &table_entry_definition_data_reference,
					 NULL );

					return( -1 );
				}
				if( libpff_table_read(
				     entry_definition->item_entry_values_table,
				     recovered,
				     io_handle,
				     table_item_entry_values_local_descriptors,
				     name_to_id_map_list,
				     table_item_entry_values_data_reference,
				     file_io_handle,
				     LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
				     debug_ascii_codepage,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_IO,
					 LIBERROR_IO_ERROR_READ_FAILED,
					 "%s: unable to read item entry values table.",
					 function );

					if( table_item_entry_values_local_descriptors != NULL )
					{
						libpff_tree_node_free(
						 &table_item_entry_values_local_descriptors,
						 &libpff_local_descriptor_values_free,
					         NULL );
					}
					libpff_data_reference_free(
					 &table_item_entry_values_data_reference,
					 NULL );
					libpff_table_free(
					 entry_definition->item_entry_values_table,
					 NULL );
					memory_free(
					 entry_definition );
					libpff_data_reference_free(
					 &table_entry_definition_data_reference,
					 NULL );

					return( -1 );
				}
				if( table_item_entry_values_local_descriptors != NULL )
				{
					if( libpff_tree_node_free(
					     &table_item_entry_values_local_descriptors,
					     &libpff_local_descriptor_values_free,
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
						 "%s: unable to free item entry values local descriptors.",
						 function );

						libpff_data_reference_free(
						 &table_item_entry_values_data_reference,
						 NULL );
						libpff_table_free(
						 entry_definition->item_entry_values_table,
						 NULL );
						memory_free(
						 entry_definition );
						libpff_data_reference_free(
						 &table_entry_definition_data_reference,
						 NULL );

						return( -1 );
					}
					table_item_entry_values_local_descriptors = NULL;
				}
				if( libpff_data_reference_free(
				     &table_item_entry_values_data_reference,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
					 "%s: unable to free table item entry values data reference.",
					 function );

					libpff_table_free(
					 entry_definition->item_entry_values_table,
					 NULL );
					memory_free(
					 entry_definition );
					libpff_data_reference_free(
					 &table_entry_definition_data_reference,
					 NULL );

					return( -1 );
				}
			}
			if( libpff_array_set_entry(
			     table_entry_definitions_array,
			     (int) entry_definition_number,
			     (intptr_t *) entry_definition,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid entry number: %" PRIu16 " value exceeds number of entries: %d.",
				 function,
				 entry_definition_number,
				 number_of_table_entry_definitions );

				libpff_table_free(
				 entry_definition->item_entry_values_table,
				 NULL );
				memory_free(
				 entry_definition );
				libpff_data_reference_free(
				 &table_entry_definition_data_reference,
				 NULL );

				return( -1 );
			}
		}
	}
#if defined( HAVE_VERBOSE_OUTPUT )
	else if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table contains no entry definitions.\n",
		 function );
	}
#endif
	if( libpff_data_reference_free(
	     &table_entry_definition_data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free entry definitions data reference.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads the bc table item entries and their values
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_bc_item_entries(
     libpff_table_t *table,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entries_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_data_reference_t *table_data_reference,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *table_entries_data_reference = NULL;
	uint8_t *table_entry_data                             = NULL;
	static char *function                                 = "libpff_table_read_bc_item_entries";
	size64_t table_entries_data_size                      = 0;
	size_t number_of_table_entries                        = 0;
	size_t table_entry_data_size                          = 0;
	uint32_t table_entry_iterator                         = 0;
	uint16_t table_entry_type                             = 0;
	uint16_t table_entry_value_type                       = 0;
	int table_entries_array_iterator                      = 0;
	int table_entry_index                                 = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry != NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid table - entries already set.",
		 function );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	for( table_entries_array_iterator = 0;
	     table_entries_array_iterator < table_entries_array->number_of_entries;
	     table_entries_array_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table entries array entry: %d\n",
			 function,
			 table_entries_array_iterator );
		}
#endif

		if( libpff_array_get_entry(
		     table_entries_array,
		     table_entries_array_iterator,
		     (intptr_t **) &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d.",
			 function,
			 table_entries_data_reference );

			return( -1 );
		}
		if( libpff_data_reference_get_data_size(
		     table_entries_data_reference,
		     &table_entries_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries data size.",
			 function );

			return( -1 );
		}
		if( ( table_entries_data_size % sizeof( pff_table_entry_bc_t ) ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries data size.",
			 function );

			return( -1 );
		}
		number_of_table_entries = (size_t) ( table_entries_data_size / sizeof( pff_table_entry_bc_t ) );

		if( number_of_table_entries > (size_t) INT_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: number of table entries value exceeds maximum.",
			 function );

			return( -1 );
		}
		if( libpff_table_entries_resize(
		     table,
		     1,
		     table->number_of_entries + (int) number_of_table_entries,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
			 "%s: unable to resize table entries.",
			 function );

			return( -1 );
		}
		for( table_entry_iterator = 0;
		     table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     (size_t) table_entry_iterator * sizeof( pff_table_entry_bc_t ),
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_iterator * sizeof( pff_table_entry_bc_t ) );

				return( -1 );
			}
			if( table_entry_data_size < sizeof( pff_table_entry_bc_t ) )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				return( -1 );
			}
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_bc_t *) table_entry_data )->item_entry_type,
			 table_entry_type );
			byte_stream_copy_to_uint16_little_endian(
			 ( (pff_table_entry_bc_t *) table_entry_data )->item_entry_value_type,
			 table_entry_value_type );

			if( libpff_table_read_entry_value(
			     table,
			     0,
			     table_entry_index,
			     (uint32_t) table_entry_type,
			     (uint32_t) table_entry_value_type,
			     (uint8_t *) ( (pff_table_entry_bc_t *) table_entry_data )->item_entry_value,
			     4,
			     table_index_array,
			     recovered,
			     io_handle,
			     file_io_handle,
			     local_descriptors,
			     name_to_id_map_list,
			     NULL,
			     NULL,
			     table_data_reference,
			     debug_item_type,
			     debug_ascii_codepage,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read entry value: %" PRIu32 ".",
				 function,
				 table_entry_index );

				return( -1 );
			}
			table_entry_index++;
		}
	}
	return( 1 );
}

/* Reads the table values array
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_values_array(
     libpff_table_t *table,
     uint32_t table_entries_reference,
     uint32_t values_array_reference,
     uint8_t table_entry_identifier_size,
     uint8_t table_entry_value_size,
     uint16_t values_array_entry_size,
     libpff_array_t *table_index_array,
     libpff_array_t *table_entry_definitions_array,
     libpff_array_t *table_entries_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_data_reference_t *table_data_reference,
     libbfio_handle_t *file_io_handle,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *table_entries_data_reference = NULL;
	libpff_data_reference_t *table_values_data_reference  = NULL;
	libpff_table_entry_definition_t *entry_definition     = NULL;
	libpff_table_index_value_t *table_index_value         = NULL;
	uint8_t *table_data                                   = NULL;
	uint8_t *table_entry_data                             = NULL;
	uint8_t *table_entry_values_data                      = NULL;
	uint8_t *table_values_data                            = NULL;
	static char *function                                 = "libpff_table_read_values_array";
	size64_t table_entries_data_size                      = 0;
	size_t number_of_table_entries                        = 0;
	size_t number_of_values_array_entries                 = 0;
	size_t table_data_size                                = 0;
	size_t table_entry_data_size                          = 0;
	size_t table_entry_size                               = 0;
	size_t table_entry_values_data_size                   = 0;
	size_t table_values_data_size                         = 0;
	uint32_t table_entry_iterator                         = 0;
	uint32_t table_entry_values_array_identifier          = 0;
	uint32_t table_entry_values_array_number              = 0;
	uint32_t table_values_array_identifier                = 0;
	uint32_t table_values_data_array_entry                = 0;
	int entry_definition_iterator                         = 0;
	int result                                            = 0;
	int table_entries_array_iterator                      = 0;
	int table_entry_index                                 = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( ( table_entry_identifier_size != 4 )
	 || ( ( table_entry_value_size != 2 )
	  && ( table_entry_value_size != 4 ) ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry identifier size: 0x%02" PRIx8 " and table entry value size: 0x%02" PRIx8 ".",
		 function,
		 table_entry_identifier_size,
		 table_entry_value_size );

		return( -1 );
	}
	if( values_array_entry_size == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_ZERO_OR_LESS,
		 "%s: invalid values array entry size value zero or less.",
		 function );

		return( -1 );
	}
	if( table_index_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table index array.",
		 function );

		return( -1 );
	}
	if( table_entry_definitions_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry definitions array.",
		 function );

		return( -1 );
	}
	if( table_entries_array == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entries array.",
		 function );

		return( -1 );
	}
	if( io_handle == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid io handle.",
		 function );

		return( -1 );
	}
	/* Check if the table contains any entries
	 */
	if( ( table_entries_reference == 0 )
	 && ( values_array_reference == 0 ) )
	{
#if defined( HAVE_VERBOSE_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table contains no entries.\n",
			 function );
		}
#endif

		return( 1 );
	}
	if( table_entries_reference == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table contains value array but no table entries.",
		 function );

		return( -1 );
	}
	if( values_array_reference == 0 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: table contains table entries but no value array.",
		 function );

		return( -1 );
	}
	table_entry_size = table_entry_identifier_size + table_entry_value_size;

	/* Determine the values array entries
	 */

	/* TODO check node type ! */
	if( ( values_array_reference & 0x0000001fUL ) != 0 )
	{
		result = libpff_descriptor_local_read(
		          values_array_reference,
		          recovered,
		          local_descriptors,
		          io_handle,
		          file_io_handle,
		          NULL,
		          &table_values_data_reference,
		          error );

		if( result == -1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_IO,
			 LIBERROR_IO_ERROR_READ_FAILED,
			 "%s: unable to read table values data array for descriptor: %" PRIu32 ".",
			 function,
			 values_array_reference );

			return( -1 );
		}
		else if( result == 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing descriptor: %" PRIu32 ".",
			 function,
			 values_array_reference );

			return( -1 );
		}
	}
	else
	{
		if( libpff_table_index_array_get_value(
		     table_index_array,
		     values_array_reference,
		     &table_index_value,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve values array entries reference.",
			 function );

			return( -1 );
		}
		if( table_index_value == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid table index value.",
			 function );

			return( -1 );
		}
		/* Retrieve the corresponding table data array entry
		 */
		if( libpff_data_reference_get_entry_data(
		     table_data_reference,
		     file_io_handle,
		     table_index_value->array_entry,
		     &table_data,
		     &table_data_size,
		     0,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table values array data array entry: %" PRIu32 ".",
			 function,
			 table_index_value->array_entry );

			return( -1 );
		}
		if( table_data == NULL )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
			 "%s: invalid table data array entry: %" PRIu32 ".",
			 function,
			 table_index_value->array_entry );

			return( -1 );
		}
		if( (size_t) table_index_value->offset >= table_data_size )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
			 "%s: table value offset exceeds table data size.",
			 function );

			return( -1 );
		}
		/* Embed data in data array structure
		 */
		if( libpff_data_reference_initialize_buffer(
		     &table_values_data_reference,
		     &( table_data[ table_index_value->offset ] ),
		     table_index_value->size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create data reference.",
			 function );

			return( -1 );
		}
	}
	if( libpff_data_reference_get_entry_data(
	     table_values_data_reference,
	     file_io_handle,
	     0,
	     &table_values_data,
	     &table_values_data_size,
	     0,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve table values data array entry: %" PRIu32 ".",
		 function,
		 table_values_data_array_entry );

		libpff_data_reference_free(
		 &table_values_data_reference,
		 NULL );

		return( -1 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( ( values_array_reference & 0x0000001fUL ) != 0 )
	{
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table values data:\n",
			 function );
			libnotify_print_data(
			 table_values_data,
			 table_values_data_size );
		}
	}
#endif
	number_of_values_array_entries = table_values_data_size / values_array_entry_size;

	for( table_entries_array_iterator = 0;
	     table_entries_array_iterator < table_entries_array->number_of_entries;
	     table_entries_array_iterator++ )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table entries array entry\t\t\t: %d\n",
			 function,
			 table_entries_array_iterator );
		}
#endif

		if( libpff_array_get_entry(
		     table_entries_array,
		     table_entries_array_iterator,
		     (intptr_t **) &table_entries_data_reference,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries array entry: %d.",
			 function,
			 table_entries_data_reference );

			return( -1 );
		}
		if( libpff_data_reference_get_data_size(
		     table_entries_data_reference,
		     &table_entries_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve table entries data size.",
			 function );

			return( -1 );
		}
		if( ( table_entries_data_size % table_entry_size ) != 0 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported table entries data size.",
			 function );

			return( -1 );
		}
		number_of_table_entries = (size_t) ( table_entries_data_size / table_entry_size );

		if( number_of_table_entries > (size_t) INT_MAX )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
			 "%s: number of table entries value exceeds maximum.",
			 function );

			return( -1 );
		}
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: number of table entries\t\t\t\t: %" PRIzd "\n",
			 function,
			 number_of_table_entries );
		}
#endif
		if( table->number_of_sets < ( table_entry_index + (int) number_of_table_entries ) )
		{
			if( libpff_table_entries_resize(
			     table,
			     table_entry_index + (int) number_of_table_entries,
			     table_entry_definitions_array->number_of_entries,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
				 "%s: unable to resize table entries.",
				 function );

				return( -1 );
			}
		}
		for( table_entry_iterator = 0;
		     table_entry_iterator < number_of_table_entries;
		     table_entry_iterator++ )
		{
			if( libpff_data_reference_get_entry_data_at_offset(
			     table_entries_data_reference,
			     file_io_handle,
			     (size_t) table_entry_iterator * table_entry_size,
			     &table_entry_data,
			     &table_entry_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry data at offset: %" PRIzd ".",
				 function,
				 (size_t) table_entry_iterator * table_entry_size );

				return( -1 );
			}
			if( table_entry_data_size < table_entry_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table entry data size value too small.",
				 function );

				return( -1 );
			}
			byte_stream_copy_to_uint32_little_endian(
			 table_entry_data,
			 table_entry_values_array_identifier );

			table_entry_data += 4;

			if( table_entry_value_size == 2 )
			{
				byte_stream_copy_to_uint16_little_endian(
				 table_entry_data,
				 table_entry_values_array_number );

				table_entry_data += 2;
			}
			else if( table_entry_value_size == 4 )
			{
				byte_stream_copy_to_uint32_little_endian(
				 table_entry_data,
				 table_entry_values_array_number );

				table_entry_data += 4;
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table entry: %03" PRIu32 " values array identifier\t: 0x%08" PRIx32 "\n",
				 function,
				 table_entry_iterator,
				 table_entry_values_array_identifier );
				libnotify_printf(
				 "%s: table entry: %03" PRIu32 " values array number\t\t: %" PRIu32 "\n",
				 function,
				 table_entry_iterator,
				 table_entry_values_array_number );
				libnotify_printf(
				 "\n" );
			}
#endif

			/* If the value array numbers are not stored sequential
			 * resize the table entries to the required size
			 */
			if( (uint32_t) table->number_of_sets < table_entry_values_array_number )
			{
				if( libpff_table_entries_resize(
				     table,
				     (int) table_entry_values_array_number,
				     table_entry_definitions_array->number_of_entries,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_RESIZE_FAILED,
					 "%s: unable to resize table entries.",
					 function );

					return( -1 );
				}
			}
			if( libpff_data_reference_get_entry_data_at_aligned_index(
			     table_values_data_reference,
			     file_io_handle,
			     table_entry_values_array_number,
			     values_array_entry_size,
			     &table_entry_values_data,
			     &table_entry_values_data_size,
			     0,
			     error ) != 1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve table entry values data for index: %" PRIu32 ".",
				 function,
				 table_entry_values_array_number );

				return( -1 );
			}
			if( table_entry_values_data_size < (size_t) values_array_entry_size )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
				 "%s: invalid table values data size value too small.",
				 function );

				return( -1 );
			}
			for( entry_definition_iterator = 0;
			     entry_definition_iterator < table_entry_definitions_array->number_of_entries;
			     entry_definition_iterator++ )
			{
				if( libpff_array_get_entry(
				     table_entry_definitions_array,
				     entry_definition_iterator,
				     (intptr_t **) &entry_definition,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve table entry definitions array entry: %d.",
					 function,
					 entry_definition_iterator );

					libpff_data_reference_free(
					 &table_values_data_reference,
					 NULL );

					return( -1 );
				}
				if( entry_definition == NULL )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
					 "%s: invalid entry definition: %d.",
					 function,
					 entry_definition_iterator );

					libpff_data_reference_free(
					 &table_values_data_reference,
					 NULL );

					return( -1 );
				}
				/* For some unknown reason when the values array is read
				 * the data array is padded with zero or remnant values
				 * therefore the values array entries do not align
				 * this check is makes sure the aligment is correct
				 */
				if( entry_definition_iterator == 0 )
				{
					if( entry_definition->entry_offset != 0 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
						 "%s: unsupported first entry definition offset.",
						 function );

						libpff_data_reference_free(
						 &table_values_data_reference,
						 NULL );

						return( -1 );
					}
					if( entry_definition->entry_size != 4 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
						 "%s: unsupported first entry definition size.",
						 function );

						libpff_data_reference_free(
						 &table_values_data_reference,
						 NULL );

						return( -1 );
					}
					byte_stream_copy_to_uint32_little_endian(
					 table_entry_values_data,
					 table_values_array_identifier );

					/* If decryption was forced reread the entry without decryption
					 */
					if( ( io_handle->force_decryption != 0 )
					 && ( table_entry_values_array_identifier != table_values_array_identifier ) )
					{
						if( libpff_data_reference_get_entry_data_at_aligned_index(
						     table_values_data_reference,
						     file_io_handle,
						     table_entry_values_array_number,
						     values_array_entry_size,
						     &table_entry_values_data,
						     &table_entry_values_data_size,
						     LIBPFF_GET_ENTRY_FLAG_IGNORE_CACHE | LIBPFF_GET_ENTRY_FLAG_IGNORE_FORCE_DECRYPTION,
						     error ) != 1 )
						{
							liberror_error_set(
							 error,
							 LIBERROR_ERROR_DOMAIN_RUNTIME,
							 LIBERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to retrieve table entry values data for index: %" PRIu32 ".",
							 function,
							 table_entry_values_array_number );

							return( -1 );
						}
						if( table_entry_values_data_size < (size_t) values_array_entry_size )
						{
							liberror_error_set(
							 error,
							 LIBERROR_ERROR_DOMAIN_RUNTIME,
							 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
							 "%s: invalid table values data size value too small.",
							 function );

							return( -1 );
						}
						byte_stream_copy_to_uint32_little_endian(
						 table_entry_values_data,
						 table_values_array_identifier );
					}
					if( table_entry_values_array_identifier != table_values_array_identifier )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_VALUE_OUT_OF_RANGE,
						 "%s: mismatch in values array identifier (%" PRIu32 " != %" PRIu32 ").",
						 function,
						 table_entry_values_array_identifier,
						 table_values_array_identifier );

						libpff_data_reference_free(
						 &table_values_data_reference,
						 NULL );

						return( -1 );
					}
#if defined( HAVE_DEBUG_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: values array entry data:\n",
						 function );
						libnotify_print_data(
						 table_entry_values_data,
						 values_array_entry_size );
					}
#endif
				}
				if( entry_definition->entry_offset > values_array_entry_size )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
					 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
					 "%s: invalid entry offset value exceeds values array entry size.",
					 function );

					return( -1 );
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: values array entry data at offset: %" PRIu16 " of size: %" PRIu16 "\n",
					 function,
					 entry_definition->entry_offset,
					 entry_definition->entry_size );
					libnotify_print_data(
					 &( table_entry_values_data[ entry_definition->entry_offset ] ),
					 (size_t) entry_definition->entry_size );
				}
#endif
				/* To prevent multiple lookups the name to id map is not passed
				 */
				if( libpff_table_read_entry_value(
				     table,
				     table_entry_index,
				     entry_definition_iterator,
				     entry_definition->entry_type,
				     entry_definition->value_type,
				     &( table_entry_values_data[ entry_definition->entry_offset ] ),
				     (size_t) entry_definition->entry_size,
				     table_index_array,
				     recovered,
				     io_handle,
				     file_io_handle,
				     local_descriptors,
				     NULL,
				     entry_definition->name_to_id_map_entry,
				     entry_definition->item_entry_values_table,
				     table_data_reference,
				     LIBPFF_DEBUG_ITEM_TYPE_DEFAULT,
				     debug_ascii_codepage,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_IO,
					 LIBERROR_IO_ERROR_READ_FAILED,
					 "%s: unable to read entry value: %" PRIu32 ".",
					 function,
					 table_entry_values_array_number );

					libpff_data_reference_free(
					 &table_values_data_reference,
					 NULL );

					return( -1 );
				}
			}
			table_entry_index++;
		}
	}
	if( libpff_data_reference_free(
	     &table_values_data_reference,
	     error ) != 1 )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free table values data array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Reads a table item entry value
 * Returns 1 if successful or -1 on error
 */
int libpff_table_read_entry_value(
     libpff_table_t *table,
     int set_index,
     int entry_index,
     uint32_t table_entry_type,
     uint32_t table_entry_value_type,
     uint8_t *table_entry_value,
     size_t table_entry_value_size,
     libpff_array_t *table_index_array,
     uint8_t recovered,
     libpff_io_handle_t *io_handle,
     libbfio_handle_t *file_io_handle,
     libpff_tree_node_t *local_descriptors,
     libpff_list_t *name_to_id_map_list,
     libpff_internal_name_to_id_map_entry_t *name_to_id_map_entry,
     libpff_table_t *item_entry_values_table,
     libpff_data_reference_t *table_data_reference,
     int debug_item_type,
     int debug_ascii_codepage,
     liberror_error_t **error )
{
	libpff_data_reference_t *table_entry_value_data_reference = NULL;
	libpff_list_element_t *list_element                       = NULL;
	libpff_table_entry_t *table_entry                         = NULL;
	libpff_table_index_value_t *table_index_value             = NULL;
	uint8_t *table_data                                       = NULL;
	uint8_t *table_entry_value_data                           = NULL;
	static char *function                                     = "libpff_table_read_entry_value";
	size_t table_data_size                                    = 0;
	size_t table_entry_value_data_size                        = 0;
	uint64_t entry_value                                      = 0;
	int list_iterator                                         = 0;
	int result                                                = 0;

	if( table == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table.",
		 function );

		return( -1 );
	}
	if( table->entry == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entries.",
		 function );

		return( -1 );
	}
	if( ( set_index < 0 )
	 || ( set_index >= table->number_of_sets ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid set index value out of range.",
		 function );

		return( -1 );
	}
	if( ( entry_index < 0 )
	 || ( entry_index >= table->number_of_entries ) )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
		 "%s: invalid entry index value out of range.",
		 function );

		return( -1 );
	}
	if( table_entry_value == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid table entry value.",
		 function );

		return( -1 );
	}
	if( table->entry[ set_index ] == NULL )
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: invalid table - missing entry set.",
		 function );

		return( -1 );
	}
	table_entry = table->entry[ set_index ];

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		libnotify_printf(
		 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry type\t\t\t: 0x%04" PRIx32 "",
		 function,
		 set_index,
		 entry_index,
		 table_entry_type );
	}
#endif

	if( ( table_entry_type >= 0x8000 )
	 || ( table_entry_type <= 0xfffe ) )
	{
		/* The corresponding name to id map entry was already determined
		 */
		if( name_to_id_map_entry != NULL )
		{
			( table_entry[ entry_index ] ).name_to_id_map_entry = name_to_id_map_entry;
		}
		else if( name_to_id_map_list != NULL )
		{
			list_element = name_to_id_map_list->first;

			for( list_iterator = 0;
			     list_iterator < name_to_id_map_list->number_of_elements;
			     list_iterator++ )
			{
				if( list_element == NULL )
				{
					break;
				}
				if( list_element->value == NULL )
				{
					continue;
				}
				if( ( (libpff_internal_name_to_id_map_entry_t *) list_element->value )->identifier == table_entry_type )
				{
					( table_entry[ entry_index ] ).name_to_id_map_entry = (libpff_internal_name_to_id_map_entry_t *) list_element->value;
				}
				list_element = list_element->next;
			}
		}
	}
	( table_entry[ entry_index ] ).identifier.format     = LIBPFF_TABLE_ENTRY_IDENTIFIER_FORMAT_MAPI_PROPERTY;
	( table_entry[ entry_index ] ).identifier.entry_type = table_entry_type;
	( table_entry[ entry_index ] ).identifier.value_type = table_entry_value_type;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		if( debug_item_type == LIBPFF_DEBUG_ITEM_TYPE_NAME_TO_ID_MAP )
		{
			libnotify_printf(
			 " (%s : %s)\n",
			 libpff_debug_get_name_to_id_map_property_type_identifier(
			  table_entry_type,
			  table_entry_value_type ),
			 libpff_debug_get_name_to_id_map_property_type_description(
			  table_entry_type,
			  table_entry_value_type ) );
		}
		else if( ( table_entry[ entry_index ] ).name_to_id_map_entry != NULL )
		{
			if( ( table_entry[ entry_index ] ).name_to_id_map_entry->type == LIBPFF_NAME_TO_ID_MAP_ENTRY_TYPE_STRING )
			{
				libnotify_printf(
				 " maps to: %s (%s : %s)\n",
				 (char *) ( table_entry[ entry_index ] ).name_to_id_map_entry->debug_string,
				 libfmapi_named_property_type_get_identifier(
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->guid,
				  (char *) ( table_entry[ entry_index ] ).name_to_id_map_entry->debug_string,
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->value_size,
				  ( table_entry[ entry_index ] ).identifier.value_type ),
				 libfmapi_named_property_type_get_description(
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->guid,
				  (char *) ( table_entry[ entry_index ] ).name_to_id_map_entry->debug_string,
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->value_size,
				  ( table_entry[ entry_index ] ).identifier.value_type ) );
			}
			else
			{
				libnotify_printf(
				 " maps to: 0x%04" PRIx32 " (%s : %s)\n",
				 ( table_entry[ entry_index ] ).name_to_id_map_entry->numeric_value,
				 libfmapi_property_type_get_identifier(
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->guid,
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->numeric_value,
				  ( table_entry[ entry_index ] ).identifier.value_type ),
				 libfmapi_property_type_get_description(
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->guid,
				  ( table_entry[ entry_index ] ).name_to_id_map_entry->numeric_value,
				  ( table_entry[ entry_index ] ).identifier.value_type ) );
			}
		}
		else
		{
			libnotify_printf(
			 " (%s : %s)\n",
			 libfmapi_property_type_get_identifier(
			  NULL,
			  ( table_entry[ entry_index ] ).identifier.entry_type,
			  ( table_entry[ entry_index ] ).identifier.value_type ),
			 libfmapi_property_type_get_description(
			  NULL,
			  ( table_entry[ entry_index ] ).identifier.entry_type,
			  ( table_entry[ entry_index ] ).identifier.value_type ) );
		}
		libnotify_printf(
		 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value type\t\t\t: 0x%04" PRIx16 " (%s : %s)\n",
		 function,
		 set_index,
		 entry_index,
		 table_entry_value_type,
		 libfmapi_value_type_get_identifier(
		  ( table_entry[ entry_index ] ).identifier.value_type ),
		 libfmapi_value_type_get_description(
		  ( table_entry[ entry_index ] ).identifier.value_type ) );
	}
#endif

	if( table_entry_value_size == 1 )
	{
		entry_value = table_entry_value[ 0 ];
	}
	else if( table_entry_value_size == 2 )
	{
		byte_stream_copy_to_uint16_little_endian(
		 table_entry_value,
		 entry_value );
	}
	else if( table_entry_value_size == 4 )
	{
		byte_stream_copy_to_uint32_little_endian(
		 table_entry_value,
		 entry_value );
	}
	else if( table_entry_value_size == 8 )
	{
		byte_stream_copy_to_uint64_little_endian(
		 table_entry_value,
		 entry_value );
	}
	else
	{
		liberror_error_set(
		 error,
		 LIBERROR_ERROR_DOMAIN_RUNTIME,
		 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
		 "%s: unsupported table entry value size: %" PRIu8 ".",
		 function,
		 table_entry_value_size );

		return( -1 );
	}
	/* Check if there is a item entry values (a5) table
	 */
	if( item_entry_values_table != NULL )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libnotify_verbose != 0 )
		{
			libnotify_printf(
			 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry values table reference\t: 0x%08" PRIx64 "\n",
			 function,
			 set_index,
			 entry_index,
			 entry_value );
		}
#endif

		/* TODO check entry value type */
		if( ( entry_value & 0x0000001fUL ) != 0 )
		{
			result = libpff_descriptor_local_read(
				  (uint32_t) entry_value,
				  recovered,
				  local_descriptors,
				  io_handle,
				  file_io_handle,
				  NULL,
				  &( ( table_entry[ entry_index ] ).value_data_reference ),
				  error );

			/* If the data descriptor could not be read mark it as missing
			 * and give it an empty value data reference
			 */
			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_IO,
				 LIBERROR_IO_ERROR_READ_FAILED,
				 "%s: unable to read item entry value data with descriptor: %" PRIu32 " - marked as missing.",
				 function,
				 (uint32_t) entry_value );

#if defined( HAVE_DEBUG_OUTPUT )
				if( ( libnotify_verbose != 0 )
				 && ( error != NULL )
				 && ( *error != NULL ) )
				{
					libnotify_print_error_backtrace(
					 *error );
				}
#endif
				liberror_error_free(
				 error );

				( table_entry[ entry_index ] ).value_data_reference = NULL;
				( table_entry[ entry_index ] ).flags               |= LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR;
				table->flags                                       |= LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA;
			}
			else if( result == 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: missing descriptor: %" PRIu32 " - marked as missing.\n",
					 function,
					 (uint32_t) entry_value );
				}
#endif
				( table_entry[ entry_index ] ).value_data_reference = NULL;
				( table_entry[ entry_index ] ).flags               |= LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR;
				table->flags                                       |= LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA;
			}
		}
		/* Check if the entry value is a value within the item entry values (a5) table
		 */
		else if( entry_value > 0 )
		{
			/* The item entry value reference needs to be transformed into a table set and entry
			 * value. Table array entries have been stored as separate sets.
			 */
			result = libpff_table_get_entry_value_by_index(
				  item_entry_values_table,
				  (int) ( ( entry_value >> 16 ) & 0xffff ),
				  (int) ( ( ( entry_value & 0xffe0 ) >> 5 ) - 1 ),
				  &table_entry_value_data_reference,
				  error );

			if( result == -1 )
			{
				liberror_error_set(
				 error,
				 LIBERROR_ERROR_DOMAIN_RUNTIME,
				 LIBERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve item entry value index: 0x%08" PRIx64 " (set: %d, entry: %d)",
				 function,
				 entry_value,
				 (int) ( ( entry_value >> 16 ) & 0xffff ),
				 (int) ( ( ( entry_value & 0xffe0 ) >> 5 ) - 1 ) );

				return( -1 );
			}
			/* A missing item entry value reference signifies an empty value (NULL)
			 */
			else if( result != 0 )
			{
				if( libpff_data_reference_get_entry_data_at_offset(
				     table_entry_value_data_reference,
				     file_io_handle,
				     0,
				     &table_entry_value_data,
				     &table_entry_value_data_size,
				     0,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve item entry value data at offset: 0.",
					 function );

					return( -1 );
				}
			}
		}
	}
	else
	{
		/* The Boolean (0x000b)
		 * is 1 byte of size in the 7c table
		 * is 4 bytes of size in the bc table
		 *
		 * the first byte contains the value
		 * the value is 0x00 if false or true otherwise
		 */
		if( table_entry_value_type == LIBPFF_VALUE_TYPE_BOOLEAN )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value\t\t\t: 0x%08" PRIx64 "\n",
				 function,
				 set_index,
				 entry_index,
				 entry_value );

				if( ( table_entry_value_size != 1 )
				 && ( table_entry_value_size != 4 ) )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
					 "%s: unsupported value type: 0x%08" PRIx32 " with value size: %" PRIu8 ".",
					 function,
					 table_entry_value_type,
					 table_entry_value_size );

					return( -1 );
				}
			}
#endif
			/* The first byte in the table entry value buffer
			 * contains the actual value of the boolean
			 */
			table_entry_value_data      = table_entry_value;
			table_entry_value_data_size = sizeof( uint8_t );
		}
		/* The Integer 16-bit signed (0x0002)
		 * is 2 bytes of size in the 7c table
		 * is 4 bytes of size in the bc table
		 *
		 * the first two bytes contains the value
		 */
		else if( table_entry_value_type == LIBPFF_VALUE_TYPE_INTEGER_16BIT_SIGNED )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value\t\t\t: 0x%08" PRIx64 "\n",
				 function,
				 set_index,
				 entry_index,
				 entry_value );

				if( ( table_entry_value_size != 2 )
				 && ( table_entry_value_size != 4 ) )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
					 "%s: unsupported value type: 0x%08" PRIx32 " with value size: %" PRIu8 ".",
					 function,
					 table_entry_value_type,
					 table_entry_value_size );

					return( -1 );
				}
			}
#endif
			table_entry_value_data      = table_entry_value;
			table_entry_value_data_size = sizeof( uint16_t );
		}
		/* The Integer 32-bit signed (0x0003)
		 *     Floating point single precision (0x0004)
		 *     Error scode (0x000a)
		 *
		 * is 4 bytes of size in the 7c and bc table
		 */
		else if( ( table_entry_value_type == LIBPFF_VALUE_TYPE_INTEGER_32BIT_SIGNED )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_FLOAT_32BIT )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_ERROR ) )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value\t\t\t: 0x%08" PRIx64 "\n",
				 function,
				 set_index,
				 entry_index,
				 entry_value );

				if( table_entry_value_size != 4 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
					 "%s: unsupported value type: 0x%08" PRIx32 " with value size: %" PRIu8 ".",
					 function,
					 table_entry_value_type,
					 table_entry_value_size );

					return( -1 );
				}
			}
#endif
			table_entry_value_data      = table_entry_value;
			table_entry_value_data_size = sizeof( uint32_t );
		}
		/* The Floating point double precision (0x0005)
		 *     Currency (64-bit) (0x0006)
		 *     Application time (64-bit) (0x0007)
		 *     Integer 64-bit signed (0x0014)
		 *     Windows Filetime (64-bit) (0x0040)
		 *
		 * is 8 bytes of size in the 7c table
		 */
		else if( ( table_entry_value_size == 8 )
		      && ( ( table_entry_value_type == LIBPFF_VALUE_TYPE_DOUBLE_64BIT )
		       || ( table_entry_value_type == LIBPFF_VALUE_TYPE_CURRENCY )
		       || ( table_entry_value_type == LIBPFF_VALUE_TYPE_APPLICATION_TIME )
		       || ( table_entry_value_type == LIBPFF_VALUE_TYPE_INTEGER_64BIT_SIGNED )
		       || ( table_entry_value_type == LIBPFF_VALUE_TYPE_FILETIME ) ) )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value\t\t\t: 0x%08" PRIx64 "\n",
				 function,
				 set_index,
				 entry_index,
				 entry_value );
			}
#endif

			table_entry_value_data      = table_entry_value;
			table_entry_value_data_size = sizeof( uint64_t );
		}
		/* These values are references in the bc table
		 */
		else if( ( table_entry_value_type == LIBPFF_VALUE_TYPE_DOUBLE_64BIT )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_CURRENCY )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_APPLICATION_TIME )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_OBJECT )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_INTEGER_64BIT_SIGNED )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_STRING_ASCII )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_STRING_UNICODE )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_FILETIME )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_GUID )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_SERVER_IDENTIFIER )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_RESTRICTION )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_RULE_ACTION )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_BINARY_DATA )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_INTEGER_16BIT_SIGNED )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_INTEGER_32BIT_SIGNED )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_FLOAT_32BIT )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_DOUBLE_64BIT )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_CURRENCY )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_APPLICATION_TIME )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_INTEGER_64BIT_SIGNED )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_STRING_ASCII )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_STRING_UNICODE )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_FILETIME )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_GUID )
		      || ( table_entry_value_type == LIBPFF_VALUE_TYPE_MULTI_VALUE_BINARY_DATA ) )
		{
			/* Check if the entry value is a referenced local descriptor
			 */
			/* TODO check entry value type */
			if( ( entry_value & 0x0000001fUL ) != 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value reference\t\t: %" PRIu64 " ",
					 function,
					 set_index,
					 entry_index,
					 entry_value );
					libpff_debug_print_node_identifier_type(
					 (uint8_t) ( entry_value & 0x0000001fUL ) );
					libnotify_printf(
					 "\n" );
				}
#endif
				if( entry_value > (uint64_t) UINT32_MAX )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
					 LIBERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
					 "%s: entry value reference value exceeds maximum.",
					 function );

					return( -1 );
				}
				result = libpff_descriptor_local_read(
				          (uint32_t) entry_value,
				          recovered,
				          local_descriptors,
				          io_handle,
				          file_io_handle,
				          NULL,
				          &( ( table_entry[ entry_index ] ).value_data_reference ),
				          error );

				/* If the data descriptor could not be read mark it as missing
				 * and give it an empty value data reference
				 */
				if( result == -1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_IO,
					 LIBERROR_IO_ERROR_READ_FAILED,
					 "%s: unable to read item entry value data with descriptor: %" PRIu32 " - marked as missing.",
					 function,
					 (uint32_t) entry_value );

#if defined( HAVE_DEBUG_OUTPUT )
					if( ( libnotify_verbose != 0 )
					 && ( error != NULL )
					 && ( *error != NULL ) )
					{
						libnotify_print_error_backtrace(
						 *error );
					}
#endif
					liberror_error_free(
					 error );

					( table_entry[ entry_index ] ).value_data_reference = NULL;
					( table_entry[ entry_index ] ).flags               |= LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR;
					table->flags                                       |= LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA;
				}
				else if( result == 0 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libnotify_verbose != 0 )
					{
						libnotify_printf(
						 "%s: missing descriptor: %" PRIu32 " - marked as missing.\n",
						 function,
						 (uint32_t) entry_value );
					}
#endif
					( table_entry[ entry_index ] ).value_data_reference = NULL;
					( table_entry[ entry_index ] ).flags               |= LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR;
					table->flags                                       |= LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA;
				}
			}
			/* Check if the entry value is empty
			 */
			else if( entry_value == 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libnotify_verbose != 0 )
				{
					libnotify_printf(
					 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value\t\t\t: <NULL>\n",
					 function,
					 set_index,
					 entry_index );
				}
#endif
			}
			/* Otherwise the entry value is a referenced table value
			 */
			else
			{
				/* Fetch the table entry value reference
				 */
				if( libpff_table_index_array_get_value(
				     table_index_array,
				     (uint32_t) entry_value,
				     &table_index_value,
				     error ) != 1 )
				{
					liberror_error_set(
					 error,
					 LIBERROR_ERROR_DOMAIN_RUNTIME,
					 LIBERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to retrieve table entry value reference: 0x%08" PRIx64 ".",
					 function,
					 entry_value );

#if defined( HAVE_DEBUG_OUTPUT )
					if( ( libnotify_verbose != 0 )
					 && ( error != NULL )
					 && ( *error != NULL ) )
					{
						libnotify_print_error_backtrace(
						 *error );
					}
#endif
					liberror_error_free(
					 error );

					( table_entry[ entry_index ] ).value_data_reference = NULL;
					( table_entry[ entry_index ] ).flags               |= LIBPFF_TABLE_ENTRY_FLAG_MISSING_DATA_DESCRIPTOR;
					table->flags                                       |= LIBPFF_TABLE_FLAG_MISSING_ENTRY_DATA;
				}
				else
				{
					if( table_index_value == NULL )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_VALUE_MISSING,
						 "%s: invalid table index value.",
						 function );

						return( -1 );
					}
					/* Retrieve the corresponding table data array entry
					 */
					if( libpff_data_reference_get_entry_data(
					     table_data_reference,
					     file_io_handle,
					     table_index_value->array_entry,
					     &table_data,
					     &table_data_size,
					     0,
					     error ) != 1 )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_RUNTIME,
						 LIBERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to retrieve table data array entry: %" PRIu32 ".",
						 function,
						 table_index_value->array_entry );

						return( -1 );
					}
					if( table_data == NULL )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
						 LIBERROR_ARGUMENT_ERROR_INVALID_VALUE,
						 "%s: invalid table data array entry: %" PRIu32 ".",
						 function,
						 table_index_value->array_entry );

						return( -1 );
					}
					if( (size_t) table_index_value->offset >= table_data_size )
					{
						liberror_error_set(
						 error,
						 LIBERROR_ERROR_DOMAIN_ARGUMENTS,
						 LIBERROR_ARGUMENT_ERROR_VALUE_OUT_OF_RANGE,
						 "%s: table value offset exceeds table data size.",
						 function );

						return( -1 );
					}
					table_entry_value_data      = &( table_data[ table_index_value->offset ] );
					table_entry_value_data_size = table_index_value->size;
				}
			}
		}
		else
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libnotify_verbose != 0 )
			{
				libnotify_printf(
				 "%s: table set: %03" PRIu32 " entry: %03" PRIu32 " item entry value (reference)\t: 0x%08" PRIx64 "\n",
				 function,
				 set_index,
				 entry_index,
				 entry_value );
				libnotify_printf(
				 "\n" );
			}
#endif

			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported value type: 0x%08" PRIx32 " with value size: %" PRIu8 ".",
			 function,
			 table_entry_value_type,
			 table_entry_value_size );

			return( -1 );
		}
	}
	if( ( table_entry[ entry_index ] ).value_data_reference == NULL )
	{
		/* Create a data reference of the table entry value data
		 */
		if( libpff_data_reference_initialize_buffer(
		     &( ( table_entry[ entry_index ] ).value_data_reference ),
		     table_entry_value_data,
		     table_entry_value_data_size,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create table entry value data reference.",
			 function );

			return( -1 );
		}
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libnotify_verbose != 0 )
	{
		if( libpff_debug_property_type_value_reference_print(
		     name_to_id_map_list,
		     ( table_entry[ entry_index ] ).identifier.entry_type,
		     ( table_entry[ entry_index ] ).identifier.value_type,
		     ( table_entry[ entry_index ] ).value_data_reference,
		     file_io_handle,
		     debug_item_type,
		     debug_ascii_codepage,
		     error ) != 1 )
		{
			liberror_error_set(
			 error,
			 LIBERROR_ERROR_DOMAIN_RUNTIME,
			 LIBERROR_RUNTIME_ERROR_PRINT_FAILED,
			 "%s: unable to print table entry value reference.",
			 function );

			return( -1 );
		}
	}
#endif
	return( 1 );
}

